/*eslint-disable */
/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */

import UIUtils from '../utils/UIUtils'  
import Constants from '../utils/Constants'
import TypedObject from '../core/TypedObject'
import Point from '../utils/Point'

/**
 * Attaches the tooltip behavior to any UI component. The behavior does not
 * automatically trigger, however, it does wrap all the functionality needed to
 * display a tooltip type control next to the requesting control.
 * @namespace flexiciousNmsp
 * @class TooltipBehavior
 * 
 * @constructor
 */
export default class TooltipBehavior extends TypedObject {
    constructor(ownerComponent) {
        super();
        /**
         * The current tooltip object.
         * @property currentTooltip
         * @type UIComponent
         */
        this.currentTooltip = null;
        /**
         * The current tooltip object trigger.
         * @property currentTooltipTrigger
         * @type UIComponent
         */
        this.currentTooltipTrigger = null;
        /**
         * A timer that watches the mouse and destroys the tooltip when the mouse is outside the tooltip and the the trigger for more than
         * tooltipWatcherTimeout msec interval
         * @property tooltipWatcher
         * @type Timer
         */
        this.tooltipWatcher = null;
        /**
         * Amount of time to wait after user has moved the mouse away from the tooltip,
         * before we destroy the tooltips. Defaults to 500 msec.
         * @property tooltipWatcherTimeout
         * @type int
         */
        this.tooltipWatcherTimeout = 1000;
        /**
         * The spinner associates itself with this component
         * @property ownerComponent
         * @type UIComponent
         */
        this.ownerComponent = ownerComponent;

    }

    getClassNames() {
        return ["TooltipBehavior", "TypedObject"];
    }

    /**
     * Displays a tooltip for the control in question. The tooltip will disappear if the mouse
     * moves over an area that is not the 'relativeTo' component or the tooltip component.
     *
     * By default, the tooltip will go away once you hover the mouse out of the trigger cell or the tooltip and stayed that way for tooltipWatcherTimeout. You may also
     * manually remove the tooltip by calling the hideToolTip() function.
     *
     * @method showTooltip
     * @param relativeTo Which component to position the popup relative to
     * @param tooltip The popup to display
     * @param dataContext If the popup has a data property, it will be set to this value
     * @param point If you specify this, the relativeTo is ignored, and the popup appears at the exact point you specify. Please ensure that the X and Y are relative to the Grid.
     * @param leftOffset Whether to shift the popup left after calculating the positions, for customizing the actual position
     * @param topOffset	Whether to shift the popup top after calculating the positions, for customizing the actual position
     * @param offScreenMath	Whether or not to adjust the popup if it appears off screen
     * @param where		One of three values, left, right or none. If left, positions to bottom left, if right, positions to bottom right, if none, positions right below the relativeTo component.
     * @param container	The holder for the tooltip, defaults to UIUtils.getTopLevelApplication(). You may need to override in multi window Air apps.
     */
    showTooltip(
        relativeTo,
        tooltip,
        dataContext,
        point,
        leftOffset,
        topOffset,
        offScreenMath,
        where,
        container) {


        if (this.currentTooltip) this.hideTooltip();
        if (tooltip.implementsOrExtends) {
            tooltip.tooltipOwner = this.ownerComponent;

        }

        this.currentTooltip = tooltip;
        this.currentTooltipTrigger = relativeTo;
        if (!container)
            container = UIUtils.getTopLevelApplication();
        this.container = container;
        const tooltipObject = tooltip;
        if (tooltipObject.typeName) {
            UIUtils.checkSetterAndApply(tooltipObject, "data", dataContext)
            UIUtils.checkSetterAndApply(tooltipObject, "grid", this.ownerComponent)
        }


        if (this.currentTooltip.move) {
            const imageOffset = UIUtils.offset(relativeTo.domElement);
            const containerOffset = UIUtils.offset(container.domElement);
            const tooltipPoint = new Point(imageOffset.left - containerOffset.left, imageOffset.top - containerOffset.top);
            this.currentTooltip.move(tooltipPoint.x, tooltipPoint.y);
        }

        //UIUtils.showTooltip(relativeTo, tooltip, dataContext, point, leftOffset, topOffset, offScreenMath, where, container, this.ownerComponent);

        container.addChild(this.currentTooltip);
        this.tooltipWatcher = new flexiciousNmsp.Timer(this.tooltipWatcherTimeout);
        this.tooltipWatcher.addEventListener(this, Constants.EVENT_TIMER,
            function (event) {
                if (!this.ownerComponent) {
                    this.tooltipWatcher.stop();
                    this.tooltipWatcher = null;
                    return;
                }
                if (!this.currentTooltip) { this.hideTooltip(); return; }
                let shouldShow = false;
                const pt = new flexiciousNmsp.Point(flexiciousNmsp.DisplayList.pageX, flexiciousNmsp.DisplayList.pageY);
                shouldShow = (
                    (this.currentTooltip.ownsPoint(pt)
                        || this.currentTooltipTrigger.ownsPoint(pt)))
                if (shouldShow) {
                    return;
                }
                this.hideTooltip();
            });

        this.tooltipWatcher.start();
    }

    /**
     * @method hideTooltip
     * Hides the current tooltip.
     */
    hideTooltip() {
        if (this.tooltipWatcher) {
            if (this.tooltipWatcher.running)
                this.tooltipWatcher.stop()
            this.tooltipWatcher.kill();
            this.tooltipWatcher = null;
        }
        if (this.currentTooltip) {
            this.container.removeChild(this.currentTooltip)
            if ("destroy" in this.currentTooltip) this.currentTooltip["destroy"].apply();
            this.currentTooltipTrigger = null;
            this.currentTooltip.kill();
            this.currentTooltip = null;
        }
    }

    /**
     * @method kill
     */
    kill() {
        this.hideTooltip();
        this.ownerComponent = null;
        this.currentTooltip = null;
        this.currentTooltipTrigger = null;
        this.tooltipWatcher = null;
        this.ownerComponent = null;
    }
}

flexiciousNmsp.TooltipBehavior = TooltipBehavior; //add to name space 
TooltipBehavior.prototype.typeName = TooltipBehavior.typeName = 'TooltipBehavior';//for quick inspection