/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */
import React from 'react';
import TypedObject from './TypedObject'

/**
 * A base class for all objects that are capable of dispatching events.
 * @constructor
 * @class EventDispatcher
 * @namespace flexiciousNmsp
 */
export default class EventDispatcher extends React.Component {
    constructor(props, arg1, arg2) {
        super(props, arg1, arg2);
        /**
         * A list of event listeners for this event dispatcher object.
         * @type {Array}
         */
        this.eventListeners = [];
        this.typeCache={};
        if(this.constructed){
            this.constructed();
        }
    }

    constructed() {

    }

    /**
     * Returns true if the class name to check is in the list of class names defined for this class.
     * @method implementsOrExtends
     * @param name Name of the class to check
     * @return {Boolean}
     */
    implementsOrExtends(name) {
        if(this.typeCache[name]!=undefined){
            return this.typeCache[name];
        }
        let i;
        let className;
        let names;
        names = this.getClassNames();
        for (i = 0; i < names.length; i += 1) {
            className = names[i];
            if (className === name) {
                this.typeCache[name]=true;
                return true;
            }
        }
        this.typeCache[name]=false;
        return false;
    }


    getClassNames() {
        return ["TypedObject", this.typeName];
    }

    /**
     * Registers an event listener object with an EventDispatcher object so that the listener receives notification of an event.
     * If you no longer need an event listener, remove it by calling removeEventListener(), or memory problems could result.
     * Event listeners are not automatically removed from memory because the garbage collector does not remove the listener as long as the dispatching object exists.
     *
     * @param context  The "this" property in the event listener function will be set to the value of this parameter.
     * @param type  The type of event.
     * @param listener  The listener function that processes the event. This function must accept an Event object as its
     * only parameter and must return nothing
     */
    addEventListener(context, type, listener) {
        if (typeof (this.eventListeners[type]) === "undefined") {
            this.eventListeners[type] = [];
        }
        const list = this.eventListeners[type];
        if(typeof list == 'function')
            return;
        for (let i = 0; i < list.length; i += 1) {
            if(list[i][0]==listener && list[i][1]==context)
            {
                return;
            }
        }
        this.eventListeners[type].push([listener, context]);
    }

    /**
     * Removes a listener from the EventDispatcher object.
     * If there is no matching listener registered with the EventDispatcher object, a call to this method has no effect.
     * @param type  The type of event.
     * @param listener  The listener function that processes the event. This function must accept an Event object as its
     * only parameter and must return nothing
     */
    removeEventListener(type, listener) {
        let i;
        let list;
        let len;
        if (this.eventListeners[type] instanceof Array) {
            list = this.eventListeners[type];
            for (i = 0; i < list.length; i += 1) {
                if (list[i][0] === listener) {
                    list.splice(i, 1);
                    break;
                }
            }
        }
    }

    /**
     * Returns true if an event listener exists
     * @param type  The type of event.
     * @return {Boolean}
     */
    hasEventListener(type) {
        if (this.eventListeners[type] instanceof Array) {
            const list = this.eventListeners[type];
            return (list && list.length > 0);
        }
        return false;
    }

    /**
     * Dispatches an event into the event flow.
     * The event target is the EventDispatcher object upon which the dispatchEvent() method is called.
     *
     * @param event The Event object that is dispatched into the event flow.
     */
    dispatchEvent(event) {
        //console.log('Dispatching' + event.type);
        let i = 0;

        let events;
        if (typeof event === "string") {
            event = { type: event };
        }
        if (!event.target) {
            event.target = this;
        }
        event.currentTarget = this;
        if (!event.type) {
            throw new Error("Event object missing 'type' property.");
        }
        if (this.eventListeners[event.type] instanceof Array) {
            events = this.eventListeners[event.type];
            for (i = 0; i < events.length; i += 1) {
                events[i][0].call(typeof events[i][1] === "undefined" ? this : events[i][1], event);
            }
        }
        return true;
    }

    /**
     * Removes all event listeners
     */
    kill() {
        //removes all event listeners
        this.eventListeners = [];
        this.dead=true;
    }

    /**
     * Given an array of event dispatchers, calls the kill method on each of them.
     * @param arr   The list of event dispatchers to kill.
     */
    killArray(arr) {
        let i;
        let col;
        for (i = 0; i < arr.length; i += 1) {
            col = arr[i];
            col.kill();
        }
    }

    /**
     * Gets the value of the prop property on this object
     */
    getStyle(prop) {
        return this[prop]
    }

    /**
     * Sets the value of the prop property on this object
     */
    setStyle(prop, val) {
        this[prop]=val;
    }
}

flexiciousNmsp.EventDispatcher = EventDispatcher;
EventDispatcher.prototype.typeName = EventDispatcher.typeName = "EventDispatcher";//for quick inspection

