
/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */
import Constants from '../../utils/Constants'
import UIUtils from '../../utils/UIUtils'
import UIComponent from '../../core/UIComponent'
import CellUtils from '../utils/CellUtils'
/**
 * FlexDataGridCell is the container component for each of the DataGrid's cells. This is a psuedo abstract class, the cells
 * that you see in the grid are actually one of the following subclasses of this class:
 * <ul>
 * <li> FlexDataGridHeaderCell</li>
 * <li> FlexDataGridFilterCell</li>
 * <li> FlexDataGridFooterCell</li>
 * <li> FlexDataGridPagerCell</li>
 * <li> FlexDataGridLevelRendererCell</li>
 * <li> FlexDataGridExpandCollapseHeaderCell</li>
 * <li> FlexDataGridExpandCollapseCell</li>
 * <li> FlexDataGridPaddingCell </li>
 * <li> FlexDataGridDataCell</li>
 * </ul>
 *
 *
 * The FlexDataGridCell is responsible for sizing, positioning (based on padding), drawing the background,
 * and drawing the borders.
 * This class has a renderer property, which is the actual component that is displayed on the UI.
 * In case of the Header,Data or Footer cells the default renderer is a UITextField or UIFTETextField.
 * For Filter, it is an instance of the IFilterControl.
 * For the Pager, it is an IPager control. For the LevelRenderer it is an instance of the Class Factory that
 * you specify in the nextLevelRenderer of the associated column Level.
 * For the ExpandCollapse cells, it will draw an instance of the expand collapse icon on basis of the disclosureIcon style property
 * All the drawing happens in the drawCell method. It seperately calls the drawBackground and drawBorder methods. Usually
 * specifying the style properties, or the cellBackgroud/rowBackground/cellBorder/rowBorder functions is sufficient, but in
 * case its needed, these methods can be overridden in a custom implementation, and this custom implementation can then
 * be hooked in via the dataCellRenderer, headerCellRenderer, footerCellRenderer, pagerCellRenderer, filterCellRender,
 * expandCollapseHeaderCellRenderer,nestIndentPaddingCellRenderer, and expandCollapseCellRenderer of the column or the level.
 *
 * Please note, that this class is recycled, so for example, as you scroll, a FlexDataGridCell object that was
 * displaying some other item in the data provider previously could be reused to display the current one.
 *
 * @constructor
 * @namespace flexiciousNmsp
 * @class FlexDataGridCell
 * 
 */
export default class FlexDataGridCell extends UIComponent {
    constructor() {
        super("div");
        /**
         * The FlexDataGridColumnLevel associated with this cell. This set on initialize. This is usually never null.
         * @property level
         * @type {FlexDataGridColumnLevel}
         * @default  null
         */
        this.level = null;

        /**
         * @private 
         */
        this._column = null;

        /**
         * The rowInfo object associated with this cell. This cell should be within the cells collection of this rowInfo object.
         * Each cell has a corresponding rowinfo object. The rowInfo, in turn has a rowPositionInfo object. The key difference
         * between these two objects is that there are only as many rowInfo objects as there visible rows on the page. RowPositionInfo
         * objects on the other hand, there are as many of these as there are total items in the data provider.
         * @type { RowInfo}
         * @property rowInfo
         * @default null
         */
        this.rowInfo = null;

        /**
         * This property stores the temporary value of a cells background color when the user hovers over it.
         * The background colors of each cell are calculated using a series of rules as defined in the
         * documentation for the getBackgroundColors method of the CellUtils class.
         *
         * @copy com.flexicious.nestedtreedatagrid.cells.CellUtils#getBackgroundColors
         * @property currentBackgroundColors
         * @default null
         */
        this.currentBackgroundColors = null;


        /**
         * This property stores the temporary value of a cells text color when the user hovers over it.
         * The text colors of each cell are calculated using a series of rules as defined in the
         * documentation for the getTextColors method of the CellUtils class.
         *
         * @copy com.flexicious.nestedtreedatagrid.cells.CellUtils#getTextColors
         * @type {Array}
         * @property currentTextColors
         * @default null
         */
        this.currentTextColors = null;

        /**
         * The component info that holds this cell in the cells collection of the associated rowInfo object.
         *
         * When a row is rendered in the FlexDataGrid, the memory structure used to hold each row is the RowInfo object.
         * If you look at grid.bodyContainer.rows, each item in this collection is an object of type RowInfo. These are the
         * rows in the dataprovider that are currently rendered, which is different than the list of items in the dataprovider.
         * In the cells collection of the rowInfo object, we have a list of ComponentInfo objects. The componentInfo object
         * stores a bunch of additional information and provides a number of handy methods to interact with the cell object.
         * It is basically a wrapper around the IFlexDataGridCell object.

         * @type { ComponentInfo}
         * @property componentInfo
         * @default null
         */
        this.componentInfo = null;

        /**
         * @private 
         */
        this._rendererFactory = null;

        /**
         * @private 
         */
        this._renderer = null;

        /**
         * Once a grid is rendered, each cell in the grid is shown in the UI. After being shown, if the user performs operations
         * like filtering, or paging, that result in the cell being taken out of view and no longer being rendered on the UI,
         * instead of removing the cell from memory, we store it in the renderer cache of the grid. Prior to decommissioning the
         * cell, we call the destroy method on the cell. This method will set the destroyed flag to true.
         * @type {Boolean}
         * @property  _renderer
         * @default null
         */
        this.destroyed = null;

        /**
         * @private
         */
        this._newlyCreated = true;
        /**
         * Used by header cells and column group cells for drag drop move. This property should not be set
         * by your code.
         * @property moving
         * @type {Boolean}
         * @default false
         */
        this.moving = false;
        /**
         * When column.enableIcon=true, this is the icon that will be associated with the cell.
         * @property colIcon
         * @type {Image}
         * @default null
         */
        this.colIcon = null;

        /**
         * A number of operations can cause the cells text and background colors to change. Resizing, mouse interactions,
         * scrolling, hovering, etc. In order to make sure we're not calling the painting and color calculating routines
         * unnecessarily, all these operations simply mark the background as dirty. The is done by calling the invalidateBackground()
         * method. This method will set the backgroundDirty flag to true and call invalidateDisplayList. In the next validation cycle
         * the cell get's its color calculation code executed and painting code will then use the calculated color values to paint
         * the cell.
         * @type {Boolean}
         * @property backgroundDirty
         * @default true
         */
        this.backgroundDirty = true;

        /**
         * @private
         */
        this._iconTimer = null;

        /**
         * In 2.7, ability was added to have any cell be a expand collapse cell. If you set enableExpandCollapse=true on the column
         * the cells that belong to that column are provided with an iExpandCollapseComponent, which defaults to an instance of
         * @see com.flexicious.nestedtreedatagrid.cells.ExpandCollapseIcon class. This property will return the current instance of this
         * class associated with this cell.
         * @type { ExpandCollapseIcon}
         * @property expandCollapseIcon
         * @default null
         */
        this.expandCollapseIcon = null;

        /**
         * @type {Boolean}
         * @property showInAutomationHierarchy
         * @default false
         */
        this.showInAutomationHierarchy = false;

        /**
         * @type {Number}
         * @property iconLeft
         * @default -1
         */
        this.iconLeft = -1;

        /**
         * @type {Number}
         * @property iconTop
         * @default -1
         */
        this.iconTop = -1;

        /**
         * @type {Number}
         * @property iconRight
         * @default -1
         */
        this.iconRight = -1;

        /**
         * @type {Number}
         * @property iconBottom
         * @default -1
         */
        this.iconBottom = -1;

        this.attachClass("flexDataGridCell");
    }

    getClassNames() {
        return ["FlexDataGridCell", "UIComponent", "IFlexDataGridCell"];
    }

    /**
     * destroy function.
     * @method destroy
     */
    destroy() {

        this.level = null;
        this.rowInfo = null;
        this.currentBackgroundColors = null;
        this.currentTextColors = null;
        this.destroyed = true;

    }

    /**
     * To set the actualsize value using width and height values.
     * @param w
     * @param h
     * @method setActualSize
     */
    setActualSize(w, h) {

        if (w != this.getWidth() || h != this.getHeight())
            this.invalidateBackground();
        super.setActualSize(w, h);
        if (this.getRenderer()) {
            this.placeComponent(this.getRenderer(), w, h);
        }

    }
    // shouldComponentUpdateCustom(){
    //     return (this._renderer && this._renderer.$$typeof ? true : false);
    // }
    /**
     * Used to refresh the contents of a Cell.
     * @method refreshCell
     */

    refreshCell() {

        if (this._renderer && this._renderer.$$typeof) {
            if(this.level.grid.enableHorizontalScrollOptimizationsBETA){
                var cell=this;
                if (cell.getIsLocked() || cell.level.grid.getBodyContainer().isInVisibleHorizontalRange(cell.getX(), cell.getWidth())) {
                    this._renderer = this.getRendererFactory().newInstance({ cell: this, row: this.rowInfo, column: this._column, level: this.level, grid: this.level.grid, key: 0 }, true);
                    this.children = [this._renderer];
                    this.setState({ timeStamp: new Date() });
                    //this.parent.invalidateState();
                }
            } 
            else
             {
                this._renderer = null;
                this.children = [];
                this.getRenderer();//this will make me a new one!
            }
        }


        if (this.rowInfo) {
            const automationName = `${this.level.grid.automationName || this.level.grid.id}_cell_${this.rowInfo.rowPositionInfo.getRowIndex() || "0"}_${this.getColumn() ? this.getColumn().getUniqueIdentifier() : this.rowInfo.cells.indexOf(this)}_${this.rowInfo.rowPositionInfo.getRowType()}`;
            this.setAutomationName(automationName);
            if (this.getRenderer() && this.getRenderer().domElement) {
                this.getRenderer().domElement.id = `${automationName}_renderer`;
            }
        }




        this.destroyed = false;
        this.currentBackgroundColors = null;
        this.currentTextColors = null;
        const mine = this.getRenderer();

        if (this.level.grid.dispatchRendererInitialized) {
            this.level.grid.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent
                (flexiciousNmsp.FlexDataGridEvent.RENDERER_INITIALIZED, this.level.grid, this.level, this.getColumn(), this));
        }
        if (this.getColumn() && this.getColumn().implementsOrExtends("FlexDataGridCheckBoxColumn") && (this.rowInfo.getIsDataRow() || this.rowInfo.getIsHeaderRow())) {
            this.initializeCheckBoxRenderer(this.getRenderer());
        }
        this._newlyCreated = false;
        if (this.getColumn() && this.getColumn().enableExpandCollapseIcon) {
            if (!this.expandCollapseIcon) {
                this.expandCollapseIcon = new flexiciousNmsp.ExpandCollapseIcon();
                this.addChild(this.expandCollapseIcon);
            }

            flexiciousNmsp.ExpandCollapseIcon.refreshCell(this.expandCollapseIcon);
            flexiciousNmsp.ExpandCollapseIcon.drawIcon(this.expandCollapseIcon);
        }
        this.placeComponent(this.getRenderer(), this.getWidth(), this.getHeight(), undefined);
        if (this.colIcon && this.implementsOrExtends('IFlexDataGridDataCell')) {
            this.colIcon.setVisible(false);
        }
        if (this.getColumn() && this.getColumn().enableIcon) {
            this.getRenderer().setVisible(!this.getColumn().hideText);
            const iconUrl = this.getIconUrl();
            if (iconUrl) {
                if (!this.colIcon) {
                    this.createColumnIcon();
                }
                this.colIcon.setToolTip(this.getColumn().iconToolTipFunction(this));
                this.colIcon.setVisible(true);
                if (this.colIcon.getSource() != iconUrl) {
                    this.colIcon.setSource(iconUrl);
                    //this.colIcon.load();
                }

                if (this.getColumn().showIconOnCellHover || this.getColumn().showIconOnRowHover)
                    this.colIcon.setVisible(false);
            } else if (this.colIcon) {
                this.colIcon.setSource(iconUrl);//there is an icon, and this cell is being recycled so we no longer should be showing an icon (iconUrl comes back undefined), then we should wipe out the icon
            }
        }


    }
    updateDisplayList(w, h) {
        if (!this.level) { return }
        this.drawCell(this.getWidth(), this.getHeight());
        if (this.level.grid.getIsRowSelectionMode() && this.level.isItemSelected(this.rowInfo.getData())) {
            this.attachClass('selectedCell');
        }
        else if (this.getColumn() && this.level.isCellSelected(this.rowInfo.getData(), this.getColumn())) {
            this.attachClass('selectedCell');
        } else {
            this.detachClass('selectedCell');
        }
        this.placeIcon();
        if (this.level.grid.dispatchCellRenderered)
            this.level.grid.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.CELL_RENDERED, this.level.grid, this.level, this.getColumn(), this, this.rowInfo.getData()))
        super.updateDisplayList(w, h);
    }
    /**
     *
     * @method createColumnIcon
     */
    createColumnIcon() {
        this.colIcon = new flexiciousNmsp.Image();
        this.colIcon.attachClass("cellIcon");
        this.addChild(this.colIcon);
        this.colIcon.addEventListener(this, Constants.EVENT_CLICK, this.onIconMouseClick);
        this.colIcon.addEventListener(this, Constants.EVENT_MOUSE_OVER, this.onIconMouseOver);
        this.colIcon.addEventListener(this, Constants.EVENT_MOUSE_OUT, this.onIconMouseOut);
        this.colIcon.setHandCursor(this.getColumn().iconHandCursor);
    }

    /**
     * when the icon is enables, handles the mouse over event of the icon.
     * @param evnt
     * @method createColumnIcon
     */
    onIconMouseOver(evnt) {

        if (!this.getColumn().useIconRollOverTimer) {
            this.onTimerComplete(null);
            return;
        }


        if (!this._iconTimer) {
            this._iconTimer = new flexiciousNmsp.Timer(this.getColumn().iconMouseOverDelay, 1);
            this._iconTimer.addEventListener(this, Constants.EVENT_TIMER, this.onTimerComplete);
        }
        this._iconTimer.start();
        const iconUrl = this.getIconUrl(true);
        if (this.colIcon.getSource() != iconUrl) {
            this.colIcon.setSource(iconUrl);
        }

    }

    /**
     * when the icon is enables, handles the mouse click event of the icon.
     * @param evt
     * @method createColumnIcon
     */
    onIconMouseClick(evt) {

        this.level.grid.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.ICON_CLICK,
            this.level.grid, this.level, this.getColumn(), this, this.rowInfo.getData(), evt));
        evt.stopPropagation();

    }

    /**
     * Once the timer completed its time, the final result will be renderered such as showing shooltips or dispatching
     * some events.
     * @param evt
     * @method createColumnIcon
     */
    onTimerComplete(evt) {

        this.level.grid.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.ICON_MOUSE_OVER,
            this.level.grid, this.level, this.getColumn(), this, this.rowInfo.getData(), evt));

        if (this.getColumn().iconTooltipRenderer != null) {
            const tooltipRenderer = this.getColumn().iconTooltipRenderer.newInstance({ cell: this, row: this.rowInfo, column: this._column, level: this.level, grid: this.level.grid, key: this.level.grid.childIndex++ }, true);
            this.level.grid.showTooltip(this.colIcon, tooltipRenderer, this.rowInfo.getData())
        }

    }

    /**
     * when the icon is enables, handles the mouse out event of the icon.
     * @param evnt
     * @method onIconMouseOut
     */
    onIconMouseOut(evnt) {

        if (!this.getColumn().useIconRollOverTimer) {
            this.level.grid.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.ICON_MOUSE_OUT, this.level.grid, this.level, this.getColumn(), this, this.rowInfo.getData()));
            return;
        }

        if (this._iconTimer.running) {
            this._iconTimer.stop();
        } else if (this._iconTimer) {
            this.level.grid.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.ICON_MOUSE_OUT, this.level.grid, this.level, this.getColumn(), this, this.rowInfo.getData()));
        }
        const iconUrl = this.getIconUrl();
        if (this.colIcon.getSource() != iconUrl) {
            this.colIcon.setSource(iconUrl);
            this.colIcon.load();
        }

    }

    /**
     *
     * @param over
     * @return {String}
     * @method getIconUrl
     */
    getIconUrl(over) {
        if (typeof over == "undefined") over = false;

        return this.getColumn().iconFunction(this, !this.getEnabled() ? "disabled" : over ? "over" : "");

    }

    /**
     * Given a renderer, sets the size on basis of whether or not there are vertical and horizotnal gridlines .
     * If there are vertical gridlines, sets the width to width minus verticalGridLineThickness
     * If there are vertical gridlines, sets the height to width minus horizontalGridLineThickness
     * @param cellRenderer
     * @param w
     * @param h
     * @method setRendererSize
     */
    setRendererSize(cellRenderer, w, h) {
        if (cellRenderer.typeName) {
            CellUtils.setRendererSize(cellRenderer, w, h);
            cellRenderer.setStyleAttribute("maxWidth", `${w}px`);
        }
    }

    /**
     * The FlexDataGridColumnLevel associated with this cell. This set on initialize. This is usually never null.
     * @return {FlexDataGridColumnLevel}
     * @method getLevel
     */
    getLevel() {
        return this.level;
    }

    /**
     *
     * @return {Number}
     * @method getLeftPadding
     */
    getLeftPadding() {

        const styleOwner = this.getColumn() ? this.getColumn() : this.level;
        return styleOwner.getStyleValue(this.getPrefix() + this.capitalizeFirstLetterIfPrefix("paddingLeft"));

    }

    /**
     *  setting the correct position of the Icon.
     * @method placeIcon
     */
    placeIcon() {

        if (!this.colIcon) return;
        const colIconHeight = this.getColumn().getStyle("iconWidth") || (this.colIcon.domElement ? this.colIcon.domElement.offsetHeight : 0) || 11;
        const colIconWidth = this.getColumn().getStyle("iconHeight") || (this.colIcon.domElement ? this.colIcon.domElement.offsetWidth : 0) || 11;

        if (!this.getColumn() || !this.getEnabled()) {
            this.colIcon.setVisible(false);
            return;
        }
        if (this.getColumn().showIconOnCellHover) {
            this.colIcon.setVisible(this.level.grid.currentCell == this)
        } else if (this.getColumn().showIconOnRowHover) {
            if (this.level.grid.currentCell == null)
                this.colIcon.setVisible(false);
            else
                this.colIcon.setVisible(this.level.grid.currentCell.rowInfo == this.rowInfo)
        }
        const left = this.iconLeft != -1 ? this.iconLeft : this.getColumn().getStyle("iconLeft");
        const top = this.iconTop != -1 ? this.iconTop : this.getColumn().getStyle("iconTop");
        const bottom = this.iconBottom != -1 ? this.iconBottom : this.getColumn().getStyle("iconBottom");
        const right = this.iconRight != -1 ? this.iconRight : this.getColumn().getStyle("iconRight");
        let iconX;
        let iconY;
        if (left > 0)
            iconX = left;
        if (top > 0)
            iconY = top;
        if (right > 0)
            iconX = this.getWidth() - right - colIconWidth;
        if (bottom > 0)
            iconY = this.getHeight() - bottom - colIconHeight;

        if (isNaN(iconX)) iconX = this.getWidth() - 2 - colIconWidth;
        if (isNaN(iconY)) iconY = Math.max(0, (this.getHeight() - colIconHeight) / 2);

        this.colIcon.move(iconX, iconY);

        if (this.getColumn().iconPlacementFunction != null) {
            this.getColumn().iconPlacementFunction(this, this.colIcon);
        }
    }

    /**
     * setting the position of ExpandCollapseIcon.
     * @method placeExpandCollapseIcon
     */

    placeExpandCollapseIcon() {

        if (!this.expandCollapseIcon) return;
        if (!this.getColumn() || !this.getColumn().enableExpandCollapseIcon) {
            this.expandCollapseIcon.setVisible(false);
            return;
        }
        this.expandCollapseIcon.setVisible(true);
        const left = this.getColumn().getStyle("expandCollapseIconLeft");
        const top = this.getColumn().getStyle("expandCollapseIconTop");
        const bottom = this.getColumn().getStyle("expandCollapseIconBottom");
        const right = this.getColumn().getStyle("expandCollapseIconRight");
        let iconX;
        let iconY;
        if (left > 0)
            iconX = left;
        if (top > 0)
            iconY = top;
        if (right > 0)
            iconX = this.getWidth() - right - this.expandCollapseIcon.getWidth();
        if (bottom > 0)
            iconY = this.getHeight() - bottom - this.expandCollapseIcon.getHeight();

        if (isNaN(iconX)) iconX = 2;
        if (isNaN(iconY)) iconY = Math.max(0, (this.getHeight() - (20)) / 2);
        if (this.getColumn().enableHierarchicalNestIndent)
            iconX += this.level.getMaxPaddingCellWidth();
        this.expandCollapseIcon.move(iconX, iconY);

    }

    /**
     * Draws the background as well as calls drawBorder on the given cell
     *
     * Each cell in Ultimate will draw its own background
     * and border. This method will use the evaluated results of a number of properties on the cell
     * object to calculate the color, width, and visibility of the background associated with this cell.
     * <ul>
     * <li>First, this method will check to see if you have defined a cellCustomBackgroundDrawFunction for the
     * associated column or level. If so, it will call this method, with the cell as a parameter. You can draw
     * a custom background in this function. If you return false, from this function, this method will not
     * execute any other code, and simply call the drawBorders method</li>
     * <li>First, this method will check to see if you have defined a cellCustomBackgroundDrawFunction for the
     * associated column or level. If so, it will call this method, with the cell as a parameter. You can draw
     * a custom background in this function. If you return false, from this function, this method will not
     * execute any other code, and simply call the drawBorders method</li>
     * </ul>
     * @param unscaledWidth
     * @param unscaledHeight
     * @method drawCell
     */
    drawCell(unscaledWidth, unscaledHeight) {

        if (!parent) return;
        this.drawBackground(unscaledWidth, unscaledHeight);

    }

    /**
     * This method will size and place the renderer within the bounds of this cell.
     * The renderer of the cell gets a width of cell.width - paddingLeft - paddingRight.
     * It gets a height of cell.height - paddingTop - paddingBottom.
     * It then gets placed at x position of paddingLeft and y position of paddingTop.
     * The padding is only applied if usePadding is set to true (default).
     *
     * The padding is calculated on basis of what type of cell this is. The following style properties are
     * used to figure out padding:
     * <ul>
     * <li>paddingLeft</li>
     * <li>paddingRight</li>
     * <li>paddingTop</li>
     * <li>paddingBottom</li>
     * <li>headerPaddingLeft</li>
     * <li>headerPaddingRight</li>
     * <li>headerPaddingTop</li>
     * <li>headerPaddingBottom</li>
     * <li>footerPaddingLeft</li>
     * <li>footerPaddingRight</li>
     * <li>footerPaddingTop</li>
     * <li>filterPaddingBottom</li>
     * <li>filterPaddingLeft</li>
     * <li>filterPaddingRight</li>
     * <li>filterPaddingTop</li>
     * <li>filterPaddingBottom</li>
     * <li>columnGroupPaddingBottom</li>
     * <li>columnGroupPaddingLeft</li>
     * <li>columnGroupPaddingRight</li>
     * <li>columnGroupPaddingTop</li>
     * <li>columnGroupPaddingBottom</li>
     * </ul>
     *
     * This method is
     * also used by the edit code, to place the editor. When placing the editor, usePadding is
     * set to falst, so the editor occupies the entire cell to prevent the text of the cells from
     * showing.
     * @param cellRenderer	The renderer to place. Can be the actual renderer, or the editor.
     * @param unscaledWidth	The width of the cell.
     * @param unscaledHeight	The height of the cell.
     * @param usePadding	Flag to apply padding while figuring out the size of the renderer
     * @return a new Point object with paddingTop and paddingLeft as y and x properties.
     * @method placeComponent
     */
    placeComponent(cellRenderer, unscaledWidth, unscaledHeight, usePadding) {
        if (typeof usePadding == "undefined") usePadding = true;

        if (!parent) return null;

        if (usePadding) {
            const paddingLeft = UIUtils.nanToZero(this.getLeftPadding());
            const paddingTop = UIUtils.nanToZero(this.getStyleValue(this.getPrefix() + this.capitalizeFirstLetterIfPrefix("paddingTop")));
            const paddingRight = UIUtils.nanToZero(this.getStyleValue(this.getPrefix() + this.capitalizeFirstLetterIfPrefix("paddingRight")));
            const paddingBottom = UIUtils.nanToZero(this.getStyleValue(this.getPrefix() + this.capitalizeFirstLetterIfPrefix("paddingBottom")));
            this.setRendererSize(cellRenderer, unscaledWidth - paddingLeft - paddingRight - 1,
                unscaledHeight - paddingTop - paddingBottom - 1);
            this.invalidateDisplayList();
            //var rend=this.getRenderer().domElement;
            //rend.style.paddingRight=paddingRight+"px";
            if(this.getIsDataCell()) {
                this.setStyleAttribute("paddingRight", paddingRight + "px");
                this.setStyleAttribute("paddingBottom", paddingBottom + "px");
            }
            if (this.getRenderer().move) {
                this.getRenderer().move(paddingLeft, paddingTop);
            } else {

            }
            this.placeIcon();
            this.placeExpandCollapseIcon();
            return new flexiciousNmsp.Point(paddingLeft, paddingTop);
        } else {
            this.setRendererSize(cellRenderer, unscaledWidth, unscaledHeight);
            return new flexiciousNmsp.Point(0, 0);
        }

    }

    /**
     * If there is a value for the prefix parameter, capitalizes the first word of the val parameter and returns it
     * @param val
     * @return {String}
     * @method capitalizeFirstLetterIfPrefix
     */
    capitalizeFirstLetterIfPrefix(val) {

        return flexiciousNmsp.CellUtils.capitalizeFirstLetterIfPrefix(this.getPrefix(), val);

    }

    /**
     * Returns true if the style property for verticalGridLines is defined for the type of the row.
     * Based upon the rowInfo.rowPositionInfo.rowType, one of the following properties are inspected:
     * <ul>
     * <li>verticalGridLines</li>
     * <li>headerVerticalGridLines</li>
     * <li>footerVerticalGridLines</li>
     * <li>pagerVerticalGridLines</li>
     * <li>filterVerticalGridLines</li>
     * <li>columnGroupVerticalGridLines</li>
     * </ul>
     * checking the availability of having VerticalGridLiines
     * @return {Boolean}
     * @method hasVerticalGridLines
     */

    hasVerticalGridLines() {

        return this.getStyleValue(this.getPrefix() + this.capitalizeFirstLetterIfPrefix("verticalGridLines"));

    }

    /**
     * Returns true if the style property for horizontalGridLines is defined for the type of the row.
     * Based upon the rowInfo.rowPositionInfo.rowType, one of the following properties are inspected:
     * <ul>
     * <li>horizontalGridLines</li>
     * <li>headerHorizontalGridLines</li>
     * <li>footerHorizontalGridLines</li>
     * <li>pagerHorizontalGridLines</li>
     * <li>filterHorizontalGridLines</li>
     * <li>columnGroupHorizontalGridLines</li>
     * </ul>
     * @return {Boolean}
     * @method hasHorizontalGridLines
     */
    hasHorizontalGridLines() {

        return this.getStyleValue(this.getPrefix() + this.capitalizeFirstLetterIfPrefix("horizontalGridLines"));

    }

    /**
     * Draws the background as well as calls drawBorder on the given cell
     *
     * Each cell in Ultimate will draw its own background
     * and border. This method will use the evaluated results of a number of properties on the cell
     * object to calculate the color, width, and visibility of the background associated with this cell.
     * <ul>
     * <li>First, this method will check to see if you have defined a cellCustomBackgroundDrawFunction for the
     * associated column or level. If so, it will call this method, with the cell as a parameter. You can draw
     * a custom background in this function. If you return false, from this function, this method will not
     * execute any other code, and simply call the drawBorders method</li>
     * <li>First, this method will check to see if you have defined a cellCustomBackgroundDrawFunction for the
     * associated column or level. If so, it will call this method, with the cell as a parameter. You can draw
     * a custom background in this function. If you return false, from this function, this method will not
     * execute any other code, and simply call the drawBorders method</li>
     * </ul>
     * Drawing the background using he width and height properties.
     * @param unscaledWidth
     * @param unscaledHeight
     * @method drawBackground
     */
    drawBackground(unscaledWidth, unscaledHeight) {

        flexiciousNmsp.CellUtils.drawBackground(this);

    }

    /**
     * If this is a fillRow, returns the value of the color property.
     * Else, returns the value of the "prop" property which is the value of one of the following style properties:
     * <ul>
     * <li>rollOverColors</li>
     * <li>headerRollOverColors</li>
     * <li>footerRollOverColors</li>
     * <li>filterRollOverColors</li>
     * <li>pagerRollOverColors</li>
     * <li>columnGroupRollOverColors</li>
     * </ul>
     * Getting the Roll Over Color Of a FlexDataGridCell
     * @return {String}
     * @method getRolloverColor
     */
    getRolloverColor() {

        return flexiciousNmsp.CellUtils.getRolloverColor(this, "rollOverColor");

    }

    /**
     * Returns the value of the textRollOverColor style property.
     * @return {String}
     * @method getRolloverTextColor
     */
    getRolloverTextColor() {

        return flexiciousNmsp.CellUtils.getRolloverColor(this, "textRollOverColor");

    }

    /**
     * The getTextColors method is responsible for evaluating the text color of the cell.
     * <ul>
     * <li>First, it checks to see if the cell is disabled. If so, it will return the value of the textDisabledColor style property</li>
     * <li>Next, it checks to see if there is a value for the currentTextColors property (based upon rollover) and if so returns that. </li>
     * <li>Next, it checks to see if there is a custom cellTextColorFunction on the column and if so, it returns the result of this function</li>
     * <li>Next, it checks to see if there is a custom rowTextColorFunction on the level and if so, it returns the result of this function</li>
     * <li>Next, it checks to see if there enableEditRowHighlight is true amd this row is in edit mode, then it returns the value of the editTextColor style property.</li>
     * <li>Next, it checks to see if the row is selected (row selection mode) or cell is selected (cell selection mode) and if so, returns textSelectedColor property.</li>
     * <li>Next, it checks to see if a columnTextColor property is specified on the associated column and if so, returns the value of this property.</li>
     * <li>Finally, it uses the alternatingTextColors array and return its 0th or 1st index based upton the rowInfo.rowPositionInfo.rowIndex.</li>
     * </ul>
     * Getting the TextColors of a FlexDataGridCell
     * @return {String}
     * @method getTextColors
     */
    getTextColors() {

        const fromGrid = flexiciousNmsp.CellUtils.getTextColorFromGrid(this);
        if (fromGrid != null) return fromGrid;
        return flexiciousNmsp.CellUtils.getTextColors(this);

    }

    /**
     * The getBackgroundColors method is responsible for evaluating the background color of the cell.
     * <ul>
     * <li>First, it checks to see if this is a filler row. If so, it will use the alternatingItemColors array and return its 0th or 1st index based upton the rowInfo.rowPositionInfo.rowIndex. </li>
     * <li>Next, it checks to see if the cell is disabled. If so, it will return the value of the selectionDisabledColor style property</li>
     * <li>Next, it checks to see if there is a custom cellBackgroundColorFunction on the column and if so, it returns the result of this function</li>
     * <li>Next, it checks to see if there is a custom rowBackgroundColorFunction on the level and if so, it returns the result of this function</li>
     * <li>Next, it checks to see if there rowInfo's data object is in the errors collection of the grid, and if so, it return the value of the errorBackgroundColor property.</li>
     * <li>Next, it checks to see if there enableEditRowHighlight is true amd this row is in edit mode, then it returns the value of the editItemColors style property.</li>
     * <li>Next, it checks to see if there is a value for the currentBackgroundColors property (based upon rollover) and if so returns that.
     * The currentBackgroundColors property is a result of the grid.highlightRow method.
     * <ul>
     * <li>If enableActiveCellHighlight=true uses the value of the activeCellColor style.</li>
     * <li>Else, uses the value of the rollOverColor style property.</li>
     * </ul>
     * </li>
     * <li>Next, it checks to see if the row is selected (row selection mode) or cell is selected (cell selection mode) and if so, returns selectionColor property.</li>
     * <li>Next, it checks to see if a backgroundColor property is specified on the associated column and if so, returns the value of this property.</li>
     * <li>Finally, it uses the alternatingItemColors array and return its 0th or 1st index based upton the rowInfo.rowPositionInfo.rowIndex.</li>
     * </ul>
     * Getting the BackgroundColors of a FlexDataGridCell
     * @return {String}
     * @method getBackgroundColors
     */
    getBackgroundColors() {

        const fromGrid = flexiciousNmsp.CellUtils.getBackgroundColorFromGrid(this);
        if (fromGrid != null) return fromGrid;
        return flexiciousNmsp.CellUtils.getBackgroundColors(this);

    }

    /**
     * Draws the right border for this cell.
     * There are a few rules on basis of which a right (or a left) border is drawn on a cell.
     * <ul>
     * <li>First, if the cell is a Pager cell, no right border is drawn, since the pager shares its left and right border with the grid.</li>
     * <li>Second, if the cell has no vertical gridlines, no border is drawn.</li>
     * <li>Else, if the cell has a column and its right locked, then a left border is drawn.</li>
     * </ul>
     * @param unscaledWidth
     * @param unscaledHeight
     * @return {*}
     * @method drawRightBorder
     */
    drawRightBorder(unscaledWidth, unscaledHeight) {

        return flexiciousNmsp.CellUtils.drawRightBorder(this);

    }

    /**
     * This method is responsible for initializing the checkbox renderer for both the header and data cells.
     * It performns the following tasks
     * <ul>
     * <li>First, it will set the radio button mode of the associated TriStateCheckBox to true if col.radioButtonMode or level.enableSingleSelect is set to true </li>
     * <li>Next, if the grid.enableSelectionExclusion flag is set to true, it will set the selected state of the associated TriStateCheckBox to the result of the getCheckBoxStateBasedOnExclusion method of the level</li>
     * <li>Next, if the grid.enableSelectionExclusion flag is set to false (default),
     * it will set the selected state of the associated TriStateCheckBox based upon the following logic (if enableTriStateCheckBox=true):
     * <ul>
     * 	<li>If the associated rowData object is in the selected objects collection of the level, sets it to checked</li>
     * 	<li>If any of the children the associated rowData object are selected, sets it to middle</li>
     * 	<li>Else sets it to unchecked</li>
     * </ul>
     * </li>
     * <li>Finally, if the enableLabelAndCheckBox flag is set to true on the column, this method will apply the result of the columns itemToLabel function
     * to the label property of the checkbox</li>
     * </ul>
     * @param renderer
     * @type DisplayObject
     * @method initializeCheckBoxRenderer
     */
    initializeCheckBoxRenderer(renderer) {

        flexiciousNmsp.CellUtils.initializeCheckBoxRenderer(this, this.level);

    }

    /**
     */
    onCheckChange(event) {


    }

    /**
     *
     */
    getStyleValue(styleProp) {

        return flexiciousNmsp.CellUtils.getStyleValue(this, styleProp)

    }

    /**
     * Marks the cell so that the background and borders are drawn in the next validation cycle. Please, note this does not
     * update or re-evaluate the content of the cell, just the border and the background. To updated the content, call the
     * refreshCell method instead.
     * @method invalidateBackground
     */
    invalidateBackground() {

        this.backgroundDirty = true;
        this.invalidateDisplayList();

    }

    /**
     * The X Position of this cell, as the user sees it. Since the Left, Right and Unlocked containers are measuremment
     * frames in themselves, each cell's X is relative to its container. This method converts the measurement of the x value
     * relative to the corresponding parent container to the measurement of X relative to the grid.
     * This is used mainly by keyboard navigation to figure out which cell is to the right or left of the given cell.
     * @method getPerceivedX
     * @return {Number}
     */
    getPerceivedX() {
        const col = this.getColumn() && this.getColumn().implementsOrExtends("FlexDataGridColumn") ? this.getColumn() : null;
        if (!col) return this.getX();
        return col.getIsLeftLocked() ? this.getX() : this.getIsRightLocked() ? col.level.grid.getLeftLockedContent().getWidth() + col.level.getUnLockedWidth() + this.getX() :
            col.level.grid.getLeftLockedContent().getWidth() + this.getX();
    }

    /**
     * What thickness to use to draw the horizontal grid lines for this cell. The value is used to draw the bottom border.
     * drawTopBorder
     * Similar to the drawTopBorder, this is a calculated property which is derived from the associated rowInfo object's row type.
     * The result is the value of one of the following based on the rowInfo.rowPositionInfo.rowType:
     * <ul>
     * <li>horizontalGridLineThickness</li>
     * <li>headerHorizontalGridLineThickness</li>
     * <li>footerHorizontalGridLineThickness</li>
     * <li>pagerHorizontalGridLineThickness</li>
     * <li>filterHorizontalGridLineThickness</li>
     * <li>columnGroupHorizontalGridLineThickness</li>
     * </ul>
     * @return {Number}
     * @method getHorizontalGridLineThickness
     */
    getHorizontalGridLineThickness() {
        return this.getStyleValue(this.getPrefix() + this.capitalizeFirstLetterIfPrefix("horizontalGridLineThickness"));
    }

    /**
     *
     * @param val
     * @method setSelectable
     */
    setSelectable(val) {
        const isIe = UIUtils.isIE();
        if (!this.getRenderer().typeName) {
            return;
        }
        if (val) {
            this.getRenderer().detachClass("unSelectableText");
            this.detachClass("unSelectableText");
            if (isIe)
                this.getRenderer().removeComponentAttribute("unselectable");
        }
        else {
            this.getRenderer().attachClass("unSelectableText");
            this.attachClass("unSelectableText");
            if (isIe)
                this.getRenderer().setAttribute("unselectable", "on");
        }
    }


    /**
     * When a cell is created for the first time, this flag is set to true. After the cell is no longer needed
     * the cell is put back into the renderer cache. If after being put in the renderer cache, the cell gets
     * called back into action and is added to the display list, this flag will be set to false.
     * @return {Boolean}
     * @method getIsNewlyCreated
     */
    getIsNewlyCreated() {
        return this._newlyCreated;
    }

    /**
     * Calcualted property that returns if this cell is a expand collapse cell.
     * @return {Boolean}
     * @method getIsExpandCollapseCell
     */
    getIsExpandCollapseCell() {
        return this.getColumn() && this.getColumn().enableExpandCollapseIcon;
    }

    /**
     * Returns true if the column associated with this cell has columnLockMode="left" or columnLockMode="right"
     * @return {Boolean}
     * @method getIsLocked
     */
    getIsLocked() {
        return (this.getColumn() && this.getColumn().getIsLocked());
    }

    /**
     * Setting the width.
     * @param value
     * @method setWidth
     */
    setWidth(value) {

        const toSet = this.getWidth() != value;
        flexiciousNmsp.UIComponent.prototype.setWidth.apply(this, [value]);
        if (toSet) {
            this.invalidateBackground();
            this.placeComponent(this.getRenderer(), this.getWidth(), this.getHeight(), true);
        }
    }

    /**
     * Calculated property that controls whether or not to draw the top border. Each cell in Ultimate will draw its own background
     * and border. In order to prevent cells from drawing double borders (cell above me drew bottom border, so I should not need
     * to draw my top border), this flag evaluates to false. But in certain cases, you may want to draw both top and bottom border.
     * The bottom border is always controlled by the horizontalGridLines style property. The top border is controlled by the drawTopBorder
     * style property. In order to provide fine grained control over border drawing functionality, there are the following style
     * properties available.
     * <ul>
     * <li>drawTopBorder</li>
     * <li>headerDrawTopBorder</li>
     * <li>footerDrawTopBorder</li>
     * <li>pagerDrawTopBorder</li>
     * <li>filterDrawTopBorder</li>
     * <li>columnGroupDrawTopBorder</li>
     * </ul>
     * To control the bottom border, the following properties are available:
     * <ul>
     * <li>horizontalGridLines</li>
     * <li>headerHorizontalGridLines</li>
     * <li>footerHorizontalGridLines</li>
     * <li>pagerHorizontalGridLines</li>
     * <li>filterHorizontalGridLines</li>
     * <li>columnGroupHorizontalGridLines</li>
     * </ul>
     * @return {Boolean}
     * @method getDrawTopBorder
     */
    getDrawTopBorder() {
        return false;
    }

    /**
     *
     * @param value
     * @method setHeight
     */
    setHeight(value) {
        //        if(this.hasHorizontalGridLines()){
        //            value = value-this.getHorizontalGridLineThickness();
        //        }
        if (value != this.getHeight()) {
            this.invalidateBackground();
        }
        flexiciousNmsp.UIComponent.prototype.setHeight.apply(this, [value]);
    }

    /**
     * The factory class used for initializing the cell. This is arrived at by calling the
     * deriveRenderer method of the associated column.
     * @see com.flexicious.nestedtreedatagrid.FlexDataGridColumn#deriveRenderer
     * @return {IFactory}
     * @method getRendererFactory
     */
    getRendererFactory() {
        if (!this._rendererFactory) {
            this._rendererFactory = UIUtils.UIComponentFactory;
        }
        return this._rendererFactory;
    }

    /**
     * The factory class used for initializing the cell. This is arrived at by calling the
     * deriveRenderer method of the associated column.
     * @see com.flexicious.nestedtreedatagrid.FlexDataGridColumn#deriveRenderer
     * @param value
     * @method setRendererFactory
     */
    setRendererFactory(value) {
        this._rendererFactory = value;
    }

    /**
     * Returns one of the following values on basis of what type of cell this is:
     * <ul>
     * <li>Blank string for data cells</li>
     * <li>header</li>
     * <li>footer</li>
     * <li>filter</li>
     * <li>columnGroup</li>
     * <li>pager</li>
     * </ul>
     * @method getPrefix
     */
    getPrefix() {
        return "";
    }

    /**
     * This flag is a used for enabling the wordWrap on the associated renderer. Please note, if you define an item renderer,
     * setting wordWrap will only have an impact if you define a renderer that exposes a wordWrap property. When you specify
     * wordWrap on the associated column, the grid will set this flag to true. It will also use a predefined item renderer
     * that has a wordWrap property, unless you specify a custom renderer. If you specify a custom renderer, then this renderer
     * must assume responsibility for exposing a wordWrap property and handling wordWrap appropriately.
     * @return {Boolean}
     * @method getWordWrap
     */
    getWordWrap() {
        return this.getRenderer().getComponentStyleAttribute("whiteSpace") == "normal";
    }

    /**
     * This flag is a used for enabling the wordWrap on the associated renderer. Please note, if you define an item renderer,
     * setting wordWrap will only have an impact if you define a renderer that exposes a wordWrap property. When you specify
     * wordWrap on the associated column, the grid will set this flag to true. It will also use a predefined item renderer
     * that has a wordWrap property, unless you specify a custom renderer. If you specify a custom renderer, then this renderer
     * must assume responsibility for exposing a wordWrap property and handling wordWrap appropriately.
     * @param val
     * @method setWordWrap
     */
    setWordWrap(val) {
        if (!this.getRenderer().typeName) {
            return;
        }
        this.getRenderer().setStyleAttribute("whiteSpace", val ? "normal" : "nowrap");
    }

    /**
     * What thickness to use to draw the horizontal grid lines for this cell. The value is used to draw the bottom border.
     * drawTopBorder
     * Similar to the drawTopBorder, this is a calculated property which is derived from the associated rowInfo object's row type.
     * The result is the value of one of the following based on the rowInfo.rowPositionInfo.rowType:
     * <ul>
     * <li>verticalGridLineThickness</li>
     * <li>headerVerticalGridLineThickness</li>
     * <li>footerVerticalGridLineThickness</li>
     * <li>pagerVerticalGridLineThickness</li>
     * <li>filterVerticalGridLineThickness</li>
     * <li>columnGroupVerticalGridLineThickness</li>
     * </ul>
     * @return {Number}
     * @method getVerticalGridLineThickness
     */
    getVerticalGridLineThickness() {
        return this.getStyleValue(this.getPrefix() + this.capitalizeFirstLetterIfPrefix("verticalGridLineThickness"));
    }

    /**
     * What color to use to draw the vertical grid lines for this cell. The value is used to draw the right border.
     * drawTopBorder
     * Similar to the drawTopBorder, this is a calculated property which is derived from the associated rowInfo object's row type.
     * The result is the value of one of the following based on the rowInfo.rowPositionInfo.rowType:
     * <ul>
     * <li>verticalGridLineColor</li>
     * <li>headerVerticalGridLineColor</li>
     * <li>footerVerticalGridLineColor</li>
     * <li>pagerVerticalGridLineColor</li>
     * <li>filterVerticalGridLineColor</li>
     * <li>columnGroupVerticalGridLineColor</li>
     * </ul>
     * @return {String}
     * @method getVerticalGridLineColor
     */
    getVerticalGridLineColor() {
        return this.getStyleValue(this.getPrefix() + this.capitalizeFirstLetterIfPrefix("verticalGridLineColor"));
    }

    /**
     *
     * @return { IExpandCollapseComponent}
     * @method getIExpandCollapseComponent
     */
    getIExpandCollapseComponent() {
        return this.expandCollapseIcon;
    }

    /**
     * This is the actual text to display in the cell. This is just a placeholder for the text. Ultimately this value
     * is applied to the renderers text property. This is why, if you specify a custom renderer that exposes a text
     * property, the grid automatically applies the result of the column.itemToLabel method on the associated rowInfo.data
     * object to the text property of the itemRenderer.
     * @return {String}
     * @method getText
     */
    getText() {
        const mine = this.getRenderer();
        if (mine.hasOwnProperty("text")) {
            return mine["text"];
        }
        return this._text;
    }

    /**
     * This is the actual text to display in the cell. This is just a placeholder for the text. Ultimately this value
     * is applied to the renderers text property. This is why, if you specify a custom renderer that exposes a text
     * property, the grid automatically applies the result of the column.itemToLabel method on the associated rowInfo.data
     * object to the text property of the itemRenderer.
     * @param val
     * @method setText
     */
    setText(val) {
        this._text = val;
        if (!this.getRenderer().typeName) {
            return;
        }
        this.getRenderer().setText(val);
    }

    /**
     * The FlexDataGridColumn associated with this cell. Can be null for cells that do not have a column (e.g. ExpandCollapse Cells).
     * For Cells that span multiple columns, it is still the column that this cell starts on.
     * @return { FlexDataGridColumn}
     * @method getColumn
     */
    getColumn() {
        return this._column;
    }

    /**
     * The rowInfo object associated with this cell. This cell should be within the cells collection of this rowInfo object.
     * @method getRowInfo
     * @return { RowInfo}
     */
    getRowInfo() {
        return this.rowInfo;
    }

    /**
     * The FlexDataGridColumnLevel associated with this cell. This set on initialize. This is usually never null.
     * @return { FlexDataGridColumnLevel}
     * @method getLevel
     */
    getLevel() {
        return this.level;
    }

    /**
     * The FlexDataGrid associated with this cell. This set on initialize. This is usually never null.
     * @return { FlexDataGrid}
     * @method getLevel
     */
    getGrid() {
        return this.getLevel().grid;
    }

    /**
     * The FlexDataGridColumnLevel associated with this cell. This set on initialize. This is usually never null.
     * @param value
     * @method setColumn
     */
    setColumn(value) {

        this._column = value;
        this.dispatchEvent(new flexiciousNmsp.BaseEvent("columnChanged"));
    }

    /**
     *
     */
    setTruncateToFit(val) {
        if (this.getRenderer().hasOwnProperty("truncateToFit")) {
            this.getRenderer()["truncateToFit"] = val;
        }
    }

    /**
     * If the cell.backgroundColors is an array, returns a string join using a comma.
     * Else, returns the result of ExtendedUIUtils.getColorName on it.
     * @return {String}
     * @method getBackgroundColorString
     */
    getBackgroundColorString() {
        return flexiciousNmsp.CellUtils.getBackGroundColorString(this);
    }

    /**
     * Returns true if this cell is not right locked or not fixed width.
     * @return {Boolean}
     * @method getIsElastic
     */
    getIsElastic() {
        return (!this.getIsRightLocked() && !(this.implementsOrExtends("IFixedWidth")) && !(this.getColumn().implementsOrExtends("IFixedWidth")));
    }

    /**
     * The actual renderer component that displays the value of the cell. For all but the FlexDataGridDataCell2 and FlexDataGridDataCell3
     * there is a separate component that does the rendering. For FlexDataGridDataCell2 and FlexDataGridDataCell3, the cell itself becomes
     * the renderer. This is why both FlexDataGridDataCell2 and FlexDataGridDataCell3 are more performant and optimized than the others.
     * These are also more numerous, since they display data values. The header, filters and footers are relatively smaller in number.
     * @return {DisplayObject}
     * @method getRenderer
     */
    getRenderer() {
        if (!this._renderer) {
            this._renderer = this.getRendererFactory().newInstance({ cell: this, row: this.rowInfo, column: this._column, level: this.level, grid: this.level.grid, key: this.children.length }, true);
            if (this.getColumn() && this.getColumn().implementsOrExtends("FlexDataGridCheckBoxColumn") && (this.rowInfo.getIsDataRow() || this.rowInfo.getIsHeaderRow())) {
                this._renderer.addEventListener(this, Constants.EVENT_CLICK, this.onCheckChange);
            }
            if (this._renderer.attachClass) {
                this._renderer.attachClass("cellRenderer");
            }

            if (this.level && this.getRenderer().parent != this && this.getRenderer() != this)
                this.addChild(this.getRenderer());
            if (this.getColumn() && this.getColumn().implementsOrExtends("FlexDataGridCheckBoxColumn") && (this.rowInfo.getIsDataRow() || this.rowInfo.getIsHeaderRow())
                && !this.getRenderer().hasEventListener(Constants.EVENT_CHANGE)) {
                this.getRenderer().addEventListener(this, Constants.EVENT_CLICK, this.onCheckChange);
            }
        }
        return this._renderer;
    }

    /**
     * What color to use to draw the horizontal grid lines for this cell. The value is used to draw the bottom border.
     * Similar to the drawTopBorder, this is a calculated property which is derived from the associated rowInfo object's row type.
     * The result is the value of one of the following based on the rowInfo.rowPositionInfo.rowType:
     * <ul>
     * <li>horizontalGridLineColor</li>
     * <li>headerHorizontalGridLineColor</li>
     * <li>footerHorizontalGridLineColor</li>
     * <li>pagerHorizontalGridLineColor</li>
     * <li>filterHorizontalGridLineColor</li>
     * <li>columnHorizontalGridLineColor</li>
     * </ul>
     * @return {String}
     * @method getHorizontalGridLineColor
     */
    getHorizontalGridLineColor() {
        return this.getStyleValue(this.getPrefix() + this.capitalizeFirstLetterIfPrefix("horizontalGridLineColor"));
    }

    /**
     * Returns true if the column associated with this cell has columnLockMode="right"
     * @return {boolean}
     * @method getIsRightLocked
     */
    getIsRightLocked() {
        return (this.getColumn() && this.getColumn().getIsRightLocked());
    }

    /**
     * If column is a checkbox column, returns the current state of the checkbox renderer
     * @method getCheckBoxState
     */
    getCheckBoxState() {
        if (this.getRenderer().implementsOrExtends("TriStateCheckBox")) {
            return this.getRenderer().getSelectedState();
        }
        return "";
    }

    /**
     * Returns the nestpDepth property of the associated level, if it is not null. If it is null, returns 1.
     * The top level column has nestDepth=1;
     * @return {*|Number}
     * @method getNestDepth
     */
    getNestDepth() {
        return this.level ? this.level.getNestDepth() : -1;
    }

    /**
     * The "Chrome" is defined as header,footer,pager,filter sections. This method returns true if this cell belongs to either of these
     * containers.
     * @return {Boolean}
     * @method getIsChromeCell
     */
    getIsChromeCell() {
        return this.rowInfo ? this.rowInfo.getIsChromeRow() : true;
    }

    /**
     * Returns true if we are a data cell or if we area a chrome cell
     * at a nest depth of greater than 1. These cells are all drawn in the
     * content area, hence have to scroll.
     * The "Chrome" is defined as header,footer,pager,filter sections. This method returns true if this cell belongs to either of these
     * containers.
     * @return {*|Boolean}
     * @method getIsContentArea
     */
    getIsContentArea() {
        return (this.getIsDataCell() || (this.getIsChromeCell() && (this.getNestDepth() > 1)));
    }

    /**
     * Returns true if the column associated with this cell has columnLockMode="left"
     * @return {Boolean}
     * @method getIsLeftLocked
     */
    getIsLeftLocked() {
        return (this.getColumn() && this.getColumn().getIsLeftLocked());
    }

    /**
     * Returns true if the rowInfo object associated with this cell is of type RowPositionInfo.ROW_TYPE_DATA.
     * @return {Boolean} 
     * @method getIsDataCell
     */
    getIsDataCell() {
        return this.rowInfo ? this.rowInfo.getIsDataRow() : false;
    }

    /**
     * Is this enabled cell
     * @param val
     * @method setEnabled
     */
    setEnabled(val) {
        if (this.colIcon) {
            this.colIcon.setSource(this.getIconUrl());
        }
        flexiciousNmsp.UIComponent.prototype.setEnabled.apply(this, [val]);
    }

    /**
     *
     */
    attachUserClass(className) {
        if (this.previouslyAttached) {
            this.detachClass(this.previouslyAttached);
        }
        if (className) {
            this.attachClass(className);
            this.previouslyAttached = className;
        }
    }

    /**
     * Clean up this cell
     * @method kill
     */
    kill() {
        if (this.domElement)
            this.destroy();
        if (this._renderer && this._renderer.kill) {
            this._renderer.kill();
        }
        flexiciousNmsp.UIComponent.prototype.kill.apply(this);
        this.level = null;
        this.rowInfo = null;
        this.column = null;
    }
    /**
     * Converts the given global Point to local co-ordinates.
     * @param ptIn {flexiciousNmsp.Point}
     * @return {flexiciousNmsp.Point}
     */
    globalToLocal(ptIn) {
        const offset = this.domElement ? UIUtils.offset(this.domElement) : { left: this.x , top: this.y };
        return new flexiciousNmsp.Point(ptIn.x - (offset ? offset.left : 0) + (this.domElement ? this.domElement.scrollLeft : 0), ptIn.y
            - (offset ? offset.top : 0) + (this.domElement ? this.domElement.scrollTop : 0));
    }
}

/**
 *
 * @type {Function}
 */
flexiciousNmsp.FlexDataGridCell = FlexDataGridCell; //add to name space
FlexDataGridCell.prototype.typeName = FlexDataGridCell.typeName = 'FlexDataGridCell';//for quick inspection