/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */

import Constants from '../../utils/Constants'
import UIUtils from '../../utils/UIUtils'
import FlexDataGridCell from './FlexDataGridCell'

/**
 * A container class for all header cells. In addition to the functionality it inherits from FlexDataGridCell,
 * it adds support for the following feautres:
 * <ul>
 * <li>Ability to trigger a column resize</li>
 * <li>Ability to render sort icons and multiColumnSort labels</li>
 * <li>Ability to render and initialize the checkbox header</li>
 * </ul>
 *
 * FlexDataGridHeaderCell inherits from FlexDataGridCell, which provides a bulk of the functionality to it.
 * @inheritDoc
 * @constructor
 * @class FlexDataGridHeaderCell
 * @namespace flexiciousNmsp
 * 
 */
export default class FlexDataGridHeaderCell extends FlexDataGridCell {
    constructor() {
        super();
        /**
         * @property resizing
         * @default false
         * @type {Boolean}
         */
        this.resizing = false;
        /**
         * @property resizingPrevious
         * @default false
         * @type {Boolean}
         */
        this.resizingPrevious = false;
        /**
         * @property sortAscending
         * @default false
         * @type {Boolean}
         */
        this.sortAscending = false;
        /**
         * @property sortable
         * @default true
         * @type {Boolean}
         */
        this.sortable = true;
        /**
         * @property icon
         * @default null
         * @type {Object}
         */
        this.icon = null;
        /**
         * @property sortLabel
         * @default null
         * @type {Object}
         */
        this.sortLabel = null;
        /**
         * @property _iconOffset
         * @default 0
         * @type {Number}
         * @private
         */
        this._iconOffset = 0;



        this.addEventListener(Constants.EVENT_CLICK, this.onHeaderClick);
        this.addEventListener(Constants.EVENT_MOUSE_MOVE, this.onMouseMove);
    }

    /**
     *
     * @return {Array}
     */
    getClassNames() {
        return ["FlexDataGridHeaderCell", "FlexDataGridCell", "UIComponent", "IFlexDataGridCell"];
    }

    /**
     * Removes the sort icon and the sort label
     */
    destroySortIcon() {
        if (this.icon && this.icon.parent != null) {
            this.icon.parent.removeChild(this.icon);
        }
        if (this.sortLabel && this.sortLabel.parent != null) {
            this.sortLabel.parent.removeChild(this.sortLabel);
            this.sortLabel.removeEventListener(Constants.EVENT_CLICK, this.onSortLabelClick, false);
        }

        this.sortLabel = null;
        this.icon = null;
        if (this.domElement && this.level) {
            this.placeComponent(this.getRenderer(), this.getWidth(), this.getHeight(), undefined);

        }
    }

    /**
     * handles the mouse move event.
     * @param event
     */
    onMouseMove(event) {

        if (this.level) {
            if (this.level.grid.enableMultiColumnSort && (event.target == this.sortLabel || event.target == this.icon)) {
                this.level.grid.getHeaderContainer().killResize();
            }
        }

    }

    /**
     *
     * @param x
     * @param y
     */
    move(x, y) {

        flexiciousNmsp.FlexDataGridCell.prototype.move.apply(this, [x, y]);
        if (this.getColumn()) {
            this.getColumn().setX(this.getX());
        }

        if (this.domElement) {
            this.placeSortIcon();
        } else {
            this.placementPending = true;
        }
    }

    /**
     * Default handler for the sort label click. (The sort label is only shown when the
     * enableMultiColumnSortFlag is set to true).
     * If the grid.getEnableSplitHeader() is true and the column is sortable, sort is handled by
     * the headerContainers onHeaderCellClicked method. Otherwise, this method will launch the
     * multiColumnSortPopUp.
     */
    onSortLabelClick(event) {

        if(this.icon && this.icon.parent!=null){
            this.removeSortIcon();
            this.icon.removeEventListener(Constants.EVENT_CLICK,this.onSortLabelClick,false);            
        }

        //this means we need to just swap the order of this
        var grid = this.level.grid;
        var existSplitHeader = grid.enableMultiColumnSort ? grid.getEnableSplitHeader() : true // assume split header exist;
        if(this.level && existSplitHeader && this.getColumn() && this.getColumn().sortable) {
            grid.getContainerForCell(this).onHeaderCellClicked(this, event, true);
        }
        else if (this.getColumn().sortable) {
            grid.multiColumnSortShowPopup();
        }

    }
    removeSortIcon(){
        flexiciousNmsp.UIUtils.removeChild(this.icon.parent,this.icon);
    }
    /**
     * Creates the sort icon. Depending whether the enableMultiColumnSort
     * flag is set to true, this method will create a sort label as well, which
     * is responsible for initializing the icon.
     */
    createSortIcon(container) {

        let ic;
        if (this.icon) {
            this.destroySortIcon();
        }
        if (this.getColumn() && !this.getColumn().sortable) return;
        const grid = this.level.grid;
        if (this.level.getEnableMultiColumnSort()) {
            var sl = new flexiciousNmsp.Label();
            sl.setAttribute("className", "sortLabel");
            sl.useHandCursor = !grid.getEnableSplitHeader();
            sl.buttonMode = !grid.getEnableSplitHeader();
            sl.mouseChildren = grid.getEnableSplitHeader();
            sl.setToolTip(grid.multiColumnSortGetTooltip(this));
            sl.addEventListener(this, Constants.EVENT_CLICK, this.onSortLabelClick, false);
            sl.setText(this.level.getSortIndex(this.getColumn()).toString());
        }

        const p = this.getIsLocked() ? this.getIsLeftLocked() ?
            grid.getLeftLockedHeader() : grid.getRightLockedHeader() : container;
        if (this.sortAscending) {
            ic = this.level.createAscendingSortIcon();
            if (ic.parent != grid) {
                p.addChild(ic);
                if (this.level.getEnableMultiColumnSort())
                    p.addChild(sl);
            }
        }
        else {
            ic = this.level.createDescendingSortIcon();
            if (ic.parent != grid) {
                p.addChild(ic);
                if (this.level.getEnableMultiColumnSort())
                    p.addChild(sl);
            }
        }
        this.icon = ic;
        //ic.propertyBag.key = new Date().getTime();
        //ic.key = new Date().getTime();

        if (this.level.getEnableMultiColumnSort())
            this.sortLabel = sl;
        if (this.domElement) {
            this.placeSortIcon();
        } else {
            this.placementPending = true;
        }

        if(this.icon) {
            this.icon.addEventListener(this, Constants.EVENT_CLICK,this.onSortLabelClick,false);
        }
    }

    /**
     * Sets the size of the renderer, accounting for the width of the split header or the icon.
     * If enableSplitHeader is set to true, reduces the renderer's width by grid.headerSortSeparatorRight
     * If sort icon is rendererd, reduces the renderer's width by the icons width.
     * @inheritDoc
     */
    setRendererSize(cellRenderer, w, h) {


        if (this.level && this.level.grid.getEnableSplitHeader() && this.getColumn() && this.getColumn().sortable) {
            w = w - this.level.grid.headerSortSeparatorRight;
        } else if (this.icon) {
            w = w - 14;
        }
        flexiciousNmsp.FlexDataGridCell.prototype.setRendererSize.apply(this, [cellRenderer, w, h]);

    }

    /**
     * handles destroy of this cell.
     */
    destroy() {

        const x = 0;
        super.destroy();
        this.resizing = false;
        this.resizingPrevious = false;
        this.sortAscending = false;
        this.sortable = false;
        this.destroySortIcon();

    }

    /**
     * handle the header click event.
     * @param event
     */
    onHeaderClick(event) {
        if (this.level)
            this.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.HEADER_CLICKED, this.level.grid, this.level, this.getColumn(), this));
    }

    /**
     * Refreshing the contents of this cell.
     */
    refreshCell() {

        flexiciousNmsp.FlexDataGridCell.prototype.refreshCell.apply(this, []);

        const fs = this.level.hasSort(this.getColumn());
        if (fs) {
            this.sortAscending = fs.isAscending;
            this.destroySortIcon();
            this.createSortIcon(this.parent);
        }

        this.setSelectable(false);
        if (this.getRenderer().typeName) {
            this.getRenderer().setData(this.rowInfo.getData());
            if (this.getColumn())
                this.getRenderer().setVisible(!this.getColumn().hideHeaderText);
            if (this.getColumn().headerTextFunction != null) {
                this.setText(this.getColumn().headerTextFunction(this.getColumn(), this));
            }
        }
        this.drawSortSeparator();

    }
    componentDidMountCustom() {
        if (this.placementPending) {
            this.placementPending = false;
            this.placeSortIcon();
            this.placeComponent(this.getRenderer(), this.getWidth(), this.getHeight(), undefined);
        }
    }
    /**
     * Places the sort icon after it is created. By default the column is placed 3 pixels from the right, half way from the top.
     * If the enableMultiColumnSort flag is set to true, the label is places above or below the icon based on if it is ascending or
     * descending respectively.
     */
    placeSortIcon() {

        if (this.icon) {
            let pt = new flexiciousNmsp.Point(this.getWidth() - 18, (this.getHeight() / 2) - 10); /*(this.getHeight() - 18) / 2 */
            if (this.icon.parent.domElement && this.domElement) {
                pt = this.icon.parent.globalToLocal(this.localToGlobal(pt));
            } else {
                pt.x += this.getX();
                pt.y += this.getY();
                this.placementPending = true;
            }
            //pt.x += this.level.grid.getHorizontalScrollPosition();
            this.icon.move(pt.getX(), pt.getY());


            if (this.sortLabel && this.level.getEnableMultiColumnSort()) {
                const right = this.level.grid.headerSortSeparatorRight;
                this.sortLabel.setWidth(this.level.grid.getStyle("multiColumnSortNumberWidth"));
                this.sortLabel.setHeight(this.level.grid.getStyle("multiColumnSortNumberHeight"));
                //pt=new flexiciousNmsp.Point(this.getWidth() - right+2, Math.max(0,((this.getHeight()-this.sortLabel.getHeight)/2)-1));
                //pt= this.icon.parent.globalToLocal(this.localToGlobal(pt));
                this.sortLabel.move(pt.getX() + 5, pt.getY() + 16);
                if (this.level && this.level.getSortIndex(this.getColumn()) > 9) {
                    this.sortLabel.move(pt.getX() - 5, pt.getY() + 4);
                }

                if(this.level && this.level && this.level.grid && this.level.grid.sortIconPlacementFunction){
                    this.level.grid.sortIconPlacementFunction(this);
                }
            }
        }
    }
    /**
     * Draws the little gray line between the header that does the actual sort
     * and the multi column sort part of the header that adds the current column to the existing sort.
     */
    drawSortSeparator() {

        if (this.level && this.level.grid.getEnableSplitHeader() && this.getColumn() && this.getColumn().sortable) {
            const right = this.level.grid.getHeaderSortSeparatorRight();
            if (!this.sortSeparator) {
                this.sortSeparator = new flexiciousNmsp.UIComponent({}, "span");
                this.addChild(this.sortSeparator);

            }
            this.sortSeparator.setActualSize(1, this.getHeight() - 6);
            this.sortSeparator.move(this.getWidth() - right, 3);
            this.sortSeparator.attachClass("sortSeparator");
        }
        if (this.domElement) {
            this.placeSortIcon();
        } else {
            this.placementPending = true;
        }

    }

    /**
     *
     * @return {String}
     */
    getBackgroundColors() {

        const fromGrid = flexiciousNmsp.CellUtils.getBackgroundColorFromGrid(this);
        if (fromGrid != null) return fromGrid;
        if (this.currentBackgroundColors) return this.currentBackgroundColors;
        return this.getStyleValue("headerColors");

    }

    /**
     *
     * @return {String}
     */
    getTextColors() {

        const fromGrid = flexiciousNmsp.CellUtils.getTextColorFromGrid(this);
        if (fromGrid != null) return fromGrid;
        if (this.currentTextColors) return this.currentTextColors;
        return this.getStyle("color");

    }

    /**
     *
     * @param renderer
     * @type DisplayObject
     */
    initializeCheckBoxRenderer(renderer) {

        const tcb = this.getRenderer().implementsOrExtends("TriStateCheckBox") ? this.getRenderer() : null;
        if (!tcb) return;
        const col = this.getColumn().implementsOrExtends("FlexDataGridCheckBoxColumn");
        if (col && col.enableLabelAndCheckBox) {
            tcb.label = col.getHeaderText(this);
        }
        if (this.level.getNestDepth() != 1) {

            flexiciousNmsp.CellUtils.initializeCheckBoxRenderer(this, this.level.getParentLevel());
            if (this.level.enableSingleSelect) {
                tcb.radioButtonMode = true;
                tcb.setEnabled(false);
            }
            return;
        }
        tcb.radioButtonMode = this.getColumn().radioButtonMode;
        tcb.enableDelayChange = false;//we dont want the delay timer
        if (this.getColumn().radioButtonMode || !this.getColumn().allowSelectAll) {
            tcb.setEnabled(false);
        }
        else {
            tcb.setEnabled(true);
            let provider = this.level.grid.getRootFlat();
            if (this.level.grid.getIsClientFilterPageSortMode())
                provider = UIUtils.filterArray(provider, this.level.grid.getRootFilter(), this.level.grid, this.level, false);
            tcb.setSelectedState(this.level.getSelectedKeysState(provider));
            if (tcb.getSelectedState() == flexiciousNmsp.TriStateCheckBox.STATE_UNCHECKED) {
                if (this.level.grid.enableSelectionExclusion) {
                    if (this.level.grid.getSelectionInfo().getHasAnySelections()) {
                        tcb.setSelectedState(flexiciousNmsp.TriStateCheckBox.STATE_MIDDLE);
                    }
                }
                else if (this.level.grid.getSelectedObjects().length > 0) {
                    tcb.setSelectedState(flexiciousNmsp.TriStateCheckBox.STATE_MIDDLE);
                }
            }

        }

    }

    /**
     * Handler for checkbox check change.
     * If the target is the top level (grid.getColumnLeve()) checkbox, dispatches the SELECT_ALL_CHECKBOX_CHANGED event, and calls
     * selectAll on the associated top level.
     * Calls level.selectRow method based on checkbox state, and whether or not grid.enableSelectionExclusion is set to true.
     */
    onCheckChange(event) {

        const cb = event.target.implementsOrExtends("TriStateCheckBox");
        if (!cb) return;
        if (this.level.getNestDepth() == 1) {
            this.level.selectAll(event.target.getSelected());
            this.level.grid.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.SELECT_ALL_CHECKBOX_CHANGED, this.level.grid, this.level, this.getColumn(), this, null, event));
        }
        else {
            if (this.level.grid.enableSelectionExclusion) {
                this.level.getParentLevel().selectRow(this.rowInfo.getData(), event.target.getSelected());
            } else {
                if (!event.target.getSelected())
                    this.level.getParentLevel().selectRow(this.rowInfo.getData(), false, false, false, false, null);
                else
                    this.level.getParentLevel().selectRow(this.rowInfo.getData(), true, false, false, false, null);
                this.selectLevel(this.level, this.level.getParentLevel().getChildren(this.rowInfo.getData()), event.target.getSelected());
            }
        }
        this.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.SELECT_ALL_CHECKBOX_CHANGED, this.level.grid, this.level, this.getColumn(), this));

    }

    /**
     *
     * @param checkLevel
     * @param items
     * @param checked
     */
    selectLevel(checkLevel, items, checked) {
        if (typeof checked == "undefined") checked = true;

        for (const item of items) {
            checkLevel.selectRow(item, checked);
        }
    }

    /**
     *
     * @return {String}
     */
    getRolloverColor() {

        return this.getStyleValue("headerRollOverColors");

    }

    /**
     * Returns icon!=null
     */
    getIsSorted() {
        return this.icon != null;
    }

    /**
     *
     * @return {String}
     */
    getPrefix() {
        return "header";
    }

    /**
     * @inheritDoc
     */
    getDrawTopBorder() {
        //if we're at a nested level, grid does not draw horizontal lines,
        //and there is no filter row, we draw a line above us to seperate ourselves from the data item above.
        return (this.level && this.level.getNestDepth() > 1 && !this.level.grid.getStyle("horizontalGridLines")
            && (this.level.getFilterRowHeight() == 0)) ||
            this.getStyleValue("headerDrawTopBorder")
    }

    /**
     *
     * @param value
     */
    setColumn(value) {
        flexiciousNmsp.FlexDataGridCell.prototype.setColumn.apply(this, [value]);

    }

    placeComponent(cellRenderer, unscaledWidth, unscaledHeight, usePadding) {
        flexiciousNmsp.FlexDataGridCell.prototype.placeComponent.apply(this, [cellRenderer, unscaledWidth, unscaledHeight, usePadding])
    }
}

flexiciousNmsp.FlexDataGridHeaderCell = FlexDataGridHeaderCell; //add to name space
FlexDataGridHeaderCell.prototype.typeName = FlexDataGridHeaderCell.typeName = 'FlexDataGridHeaderCell';//for quick inspection
/**
 * Dispatched when the cell is clicked.
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 */

FlexDataGridHeaderCell.EVENT_HEADERCLICKED = 'headerClicked';
