/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */

import UIUtils from '../../utils/UIUtils'
import Constants from '../../utils/Constants'
import TypedObject from '../../core/TypedObject'
import ExportEvent from './ExportEvent'
import ExportOptions from './ExportOptions'

/**
 * Main class responsible for exporting data.
 * @constructor
 * @class ExportController
 * @namespace flexiciousNmsp
 * 
 */

export default class ExportController extends TypedObject {
    constructor() {
        super()
        this._exportOptions = null;
        this._grid = null;
    }

    getClassNames() {
        return ["ExportController", "TypedObject"];
    }

    static instance() {
        return ExportController._instance;
    }

    /**
     * Exports the grid on basis of the provided
     * Export options
     * @param grid		The IExtendedDataGrid to export
     * @param exportOptions	The Export Options (See com.flexicious.export.ExportOptions)
     * @see com.flexicious.export.ExportOptions
     */
    doexport(grid, exportOptions) {
        if (typeof exportOptions == "undefined") exportOptions = null;

        if (!exportOptions)
            exportOptions = new flexiciousNmsp.ExportOptions();
        exportOptions.availableColumns = [];
        exportOptions.columnsToExport = [];
        exportOptions.printExportParameters = grid.printExportParameters;
        exportOptions.grid = grid;
        if (exportOptions.grid) {
            const expEvent = new flexiciousNmsp.ExportEvent(flexiciousNmsp.ExportEvent.BEFORE_EXPORT);
            expEvent.exportOptions = exportOptions;
            exportOptions.grid.dispatchEvent(expEvent);
        }
        return this.exportWithOptions(grid, exportOptions);

    }

    /**
     * Exports the grid on basis of the provided
     * Export options
     * @param grid		The IExtendedDataGrid to export
     * @param exportOptions	The Export Options (See com.flexicious.export.ExportOptions)
     * @see com.flexicious.export.ExportOptions
     */
    exportWithOptions(grid, exportOptions) {
        let i;
        let items;
        this._exportOptions = exportOptions;
        this._grid = grid;
        exportOptions.availableColumns = [];
        if (exportOptions.useExcelExporter || grid.nativeExcelExporter) {
            items = exportOptions.exporters;
            for (i = 0; i < items.length; i++) {
                const exporter = items[i];
                if (exporter.implementsOrExtends("CsvExporter")) {
                    const expIndex = items.indexOf(exporter);
                    items.splice(expIndex, 1, grid.nativeExcelExporter || new flexiciousNmsp.ExcelExporter());
                    if (exportOptions.exporter.implementsOrExtends("CsvExporter"))
                        exportOptions.exporter = exportOptions.exporters[expIndex];
                    break;
                }
            }
        }
        items = grid.getExportableColumnsAtAllLevel(exportOptions);
        for (let j = 0; j < items.length; j++) {
            var col = items[j];
            exportOptions.availableColumns.push({ 'name': col.getUniqueIdentifier(), 'headerText': col.getConcatenatedHeaderText() });
        }
        if (exportOptions.hideHiddenColumns) {
            for (i = 0; i < items.length; i++) {
                const col2 = items[i];
                if (col2.getVisible()) {
                    exportOptions.columnsToExport.push({ 'name': col2.getUniqueIdentifier(), 'headerText': col.getConcatenatedHeaderText() });
                }
            }
        }

        if (exportOptions.showColumnPicker) {
            const exportOptionsView = exportOptions.exportOptionsRenderer.newInstance();
            exportOptions.exportOptionsView = exportOptionsView;
            exportOptionsView.setGrid(grid);
            exportOptionsView.exportOptions = exportOptions;
            exportOptionsView.addEventListener(this, Constants.EVENT_CLOSE, this.onExportOptionsClose);
            exportOptionsView.addEventListener(this, Constants.EVENT_CANCEL, this.onExportOptionsCancel);
            let popupParent = exportOptions.popupParent;
            if (!popupParent) popupParent = UIUtils.getTopLevelApplication();
            //UIUtils.addPopUp(exportOptionsView, popupParent, exportOptions.modalWindows, null, exportOptions.exportOptionsWindowTitle);
            exportOptions.exportOptionsView.showDialog();
        }
        else {
            this.doExport();
        }
    }

    /**
     * handles onExportOptionsClose event.
     * @param event
     * @method onExportOptionsClose
     */
    onExportOptionsClose(event) {

        (event.target).removeEventListener(Constants.EVENT_CLOSE, this.onExportOptionsClose);
        const exportView = event.target;
        if (event.detail == Constants.ALERT_OK) {
            this._exportOptions = exportView.exportOptions;
            this.doExport();
        }
    }

    /**
     *  handles onExportOptionsCancel event.
     * @param event 
     */
    onExportOptionsCancel(event) {
        (event.target).removeEventListener(Constants.EVENT_CANCEL, this.onExportOptionsCancel);

        const grid = this._grid;
        grid.removeEventListener(flexiciousNmsp.PrintExportDataRequestEvent.PRINT_EXPORT_DATA_RECD, this.onPrintRequestDataRecieved);
    }

    /**
     * handles doExport.
     */
    doExport() {

    }

    /**
     * handles dispatchDataRequest event.
     */
    dispatchDataRequest() {

        const exportOptions = this._exportOptions;
        if (exportOptions.saveFileMessage)
            exportOptions.useSaveFileMessage = true;
        const grid = this._grid;
        const printExportFilter = new flexiciousNmsp.PrintExportFilter();
        printExportFilter.copyFrom(grid.createFilter());
        printExportFilter.printExportOptions = exportOptions;
        grid.addEventListener(this, flexiciousNmsp.PrintExportDataRequestEvent.PRINT_EXPORT_DATA_RECD, this.onPrintRequestDataRecieved);
        const printExportDataRequestEvent = new flexiciousNmsp.PrintExportDataRequestEvent(flexiciousNmsp.PrintExportDataRequestEvent.PRINT_EXPORT_DATA_REQUEST, printExportFilter);
        grid.dispatchEvent(printExportDataRequestEvent);
        return; //we will continue processing when we get all data from the server.

    }

    /**
     * handles onCancelExport event
     * @param event 
     */
    onCancelExport(event) {
        this._grid.removeEventListener(flexiciousNmsp.PrintExportDataRequestEvent.PRINT_EXPORT_DATA_RECD, this.onPrintRequestDataRecieved);
        this.runExport(this._grid.getprintExportData(), true);
    }

    /**
     * handles onPrintRequestDataRecieved
     * @param event
     */
    onPrintRequestDataRecieved(event) {

        this._grid.removeEventListener(flexiciousNmsp.PrintExportDataRequestEvent.PRINT_EXPORT_DATA_RECD, this.onPrintRequestDataRecieved);
        this.runExport(this._grid.getprintExportData(), true);

    }

    /**
     * Just build up the doc header, body, footer and ask server
     * to buffer it back
     * @param iCollectionView
     * @param allOrSelectedPages
     */
    runExport(iCollectionView, allOrSelectedPages) {

    }
}

flexiciousNmsp.ExportController = ExportController; //add to name space
ExportController.prototype.typeName = ExportController.typeName = 'ExportController';//for quick inspection
ExportController._instance = new ExportController();