/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */

import Constants from '../../utils/Constants'
import UIUtils from '../../utils/UIUtils' 
import Label from '../../controls/Label'
import TextInput from '../../controls/TextInput'
import ComboBox from '../../controls/ComboBox'
import DateComboBox from '../../controls/DateComboBox'
import DatePicker from '../../controls/DatePicker'


import EventDispatcher from '../../core/EventDispatcher'
import FilterExpression from './filters/FilterExpression'
import FlexDataGridHeaderCell from '../cells/FlexDataGridHeaderCell'
import FlexDataGridFooterCell from '../cells/FlexDataGridFooterCell'
import FlexDataGridFilterCell from '../cells/FlexDataGridFilterCell'
import FlexDataGridPagerCell from '../cells/FlexDataGridPagerCell'
import FlexDataGridDataCell from '../cells/FlexDataGridDataCell'
import FlexDataGridLevelRendererCell from '../cells/FlexDataGridLevelRendererCell'
import FlexDataGridExpandCollapseCell from '../cells/FlexDataGridExpandCollapseCell'
import FlexDataGridColumnGroupCell from '../cells/FlexDataGridColumnGroupCell'
import FlexDataGridExpandCollapseHeaderCell from '../cells/FlexDataGridExpandCollapseHeaderCell'
import FlexDataGridHeaderCellHS from '../cells/FlexDataGridHeaderCellHS';


/**
 * The FlexDataGridColumn class describes a column in a FlexDataGrid Column control.
 *
 * @constructor
 * @class FlexDataGridColumn
 * @namespace flexiciousNmsp
 */
export default class FlexDataGridColumn extends EventDispatcher {
    constructor() {
        super();
        /**
         * A function that determines the header text for a column.
         * Should take a column.
         * @type {Function}
         * @property headerTextFunction
         * @default null
         * @example 
         * setHeaderLabel() {
         *   return 'demo';
         * }
         */
        this.headerTextFunction = null;
        /**
         * The default function for excel export. By default, we will return the header text
         * If the parent is a column group, we will concatenate its text to the excel export.
         * @type {Function}
         * @property headerTextForExportFunction
         * @default {getConcatenatedHeaderText}
         */
        this.headerTextForExportFunction = this.getConcatenatedHeaderText;


        /**
         * Used by the doFormat method. One of the following values:
         * FlexDataGridColumn.FORMAT_NONE="none";
         * FlexDataGridColumn.FORMAT_DATE="date";
         * FlexDataGridColumn.FORMAT_DATE_TIME="datetime";
         * FlexDataGridColumn.FORMAT_TIME="time";
         * FlexDataGridColumn.FORMAT_CURRENCY="currency";
         * FlexDataGridColumn.FORMAT_NUMBER="number";
         * @type {String}
         * @property format
         * @default FlexDataGridColumn.FORMAT_NONE
         * @see http://www.reactdatagrid.com/demo/#/itemeditor
         */
        this.format = FlexDataGridColumn.FORMAT_NONE;
        /**
         * For number and currency formatters, the default number of digits after the decimal point. Defaults to 2
         * @property  formatterPrecision
         * @type {number}
         * @default 2
         */
        this.formatterPrecision = 2;
        /**
         * For currency formatter, the currency symbol. Defaults to blank value.
         * @property formatterCurrencySymbol
         * @type {String}
         * @default null
         */
        this.formatterCurrencySymbol = "";
        /**
         * For date formatters, the date format. Defaults to blank value. Initialized on basis of date format.
         * If you wish to override the default format, you can use this string.
         * @property formatterDateFormatString
         * @type {String}
         * @default null
         */
        this.formatterDateFormatString = "";
        /**
         * When you set to true, a custom sortCompareFunction is applied that uses the labelFunction instead
         * of the dataField.
         * @property useLabelFunctionForSortCompare
         * @type {boolean}
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/grouped2
         */
        this.useLabelFunctionForSortCompare = false;


        /**
         * @private
         */        
        this._naturalLastColWidth = 0;

        /**
         * Label for the footer
         * @type {null}
         * @property footerLabel
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.footerLabel = null;

        /**
         * A function that will return the footer label
         * @property footerLabelFunction
         * @type {Function}
         * @default null
         * @example 
         * getSalaryFooter(col) {
         *       const grid = col.level.grid;
         *       return <div>
         *       <div style={{textAlign:'right'}}> Avg: {flexiciousNmsp.UIUtils.formatCurrency(flexiciousNmsp.UIUtils.average(grid.getDataProvider(), "annualSalary"))}</div>
         *       <div style={{textAlign:'right'}}> Min: {flexiciousNmsp.UIUtils.formatCurrency(flexiciousNmsp.UIUtils.min(grid.getDataProvider(), "annualSalary"))}</div>
         *       <div style={{textAlign:'right'}}> Max: {flexiciousNmsp.UIUtils.formatCurrency(flexiciousNmsp.UIUtils.max(grid.getDataProvider(), "annualSalary"))}</div>
         *     </div> 
         *   };
         * @see http://www.reactdatagrid.com/demo/#/customfooter
         */
        this.footerLabelFunction = null;

        /**
         * A renderer for the footer
         * @property footerRenderer
         * @type ClassFactory
         * @default null
         */
        this.footerRenderer = null;

        /**
         * The operation to apply to the footer. One of the following values:
         * <ul>
         * 	<li>average</li>
         * 	<li>sum</li>
         * 	<li>min</li>
         * 	<li>max</li>
         * 	<li>count</li>
         * </ul>
         * @property footerOperation
         * @type {String}
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.footerOperation = null;


        /**
         * Formatter for the footer label.
         * @property footerFormatter
         * @type {Formatter}
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.footerFormatter = null;

        /**
         * Precision for the footer label
         * @type {number}
         * @property footerOperationPrecision
         * @default 2
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.footerOperationPrecision = 2;

        /**
         * The alignment of the footer label
         * @type {String}
         * @property footerAlign
         * @default "left"
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.footerAlign = "left";

        /**
         * Dataprovider to use to build the list of values to dipsplay in
         * the filter control, only applicable if the filterControl is a
         * @see com.flexicious.controls.interfaces.filters.ISelectFilterControl
         * Note:
         * @see com.flexicious.controls.interfaces.filters.IMultiSelectFilterControl,
         * @see com.flexicious.controls.interfaces.filters.ISingleSelectFilterControl
         * inherit from ISelectFilterControl, and
         * @see com.flexicious.controls.MultiSelectComboBox
         * @see com.flexicious.controls.RadioButtonList
         * @see com.flexicious.controls.ComboBox
         * @see com.flexicious.controls.DateComboBox all inherently implement
         * @see com.flexicious.controls.interfaces.filters.ISelectFilterControl.
         *
         * Note 2:
         * It is not required to set this field. You may set the @see filterComboBoxBuildFromGrid
         * property to true which will automatically build this collection on basis
         * of distinct values in the grid.
         * @type {object}
         * @property filterComboBoxDataProvider
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.filterComboBoxDataProvider = null;

        /**
         * Used in conjunction with the @filterComboBoxDataProvider field, used to
         * set the value of the label field for the associated ISelectFilterControl
         * @type {String}
         * @property filterComboBoxLabelField
         * @see http://www.reactdatagrid.com/demo/#/grouped
         */
        this.filterComboBoxLabelField = "label";

        /**
         * Used in conjunction with the @filterComboBoxDataProvider field, used to
         * set the value of the data field for the associated ISelectFilterControl
         * @type {String}
         * @property filterComboBoxDataField
         * @see http://www.reactdatagrid.com/demo/#/grouped
         */
        this.filterComboBoxDataField = "data";

        /**
         * Flag, if set to true, will automatically build this
         * collection on basis of distinct values in the grid.
         * @type {boolean}
         * @property filterComboBoxBuildFromGrid
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/nested
         */
        this.filterComboBoxBuildFromGrid = false;

        /**
         * @property filterComboBoxLabelFunction
         * @default null
         */
        this.filterComboBoxLabelFunction = null;

        /**
         * Applicable only when the filtercontrol is a
         * @see com.flexicious.controls.interfaces.filters.IDateComboBox
         * This has to be a list of Date Range Strings from the below:
         *     DateRange.DATE_RANGE_LAST_SIXTY_MINTUES= "Last 60 Minutes";
         *     DateRange.DATE_RANGE_LAST_12_HOURS= "Last 12 Hours";
         *     DateRange.DATE_RANGE_LAST_24_HOURS= "Last 24 Hours";
         *     DateRange.DATE_RANGE_LAST_7_DAYS= "Last 7 Days";
         *     DateRange.DATE_RANGE_THISHOUR= "This Hour";
         *     DateRange.DATE_RANGE_LASTHOUR= "Last Hour";
         *     DateRange.DATE_RANGE_NEXTHOUR= "Next Hour";
         *     DateRange.DATE_RANGE_TODAY= "Today";
         *     DateRange.DATE_RANGE_YESTERDAY= "Yesterday";
         *     DateRange.DATE_RANGE_TOMORROW= "Tomorrow";
         *     DateRange.DATE_RANGE_THISWEEK= "This Week";
         *     DateRange.DATE_RANGE_LASTWEEK= "Last Week";
         *     DateRange.DATE_RANGE_NEXTWEEK= "Next Week";
         *     DateRange.DATE_RANGE_THISMONTH= "This Month";
         *     DateRange.DATE_RANGE_LASTMONTH= "Last Month";
         *     DateRange.DATE_RANGE_NEXTMONTH= "Next Month";
         *     DateRange.DATE_RANGE_THISYEAR= "This Year";
         *     DateRange.DATE_RANGE_LASTYEAR= "Last Year";
         *     DateRange.DATE_RANGE_NEXTYEAR= "Next Year";
         *     DateRange.DATE_RANGE_CUSTOM= "Custom";
         *     DateRange.DATE_RANGE_THISQUARTER= "This Quarter";
         *     DateRange.DATE_RANGE_NEXTQUARTER= "Next Quarter";
         *     DateRange.DATE_RANGE_LASTQUARTER= "Last Quarter";
         *     * See the @com.flexicious.utils.DateRange class for details
         * @type {Array}
         * @property filterDateRangeOptions
         * @default []
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.filterDateRangeOptions = [];

        /**
         * @private 
         */
        this._filterRenderer = null;

        /**
         * @private 
         */
        this._filterControlClass = "";
        /**
         * The operator to apply when doing the conversion.
         * See FILTER_OPERATION_TYPE constant values from
         * com.flexicious.grids.filters.FilterExpression. Here are a list of options:
         * <ul>
         * <li>Equals</li>
         * <li>NotEquals</li>
         * <li>BeginsWith</li>
         * <li>EndsWith</li>
         * <li>Contains</li>
         * <li>DoesNotContain</li>
         * <li>GreaterThan</li>
         * <li>LessThan</li>
         * <li>GreaterThanEquals</li>
         * <li>LessThanEquals</li>
         * <li>InList</li>
         * <li>NotInList</li>
         * <li>Between</li>
         * </ul>
         * @type {String}
         * @property  filterOperation
         * @default FilterExpression.FILTER_OPERATION_TYPE_EQUALS
         * @see http://www.reactdatagrid.com/demo/#/outlook
         */
        this.filterOperation = FilterExpression.FILTER_OPERATION_TYPE_EQUALS;
        /**
         * This is usually automatically set, you dont have to manually set it,
         * unless you're sending strings as Date objects. When set, will attempt
         * to first convert the current value to the type you specified and then
         * do the conversion.
         * Values : auto,string,number,boolean,date
         * @type {String}
         * @property  filterComparisionType
         * @default "auto"
         */
        this.filterComparisionType = "auto";

        /**
         * @private 
         */
        this._searchField = "";

        /**
         * @private 
         */
        this._sortField = "";
        /**
         * The event that the filter triggers on. Defaults to "change", or if the
         * filterRenderer supports com.flexicious.controls.interfaces.IDelayedChange, then
         * the delayedChange event. "none,change,enterKeyUp"
         * @see com.flexicious.controls.interfaces.IDelayedChange
         * @see http://www.reactdatagrid.com/demo/#/simple
         * @type {String}
         * @property filterTriggerEvent
         * @default "change"
         */
        this.filterTriggerEvent = "change";

        /**
         * If set to true, this column wont be included in the
         * quick find mechanism by default.
         * @type {boolean}
         * @property excludeFromQuickFind
         * @default false
         */
        this.excludeFromQuickFind = false;

        /**
         * If set to true, this column wont be included in the
         * print mechanism by default.
         * @type {boolean}
         * @property excludeFromPrint
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.excludeFromPrint = false;

        /**
         * If set to true, this column wont be included in the settings popup by default.
         * @type {boolean}
         * @property excludeFromSettings
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.excludeFromSettings = false;

        /**
         * If set to true, this column wont be included in the
         * export mechanism by default.
         * @type {boolean}
         * @property excludeFromExport
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.excludeFromExport = false;

        /**
         * @type {String}
         * @property filterWaterMark
         * @default ""
         * @see http://www.reactdatagrid.com/demo/#/nested
         */
        this.filterWaterMark = "";

        /**
         * Flag, that defaults to true, indicates that in client filter mode,
         * use the label function to perform the search (opposed) comparison to using the
         * raw value of the "searchField" property of the column.
         * @type {boolean}
         * @property useLabelFunctionForFilterCompare
         * @default false
         */
        this.useLabelFunctionForFilterCompare = false;

        /**
         * @deprecated
         * @property linkText
         * @default null
         */
        this.linkText = null;

        /**
         * @private 
         */
        this._cellText = null;

        /**
         * On creation complete, will use this to distribute
         * the width of the grid among the visible columns in
         * the specified ratio.
         * @type {number}
         * @property percentWidth
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/variablerowheight
         */
        this.percentWidth = null;

        /**
         * Uses a special itemrenderer, which causes the mouse to
         * turn into a hand cursor on hover.
         * @type {boolean}
         * @property useHandCursor
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/nested
         */
        this.useHandCursor = null;

        /**
         * When true, if the renderer supports the text decoration style property, sets it to underline.
         * @type {boolean}
         * @property useUnderLine
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/nested
         */
        this.useUnderLine = null;


        /**
         *  The class factory for item renderer instances that display the
         *  data for each item in the column.
         *  You can specify a ClassFactory ustom item renderer component as the
         *  value of this property.
         *
         *  <p>The default item renderer is a TextView class,
         *  which displays the item data as text. </p>
         * Each Data Cell in the grid is contains an instance of this class.
         * @type {flexiciousNmsp.ClassFactory}
         * @property itemRenderer
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.itemRenderer = null;

        // /**
        //  * By default, we assume that the item renderers are either built in Flexicious components
        //  * or stateless functional components for performance reasons. If you are using custom react
        //  * components, then we need to set this property to true - we will then call the following 
        //  * React lifecycle events: componentDidMount, shouldComponentUpdate
        //  * @property itemRendererIsComponent
        //  */
        // this.itemRendererIsReactComponent = false;

        /**
         *  A class factory for the instances of the item editor to use for the
         *  column, when it is editable.
         * @type {flexiciousNmsp.ClassFactory}
         * @default FlexDataGridColumn.static_TextInput
         * @property itemEditor
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.itemEditor = FlexDataGridColumn.static_TextInput;

        /**
         * Indicates that the item editors manage their own persitence and populating the dataprovider
         * back with their changes, so the editorDataField is ignored and the DataGrid does not write
         * back the value from the editor into the data provider.
         * When you set itemEditorManagesPersistence, the itemEditorValidatorFunction or itemEditorApplyValueOnCommit are ignored,
         * since we completely handover the responsibility of validating and persisting the data from the editor back into the dataprovider to the editor
         * @type {boolean}
         * @property itemEditorManagesPersistence
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/itemeditor
         */
        this.itemEditorManagesPersistence = false;

        /**
         * Function to call before committing the value of the item editor
         * to the data provider. Should take a View with is the
         * item editor, and return a true or a false. If false, value is not committed
         * If true, value is committed.
         * @type {Function}
         * @property itemEditorValidatorFunction
         * @default null
         * @example 
         * validate(editor) {
         * 
         *      const grid = this.grid;
         *      var cell = grid.getCurrentEditCell();
         *      var txt = editor;
         *      if (txt.getValue().length < 3) {
         *          const errMsg = "Legal name must be greater than 3 characters";
         *          UIUtils.showMessage(errMsg);
         *          grid.setErrorByObject(cell.getRowInfo().getData(), cell.getColumn().getDataField(), errMsg);
         *      } else {
         *          grid.clearErrorByObject(cell.getRowInfo().getData());
         *      }
         * 
         *      return (txt.getValue().length >= 3);
         * };
         * @see http://www.reactdatagrid.com/demo/#/errorhandling
         */
        this.itemEditorValidatorFunction = null;

        /**
         *  The name of the property of the item editor that contains the new
         *  data for the list item.
         * @property editorDataField
         * @see http://www.reactdatagrid.com/demo/#/itemeditor
         */
        this.editorDataField = "value";

        /**
         * @private 
         */
        this._dataCellRenderer = FlexDataGridColumn.static_FlexDataGridDataCell;

        /**
         * Container for the header renderer. Needs to inherit from FlexDataGridHeaderCell.
         * Defaults to FlexDataGridHeaderCell
         * @type {flexiciousNmsp.ClassFactory}
         * @property headerCellRenderer
         * @default FlexDataGridColumn.static_FlexDataGridHeaderCell
         */
        this.headerCellRenderer = FlexDataGridColumn.static_FlexDataGridHeaderCell;

        /**
         * The alignment of the header label enumeration="left,right,center", defaultValue="left"
         * @type {String}
         * @property headerAlign
         * @default "left"
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.headerAlign = "left";

        /**
         * Flag that controls whether or not to show variable sized headers. The height of the header section is increased
         * until the longest text is fully shown
         * @type {boolean}
         * @property headerWordWrap
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.headerWordWrap = false;

        /**
         * Flag that controls whether or not to show variable sized footers. The height of the footer section is increased
         * until the longest text is fully shown
         * @type {boolean}
         * @property footerWordWrap
         * @default false
         */
        this.footerWordWrap = false;

        /**
         * Flag to indicate that the text in this column can flow into multiple lines, based
         * upon the width of the column. This flag is usually set in combination with the
         * variableRowHeight flag on the grid. Below are the docs for the variableRowHeight tag:
         * @copy com.flexicious.nestedtreedatagrid.FlexDataGrid#variableRowHeight
         * Please note, if you set an itemRenderer, wordWrap will not be applied to it automatically.
         * When an itemrenderer is specified, and variableRowHeight is true, the grid will still allocate
         * only the amount of space required to render the text from the dataField.
         * If you want to use the itemRenderer for RowHeight calculation, set variableRowHeightUseRendererForCalculation=true.
         * @type {boolean}
         * @property wordWrap
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/rowspancolspan
         */
        this.wordWrap = false;

        /**
         * Container for the footer renderer. Needs to inherit from FlexDataGridFooterCell.
         * Defaults to FlexDataGridFooterCell
         * @type {flexiciousNmsp.ClassFactory}
         * @property footerCellRenderer
         * @default FlexDataGridColumn.static_FlexDataGridFooterCell
         */
        this.footerCellRenderer = FlexDataGridColumn.static_FlexDataGridFooterCell;

        /**
         * Container for the footer renderer. Needs to inherit from FlexDataGridFooterCell.
         * Defaults to FlexDataGridFooterCell
         * @type {flexiciousNmsp.ClassFactory}
         * @property filterCellRenderer
         * @default FlexDataGridColumn.static_FlexDataGridFilterCell
         */
        this.filterCellRenderer = FlexDataGridColumn.static_FlexDataGridFilterCell;
        /**
         *  The name of the field or property in the data provider item associated
         *  with the column.
         *  Each DataGridColumn control
         *  requires this property and/or the <code>labelFunction</code> property
         *  to be set in order to calculate the displayable text for the item
         *  renderer.
         *  If the <code>dataField</code>
         *  and <code>labelFunction</code> properties are set,
         *  the data is displayed using the <code>labelFunction</code> and sorted
         *  using the <code>dataField</code>.  If the property named in the
         *  <code>dataField</code> does not exist, the
         *  <code>sortCompareFunction</code> must be set for the sort to work
         *  correctly.
         *
         *  <p>This value of this property is not necessarily the String that
         *  is displayed in the column header.  This property is
         *  used only to access the data in the data provider.
         *  For more information, see the <code>headerText</code> property.</p>
         *
         * If you specify a complex property ,the grid takes over the sortCompareFunction, and the sortField
         * property is ignored.
         * @see #headerText
         * @property dataField
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/simple
         */

        this.dataField = null;


        /**
         * A function that takes a IFlexDataGridCell, and returns true if this cell is editable.
         * Use this in scenarios where some cells out of a column are not editable.
         * @deprecated Use grid.cellEditableFunction
         * @property isEditableFunction
         * @type {Function}
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/largedynamicgrid
         */
        this.isEditableFunction = null;

        /**
         * @private 
         */
        this._editable = true;

        /**
         * Flag that indicates whether the items in the column are selectable.
         * @type {boolean}
         * @property selectable
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/itemeditor
         */
        this.selectable = false;

        /**
         * Flag to enable selection of a row when a cell within that row is clicked.
         * You may choose to set this to false when you set selectable =true.
         * When the user clicks on a cell, the row is automatically selected when the
         * selection mode is either single row or multiple rows. When you have cells
         * that are selectable, the user may wish to select the text in the (opposed) cell
         * to selecting the row. Setting this flag to false will disable the row selection
         * when any cell within this column is clicked.
         * @type {boolean}
         * @property enableCellClickRowSelect
         * @default true
         * @see http://www.reactdatagrid.com/demo/#/selectionui1
         */
        this.enableCellClickRowSelect = true;

        /**
         * For grouped datagrids, when a single column represents entities from various nest levels, its nice
         * to visually indent them. Set this parameter to true for such columns
         * If you wish to programatically control the padding, you could do so via the rendererInitialized event on the grid.
         * @default true
         * @property enableHierarchicalNestIndent
         * @type {boolean}
         * @see http://www.reactdatagrid.com/demo/#/selectionui2
         */
        this.enableHierarchicalNestIndent = false;


        /**
         * If this propery is true, the text that does not fit in the available size
         * is truncated with "...".
         * @type {boolean}
         * @property truncateToFit
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.truncateToFit = false;
        /**
         *
         * Storage for the headerRenderer property.
         * @type {flexiciousNmsp.ClassFactory}
         * @property headerRenderer
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/itemrenderer
         */
        this.headerRenderer = null;
        /**
         * The class factory for item renderer instances that display the
         * column header for the column.
         * Storage for the headerText property.
         * @private 
         * @type {flexiciousNmsp.ClassFactory}
         * @default null
         */
        this._headerText = null;

        /**
         * @private
         */
        this._uniqueIdentifier = "";
        /**
         *
         * Storage for the labelFunction property.
         * @private
         */

        this._labelFunction = null;
        /**
         * @private 
         * Storage for the labelFunction property.
         */
        this._labelFunction2 = null;

        /**
         * Flag that indicates whether the user is allowed to resize
         * the width of the column. Please note, left and right locked columns are not resizable
         * @default true
         * @property resizable
         * @type {boolean}
         * @see http://www.reactdatagrid.com/demo/#/itemrenderer
         */
        this.resizable = true;

        /**
         * Flag that indicates whether the user is allowed to resize
         * the width of the column. Please note, left and right locked columns are not resizable
         * @default true
         * @property draggable
         * @type {boolean}
         */
        this.draggable = true;
        /**
         * Flag that indicates whether the user can click on the
         * header of this column to sort the data provider.
         * @property sortable
         * @default true
         * @type {boolean}
         * @see http://www.reactdatagrid.com/demo/#/itemrenderer
         */
        this.sortable = true;
        /**
         * A callback function that gets called when sorting the data in
         * the column.
         * @type {Function}
         * @property sortCompareFunction
         * @default null
         * @example
         *   amountSortCompareFunction(obj1, obj2) {
         *      if (obj1.implementsOrExtends('Organization') && obj2.implementsOrExtends('Organization')) {
         *          return UIUtils.numericCompare(obj1.getRelationshipAmount(), obj2.getRelationshipAmount());
         *      }
         *      else if (obj1.implementsOrExtends('Deal') && obj2.implementsOrExtends('Deal')) {
         *          return UIUtils.numericCompare(obj1.getDealAmount(), obj2.getDealAmount());
         *      }
         *      else if (obj1.implementsOrExtends('Invoice') && obj2.implementsOrExtends('Invoice')) {
         *          return UIUtils.numericCompare(obj1.getInvoiceAmount(), obj2.getInvoiceAmount());
         *      }
         *      return 0;
         *  }
         * @see http://www.reactdatagrid.com/demo/#/outlook
         */

        this.sortCompareFunction = null;
        /**
         * @private 
         *  Storage for the visible property.
         */
        this._visible = true;

        /**
         * @private 
         * @type {number}
         */
        this._x = null;
        /**
         *
         *  Storage for the width property.
         * @private 
         */
        this._width = FlexDataGridColumn.DEFAULT_COLUMN_WIDTH;

        /**
         * The mininum width that a column can receive.
         *
         * After all calculations are made, it could result in some columns being squished beyond a certain value, where they're hardly visible.
         * Setting a value for this property ensure this does not happen.
         * @type {number}
         * @property minWidth
         * @default 20
         * @see http://www.reactdatagrid.com/demo/#/nested
         */
        this.minWidth = 20;

        /**
         * @type {Function}
         * @property footerLabelFunction2
         * @default null
         * @example 
         *   getFooterLabel(cell) {
         * 
         *     if (_footerData[cell.getRowInfo().getData()]) {
         *       if (cell.getColumn().dataField == "invoiceAmount" || cell.getColumn().dataField == "lineItemAmount" || cell.getColumn().dataField == "dealAmount")
         *           return "Total: " + UIUtils.formatCurrency(_footerData[cell.getRowInfo().getData()].total);
         *       else
         *           return "Count: " + _footerData[cell.getRowInfo().getData()].count;
         *       }
         *       return "";
         *  };
         * @see http://www.reactdatagrid.com/demo/#/xmlgroupeddata
         */
        this.footerLabelFunction2 = null;

        /**
         * @type {Function}
         * @property headerLabelFunction2
         * @default null
         */
        this.headerLabelFunction2 = null;

        /**
         * @private
         */
        this._persistenceKey = "";

        /**
         * Returns the width of the dropdown. Applicable only when the
         * filter controls are comboboxes
         * @type {number}
         * @property filterComboBoxWidth
         * @default 0
         * @see http://www.reactdatagrid.com/demo/#/levelrenderer
         */
        this.filterComboBoxWidth = 0;

        /**
         * Not supported
         * @type {boolean}
         * @property useCurrentDataProviderForFilterComboBoxValues
         * @default null
         */
        this.useCurrentDataProviderForFilterComboBoxValues = null;

        /**
         * A function that can be used to control the enabled flag of each cell in this column.
         * If this function is null or returns null, the cell will be enabled, selectable, and rollovers
         * will work. If this function returns false, the cell will not be selectable, the item renderers will
         * be disabled and it will have no rollovers.
         * This function should take a IFlexDataGridDataCell
         * object, which has a pointer to the row data ((well) data) as other related information
         * found in the documentation of the FlexDataGridDataCell class. This function
         * should true or false.
         * @type {Function}
         * @property cellDisabledFunction
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/grouped2
         */
        this.cellDisabledFunction = null;

        /**
         * A function that can be used to control the Text color of each cell in this column.
         * If this function is null or returns null, the cell Text will
         * use the alternatingTextColors style property. This function should take a IFlexDataGridDataCell
         * object, which has a pointer to the row data ((well) data) as other related information
         * found in the documentation of the FlexDataGridDataCell class. This function
         * should return  a single color hex code (uint).
         *
         * @deprecated. Use the grid.cellTextColorFunction instead.
         * @type {Function}
         * @property cellTextColorFunction
         * @default null
         * @example 
         * getColumnTextColor(cell) {
         *      var val = UIUtils.resolveExpression(cell.getRowInfo().getData(), cell.getColumn().dataField);
         *      if (val < 10000) {
         *          return 0xFFFFFF;
         *      } else if (val > 50000) {
         *          return 0x000000;
         *      }
         *       else {
         *           return 0x000000;
         *       }
         * }
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.cellTextColorFunction = null;

        /**
         * A function that lets you control the border drawing mechanism for  cell.
         * By default, each cell will draw a 1px wide line to its right (if XXXXVerticalGridLines=true)
         * and bottom (if XXXXhorizontalGridLines=true) ends.
         *
         * This function offers you the opportunity to hook into this mechanism, and draw
         * your own border using the graphics property of the FlexDataGridCell object,
         * that is passed (a) in parameter to this function. If this function returns false,
         * the default border is not drawn.
         * @type {Function}
         * @property cellBorderFunction
         * @default null
         * @example 
         * drawCellBorder(cell) {
         *      if (!cell.getIsDataCell()) {
         *           return true; //let the grid handle this
         *      }
         *      if (cell.level.getNestDepth() == 1) {
         *           cell.setStyleAttribute("borderBottom", "solid 1px #ececec");
         *           return false;
         *      }
         *      else if (cell.level.getNestDepth() == 2) {
         *           cell.setStyleAttribute("borderBottom", "solid 1px #afafaf");
         *           return false;
         *      }
         *    return true;//let the grid handle this
         * }
         */
        this.cellBorderFunction = null;

        /**
         * A function that can be used to control the background color of each cell in this column.
         * If this function is null or returns null, the cell  will
         * use the alternatingItemColors style property. This function should take a IFlexDataGridDataCell
         * object, which has a pointer to the row data ((well) data) as other related information
         * found in the documentation of the FlexDataGridDataCell class. This function
         * should return an array of colors for a gradient fill, or a single color
         * hex code (uint) for a single color fill.
         *
         * @deprecated Use the grid.cellBackgroundColorFunction instead.
         * @type {Function}
         * @property cellBackgroundColorFunction
         * @default null
         * @example 
         * cellBackgroundColorFunction(cell) {
         *       if (!cell.getIsDataCell()) {
         *           return null;
         *       }
         *       if (cell.level.getNestDepth() == 1) {
         *           return 0xececec;
         *       }
         *       return null;
         *   }
         * @see http://www.reactdatagrid.com/demo/#/rowspancolspan
         */
        this.cellBackgroundColorFunction = null;
        /**
         * @type {Function}
         * @property rowBackgroundColorFunction
         * @default null
         * @example 
         * getRowBackground(cell) {
         * 
         * if (cell.rowInfo.getData().headquarterAddress.state.name == "New York") {
         *      return 0xCFCFCF;
         * } else if (cell.rowInfo.getIsFillRow()) {
         *      return [0xCFCFCF, 0xFFFFFF];
         * }
         * return null;
         * }
         * @see http://www.reactdatagrid.com/demo/#/programaticcellformatting
         */
        this.rowBackgroundColorFunction = null;
        /**
         * A function that lets you control the background drawing mechanism for  cell.
         * By default, each cell will draw a background on basis of the XXXXColors style property
         * and XXXXRollOverColors style property, where XXXX= header,filter,pager,footer,renderer or blank.
         *
         * This function offers you the opportunity to hook into this mechanism, and draw
         * your own background using the graphics property of the FlexDataGridCell object,
         * that is passed (a) in parameter to this function. If this function returns false,
         * the default background is not drawn.
         * @type {Function}
         * @property cellCustomDrawFunction
         * @default null
         */
        this.cellCustomDrawFunction = null;

        /**
         * @private 
         */
        this._columnLockMode = "none";
        /**
         * @private
         */
        this.calcualtedMeasurements = {};

        /**
         * Whether or not to include the header text in the calculation of the column width.
         * @default false
         * @property columnWidthModeFitToContentExcludeHeader
         * @type {boolean}
         * @see http://www.reactdatagrid.com/demo/#/levelrenderer
         */
        this.columnWidthModeFitToContentExcludeHeader = false;

        /**
         * A function that takes column, data provider and returns a number that represents, in pixel terms
         * the width that needs to be applied to the column. The dataprovider is the subset of the dataset upto 
         * this.columnWidthModeFitToContentSampleSize
         * @default null
         * @type {Function}
         * @property columnWidthModeFitToContentWidthCalculator
         */
        this.columnWidthModeFitToContentWidthCalculator = null;
        /**
         * A value in pixels, that may be used to add/remove width
         * that gets calculated when you use the columnWidthMode property
         * Use if you have custom renderers or icons.
         * @type {boolean}
         * @property columnWidthOffset
         * @default 10
         */
        this.columnWidthOffset = 10;
        /**
         * The columnWidthMode property on the column specifies how the column widths are applied. This property defaults to "none".
         *
         * Flexicious Ultimate provides a rich mechanism to control column widths. Column widths are a complicated
         * topic because there are a number of scenarios and rules that we need to account for
         *
         * <ul>
         * <li>When there is a horizontal scroll (horizontalScrollPolicy=on or auto):
         * In this case, the columns are free to (much) take width as they need. Below is how the column widths should
         * handle in this case:
         * <ul>
         * <li>When columnWidthMode=none or fixed: The column will basically take the width specified by the width property</li>
         * <li>When columnWidthMode=fitToContent The column will take the width calculated by its contents. The grid identifies the longest
         * string to be displayed in this column, and sets the width of the column to this value.</li>
         * <li>When columnWidthMode=percent. This is not a valid setting when horizontalScrollPolicy is on or auto.
         * The setting will be ignored and the column will take the width specified by the width property.
         * When horizontalScrollPolicy is set to auto or on,
         * columnWidthMode=percent holds no meaning, since there are no fixed bounds to squish the columns within.</li>
         * </ul>
         * </li>
         * <li>When there is no horizontal scroll (horizontalScrollPolicy=off - this is the default):
         * <ul>
         * <li>When columnWidthMode=none: The column will take the width specified by the width property, and adjust for width (see sum of Column Widths exceeds Grid Width below).</li>
         * <li>When columnWidthMode=fixed: The column will  take the width specified by the width property, and not adjust for width.</li>
         * <li>When columnWidthMode=fitToContent: The column will take the width calculated by its contents, and adjust for width (see sum of Column Widths exceeds Grid Width below).</li>
         * <li>When columnWidthMode=percent: For these columns, the grid divides the remaining width after allocating
         * all fixed and fitToContent columns, on a percentage basis among all columns that have
         * columnWidthMode set to percent. PLEASE NOTE : If you set columnWidthMode='percent', also
         * set percentWidth. Also, ensure that the percentWidth of the columns adds up to a 100.</li>
         * </ul>
         * </li>
         * </ul>
         *
         * Finally, there are the below calculations once the column widths are allocated:
         *
         * <ul>
         * <li>Grid Width exceeds Sum of Column Widths:
         * The situation where the calcualted column widths do not add upto the grid with is also handled on basis of
         * the horizontalScrollPolicy.
         * <ul>
         * <li>When there is a horizontal scroll (horizontalScrollPolicy=on or auto): The last column extends to fill up all the remaining space. If you do not want your last column to resize, add a dummy column that has the following property (order is important) minWidth="1" width="1" paddingLeft="0" paddingRight="0"</li>
         * <li>When there is no horizontal scroll (horizontalScrollPolicy=off - this is the default): The extra width is divided proportionally between all the columns where columnWidthMode does not equal fixed.</li>
         * </ul>
         * </li>
         * </ul>
         *
         * <ul>
         * <li>Sum of Column Widths exceeds Grid Width:
         * Similarly, The situation where the allocated column widths exceed the width of the grid with is also handled on basis of
         * the horizontalScrollPolicy.
         * <ul>
         * <li>When there is a horizontal scroll (horizontalScrollPolicy=on or auto): We simply show a scroll bar, and no column widths are changed.</li>
         * <li>When there is no horizontal scroll (horizontalScrollPolicy=off - this is the default): This excess width
         * is reduced proportionally between columns where columnWidthMode does not equal fixed.</li>
         * </ul>
         * </li>
         * </ul>
         * Left and right locked columns do not support column width mode, it is ignored for these.
         * <br/>
         * Finally, with multi level grids, if the hierarchy's columns width in the top level is smaller than the next level's width
         * (and the horizontal scroll policy of the grid is "on"/"auto"), the horizontal scroller will be calculated only by the top
         * level's width, making some columns in the next level unreachable. The recommendation is to give a large column width to the last top level column,
         * which is large enough so that sum of column widths at top level is larger than the sum of column widths at the bottom level.
         * <br/>
         * Values : none,fixed,percent,fitToContent
         * @default auto
         */
        this._columnWidthMode = FlexDataGridColumn.COLUMN_WIDTH_MODE_NONE;
        /**
         *
         * @private
         * @see http://www.reactdatagrid.com/demo/#/dynamiccolumngroups
         */
        this.columnGroup = null;

        /**
         * @private
         */
        this.styleManager = null;

        /**
         * @type {boolean}
         * @property clearFilterOnIconClick
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/nested
         */
        this.clearFilterOnIconClick = false;
        /**
         * For filterControl=textInput, if you specify an Filter Icon Style, setting this variable to
         * true will clear show the icon only when there is text in the box.
         * @type {boolean}
         * @property showClearIconWhenHasText
         * @default false
         */
        this.showClearIconWhenHasText = false;
        /*if(value){
         filterTriggerEvent="enterKeyUp";
         }*/

        /**
         * When filterControl=textInput, setting this flag to true will set the enableAutoComplete flag on the associated TextInput control
         * Not Supported
         * @type {boolean}
         * @property enableFilterAutoComplete
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.enableFilterAutoComplete = false;

        /**
         * For list based item editors, setting this flag will use the dataprovider from the
         * filter mechanism for the item editor.
         * @type {boolean}
         * @property useFilterDataProviderForItemEditor
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/nested
         */
        this.useFilterDataProviderForItemEditor = false;

        /**
         * In Edit mode, we only apply the new value to the model object when the user
         * either hits the enter key or the tab key after typing input. This works fine
         * for most cases, but for ItemEditors like date fields or Select Boxes, you may wish to
         * apply the value on Change. In this case, set this flag to true.
         * @type {boolean}
         * @property itemEditorApplyOnValueCommit
         * @default true
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.itemEditorApplyOnValueCommit = true;

        /**
         * Flag to enable/disable icons within this cell.<br/>
         * For columns that have icons (with or without text), you can <br>
         * <ul>
         * <li>Set enableIcons=true.</li>
         * <li>If it will be the same icon for  cell, just set the "icon" style property on the column.</li>
         * <li>If it will be different icons based on some logic, you can either provide that via the iconField property, or specify an iconFunction.</li>
         * <li>If the cell will contain just an icon and no text, then set hideText/hideHeaderText to true.</li>
         * <li>If the header also needs an icon, use the headerIconUrl property.</li>
         * <li>Specify (iconLeft or iconRight) and (iconTop or iconBottom). If not specified, icon will be placed to the right middle of the cell.</li>
         * <li>If needed, Wire up the iconClick or iconMouseOver (delay)/iconMouseOut, and specify iconMouseOverDelay (default 250 msec).</li>
         * <li>Set enableIconHandCursor if needed.</li>
         * <li>Set the iconTooltip, iconTooltipFunction, iconTooltipRenderer (for custom interactive (needed) popovers).</li>
         * <li>Set the showIconOnRowHover and (needed) showIconOnCellHover (If you only wish to show the icon when the user hovers over the cell or the row.</li>
         * </ul>
         * @type {boolean}
         * @property enableIcon
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/iconcolumns
         */
        this.enableIcon = false;

        /**
         * @private 
         */
        this._iconField = "";
        /**
         * If enableIcon=true, icon function for the icon.
         * Default implementation has the following logic:<br/>
         * 1) If this is a header cell, returns the value of the headerIcon style property<br/>
         * 2) Else if this is a data cell, and the value of either the icon,overIcon or disabledIcon style property on basis of the passed in state is available, returns it.<br/>
         * 3) Else if this is a data cell, and the iconField is not empty, returns the value of the iconField for the row data<br/>
         * Should take an IFlexDataGridCell object.<br/>
         * @type {Function}
         * @property iconFunction
         * @default {defaultIconFunction}
         * @example 
         *   defaultIconFunction(cell, state) {
         *       if (typeof state == "undefined") state = "";
         * 
         *       if (cell.rowInfo.getIsHeaderRow()) {
         *           return this.getStyle("headerIcon");
         *       }
         *       return null;
         *  }
         * @see http://www.reactdatagrid.com/demo/#/iconcolumns
         */
        this.iconFunction = this.defaultIconFunction;
        /**
         * A function that allows you to position the icon when you provide an iconFunction or iconUrl.
         * By default, we use the iconLeft, iconRight, iconTop and iconBottom style properties. However incase
         * of hierarchical grids, you may want to control the positioning of the icon.
         * Takes a FlexDataGridCell object, and a Image object. The image object is already positioned x,y at the
         * result of the calculation based upon iconLeft, Right, Top and Bottom properties for you to make additional
         * adjustments.
         * @type {Function}
         * @property iconPlacementFunction
         * @default null
         */
        this.iconPlacementFunction = null;

        /**
         * The delay to wait until dispatching the iconMouseOver event when the user mouse the mouse over the icon.
         * Defaults to 500
         * @type {number}
         * @property iconMouseOverDelay
         * @default 500
         */
        this.iconMouseOverDelay = 500;
        /**
         * If enableIcon=true, enables hand cursor on the icon
         * @type {boolean}
         * @property iconHandCursor
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/iconcolumns
         */
        this.iconHandCursor = false;
        /**
         * If enableIcon=true, tooltip for the icon
         * @type {string}
         * @property iconToolTip
         * @default ""
         * @see http://www.reactdatagrid.com/demo/#/iconcolumns
         */
        this.iconToolTip = "";
        /**
         * If enableIcon=true, tooltip function for the icon. Default implementation returns the value of the iconToolTip. Should take an IFlexDataGridCell object.
         * @type {Function}
         * @property iconToolTipFunction
         * @default {defaultIconTooltipFunction}
         */
        this.iconToolTipFunction = this.defaultIconTooltipFunction;
        /**
         * The renderer factory for the icon tooltip. Defaults to null.
         * If null, we simply show a string tooltip, else we call the showToolTip function on the grid passing in an instance of the class factory specified here.
         * @type {flexiciousNmsp.ClassFactory}
         * @property iconTooltipRenderer
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/iconcolumns
         */
        this.iconTooltipRenderer = null;
        /**
         * If set to true, icon only shows when user hovers over the row.
         * @type {boolean}
         * @property showIconOnRowHover
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/iconcolumns
         */
        this.showIconOnRowHover = false;
        /**
         * If set to true, icon only shows when user hovers over the cell.
         * @type {boolean}
         * @property showIconOnCellHover
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/iconcolumns
         */
        this.showIconOnCellHover = false;
        /**
         * In scenarios where you only want to display the icon, set enableIcon=true, and hideText=true;
         * @type {boolean}
         * @property hideText
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/iconcolumns
         */
        this.hideText = false;
        /**
         * The header text is used in the column picker, export, etc. However, if you do not wish for it to be displayed in the view, set this to true. Used for IconColumns.
         * @type {boolean}
         * @property hideHeaderText
         * @default false
         */
        this.hideHeaderText = false;
        /**
         * To sort this column insensitive of case. Only set this to true for string data, if you set this to true for numeric or date data, the sort results will be incorrect.
         * @type {boolean}
         * @property sortCaseInsensitive
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.sortCaseInsensitive = false;
        /**
         * To sort this column numerically.
         * @type {boolean}
         * @property sortNumeric
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.sortNumeric = false;
        /**
         * For Grouped datagrids, searching on this column will cause match current item, as well as any of its children. If any children match, this item matches as well.
         * @type {boolean}
         * @property enableRecursiveSearch
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/selectionui2
         */
        this.enableRecursiveSearch = false;
        /**
         * For hierarchical datagrids, instead of having a separate column for icons,
         * this flag will be used to embed an icon based on disclosureOpenIcon and disclosureClosedIcon
         * within this column that will control the expand collapse of the row.
         * @type {boolean}
         * @property enableExpandCollapseIcon
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/selectionui2
         */
        this.enableExpandCollapseIcon = false;
        /**
         *  Performance optimization, is set to true if there is a complex fieldname or a labelFunction or labelFunction2
         * @type {boolean}
         * @property hasComplexDisplay
         * @default false
         */
        this.hasComplexDisplay = false;
        /**
         * A function that takes an item, and a filterExpression, and returns true or false on basis of whether the provided
         * item matches the filterExpression. Please note, the filterExpression has a pointer to the iFilterControl that it was created from.
         * @type {Function}
         * @property filterCompareFunction
         * @default null
         */
        this.filterCompareFunction = null;
        /**
         * A function that takes an item, and returns a processed value for comparison. For example, if your backend sends down strings that represents a date,
         * wire this function up and return a date object for comparison purposes.
         * @type {Function}
         * @property filterConverterFunction
         * @default null
         * @example 
         *   convertDate(item,col) {
         *       var dt = flexiciousNmsp.UIUtils.resolveExpression(item,col.getDataField()).toString();
         *       var date =flexiciousNmsp.UIUtils.getDateValue(dt,flexiciousNmsp.Constants.YMD_MASK); //will need to change this for
         *       return date;
         *   }
         * @see http://www.reactdatagrid.com/demo/#/xmldata
         */
        this.filterConverterFunction = null;
        /**
         * Prior to dispatching the icon roll over, we use a timer so accidental roll over does not cause popups.  But sometimes you need this behavior,
         * for example for custom cursors over icons. set this to false in that case
         * @type {boolean}
         * @property useIconRollOverTimer
         * @default true
         */
        this.useIconRollOverTimer = true;

        /**
         * Used by the doFormat method.
         * Formatter to use when format not equals none. On basis of format, a new formatter
         * is initialized. If you wish to override the format of the default formatter,
         * you can provide your own formatter here.
         * @type {flexiciousNmsp.Formatter}
         * @property formatter
         * @default null
         */
        this.formatter = null;

        /**
         * @type {flexiciousNmsp.FlexDataGridColumnLevel}
         * @property level
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/xmlgroupeddata
         */
        this.level = null;
        /**
         * Added to avoid flicker
         * @type {boolean}
         * @property enableLocalStyles
         * @default true
         */
        this.enableLocalStyles = true;
        /**
         * Value for the blank item in multi select combobox filter control
         * Please note, there is a space in the front of this string so it shows up
         * right below the "All" option.
         * @type {string}
         * @property blankValuesLabel
         * @default " [None] "
         */
        this.blankValuesLabel = " [None] ";

        /**
         * [Style(name="backgroundColor",  type="Array", arrayType="uint", format="Color", inherit="no")]
         *  Background color, the color of the fill area.
         *	@default null
         *  @type Array
         *  @property  backgroundColor
         *
         */

        this.backgroundColor = null;
        /**
         * [Style(name="color", type="uint", format="Color", inherit="yes")]
         *  Color of text in the component, including the component label.
         *
         *  The default value for the Halo theme is <code>0x0B333C</code>.
         *  The default value for the Spark theme is <code>0x000000</code>.
         *
         *	@default null
         *  @type uint
         *  @property  color
         *  @see http://www.reactdatagrid.com/demo/#/dynamiccolumngroups
         */

        this.color = null;
        /**
         * [Style(name="disabledColor", type="uint", format="Color", inherit="yes")]
         *  Color of text in the component if it is disabled.
         *
         *  @default 0xAAB3B3
         *  @type uint
         *  @property  disabledColor
         */

        this.disabledColor = null;
        /**
         * [Style(name="textAlign", type="String", enumeration="left,center,right", inherit="yes")]
         *  Alignment of text within a container.
         *  Possible values are <code>"left"</code>, <code>"right"</code>,
         *  or <code>"center"</code>.
         *
         *  <p>The default value for most components is <code>"left"</code>.
         *  For the FormItem component,
         *  the default value is <code>"right"</code>.
         *  For the Button, LinkButton, and AccordionHeader components,
         *  the default value is <code>"center"</code>, and
         *  this property is only recognized when the
         *  <code>labelPlacement</code> property is set to <code>"left"</code> or
         *  <code>"right"</code>.
         *  If <code>labelPlacement</code> is set to <code>"top"</code> or
         *  <code>"bottom"</code>, the text and any icon are centered.</p>
         *  @default null
         *  @type {String}
         *  @property  textAlign
         *  @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.textAlign = null;

        /**
         * @type {number}
         * @property multiColumnSortNumberWidth
         * @default null
         */
        this.multiColumnSortNumberWidth = null;
        /**
         * [Style(name="multiColumnSortNumberHeight", type="Number", inherit="no")]
         *  The height of  the numeric value representing the order of the column sort.
         *  @default 15
         *  @type Number
         *  @property  multiColumnSortNumberHeight
         *
         */

        this.multiColumnSortNumberHeight = null;
        /**
         * [Style(name="multiColumnSortNumberStyleName", type="String", inherit="no")]
         *  The name of a CSS style declaration for controlling other aspects of
         *  the appearance of the numeric value representing the order of the column sort.
         *  @default "multiColumnSortNumberStyle"
         *  @type String
         *  @property  multiColumnSortNumberStyleName
         *
         */

        this.multiColumnSortNumberStyleName = null;
        /**
         * [Style(name="columnGroupStyleName", type="String", inherit="no")]
         *  The name of a CSS style declaration for controlling other aspects of
         *  the appearance of the column groups.
         *  @default "columnGroupStyle"
         *  @type String
         *  @property  columnGroupStyleName
         *  @see http://www.reactdatagrid.com/demo/#/rowspancolspan
         */

        this.columnGroupStyleName = null;
        /**
         * [Style(name="headerStyleName", type="String", inherit="no")]
         *  The name of a CSS style declaration for controlling other aspects of
         *  the appearance of the column headers.
         *  @default "dataGridStyles"
         *  @type String
         *  @property  headerStyleName
         *  @see http://www.reactdatagrid.com/demo/#/rowspancolspan
         */

        this.headerStyleName = null;
        /**
         * [Style(name="footerStyleName", type="String", inherit="no")]
         *  The name of a CSS style declaration for controlling other aspects of
         *  the appearance of the column headers.
         *  @default "dataGridStyles"
         *  @type String
         *  @property  footerStyleName
         */

        this.footerStyleName = null;
        /**
         * [Style(name="verticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
         *  @property  verticalGridLines
         */

        this.verticalGridLines = null;
        /**
         * [Style(name="horizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
         *  @property  horizontalGridLines
         */

        this.horizontalGridLines = null;
        /**
         * [Style(name="verticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
         *  @property  verticalGridLineColor
         */

        this.verticalGridLineColor = null;
        /**
         * [Style(name="horizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
         *  @default null
         *  @type uint
         *  @property  horizontalGridLineColor
         */

        this.horizontalGridLineColor = null;
        /**
         * [Style(name="horizontalGridLineThickness", type="Number", format="Length", inherit="yes")]
         *  Thickness of the horizontal grid lines.
         *  @default 1
         *  @type Number
         *  @property  horizontalGridLineThickness
         */

        this.horizontalGridLineThickness = null;
        /**
         * [Style(name="verticalGridLineThickness", type="Number", format="Length", inherit="yes")]
         *  Thickness of the vertical grid lines.
         *  @default 1
         *  @type Number
         *  @property  verticalGridLineThickness
         */

        this.verticalGridLineThickness = null;
        /**
         * [Style(name="headerVerticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
         *  @property  headerVerticalGridLines
         */

        this.headerVerticalGridLines = null;
        /**
         * [Style(name="headerHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
         *  @property  headerHorizontalGridLines
         */

        this.headerHorizontalGridLines = null;
        /**
         * [Style(name="headerVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
         *  @property  headerVerticalGridLineColor
         */

        this.headerVerticalGridLineColor = null;
        /**
         * [Style(name="headerHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
         *  @default null
         *  @type uint
         *  @property  headerHorizontalGridLineColor
         */

        this.headerHorizontalGridLineColor = null;
        /**
         * [Style(name="headerHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the header horizontal grid lines.
         *  @default 1
         *  @type Number
         *  @property  headerHorizontalGridLineThickness
         */

        this.headerHorizontalGridLineThickness = null;
        /**
         * [Style(name="headerVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the header vertical grid lines.
         *  @default 1
         *  @type Number
         *  @property  headerVerticalGridLineThickness
         */

        this.headerVerticalGridLineThickness = null;
        /**
         * [Style(name="headerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border
         *  @default false
         *  @type Boolean
         *  @property  headerDrawTopBorder
         */

        this.headerDrawTopBorder = null;
        /**
         * [Style(name="columnGroupVerticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
         *  @property  columnGroupVerticalGridLines
         */

        this.columnGroupVerticalGridLines = null;
        /**
         * [Style(name="columnGroupHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
         *  @property  columnGroupHorizontalGridLines
         */

        this.columnGroupHorizontalGridLines = null;
        /**
         * [Style(name="columnGroupVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
         *  @property  columnGroupVerticalGridLineColor
         */

        this.columnGroupVerticalGridLineColor = null;
        /**
         * [Style(name="columnGroupHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
         *  @default 0x666666
         *  @type uint
         *  @property  columnGroupHorizontalGridLineColor
         */

        this.columnGroupHorizontalGridLineColor = null;
        /**
         * [Style(name="columnGroupHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the header horizontal grid lines.
         *  @default 1
         *  @type Number
         *  @property  columnGroupHorizontalGridLineThickness
         */

        this.columnGroupHorizontalGridLineThickness = null;
        /**
         * [Style(name="columnGroupVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the header vertical grid lines.
         *  @default 1
         *  @type Number
         *  @property  columnGroupVerticalGridLineThickness
         */

        this.columnGroupVerticalGridLineThickness = null;
        /**
         * [Style(name="columnGroupDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border
         *  @default false
         *  @type Boolean
         *  @property  columnGroupDrawTopBorder
         */

        this.columnGroupDrawTopBorder = null;
        /**
         * [Style(name="filterVerticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
         *  @property  filterVerticalGridLines
         */

        this.filterVerticalGridLines = null;
        /**
         * [Style(name="filterHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
         *  @property  filterHorizontalGridLines
         */

        this.filterHorizontalGridLines = null;
        /**
         * [Style(name="filterVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
         *  @property  filterVerticalGridLineColor
         */

        this.filterVerticalGridLineColor = null;
        /**
         * [Style(name="filterHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
         *  @default null
         *  @type uint
         *  @property  filterHorizontalGridLineColor
         */

        this.filterHorizontalGridLineColor = null;
        /**
         * [Style(name="filterHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the filter horizontal grid lines.
         *  @default 1
         *  @type Number
         *  @property  filterHorizontalGridLineThickness
         */

        this.filterHorizontalGridLineThickness = null;
        /**
         * [Style(name="filterVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the filter vertical grid lines.
         *  @default 1
         *  @type Number
         *  @property  filterVerticalGridLineThickness
         */

        this.filterVerticalGridLineThickness = null;
        /**
         * [Style(name="filterDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
         *  @property  filterDrawTopBorder
         */

        this.filterDrawTopBorder = null;
        /**
         * [Style(name="footerVerticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
         *  @property  footerVerticalGridLines
         */

        this.footerVerticalGridLines = null;
        /**
         * [Style(name="footerHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
         *  @property  footerHorizontalGridLines
         */

        this.footerHorizontalGridLines = null;
        /**
         * [Style(name="footerVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
         *  @property  footerVerticalGridLineColor
         */

        this.footerVerticalGridLineColor = null;
        /**
         * [Style(name="footerHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
         *  @default null
         *  @type uint
         *  @property  footerHorizontalGridLineColor
         */

        this.footerHorizontalGridLineColor = null;
        /**
         * [Style(name="footerHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the footer horizontal grid lines.
         *  @default 1
         *  @type Number
         *  @property  footerHorizontalGridLineThickness
         */

        this.footerHorizontalGridLineThickness = null;
        /**
         * [Style(name="footerVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the footer vertical grid lines.
         *  @default 1
         *  @type Number
         *  @property  footerVerticalGridLineThickness
         */

        this.footerVerticalGridLineThickness = null;
        /**
         * [Style(name="footerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
         *  @property  footerDrawTopBorder
         */

        this.footerDrawTopBorder = null;
        /**
         * [Style(name="pagerVerticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
         *  @property  pagerVerticalGridLines
         */

        this.pagerVerticalGridLines = null;
        /**
         * [Style(name="pagerHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
         *  @property  pagerHorizontalGridLines
         */

        this.pagerHorizontalGridLines = null;
        /**
         * [Style(name="pagerVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
         *  @property  pagerVerticalGridLineColor
         */

        this.pagerVerticalGridLineColor = null;
        /**
         * [Style(name="pagerHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
         *  @default null
         *  @type uint
         *  @property  pagerHorizontalGridLineColor
         */

        this.pagerHorizontalGridLineColor = null;
        /**
         * [Style(name="pagerHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the pager horizontal grid lines.
         *  @default 1
         *  @type Number
         *  @property  pagerHorizontalGridLineThickness
         */

        this.pagerHorizontalGridLineThickness = null;
        /**
         * [Style(name="pagerVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the pager vertical grid lines.
         *  @default 1
         *  @type Number
         *  @property  pagerVerticalGridLineThickness
         */

        this.pagerVerticalGridLineThickness = null;
        /**
         * [Style(name="pagerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
         *  @property  pagerDrawTopBorder
         */

        this.pagerDrawTopBorder = null;
        /**
         * [Style(name="rendererVerticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
         *  @property  rendererVerticalGridLines
         */

        this.rendererVerticalGridLines = null;
        /**
         * [Style(name="rendererHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
         *  @property  rendererHorizontalGridLines
         */

        this.rendererHorizontalGridLines = null;
        /**
         * [Style(name="rendererVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
         *  @property  rendererVerticalGridLineColor
         */

        this.rendererVerticalGridLineColor = null;
        /**
         * [Style(name="rendererHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
         *  @default null
         *  @type uint
         *  @property  rendererHorizontalGridLineColor
         */

        this.rendererHorizontalGridLineColor = null;
        /**
         * [Style(name="rendererHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the renderer horizontal grid lines.
         *  @default 1
         *  @type Number
         *  @property  rendererHorizontalGridLineThickness
         */

        this.rendererHorizontalGridLineThickness = null;
        /**
         * [Style(name="rendererVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the renderer vertical grid lines.
         *  @default 1
         *  @type Number
         *  @property  rendererVerticalGridLineThickness
         */

        this.rendererVerticalGridLineThickness = null;
        /**
         * [Style(name="rendererDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
         *  @property  rendererDrawTopBorder
         */

        this.rendererDrawTopBorder = null;
        /**
         * [Style(name="headerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
         *  @property  headerDrawTopBorder
         */

        this.headerDrawTopBorder = null;
        /**
         * [Style(name="filterDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
         *  @property  filterDrawTopBorder
         */

        this.filterDrawTopBorder = null;
        /**
         * [Style(name="footerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
         *  @property  footerDrawTopBorder
         */

        this.footerDrawTopBorder = null;
        /**
         * [Style(name="pagerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
         *  @property  pagerDrawTopBorder
         */

        this.pagerDrawTopBorder = null;
        /**
         * [Style(name="rendererDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
         *  @property  rendererDrawTopBorder
         */

        this.rendererDrawTopBorder = null;
        /**
         * [Style(name="selectionColor", type="uint", format="Color", inherit="yes")]
         *  The color of the background for the row when the user selects
         *  an item renderer in the row.
         *  @default null
         *  @type uint
         *  @property  selectionColor
         *  The default value for the Halo theme is <code>0x7FCEFF</code>.
         *  The default value for the Spark theme is <code>0xA8C6EE</code>.
         */

        this.selectionColor = null;
        /**
         * [Style(name="color", type="uint", format="Color", inherit="yes")]
         *  Color of text in the component, including the component label.
         *
         *  The default value for the Halo theme is <code>0x0B333C</code>.
         *  The default value for the Spark theme is <code>0x000000</code>.
         *  @default null
         *  @type uint
         *  @property  color
         *
         */

        this.color = null;
        /**
         * [Style(name="sortArrowSkin", type="Class", inherit="no")]
         *  The class to use as the skin for the arrow that indicates the column sort
         *  direction.
         *  The default value for the Halo theme is <code>mx.skins.halo.DataGridSortArrow</code>.
         *  The default value for the Spark theme is <code>mx.skins.spark.DataGridSortArrow</code>.
         *  @default null
         *  @type Class
         *  @property  sortArrowSkin
         */

        this.sortArrowSkin = null;
        /**
         * [Style(name="paddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
         *  @type Number
         *  @property  paddingLeft
         *  @see http://www.reactdatagrid.com/demo/#/nested
         */

        this.paddingLeft = null;
        /**
         * [Style(name="paddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
         *  @property  paddingRight
         *  @see http://www.reactdatagrid.com/demo/#/selectionui2
         */

        this.paddingRight = null;
        /**
         * [Style(name="paddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  paddingTop
         */

        this.paddingTop = null;
        /**
         * [Style(name="paddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  paddingBottom
         *  @see http://www.reactdatagrid.com/demo/#/itemrenderer
         */

        this.paddingBottom = null;
        /**
         * [Style(name="headerPaddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
         *  @type Number
         *  @property  headerPaddingLeft
         *  @see http://www.reactdatagrid.com/demo/#/selectionui2
         */

        this.headerPaddingLeft = null;
        /**
         * [Style(name="headerPaddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
         *  @property  headerPaddingRight
         */

        this.headerPaddingRight = null;
        /**
         * [Style(name="headerPaddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  headerPaddingTop
         */

        this.headerPaddingTop = null;
        /**
         * [Style(name="headerPaddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  headerPaddingBottom
         */

        this.headerPaddingBottom = null;
        /**
         * [Style(name="footerPaddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
         *  @type Number
         *  @property  footerPaddingLeft
         */

        this.footerPaddingLeft = null;
        /**
         * [Style(name="footerPaddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
         *  @property  footerPaddingRight
         */

        this.footerPaddingRight = null;
        /**
         * [Style(name="footerPaddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  footerPaddingTop
         */

        this.footerPaddingTop = null;
        /**
         * [Style(name="footerPaddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  footerPaddingBottom
         */

        this.footerPaddingBottom = null;
        /**
         * [Style(name="filterPaddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
         *  @type Number
         *  @property  filterPaddingLeft
         *  @see http://www.reactdatagrid.com/demo/#/selectionui2
         */

        this.filterPaddingLeft = null;
        /**
         * [Style(name="filterPaddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
         *  @property  filterPaddingRight
         */

        this.filterPaddingRight = null;
        /**
         * [Style(name="filterPaddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  filterPaddingTop
         */

        this.filterPaddingTop = null;
        /**
         * [Style(name="filterPaddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  filterPaddingBottom
         */

        this.filterPaddingBottom = null;
        /**
         * [Style(name="pagerPaddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
         *  @type Number
         *  @property  pagerPaddingLeft
         */

        this.pagerPaddingLeft = null;
        /**
         * [Style(name="pagerPaddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
         *  @property  pagerPaddingRight
         */

        this.pagerPaddingRight = null;
        /**
         * [Style(name="pagerPaddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  pagerPaddingTop
         */

        this.pagerPaddingTop = null;
        /**
         * [Style(name="pagerPaddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  pagerPaddingBottom
         */

        this.pagerPaddingBottom = null;
        /**
         * [Style(name="rendererPaddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
         *  @type Number
         *  @property  rendererPaddingLeft
         */

        this.rendererPaddingLeft = null;
        /**
         * [Style(name="rendererPaddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
         *  @property  rendererPaddingRight
         */

        this.rendererPaddingRight = null;
        /**
         * [Style(name="rendererPaddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  rendererPaddingTop
         */

        this.rendererPaddingTop = null;
        /**
         * [Style(name="rendererPaddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
         *  @property  rendererPaddingBottom
         */

        this.rendererPaddingBottom = null;
        /**
         * [Style(name="alternatingItemColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The colors to use for the backgrounds of the items in the grid.
         *  The value is an array of two or more colors.
         *  The backgrounds of the list items alternate among the colors in the array.
         *  @default undefined
         *  @type Array
         *  @property  alternatingItemColors
         */

        this.alternatingItemColors = null;
        /**
         * [Style(name="alternatingTextColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The colors to use for the text of the items in the grid.
         *  The value is an array of two colors.
         *  The text color of the list items alternate among the colors in the array.
         *  @default [ #000000, #000000]
         *  @type Array
         *  @property  alternatingTextColors
         */

        this.alternatingTextColors = null;
        /**
         * [Style(name="editItemColor", type="Array", arrayType="uint", format="Color")]
         *  The colors to use for the backgrounds of the items in the grid in the editable mode.
         *  The value is an array two colors.
         *  @default undefined
         *  @type Array
         *  @property  editItemColor
         */

        this.editItemColor = null;
        /**
         * [Style(name="editTextColor", format="Color")]
         *  The colors to use for the text of the items in the editable grid.
         *  @default null
         *  @type null
         *  @property  editTextColor
         */

        this.editTextColor = null;
        /**
         * [Style(name="selectionColor", type="uint", format="Color", inherit="yes")]
         *  The color of the background of a renderer when the user selects it.
         *
         *  The default value for the Halo theme is <code>0x7FCEFF</code>.
         *  The default value for the Spark theme is <code>0xA8C6EE</code>.
         *
         *  @default null
         *  @type uint 
         *  @property  selectionColor
         */

        this.selectionColor = null;
        /**
         * [Style(name="selectionDisabledColor", type="uint", format="Color", inherit="yes")]
         *  The color of the background of a renderer when the component is disabled.
         *  @default null
         *  @type uint
         *  @property  selectionDisabledColor
         */

        this.selectionDisabledColor = null;
        /**
         * [Style(name="disclosureOpenIcon", type="Class", format="EmbeddedFile", inherit="no")]
         *  The icon that is displayed next to an open branch node of the navigation tree.
         *  The default value is <code>TreeDisclosureOpen</code> in the assets.swf file.
         *  @default null
         *  @type Class
         *  @property  disclosureOpenIcon
         */

        this.disclosureOpenIcon = null;
        /**
         * [Style(name="disclosureClosedIcon", type="Class", format="EmbeddedFile", inherit="no")]
         *  The icon that is displayed next to a closed branch node of the navigation tree.
         *
         *  The default value is <code>TreeDisclosureClosed</code> in the assets.swf file.
         *  @default null
         *  @type Class
         *  @property  disclosureClosedIcon
         */

        this.disclosureClosedIcon = null;
        /**
         * [Style(name="columnGroupOpenIcon", type="Class", format="EmbeddedFile", inherit="no")]
         *  The icon that is displayed next to an open column group.
         *
         *  The default value is <code>TreeDisclosureOpen</code> in the assets.swf file.
         *  @default null
         *  @type Class
         *  @property  columnGroupOpenIcon
         */

        this.columnGroupOpenIcon = null;
        /**
         * [Style(name="columnGroupClosedIcon", type="Class", format="EmbeddedFile", inherit="no")]
         *  The icon that is displayed next to a closed column group.
         *
         *  The default value is <code>TreeDisclosureClosed</code> in the assets.swf file.
         *  @default null
         *  @type Class
         *  @property  columnGroupClosedIcon
         */

        this.columnGroupClosedIcon = null;
        /**
         * [Style(name="rollOverColor", type="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the row.
         *
         *  The default value for the Halo theme is <code>0xB2E1FF</code>.
         *  The default value for the Spark theme is <code>0xCEDBEF</code>.
         *  @default null
         *  @type uint
         *  @property  rollOverColor
         */

        this.rollOverColor = null;
        /**
         * [Style(name="activeCellColor", type="uint", format="Color", inherit="yes")]
         *  The color of the cell directly under the mouse or if using keyboard navigation, current keyboard seed.
         *  @default #000000
         *  @type uint
         *  @property  activeCellColor
         */

        this.activeCellColor = null;
        /**
         * [Style(name="footerColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  An array of two colors used to draw the footer background gradient.
         *  The first color is the top color.
         *  The second color is the bottom color.
         *  @default [0xCFCFCF, 0xCCCCCC]
         *  @type Array
         *  @property  footerColors
         */

        this.footerColors = null;
        /**
         * [Style(name="footerRollOverColors",  type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the footer.
         *  The default value is [0xCCCCCC,0xCFCFCF]
         *  @default null
         *  @type Array
         *  @property  footerRollOverColors
         */

        this.footerRollOverColors = null;
        /**
         * [Style(name="filterColors",  type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background for the filter.
         *  The default value is [0xCFCFCF,0xCFCFCF]
         *  @default null
         *  @type Array
         *  @property  filterColors
         */

        this.filterColors = null;
        /**
         * [Style(name="filterRollOverColors",  type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the filter.
         *  The default value is [0xCFCFCF,0xCFCFCF]
         *  @default null
         *  @type Array
         *  @property  filterRollOverColors
         */

        this.filterRollOverColors = null;
        /**
         * [Style(name="headerColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  An array of two colors used to draw the header background gradient.
         *  The first color is the top color.
         *  The second color is the bottom color.
         *  @default [0xFFFFFF, 0xE6E6E6]
         *  @type Array
         *  @property  headerColors
         */

        this.headerColors = null;
        /**
         * [Style(name="headerRollOverColors",  type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the header.
         *
         *  The default value is [0xE6E6E6,0xFFFFFF]  
         *	@default null
         *  @type Array
         *  @property  headerRollOverColors
         */

        this.headerRollOverColors = null;
        /**
         * [Style(name="columnGroupColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  An array of two colors used to draw the Column Groups background gradient.
         *  The first color is the top color.
         *  The second color is the bottom color.
         *  @default [0xFFFFFF, 0xE6E6E6]
         *  @type Array
         *  @property  columnGroupColors
         */

        this.columnGroupColors = null;
        /**
         * [Style(name="columnGroupRollOverColors",  type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the Column Groups.
         *
         *  The default value is [0xE6E6E6,0xFFFFFF]
         *	@default null
         *  @type Array
         *  @property  columnGroupRollOverColors
         */

        this.columnGroupRollOverColors = null;
        /**
         * [Style(name="rendererColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  An array of two colors used to draw the renderer background gradient.
         *  The first color is the top color.
         *  The second color is the bottom color.
         *  @default [0xFFFFFF, 0xFFFFFF]
         *  @type Array
         *  @property  rendererColors
         */

        this.rendererColors = null;
        /**
         * * [Style(name="rendererRollOverColors", type="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over a level renderer.
         *  The default value is [0xFFFFFF,0xFFFFFF]
         *  @default [0xFFFFFF, 0xFFFFFF]
         *  @type Array
         *  @property  rendererRollOverColors
         */

        this.rendererRollOverColors = null;
        /**
         * [Style(name="pagerColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  An array of two colors used to draw the pager background gradient.
         *  The first color is the top color.
         *  The second color is the bottom color.
         *  @default [0xCCCCCC, 0xCCCCCC]
         *  @type Array
         *  @property  pagerColors
         */

        this.pagerColors = null;
        /**
         * [Style(name="pagerRollOverColors",type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the pager.
         *  The default value is [0xE6E6E6,0xFFFFFF]
         *  @default [0xE6E6E6,0xFFFFFF]
         *  @type Array
         *  @property   pagerRollOverColors
         */

        this.pagerRollOverColors = null;
        /**
         * [Style(name="textSelectedColor", type="uint", format="Color", inherit="yes")]
         *  Color of the text when the user selects a row.
         *  @default 0x000000
         *  @type uint
         *  @property  textSelectedColor
         */

        this.textSelectedColor = null;
        /**
         * [Style(name="textRollOverColor", type="uint", format="Color", inherit="yes")]
         *  Color of the text when the user rolls over a row.
         *
         *  @default 0x000000
         *  @type uint
         *  @property  textRollOverColor
         */

        this.textRollOverColor = null;
        /**
         * [Style(name="textDisabledColor", type="uint", format="Color", inherit="yes")]
         *  The color of the text of a renderer when the component is disabled.
         *  @default 0xDDDDDD
         *  @type uint
         *  @property  textDisabledColor
         */

        this.textDisabledColor = null;
        /**
         * [Style(name="columnTextColor", type="uint", format="Color", inherit="yes")]
         *  The color of the cell text of each cell in this column.
         *  Usually these colors are controlled via the alternatingTextColors
         *  style property of the level or the grid. In case you wish to have this
         *  column have a specific text color, you may use this property.
         *  The default value is null
         *  @default null
         *  @type uint
         *  @property  columnTextColor
         */

        this.columnTextColor = null;
        /**
         * [Style(name="filterIcon")]
         *  When filterControl=textInput, the icon that is displayed inside the textbox, based on the insideIconPosition property.
         *  @default null
         *  @type null
         *  @property  filterIcon
         */

        this.filterIcon = null;
        /**
         * [Style(name="filterIconPosition", type="String", enumeration="left,right")]
         *  When filterControl=textInput, the position of the icon that is displayed inside the textbox.
         *  Default 'right'
         *  @default null
         *  @type String
         *  @property  filterIconPosition
         */

        this.filterIconPosition = null;
        /**
         * * [Style(name="icon")]
         * When enableIcon=true, this property specifies the icon to display.
         * If the icon is dynamic, use the dataField to specify the url instead.
         *  @default null
         *  @type null
         *  @property  icon
         */

        this.icon = null;
        /**
         * * [Style(name="overIcon")]
         * When enableIcon=true, this property specifies the icon to display.
         * If the icon is dynamic, use the dataField to specify the url instead.
         *  @default null
         *  @type null
         *  @property  overIcon
         */

        this.overIcon = null;
        /**
         * * [Style(name="disabledIcon")]
         * When enableIcon=true, this property specifies the icon to display when enabled=false.
         * If the icon is dynamic, use the dataField to specify the url instead.
         *  @default null
         *  @type null
         *  @property  disabledIcon
         */

        this.disabledIcon = null;
        /**
         * * [Style(name="headerIcon")]
         * When enableIcon=true, this property specifies the icon to display.
         * If the icon is dynamic, use the dataField to specify the url instead.
         *  @default null
         *  @type null
         *  @property  headerIcon
         */

        this.headerIcon = null;
        /**
         * [Style(name="iconLeft", type="Number")]
         * When enableIcon=true, To position the icon correctly, you can set either iconLeft or iconRight,
         * and iconTop or iconBottom
         *  @default null
         *  @type Number
         *  @property  iconLeft
         */

        this.iconLeft = null;
        /**
         * [Style(name="iconWidth", type="Number")]
         * When enableIcon=true, To position the icon correctly, we need to know the height and width of the icon
         * Since the icon is not yet drawn, sometimes we do not know how big the icon will be. This property
         * lets you specify that.
         *  @default null
         *  @type Number
         *  @property  iconWidth
         */

        this.iconWidth = null;
        /**
         * [Style(name="iconHeight", type="Number")]
         * When enableIcon=true, To position the icon correctly, we need to know the height and width of the icon
         * Since the icon is not yet drawn, sometimes we do not know how big the icon will be. This property
         * lets you specify that.
         *  @default null
         *  @type Number
         *  @property  iconHeight
         */

        this.iconHeight = null;
        /**
         * [Style(name="iconRight", type="Number")]
         * When enableIcon=true, To position the icon correctly, you can set either iconLeft or iconRight,
         * and iconTop or iconBottom
         *  @default null
         *  @type Number
         *  @property  iconRight
         */

        this.iconRight = null;
        /**
         * [Style(name="iconTop", type="Number")]
         * When enableIcon=true, To position the icon correctly, you can set either iconLeft or iconRight,
         * and iconTop or iconBottom
         *  @default null
         *  @type Number
         *  @property  iconTop
         */

        this.iconTop = null;
        /**
         * [Style(name="iconBottom", type="Number")]
         * When enableIcon=true, To position the icon correctly, you can set either iconLeft or iconRight,
         * and iconTop or iconBottom
         *  @default null
         *  @type Number
         *  @property  iconBottom
         */

        this.iconBottom = null;
        /**
         * [Style(name="expandCollapseIconLeft", type="Number")]
         * When enableExpandCollapseIcon=true, To position the icon correctly, you can set either expandCollapseIconLeft or expandCollapseIconRight,
         * and expandCollapseIconTop or expandCollapseIconBottom
         *  @default null
         *  @type Number
         *  @property  expandCollapseIconLeft
         */

        this.expandCollapseIconLeft = null;
        /**
         * [Style(name="expandCollapseIconRight", type="Number")]
         * When enableExpandCollapseIcon=true, To position the icon correctly, you can set either expandCollapseIconLeft or expandCollapseIconRight,
         * and expandCollapseIconTop or expandCollapseIconBottom
         *  @default null
         *  @type Number
         *  @property  expandCollapseIconRight
         */

        this.expandCollapseIconRight = null;
        /**
         * [Style(name="expandCollapseIconTop", type="Number")]
         * When enableExpandCollapseIcon=true, To position the icon correctly, you can set either expandCollapseIconLeft or expandCollapseIconRight,
         * and expandCollapseIconTop or expandCollapseIconBottom
         *  @default null
         *  @type Number
         *  @property  expandCollapseIconTop
         */

        this.expandCollapseIconTop = null;
        /**
         * [Style(name="expandCollapseIconBottom", type="Number")]
         * When enableExpandCollapseIcon=true, To position the icon correctly, you can set either expandCollapseIconLeft or expandCollapseIconRight,
         * and expandCollapseIconTop or expandCollapseIconBottom
         *  @default null
         *  @type Number
         *  @property  expandCollapseIconBottom
         */

        this.expandCollapseIconBottom = null;
        /**
         * [Style(name="editorStyleName", type="String")]
         * Style to apply to the item editor.
         *  @default null
         *  @type String
         *  @property  editorStyleName
         */

        this.editorStyleName = null;

        /**
         * in which priority level column should be hidden.
         * if -1 then column never hide from the responsive behaviour.
         * if 1 then column hide if required, when grid width < ResponsiveBehavior.columnPriority1WidthTreshold
         * if 2 then column hide if required, when grid width < ResponsiveBehavior.columnPriority2idthTreshold
         * if 3 then column hide if required, when grid width < ResponsiveBehavior.columnPriority3WidthTreshold
         * if 4 then column hide if required, when grid width < ResponsiveBehavior.columnPriority4WidthTreshold
         * if 5 then column hide if required, when grid width < ResponsiveBehavior.columnPriority5WidthTreshold
         * @type {number}
         * @property responsivePriority
         * @default -1
         */
        this.responsivePriority = -1;

        /**
         * A flag used to the identify that the responsive behaviour touched this columns visiblity,
         * So later we can restore the column visiblity on grid resize.
         * 
         * @type {boolean}
         * @property responseVisibleDirty
         * @default false
         */
        this.responseVisibleDirty = false;

        /**
         * Exclude this column from transposed mode.
         * 
         * @type {boolean}
         * @property excludeFromTransposedMode
         * @default false
         */
        this.excludeFromTransposedMode = false;


    }

    /**
     * @method getClassNames
     */
    getClassNames() {
        return ["FlexDataGridColumn", "CSSStyleDeclaration", "IDataGridFilterColumn"];
    }

    /**
     * The doFormat method returns formatted value on basis of the format properties (See description below)<br/>
     * If the input is null, returns null.
     * If there is a formatter specified, uses the formatter's format method to return a formatted value.
     * If the format property is specified and no formatter is specified, the column will instantiate
     * a formatter on basis of the value of the format property. The property can be one of 5 constants.
     * <ul>
     * <li>date (FORMAT_DATE) - Uses the DateUtils.MEDIUM_DATE_MASK format MMM D, YYYY - Can be customized using formatterDateString Property</li>
     * <li>datetime (FORMAT_DATE_TIME) - Uses the DateUtils.LONG_DATE_MASK+ " " + DateUtils.MEDIUM_TIME_MASK format MMM D, YYYY L:NN:SS A - Can be customized using formatterDateString Property</li>
     * <li>time (FORMAT_TIME) - Uses the DateUtils.LONG_TIME_MASK format. L:NN:SS A TZD - Can be customized using formatterDateString Property</li>
     * <li>currency (FORMAT_CURRENCY) - Uses formatterCurrencySymbol and formatterPrecision properties for the currency formatter. </li>
     * <li>number (FORMAT_NUMBER) - Uses formatterPrecision properties for the number formatter.</li>
     * </ul>
     * Based on the formatter created above, the value will be returned.
     *
     * @param retVal	String to format
     * @return 			Formatted string.
     * @method doFormat
     */
    doFormat(retVal) {

        if (retVal == null) return "";
        let fmt = this.formatter;
        if (fmt != null)
            return fmt.format(retVal);

        if (this.format != FlexDataGridColumn.FORMAT_NONE) {
            if (this.format == FlexDataGridColumn.FORMAT_DATE) {
                const dateFormat = new flexiciousNmsp.DateFormatter();
                dateFormat.formatString = this.formatterDateFormatString || Constants.MEDIUM_DATE_MASK;
                fmt = dateFormat;
            } else if (this.format == FlexDataGridColumn.FORMAT_DATE_TIME) {
                const dateFormat1 = new flexiciousNmsp.DateFormatter();
                dateFormat1.formatString = this.formatterDateFormatString || (`${Constants.LONG_DATE_MASK} ${Constants.MEDIUM_TIME_MASK}`);
                fmt = dateFormat1;
            } else if (this.format == FlexDataGridColumn.FORMAT_TIME) {
                const dateFormat2 = new flexiciousNmsp.DateFormatter();
                dateFormat2.formatString = this.formatterDateFormatString || Constants.LONG_TIME_MASK;
                fmt = dateFormat2;
            } else if (this.format == FlexDataGridColumn.FORMAT_CURRENCY) {
                retVal = UIUtils.toStringSafe(retVal);
                const cf = new flexiciousNmsp.CurrencyFormatter();
                cf.precision = this.formatterPrecision;
                cf.currencySymbol = this.formatterCurrencySymbol;
                fmt = cf;
            } else if (this.format == FlexDataGridColumn.FORMAT_NUMBER) {
                retVal = UIUtils.toStringSafe(retVal);
                const nf = new flexiciousNmsp.NumberFormatter();
                nf.precision = this.formatterPrecision;
                fmt = nf;
            }
            return fmt.format(retVal);
        }
        return retVal;

    }

    /**
     * @method labelFunctionSortCompare
     * @param {object} obj1
     * @param {object} obj2
     * @return {boolean} 
     */
    labelFunctionSortCompare(obj1, obj2) {

        const value1 = this.itemToLabel(obj1);
        const value2 = this.itemToLabel(obj2);
        return (value1 < value2);

    }

    /**
     *  Returns the String that the item renderer displays for the given data object.
     *  0) If there is a linktext property specified on the column, this function returns that.
     *  1) If the DataGridColumn or its DataGrid control
     *  has a non-null <code>labelFunction</code>
     *  property, it applies the function to the data object.
     *  2) If the DataGridColumn or its DataGrid control
     *  has a non-null <code>labelFunction2</code>
     *  property, it applies the function to the data object.
     *  3) Otherwise, the method extracts the contents of the field specified by the
     *  <code>dataField</code> property, or gets the string value of the data object.
     *  4) If the method cannot convert the parameter to a String, it returns a
     *  single space.
     *  @return Displayable String based on the data.
     *  @param {object} data
     *  @param {flexiciousNmsp.FlexDataGridCell} cell
     *  @method itemToLabel
     */
    itemToLabel(data, cell) {
        if (typeof cell == "undefined") cell = null;

        if (!data)
            return "";
        if (!this.hasComplexDisplay && this.getDataField()) {
            try {
                return this.doFormat(data[this.getDataField()]);
            } catch (e) {
                return "";
            }
        }
        if (this.linkText)
            return this.linkText;
        if (this._labelFunction != null)
            return this.doFormat(this._labelFunction(data, this));
        if (this._labelFunction2 != null)
            return this.doFormat(this._labelFunction2(data, this, cell));
        if (this.getDataField())
            return this.doFormat(UIUtils.resolveExpression(data, this.getDataField()));
        return "";

    }

    /**
     * @method getAdjacentColumn
     * @param {number} num
     * @return {flexiciousNmsp.FlexDataGridColumn} 
     */
    getAdjacentColumn(num) {

        const cols = this.level.getVisibleColumns();
        const newIndex = this.getColIndex() + num;
        if (newIndex > 0 && newIndex < cols.length) {
            return cols[newIndex];
        }
        return null;

    }

    /**
     * Gets the style value for the provided style property at the current
     * level. If not defined at current level, gets the one defined at
     * the grid level.
     * @method getStyleValue
     * @param {string} styleProp
     * @return {object} 
     */
    getStyleValue(styleProp) {

        let result = this.getStyle(styleProp);
        if (result == undefined)
            result = this.level.getStyleValue(styleProp);
        return result;

    }

    /**
     *
     * @method getIndex
     * @return {number} 
     */
    getIndex() {

        if (this.calcualtedMeasurements["getIndex"] != null) {
            return this.calcualtedMeasurements["getIndex"];
        }
        this.calcualtedMeasurements["getIndex"] = this.level.getColumns().indexOf(this);
        return this.level.getColumns().indexOf(this);

    }

    /**
     * Depending on the specified chrome type, gets the appropriate renderer.
     * If truncateToFit or selectable is true, returns a label factory
     * else returns a text field factory.
     * @method deriveRenderer
     * @param {*} chromeType
     * @return {*}
     */
    deriveRenderer(chromeType) {
        if (typeof chromeType == "undefined") chromeType = 5/*RowPositionInfo.ROW_TYPE_DATA*/;

        const renderer = (chromeType == flexiciousNmsp.RowPositionInfo.ROW_TYPE_DATA) ? this.itemRenderer
            : (chromeType == flexiciousNmsp.RowPositionInfo.ROW_TYPE_FILL) ? FlexDataGridColumn.static_UIComponent
                : (chromeType == flexiciousNmsp.RowPositionInfo.ROW_TYPE_HEADER ||
                    chromeType == flexiciousNmsp.RowPositionInfo.ROW_TYPE_COLUMN_GROUP) ? this.headerRenderer
                    : (chromeType == flexiciousNmsp.RowPositionInfo.ROW_TYPE_FOOTER) ? this.footerRenderer
                        : null;
        return (renderer == null) ? FlexDataGridColumn.static_UIComponent : renderer;
    }

    /**
     * @method hasCustomItemRenderer
     * @return {boolean}
     */

    hasCustomItemRenderer() {
        return (this.itemRenderer != null);
    }

    /**
     * Whenever a tree datagrid event is dispatched at any column,
     * we also dispatch it from the level so if someone wants
     * to listen for an event at all columns...
     * @method dispatchEvent
     * @param event
     */
    dispatchEvent(event) {

        if (event.implementsOrExtends(('FlexDataGridEvent')) && this.level) {
            return this.level.dispatchEvent(event);
        }
        return flexiciousNmsp.EventDispatcher.prototype.dispatchEvent.apply(this, [event]);
    }

    /**
     * Returns this list of distinct values for this column,
     * in a list of arraycollection which will have properties
     * data and label.
     * @method getDistinctValues
     * @param {object} dp
     * @param {Array} collection
     * @param {Array} addedCodes
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} lvl
     * @return {Array}
     */
    getDistinctValues(dp, collection, addedCodes, lvl) {
        if(typeof collection == "undefined") collection = null;
        if(typeof addedCodes == "undefined") addedCodes = null;
        if(typeof lvl == "undefined") lvl = null;

        addedCodes=addedCodes==null?[]:addedCodes;
        collection=collection==null?[]:collection;
        lvl=lvl==null?this.level:lvl;
        for(var i=0;i<dp.length;i++)
        {
            var item=dp[i];
            if(lvl.nextLevel && lvl.nextLevel.reusePreviousLevelColumns){
                this.getDistinctValues(lvl.getChildren(item,
                    this.level && 
                    this.level.grid && 
                    this.level.grid.filterComboBoxDataProviderEnableReduce),collection,addedCodes,lvl.nextLevel);
            }
            var itemLabel;
            var collectionItem=null;
            var label=this.itemToLabel(item);
            if(label || (this.blankValuesLabel && item))
            {
                itemLabel = (!label?this.blankValuesLabel:this.filterComboBoxLabelFunction?this.filterComboBoxLabelFunction(item):label);
                collectionItem={"label":itemLabel, "data":label};
            }
            if(collectionItem!=null)
            {
                if(addedCodes.indexOf(itemLabel)<0){
                    collection.push(collectionItem);
                    addedCodes.push(itemLabel)
                }
            }

        }
        if(lvl.getNestDepth()==1){
            flexiciousNmsp.UIUtils.sortOn(collection,"label");
        }
        return collection;
    }

    /**
     * @method itemToLabelCommon
     * @param {object} item
     * @return {String}
     */
    itemToLabelCommon(item) {

        return this.itemToLabel(item);

    }

    /**
     * A function that is used when we have complex properties in the datafield.
     * @method isValidStyleValue
     * @param {*} val
     * @return {boolean} 
     */

    isValidStyleValue(val) {
        return val != null
    }

    /**
     * Default Icon tooltip function. Returns value of the iconToolTip string.
     * @method defaultIconFunction
     * @param {flexiciousNmsp.FlexDataGridDataCell} cell
     * @param {string} state
     */
    defaultIconFunction(cell, state) {
        if (typeof state == "undefined") state = "";

        if (cell.rowInfo.getIsHeaderRow()) {
            return this.getStyle("headerIcon");
        } else if (cell.rowInfo.getIsDataRow()) {
            const styleProp = state == "over" && this.getStyle("overIcon") ? "overIcon" : state == "disabled"
                && this.getStyle("disabledIcon") ? "disabledIcon" : "icon";
            const retVal = this.getStyle(styleProp);
            return retVal ? retVal : UIUtils.resolveExpression(cell.rowInfo.getData(), this.getIconField());
        }
        return null;

    }

    /**
     * Default Icon tooltip function. Returns value of the iconToolTip string.
     * @method defaultIconTooltipFunction
     * @param {flexiciousNmsp.FlexDataGridDataCell} cell
     * @return {string} 
     */
    defaultIconTooltipFunction(cell) {
        return this.iconToolTip;
    }

    /**
     *
     * @method setStyle
     * @param {string} styleProp
     * @param {*} value
     */
    setStyle(styleProp, value) {

        this[styleProp] = value;
    }

    /**
     * Returns true if the column is left or right locked.
     * @method getIsLocked
     * @return {boolean}
     */
    getIsLocked() {
        return this._columnLockMode != FlexDataGridColumn.LOCK_MODE_NONE;
    }

    /**
     * MX and Spark Grids implement the sortable differently, so to match the interface.
     * @method getIsSortable
     * @return {boolean} 
     */
    getIsSortable() {
        return this.sortable;
    }

    /**
     * @method getIsFirstRightLocked
     * @return {boolean}
     */
    getIsFirstRightLocked() {
        if (this.calcualtedMeasurements["isFirstRightLocked"] != null) {
            return this.calcualtedMeasurements["isFirstRightLocked"];
        }
        const arr = this.level.getRightLockedColumns();
        if (arr.length == 0) return false;
        const isFirst = arr.indexOf(this) == 0;
        this.calcualtedMeasurements["isFirstRightLocked"] = isFirst;
        return isFirst;
    }

    /**
     * @method getIsLastUnLocked
     * @return {boolean}
     */
    getIsLastUnLocked() {
        if (this.calcualtedMeasurements["isLastUnLocked"] != null) {
            return this.calcualtedMeasurements["isLastUnLocked"];
        }
        const arr = this.level.getUnLockedColumns();
        if (arr.length == 0) return false;
        const isLast = arr.indexOf(this) == arr.length - 1;
        this.calcualtedMeasurements["isLastUnLocked"] = isLast;
        return isLast;
    }

    /**
     * A control that implements IFilterControl
     * @return 
     * @method getFilterControl
     */
    getFilterControl() {
        return this._filterControlClass;
    }

    /**
     * The filter control to associate with this column. There are a number of built in filter controls, and you can even write your own by implementing
     * the IFilterControl interface. To be implemented by any control that can participate in the filtering mechanism.
     *
     * There are the following controls available out of the box:
     * <ul>
     * <li>TextInput</li>
     * <li>TriStateCheckBox</li>
     * <li>ComboBox</li>
     * <li>MultiSelectComboBox</li>
     * <li>DateComboBox</li>
     * <li>DateRangeBox</li>
     * <li>NumericRangeBox</li>
     * <li>NumericTextInput</li>
     * </ul>
     * You can write your own custom text input controls by extending any of these, or by implementing the IFilterControl, ICustomMatchFilterControl, or IDynamicFilterControl.
     * For an example of this, please look at http://blog.flexicious.com/post/ICustomMatchFilterControl-example.aspx
     * @method setFilterControl
     * @param {*} val
     */
    setFilterControl(val) {
        this._filterControlClass = val;
    }

    /**
     * @method getIsLastLeftLocked
     * @return {boolean} 
     */
    getIsLastLeftLocked() {
        return this.getIsLastleftLocked();
    }

    /**
     * @method getIsLastleftLocked
     * @return {boolean} 
     */
    getIsLastleftLocked() {
        if (this.calcualtedMeasurements["isLastleftLocked"] != null) {
            return this.calcualtedMeasurements["isLastleftLocked"];
        }
        const arr = this.level.getLeftLockedColumns();
        if (arr.length == 0) return false;
        const isLast = arr.indexOf(this) == arr.length - 1;
        this.calcualtedMeasurements["isLastleftLocked"] = this.getIsLeftLocked() && isLast;
        return this.getIsLeftLocked() && isLast;
    }

    /**
     * Returns true if the column is right locked.
     * @method getIsRightLocked
     * @return {boolean} 
     */
    getIsRightLocked() {
        return this._columnLockMode == FlexDataGridColumn.LOCK_MODE_RIGHT;
    }

    /**
     * Current X position of this column relative to the container for its cells.
     * @method getX
     * @return {number}
     */
    getX() {
        if (this.level._wxInvalid) {
            this.level.setColumnLevel();
        }
        return this._x;
    }

    /**
     *
     * @method setX
     * @param {number} value
     */
    setX(value) {
        if (this._x != value) {
            this._x = value;

            if (this.level) {
                const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.COLUMN_X_CHANGED,
                    this.level.grid, this.level, this);
                this.dispatchEvent(evt);
            }
        }
    }

    /**
     * In addition to the left locked and unlocked columns supported by the regular Flex DataGrid, Flexicious Ultimate supports right locked columns.<br/>
     * The grid supports the following selection modes:<br/>
     * 1) LOCK_MODE_LEFT<br/>
     * 2) LOCK_MODE_RIGHT<br/>
     * 3) LOCK_MODE_NONE<br/>
     * If set to left, the column will be locked to the left and not scroll horizontally.<br/>
     * If set to right, the column will be locked to the right, and not scroll horizontally.<br/>
     * Please note, column lock modes only make sense if the horizontalScrollPolicy of the grid is set
     * to "on" or "auto", since if there is no horizontal scrollbar, locked columns behave like unlocked columns,
     *  because there is no horizontal scroll. This horizontal scroll policy of the grid defaults to "off",
     * and changing this to "on" or "auto" affects the calculations that we perform in deciding the optimal
     *  column widths, in that columns are no longer "squished" or "expanded" to fit the visible width of the grid.
     * The left and right locked section for grids with flat data are wide enough to accommodate the left and right
     * locked columns respectively. For nested grids, both left and right locked sections are wide enough to accommodate
     * the widest left and right locked columns at each level, and for the leftLocked section, we add the width of the
     * disclosure cells which is set via the nestIndent property for each level (it gets multiplied by the nestDepth for a tree look).<br/>
     * There are a few restrictions with left and right locked columns. They cannot be re-ordered. <br/>
     * Resizing them results in resizing the section that they are in. <br/>
     * Finally, resizing them is very slightly slower than resizing the unlocked columns,
     * because we have to redraw the grid on basis of updated section widths.
     * @method getColumnLockMode
     * @return
     */
    getColumnLockMode() {
        return this._columnLockMode;
    }

    /**
     *
     * @param val
     * @method setColumnLockMode
     *
     */
    setColumnLockMode(val) {
        if (this._columnLockMode != val && this.level) {
            this.level.invalidateCache();
        }
        this._columnLockMode = val;
    }

    /**
     * A key used to uniquely identify a column. Defaults to the header text
     * Used in print and preference persistence to identify columns.
     * @return
     * @method  getUniqueIdentifier
     */
    getUniqueIdentifier() {
        return this._uniqueIdentifier ? this._uniqueIdentifier : this.getHeaderText() ? this.getHeaderText() : this.getDataField() ? this.getDataField() :
            (`Column ${this.getColIndex() + 1}`);
    }

    /**
     *
     * @method setUniqueIdentifier
     * @param {string} value
     */
    setUniqueIdentifier(value) {
        this._uniqueIdentifier = value;
    }

    /**
     * Similar to searchField, if you want to override what the database
     * actually sorts when you sort this column
     * @method getSortField
     * @return {string}
     */
    getSortField() {
        return this._sortField.length == 0 ? this.getDataField() : this._sortField;
    }

    /**
     *
     * @param val
     * @method setSortField
     *
     */
    setSortField(val) {
        this._sortField = val;
    }

    /**
     * The property on the model object that returns the url to the icon,
     * if you wish to show different icons for each cell.
     *
     * Defaults to data field. If there are scenarios where you wish to
     * show both the icon as well as the text, you can specify the iconField
     * and the dataField. Otherwise, where you show just the icon, you can leave the
     * iconField blank and have the data field specify the value of the icon
     * url.
     * @method getIconField
     * @return {string}
     */
    getIconField() {
        return this._iconField ? this._iconField : this.getDataField();
    }

    /**
     *
     * @method setIconField
     * @param  {string} value
     */
    setIconField(value) {
        this._iconField = value;
    }

    /**
     *
     * @method getColIndex
     * @return {number}
     */
    getColIndex() {
        if (this.calcualtedMeasurements["colIndex"] != null) {
            return this.calcualtedMeasurements["colIndex"];
        }
        this.calcualtedMeasurements["colIndex"] = this.level.getVisibleColumns().indexOf(this);
        return this.level.getVisibleColumns().indexOf(this);
    }

    /**
     * @method getOwner
     * @return {flexiciousNmsp.FlexDataGrid} 
     */
    getOwner() {
        return this.level.grid;
    }

    /**
     *  The width of the column, in pixels.
     *  @default 100
     *  @method getWidth
     * @return {number}
     *
     */
    getWidth() {

        return this._width;
    }

    /**
     *
     * @method setWidth
     * @param {number} value
     */
    setWidth(value) {
        if (this._width != value) {
            if (value < this._minWidth) return;
            this._width = value;
            if (this.level) {
                this.level._wxInvalid = true;
                if (this.getIsLocked() || (this.headerWordWrap && this.level.grid.variableHeaderHeight)) {
                    this.level.grid.setPreservePager(true);
                    this.level.grid.reDraw();
                } else {
                    this.level.grid.invalidateCellWidths();
                }
                this.level.grid.invalidateFiller();
            }

            const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.COLUMN_RESIZED, this.level ? this.level.grid : null, this.level, this);
            this.dispatchEvent(evt);
        }

    }

    /**
     * In addition to the left locked and unlocked columns supported by the regular Flex DataGrid, Flexicious Ultimate supports right locked columns.<br/>
     * The grid supports the following selection modes:<br/>
     * 1) LOCK_MODE_LEFT<br/>
     * 2) LOCK_MODE_RIGHT<br/>
     * 3) LOCK_MODE_NONE<br/>
     * If set to left, the column will be locked to the left and not scroll horizontally.<br/>
     * If set to right, the column will be locked to the right, and not scroll horizontally.<br/>
     * Please note, column lock modes only make sense if the horizontalScrollPolicy of the grid is set
     * to "on" or "auto", since if there is no horizontal scrollbar, locked columns behave like unlocked columns,
     *  because there is no horizontal scroll. This horizontal scroll policy of the grid defaults to "off",
     * and changing this to "on" or "auto" affects the calculations that we perform in deciding the optimal
     *  column widths, in that columns are no longer "squished" or "expanded" to fit the visible width of the grid.
     * The left and right locked section for grids with flat data are wide enough to accommodate the left and right
     * locked columns respectively. For nested grids, both left and right locked sections are wide enough to accommodate
     * the widest left and right locked columns at each level, and for the leftLocked section, we add the width of the
     * disclosure cells which is set via the nestIndent property for each level (it gets multiplied by the nestDepth for a tree look).<br/>
     * There are a few restrictions with left and right locked columns. They cannot be re-ordered. <br/>
     * Resizing them results in resizing the section that they are in. <br/>
     * Finally, resizing them is very slightly slower than resizing the unlocked columns,
     * because we have to redraw the grid on basis of updated section widths.
     * @method getColumnLockMode
     * @return 
     */
    getColumnLockMode() {
        return this._columnLockMode;
    }

    /**
     *
     * @param val
     * @method setColumnLockMode
     *
     */
    setColumnLockMode(val) {
        if (this._columnLockMode != val && this.level) {
            this.level.invalidateCache();
        }
        this._columnLockMode = val;
    }

    /**
     * Gets the nest depth of the associated level.
     * @return
     * @method getNestDepth
     */
    getNestDepth() {
        return this.level ? this.level.getNestDepth() : 1;
    }

    /**
    * @method getIsFirstUnLocked
    * @return {boolean}
    */
    getIsFirstUnLocked() {
        if (this.calcualtedMeasurements["isFirstUnLocked"] != null) {
            return this.calcualtedMeasurements["isFirstUnLocked"];
        }
        const arr = this.level.getUnLockedColumns();
        if (arr.length == 0) return false;
        const isLast = arr.indexOf(this) == 0;
        this.calcualtedMeasurements["isFirstUnLocked"] = isLast;
        return isLast;
    }

    /**
     * The field that the filter should search on. Defaults to the
     * dataField property from mx.controls.dataGridClasses.DataGridColumn.
     * But can be overridden if needed.
     *
     * For example, consider a grid bound to a list of Employee objects, with a departmentName
     * and departmentId property. This column shows the departmentName, with a combobox filter
     * or a multi-select filter using a list of departments as filterComboBoxDataProvider
     *
     * By default, the comparison will occur using the selected label of the filter control
     * being compared to the
     * departmentName property of each employee. This is allright if the
     * filterPageSortMode property of the com.flexicious.grids.ExtendedDataGrid is
     * set to "client", but if it is "server" ,
     * you would ideally want to have the ID of the selected department(s). In that case, you would
     * set the searchField property to "departmentId", filterComboBoxDataField to "departmentId",
     * filterComboBoxLabelField to "departmentName".
     * @method getSearchField
     * @return {string}
     *
     *
     */
    getSearchField() {
        return this._searchField.length == 0 ? this.getDataField() : this._searchField;
    }

    /**
     *
     * @param val
     *
     * @method setSearchField
     */
    setSearchField(val) {
        this._searchField = val;
    }

    /**
     *  Flag that indicates whether the column is visible.
     *  If <code>true</code>, the column is visible.
     *  @default true
     * @method getVisible
     * @return {boolean}
     */
    getVisible() {
        return this._visible;
    }

    /**
     *
     * @method setVisible
     * @param {boolean} value
     */
    setVisible(value) {
        if (this._visible != value) {

            this._visible = value;
            if (this.level) {
                for (const col of this.level._columns) {
                    col.calcualtedMeasurements = {};
                }

                this.level.invalidateCache();
                this.level._wxInvalid = true;
                this.level.grid.reDraw();

                var evt= new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.COLUMN_VISIBLE_CHANGED,this.level.grid,this.level,this);
                this.dispatchEvent(evt);
            }
        }
    }

    /**
     * For datagrids where there are multiple columns with the same header text,
     * this field may be specified so that the persistence mechanism does not
     * overwrite it.
     * @method getPersistenceKey
     * @return {string}
     * @deprecated - now defaults to the uniqueIdentifier.
     */
    getPersistenceKey() {
        return this.getUniqueIdentifier();
    }

    /**
     *
     * @param val
     * @method setPersistenceKey
     */
    setPersistenceKey(val) {
        this._persistenceKey = val;
    }

    /**
     * The actual control
     * to render inside the header column. This control must implement
     * the com.flexicious.controls.interfaces.filters.IFilterControl interface.
     * @return
     * @method getFilterRenderer
     */
    getFilterRenderer() {
        if (this._filterRenderer == null && this._filterControlClass.length > 0) {
            const c = eval(`flexiciousNmsp.${this._filterControlClass}`);
            this._filterRenderer = new flexiciousNmsp.ClassFactory(c, undefined);
            if (c.typeName == "MultiSelectComboBox")
                this.filterOperation = flexiciousNmsp.FilterExpression.FILTER_OPERATION_TYPE_IN_LIST;
        }
        else if (this._filterRenderer == null) {
            this._filterRenderer = UIUtils.UIComponentFactory;
        }
        return this._filterRenderer;
    }

    /**
     *
     * @param value
     * @method setFilterRenderer
     *
     */
    setFilterRenderer(value) {
        this._filterRenderer = value;
        this.dispatchEvent(new flexiciousNmsp.BaseEvent("filterRendererChanged"));
    }

    /**
     * Container for the item renderer. Needs to implement IFlexDataGridDataCell.
     * Defaults to FlexDataGridDataCell
     * @return
     * @method getDataCellRenderer
     */
    getDataCellRenderer() {
        return this._dataCellRenderer;
    }

    /**
     * @param value
     * @method  setDataCellRenderer
     */
    setDataCellRenderer(value) {
        this._dataCellRenderer = value;
    }

    /**
     * Returns true if the column is left locked.
     * @method  getIsLeftLocked
     * @return {boolean}
     */
    getIsLeftLocked() {
        return this._columnLockMode == FlexDataGridColumn.LOCK_MODE_LEFT;
    }

    /**
     *  Flag that indicates whether the items in the column are editable.
     * @method  getEditable
     * @return {boolean}
     */
    getEditable() {
        return this._editable && this.level.grid.editable;
    }

    /**
     * @method  setEditable
     * @param {boolean} value
     */
    setEditable(value) {
        this._editable = value;
    }

    /**
     * Unused, for legacy purposes only
     * @method  getClickBehavior
     */
    getClickBehavior() {
        return null;
    }

    /**
     * Similar to searchField, if you want to override what the database
     * actually sorts when you sort this column
     * @method  getSortFieldName
     * @return {string}
     */
    getSortFieldName() {
        return this.getSortField();
    }

    /**
     * @method  getIsLastrightLocked
     * @return {boolean}
     */
    getIsLastrightLocked() {
        if (this.calcualtedMeasurements["isLastrightLocked"] != null) {
            return this.calcualtedMeasurements["isLastrightLocked"];
        }
        const arr = this.level.getRightLockedColumns();
        if (arr.length == 0) return false;
        const isLast = arr.indexOf(this) == arr.length - 1;
        this.calcualtedMeasurements["isLastrightLocked"] = this.getIsRightLocked() && isLast;
        return this.getIsRightLocked() && isLast;
    }

    /**
     *  Text for the header of this column. By default, the DataGrid
     *  control uses the value of the <code>dataField</code> property
     *  as the header text. IF a headerTextFunction is specified, that
     * is used first.
     * @param {flexiciousNmsp.FlexDataGridDataCell} cell
     * @return {string}
     *
     */
    getHeaderText(cell) {
        return (this.headerLabelFunction2 != null) ? this.headerLabelFunction2(this, cell) :
            (this.headerTextFunction != null) ? this.headerTextFunction(this) : (this._headerText != null) ? this._headerText : this.getDataField();
    }

    /**
     *
     * @param {string} value
     * @method  setHeaderText
     */
    setHeaderText(value) {
        this._headerText = value ? value.toString() : "";
    }

    /**
     * Returns true if the column is not right locked or IFixedWidth.
     * @method  getIsElastic
     * @return {boolean}
     */
    getIsElastic() {
        return (!this.getIsRightLocked() && !(this.implementsOrExtends('IFixedWidth)')))
    }

    /**
     * For support in pickers
     * @return {string}
     * @method  getLabel
     *
     */
    getLabel() {
        return this.getHeaderText();
    }

    /**
     * The columnWidthMode property on the column specifies how the column widths are applied. This property defaults to "none".
     *
     * Flexicious Ultimate provides a rich mechanism to control column widths. Column widths are a complicated
     * topic because there are a number of scenarios and rules that we need to account for
     *
     * <ul>
     * <li>When there is a horizontal scroll (horizontalScrollPolicy=on or auto):
     * In this case, the columns are free to (much) take width as they need. Below is how the column widths should
     * handle in this case:
     * <ul>
     * <li>When columnWidthMode=none or fixed: The column will basically take the width specified by the width property</li>
     * <li>When columnWidthMode=fitToContent The column will take the width calculated by its contents. The grid identifies the longest
     * string to be displayed in this column, and sets the width of the column to this value.</li>
     * <li>When columnWidthMode=percent. This is not a valid setting when horizontalScrollPolicy is on or auto.
     * The setting will be ignored and the column will take the width specified by the width property.
     * When horizontalScrollPolicy is set to auto or on,
     * columnWidthMode=percent holds no meaning, since there are no fixed bounds to squish the columns within.</li>
     * </ul>
     * </li>
     * <li>When there is no horizontal scroll (horizontalScrollPolicy=off - this is the default):
     * <ul>
     * <li>When columnWidthMode=none: The column will take the width specified by the width property, and adjust for width (see sum of Column Widths exceeds Grid Width below).</li>
     * <li>When columnWidthMode=fixed: The column will  take the width specified by the width property, and not adjust for width.</li>
     * <li>When columnWidthMode=fitToContent: The column will take the width calculated by its contents, and adjust for width (see sum of Column Widths exceeds Grid Width below).</li>
     * <li>When columnWidthMode=percent: For these columns, the grid divides the remaining width after allocating
     * all fixed and fitToContent columns, on a percentage basis among all columns that have
     * columnWidthMode set to percent. PLEASE NOTE : If you set columnWidthMode='percent', also
     * set percentWidth. Also, ensure that the percentWidth of the columns adds up to a 100.</li>
     * </ul>
     * </li>
     * </ul>
     *
     * Finally, there are the below calculations once the column widths are allocated:
     *
     * <ul>
     * <li>Grid Width exceeds Sum of Column Widths:
     * The situation where the calcualted column widths do not add upto the grid with is also handled on basis of
     * the horizontalScrollPolicy.
     * <ul>
     * <li>When there is a horizontal scroll (horizontalScrollPolicy=on or auto): The last column extends to fill up all the remaining space. If you do not want your last column to resize, add a dummy column that has the following property (order is important) minWidth="1" width="1" paddingLeft="0" paddingRight="0"</li>
     * <li>When there is no horizontal scroll (horizontalScrollPolicy=off - this is the default): The extra width is divided proportionally between all the columns where columnWidthMode does not equal fixed.</li>
     * </ul>
     * </li>
     * </ul>
     *
     * <ul>
     * <li>Sum of Column Widths exceeds Grid Width:
     * Similarly, The situation where the allocated column widths exceed the width of the grid with is also handled on basis of
     * the horizontalScrollPolicy.
     * <ul>
     * <li>When there is a horizontal scroll (horizontalScrollPolicy=on or auto): We simply show a scroll bar, and no column widths are changed.</li>
     * <li>When there is no horizontal scroll (horizontalScrollPolicy=off - this is the default): This excess width
     * is reduced proportionally between columns where columnWidthMode does not equal fixed.</li>
     * </ul>
     * </li>
     * </ul>
     * Left and right locked columns do not support column width mode, it is ignored for these.
     * <br/>
     * Finally, with multi level grids, if the hierarchy's columns width in the top level is smaller than the next level's width
     * (and the horizontal scroll policy of the grid is "on"/"auto"), the horizontal scroller will be calculated only by the top
     * level's width, making some columns in the next level unreachable. The recommendation is to give a large column width to the last top level column,
     * which is large enough so that sum of column widths at top level is larger than the sum of column widths at the bottom level.
     * <br/>
     * Values : none,fixed,percent,fitToContent
     * @default auto
     * @method  getColumnWidthMode
     */
    getColumnWidthMode() {
        return (this._columnWidthMode);
    }

    /**
     *
     * @param val
     * @method  setColumnWidthMode
     */
    setColumnWidthMode(val) {
        return (this._columnWidthMode = val);
    }

    /**
     * Formatter for the footer label
     * @method  setFooterFormatter
     */
    setFooterFormatter(val) {
        return this.footerFormatter = val;
    }

    /**
     *  The name of the field or property in the data provider item associated
     *  with the column.
     *  Each DataGridColumn control
     *  requires this property and/or the <code>labelFunction</code> property
     *  to be set in order to calculate the displayable text for the item
     *  renderer.
     *  If the <code>dataField</code>
     *  and <code>labelFunction</code> properties are set,
     *  the data is displayed using the <code>labelFunction</code> and sorted
     *  using the <code>dataField</code>.  If the property named in the
     *  <code>dataField</code> does not exist, the
     *  <code>sortCompareFunction</code> must be set for the sort to work
     *  correctly.
     *
     *  <p>This value of this property is not necessarily the String that
     *  is displayed in the column header.  This property is
     *  used only to access the data in the data provider.
     *  For more information, see the <code>headerText</code> property.</p>
     *
     *
     * If you specify a complex property ,the grid takes over the sortCompareFunction, and the sortField
     * property is ignored.
     * @method  getDataField
     * @return {string}
     */
    getDataField() {
        return this.dataField;
    }

    /**
     * @method  setDataField
     * @param value
     */
    setDataField(value) {
        this.dataField = value;
        if (value && value.indexOf(".") > 0) {
            //removed - SH 1/21/2016 - we actually dont need getNestedSortCompareFunction because it is not used in ExtendedUIUtils.sortArray.
            //this.sortCompareFunction=this.getNestedSortCompareFunction();
            this.hasComplexDisplay = true;
        }
    }

    /**
     *  A function that determines the text to display in this column.  By default
     *  the column displays the text for the field in the data that matches the
     *  column name.  However, sometimes you want to display text based on
     *  more than one field in the data, or display something that does not
     *  have the format that you want.
     *  In such a case you specify a callback function using <code>labelFunction</code>.
     *
     *  <p>For the DataGrid control, the method signature has the following form:</p>
     *
     *  <pre>labelFunction(item:Object, column:DataGridColumn):String</pre>
     *
     *  <p>Where <code>item</code> contains the DataGrid item object, and
     *  <code>column</code> specifies the DataGrid column.</p>
     *
     *  <p>A callback function might concatenate the firstName and
     *  lastName fields in the data, or do some custom formatting on a Date,
     *  or convert a number for the month into the string for the month.</p>
     *
     * @method  getLabelFunction
     * @return {Function}
     */
    getLabelFunction(value) {
        return this._labelFunction;
    }

    /**
     * @method setLabelFunction
     * @param {Function} value
     */
    setLabelFunction(value) {
        this._labelFunction = value;
        this.hasComplexDisplay = true;
    }

    /**
     * A specialized function that determines the text to display in this column, and takes the actual cell being (a) evaluated parameter.
     *
     * As opposed to the regular label function, this actually takes a FlexDataGridCell (a) object parameter,
     * which contains the level information if you need to perform additonal logic. Please note, the cell parameter will be null in certain cases,
     * like from the quickFind() method, where the cell is not currently drawn.
     *
     * Example:
     * labelFunction="myLabelFunc"
     *
     * private String myLabelFunc(Object item, FlexDataGridColumn col, FlexDataGridCell cell ){
     * }
     * @method  getLabelFunction2
     * @return {Function}
     */
    getLabelFunction2(value) {
        return this._labelFunction2;
    }

    /**
     * @method  setLabelFunction2
     * @param {Function} value
     */
    setLabelFunction2(value) {
        this._labelFunction2 = value;
        this.hasComplexDisplay = true;
    }

    /**
     * @method  getLevel
     * @return {flexiciousNmsp.FlexDataGridColumnLevel} level
     */

    getLevel(value) {
        return this.level;
    }

    /**
     * @method  kill
     */
    kill() {
        flexiciousNmsp.EventDispatcher.prototype.kill.apply(this);
        this.level = null;
        this.grid = null;
    }

    /**
     * @private
     */
    getConcatenatedHeaderText() {
        if (this.columnGroup == null) {
            return this.getHeaderText();
        }
        else {
            return `${this.columnGroup.getConcatenatedHeaderText()}-${this.getHeaderText()}`;
        }
    }

    /**
     * @private
     */
    setUseLabelFunctionForSortCompare(value) {
        this.useLabelFunctionForSortCompare = value;
        if (value) {

            const that = this;
            this.sortCompareFunction = (obj1, obj2) => {
                const value1 = that.itemToLabel(obj1);
                const value2 = that.itemToLabel(obj2);

                return (value1 > value2) ? 1 : (value1 < value2) ? -1 : 0;
            };

        }
    }

    /**
     * a.	The default text of the footer is calculated on basis of the following logic: <br/>
     * i.	First we check to see if a footerLabelFunction2 is specified on the column. If so, we use it.<br/>
     * ii.	Second, we check to see if there is a footerLabelFunction specified on the column. If so, we use it.<br/>
     * iii.	Finally, if neither of these is specified, we use a default footer label function, which performs the following logic:<br/>
     * 1.	Checks to see the value of the columns footerOperation. The valid values for this property are sum, min, max, count and average. On basis of this , it computes the value.<br/>
     * 2.	It calls toFixed method passing in the footerOperationPrecision value to give it the appropriate number of decimal places<br/>
     * 3.	If there is a footerFormater specified, it calls the format method of the footerFormatter passing in the value and displays the result.<br/>
     * 4.	If there is a footerLabel specified, it will concatenate the value of the footer label to the result of the calculated value in step 3.<br/>
     * @method calculateValue
     * @param dataProvider
     * @param {flexiciousNmsp.FlexDataGridDataCell} cell
     * @return {String}
     */
    calculateFooterValue(dataProvider, cell) {
        if (typeof cell == "undefined") cell = null;

        if (this.footerLabelFunction2 != null) {
            if (this.footerLabelFunction2.length == 3)
                return this.footerLabelFunction2(this, cell, dataProvider);
            else
                return this.footerLabelFunction2(this, cell);
        }
        else if (this.footerLabelFunction != null)
            return this.footerLabelFunction(this);
        else
            return this.defaultLabelFunction(this, dataProvider);

    }



    /**
     * 1.	Checks to see the value of the columns footerOperation. The valid values for this property are sum, min, max, count and average. On basis of this , it computes the value.<br/>
     * 2.	It calls toFixed method passing in the footerOperationPrecision value to give it the appropriate number of decimal places<br/>
     * 3.	If there is a footerFormater specified, it calls the format method of the footerFormatter passing in the value and displays the result.<br/>
     * 4.	If there is a footerLabel specified, it will concatenate the value of the footer label to the result of the calculated value in step 3.<br/>
     * @param {flexiciousNmsp.FlexDataGridDataCell} col
     * @param dataProvider
     * @method defaultLabelFunction
     * @return {String}
     */
    defaultLabelFunction(col, dataProvider) {
        const uiUtil = flexiciousNmsp.UIUtils;
        let value = null;

        switch (col.footerOperation) {
            case flexiciousNmsp.FlexDataGridColumn.SUM:
                value = UIUtils.sum(dataProvider, col.dataField);
                break;
            case flexiciousNmsp.FlexDataGridColumn.AVERAGE:
                value = UIUtils.average(dataProvider, col.dataField);
                break;
            case flexiciousNmsp.FlexDataGridColumn.MIN:
                value = UIUtils.min(dataProvider, col.dataField);
                break;
            case flexiciousNmsp.FlexDataGridColumn.MAX:
                value = UIUtils.max(dataProvider, col.dataField);
                break;
            case flexiciousNmsp.FlexDataGridColumn.COUNT:
                value = UIUtils.getLength(dataProvider);
                break;
        }

        let val = (value != null) ? isNaN(value) ? value.toString() : value.toFixed(col.footerOperationPrecision) : '';

        if (col.footerFormatter != null && val != null && val != undefined && val != "") {
            if (!col.footerFormatter["format"]) {
                col.footerFormatter = new col.footerFormatter();
            }
            val = col.footerFormatter.format(val);
        }
        if (val == "0.") val = "0";
        return (col.footerLabel != null ? col.footerLabel : "") + val;

    }
}

flexiciousNmsp.FlexDataGridColumn = FlexDataGridColumn; //add to name space 
FlexDataGridColumn.prototype.typeName = FlexDataGridColumn.typeName = 'FlexDataGridColumn';//for quick inspection
FlexDataGridColumn.FORMAT_NONE = "none";
FlexDataGridColumn.FORMAT_DATE = "date";
FlexDataGridColumn.FORMAT_DATE_TIME = "datetime";
FlexDataGridColumn.FORMAT_TIME = "time";
FlexDataGridColumn.FORMAT_CURRENCY = "currency";
FlexDataGridColumn.FORMAT_NUMBER = "number";
FlexDataGridColumn.static_UIComponent = new flexiciousNmsp.ClassFactory(flexiciousNmsp.Label, undefined);
FlexDataGridColumn.static_TextInput = new flexiciousNmsp.ClassFactory(flexiciousNmsp.TextInput, undefined);
FlexDataGridColumn.static_FlexDataGridDataCell = new flexiciousNmsp.ClassFactory(flexiciousNmsp.FlexDataGridDataCell, undefined);
FlexDataGridColumn.static_FlexDataGridHeaderCell = new flexiciousNmsp.ClassFactory(flexiciousNmsp.FlexDataGridHeaderCell, undefined);
FlexDataGridColumn.static_FlexDataGridHeaderCellHS = new flexiciousNmsp.ClassFactory(FlexDataGridHeaderCellHS, undefined);
FlexDataGridColumn.static_FlexDataGridFooterCell = new flexiciousNmsp.ClassFactory(flexiciousNmsp.FlexDataGridFooterCell, undefined);
FlexDataGridColumn.static_FlexDataGridFilterCell = new flexiciousNmsp.ClassFactory(flexiciousNmsp.FlexDataGridFilterCell, undefined);
FlexDataGridColumn.LOCK_MODE_LEFT = "left";
FlexDataGridColumn.LOCK_MODE_RIGHT = "right";
FlexDataGridColumn.LOCK_MODE_NONE = "none";
FlexDataGridColumn.DEFAULT_COLUMN_WIDTH = 150;
FlexDataGridColumn.COLUMN_WIDTH_MODE_NONE = "none";
FlexDataGridColumn.COLUMN_WIDTH_MODE_FIXED = "fixed";
FlexDataGridColumn.COLUMN_WIDTH_MODE_FIT_TO_CONTENT = "fitToContent";
FlexDataGridColumn.COLUMN_WIDTH_MODE_PERCENT = "percent";
/**
 * Dispatched when the column width changes for any reason, including programmatic recalculation of column widths
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event columnResized
 */

FlexDataGridColumn.EVENT_COLUMNRESIZED = 'columnResized';
/**
 * Dispatched when the column x changes for any reason, including programmatic recalculation of column x
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event columnXChanged
 */

FlexDataGridColumn.EVENT_COLUMNXCHANGED = 'columnXChanged';


FlexDataGridColumn.EVENT_COLUMNVISIBLECHANGED = "columnVisibleChanged";


FlexDataGridColumn.SUM = 'sum';
FlexDataGridColumn.MIN = 'min';
FlexDataGridColumn.MAX = 'max';
FlexDataGridColumn.AVERAGE = 'average';
FlexDataGridColumn.COUNT = 'count';
