/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */

import EventDispatcher from '../../core/EventDispatcher'

import Constants from '../../utils/Constants'
import UIUtils from '../../utils/UIUtils' 
/**
 * Class responsible for storing information about grouped columns
 * @constructor
 * @class FlexDataGridColumnGroup
 * @namespace flexiciousNmsp
 * 
 */
export default class FlexDataGridColumnGroup extends EventDispatcher {
    constructor() {
        super();

        /**
         * @type {flexiciousNmsp.ClassFactory}
         * @property columnGroupCellRenderer
         * @default FlexDataGridColumnGroup.static_FlexDataGridColumnGroupCell
         */
        this.columnGroupCellRenderer =  FlexDataGridColumnGroup.static_FlexDataGridColumnGroupCell;

        /**
         * @property parentGroup
         * @default null
         */
        this.parentGroup=null;

        /**
         * @type {Array}
         * @property columnGroups
         * @default []
         * @see http://www.reactdatagrid.com/demo/#/largedynamicgrid
         */
        this.columnGroups =  [];

        /**
         * @private 
         */
        this._startColumn = null;

        /**
         * @property calculatedStart
         * @default null
         */
        this.calculatedStart = null;

        /**
         * @property calculatedEnd
         * @default null
         */
        this.calculatedEnd = null;

        /**
         * @private
         */
        this._columns = [];

        /**
         * @private
         */
        this._endColumn = null;

        /**
         *  Storage for the headerText property.
         *  @type {String}
         *  @property headerText
         *  @default null
         *  @see http://www.reactdatagrid.com/demo/#/nested
         */
        this.headerText = null;
        
        /**
         * @type {flexiciousNmsp.FlexDataGridColumnLevel}
         * @property level
         * @default null
         */
        this.level = null;

        /**
         * @private
         */
        this._depth = 0;

        /**
         * @private
         */
        this._y = 0;

        /**
         * Storage for the headerRenderer property.
         * @type {flexiciousNmsp.ClassFactory}
         * @property columnGroupRenderer
         * @default null
         */
        this.columnGroupRenderer = null;

        /**
         * @type {Boolean}
         * @property enableExpandCollapse
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/largedynamicgrid
         */
        this.enableExpandCollapse = null;

        /**
         * @type {Function}
         * @property expandCollapsePositionFunction
         * @example 
         *    defaultPositionFunction(cell) {
         *      return new flexiciousNmsp.Point(2,cell.getHeight()/4);
         *    }
         */
        this.expandCollapsePositionFunction = this.defaultPositionFunction;

        /**
         * Tooltip to display when user hovers over expand icon
         * @type {String}
         * @property expandTooltip
         * @default "Expand"
         */
        this.expandTooltip="Expand";

        /**
         * Tooltip to display when user hovers over collapse icon
         * @type {String}
         * @property collapseTooltip
         * @default "Collapse"
         */
        this.collapseTooltip="Collapse";
        /**
         * Default function will return the value of expandTooltip or collapseTooltip. Should take a IFlexDataGridCell object
         * @type {Function}
         * @property expandCollapseTooltipFunction
         * @example 
         *     expandCollapseTooltipFunction(cell) {
         *           if(!this.enableExpandCollapse) return "";
         *           if (this.isClosed()){
         *               return `${this.expandTooltip} ${this.getHeaderText()}`;
         *           }else{
         *               return `${this.collapseTooltip} ${this.getHeaderText()}`;
         *           }
         *       }
         */
        this.expandCollapseTooltipFunction=this.defaultExpandCollapseTooltipFunction;

        /**
         * @private
         */
        this._groupedColumns = [];

        /**
         * The column to show when the column group is collapsed. Defaults to the first column.
         * @type {flexiciousNmsp.FlexDataGridColumn}
         * @property collapseStateColumn
         * @default null
         */
        this.collapseStateColumn=null;
        /**
         * When this flag is set to true, every time the columns are set, the last column is used
         * as the collapseStateColumn
         * 
         * @type {boolean}
         * @property useLastColumnAsCollapseStateColumn
         * @default null
         */
        this.useLastColumnAsCollapseStateColumn=false;

    }

    getClassNames() {
        return ["FlexDataGridColumnGroup","EventDispatcher"];
    }

    /**
     * @method getColumns
     */
    getColumns() {
        return this._columns;
    }

    /**
     * Will recalculate the columns at extremities in the next round. Will recursively go up the column group hierarchy
     * invalidating its extremities
     * @method invalidateCalculations
     */
    invalidateCalculations() {
        if(this.calculatedEnd || this.calculatedStart ){
            this.calculatedEnd=this.calculatedStart=null;
            if(this.parentGroup)this.parentGroup.invalidateCalculations();
        }

    }

    /**
     * Will recalculate the columns at extremities in the next round. Will recursively go down the column group hierarchy
     * invalidating child extremities
     * @method invalidateCalculationsDown
     */
    invalidateCalculationsDown() {
        this.calculatedEnd=this.calculatedStart=null;

        for (const cg of this.columnGroups) {
            cg.invalidateCalculationsDown();
        }
    }

    /**
     *  If isClosed(),
     *  return the value of collapseTooltip+ headerText else expandTooltip + headerText
     * 
     * @method defaultExpandCollapseTooltipFunction
     * 
     * @param {flexiciousNmsp.FlexDataGridCell} cell
     * @return {String}
     */
    defaultExpandCollapseTooltipFunction(cell) {

        if(!this.enableExpandCollapse)return "";
        if (this.isClosed()){
            return `${this.expandTooltip} ${this.getHeaderText()}`;
        }else{
            return `${this.collapseTooltip} ${this.getHeaderText()}`;
        }

    }

    /**
     *
     * @method defaultPositionFunction
     * 
     * @param {flexiciousNmsp.FlexDataGridCell} cell
     * @return {flexiciousNmsp.Point}
     */
    defaultPositionFunction(cell) {

        return new flexiciousNmsp.Point(2,cell.getHeight()/4);

    }

    /**
     * Returns true if all but the collapseStateColumn column in this column group is visible.
     * @method isClosed
     * @return {Boolean}
     */
    isClosed() {

        const cols=this.getAllColumns();
        if(cols.length==0)return false;
        if(!this.collapseStateColumn)this.collapseStateColumn=cols[0];
        for(let i=0;i<cols.length;i++){
            if(cols[i].getVisible() && cols[i]!=this.collapseStateColumn){
                return false
            }
        }
        return true;

    }

    /**
     * Sets the visibility of all columns to true
     * @method openColumns
     */
    openColumns() {
        const cols=this.getAllColumns();

        for (const col of cols) {
            col._visible=true;
            col.calcualtedMeasurements={};
        }

        const startCol=this.getStartColumn();
        startCol.level.invalidateCache();
        startCol.level._wxInvalid=true;
        startCol.level.grid.reDraw();
    }

    /**
     * Sets the visibility of all but the first column to false
     * @method closeColumns
     */
    closeColumns() {
        if(!this.collapseStateColumn)this.collapseStateColumn=this.getStartColumn();
        const cols = this.getAllColumns();

        for (const col of cols) {
            col._visible=(col==this.collapseStateColumn);
            col.calcualtedMeasurements={};
        }

        this.collapseStateColumn.level.invalidateCache();
        this.collapseStateColumn.level._wxInvalid=true;
        this.collapseStateColumn.level.grid.reDraw();
    }

    /**
     *
     * @method initializeGroup
     */
    initializeGroup() {
        let i;
        for(i=0;i<this.columnGroups.length;i++){
            const cg=this.columnGroups[i];
            cg.parentGroup=this;
            cg.level=this.level;
            cg.initializeGroup();
        }
        let hit=false;
        if(this.columnGroups.length==0){
            const cols=this.level.getColumns();
            for(i=0;i<cols.length;i++){
                const col=cols[i];
                if(col==this.getStartColumn()){
                    hit=true;
                }
                else if (col==this.getEndColumn()){
                    col.columnGroup=this;
                    return;
                }
                if(this.getStartColumn()==this.getEndColumn()){
                    return;
                }
                if(hit){
                    col.columnGroup=this;
                }
            }
        }

    }

    /**
     * @method initializeDepthY
     * 
     * @param {Number} dpIn 
     * @param {Number} yIn 
     */
    initializeDepthY(dpIn, yIn) {
        if(typeof dpIn=="undefined")dpIn=1;
        if(typeof yIn=="undefined")yIn=0;

        const w=0;
        const x=-1;
        if(this._y==0)
            this._y=yIn;
        if(this._depth==0)
            this._depth=dpIn;

        dpIn=dpIn+1;
        yIn=yIn+this.getHeight();
        if(this.columnGroups.length>0){
            for (const cg of this.columnGroups) {
                if(!cg.level)cg.level=this.level;
                cg.initializeDepthY(dpIn,yIn);
            }
        }

    }

    /**
     * Returns an array of Width and X for this column group.
     * @method getWX
     * @return {Array}
     */
    getWX() {
        if(!this.getStartColumn())return [-1,-1];
        return this.getEndColumn()&&this.getEndColumn().getVisible()?
            [this.getEndColumn().getX()+ this.getEndColumn().getWidth() - this.getStartColumn().getX(),this.getStartColumn().getX()]:
            [this.getStartColumn().getWidth(),this.getStartColumn().getX()];
    }

    /**
     * Iterates through all child groups and gets columns
     * @method getAllColumns
     * @param result
     * @return {Array}
     *
     */
    getAllColumns(result) {
        if(typeof result=="undefined")result=null;
        let i;
        if(!result)result=[];
        for(i=0;i<this._columns.length;i++){
            const col=this._columns[i];
            if(!result.includes(col))
                result.push(col)
        }
        for(i=0;i<this.columnGroups.length;i++){
            const colG=this.columnGroups[i];
            colG.getAllColumns(result);
        }
        return result;


    }

    /**
     * Gets the left most or right most column
     * 
     * @method getColumnAtExtremity
     * @param {Boolean} left
     * 
     * @return {flexiciousNmsp.FlexDataGridColumn}
     */
    getColumnAtExtremity(left) {
        if(typeof left=="undefined")left=true;

        if(left && this.calculatedStart)
            return this.calculatedStart;
        if(!left && this.calculatedEnd)
            return this.calculatedEnd;

        let least;
        let i;
        let col;
        if(this.columnGroups.length>0 && this.getColumns().length==0){
            for(i=0;i<this.columnGroups.length;i++){
                const colGroup=this.columnGroups[i];
                col = colGroup.getColumnAtExtremity(left);
                if(!least&&col)least=col;
                if(col){
                    if(left && least.getColIndex()>col.getColIndex())least=col;
                    if(!left && least.getColIndex()<col.getColIndex())least=col;
                }
            }
        }
        else if(left){
            for(let j=0;j<this._columns.length;j++){
                col=this._columns[j];
                if(col.getVisible()){
                    least=col;
                    break;
                }
            }

        }else if(!left){
            const items=this._columns.slice().reverse();
            for(let k=0;k<items.length;k++){
                col=items[k];
                if(col.getVisible()){
                    least=col;
                    break;
                }
            }
        }
        if(left)
            this.calculatedStart=least;
        else
            this.calculatedEnd=least;
        return least;
    }

    /**
     * Clone this Group. Copies over the start and end groups.
     * 
     * @method clone
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} newLevel
     * @return {flexiciousNmsp.FlexDataGridColumnGroup}
     */
    clone(newLevel) {
        const newCg = new flexiciousNmsp.FlexDataGridColumnGroup();
        if(this._startColumn){
            const startColIndex=this.level.getPrintableColumns().indexOf(this._startColumn);
            newCg.setStartColumn(newLevel.getPrintableColumns()[startColIndex]);
        }
        if(this._endColumn){
            const endColIndex=this.level.getPrintableColumns().indexOf(this._endColumn);
            newCg.setEndColumn(newLevel.getPrintableColumns()[endColIndex]);
        }

        for (const cg of this.columnGroups) {
            newCg.columnGroups.push(cg.clone(newLevel));
        }

        newCg.setColumns(this._columns);
        newCg.setHeaderText(this.getHeaderText());
        newCg.level=newLevel;

        return newCg;
    }

    /**
     * The level of nesting for this column group
     * @method getDepth
     * @return {Number}
     */
    getDepth() {
        return this._depth;
    }

    /**
     * The level of nesting for this column group
     * @method getY
     * @return {Number}
     */
    getY() {
        return this._y;
    }

    /**
     * @method getChildren
     * @return {Array}
     */
    getChildren() {
        return this.getAllColumns();
    }

    /**
     * @method setChildren 
     * @param val 
     */
    setChildren(val) {
        this.setColumns(val);
    }

    /**
     * @method getIsColumnOnly
     * @return {Boolean}
     */
    getIsColumnOnly() {
        return this.columnGroups.length==0&&this._columns.length>0;
    }

    /**
     * If i have a parent group return his group, else return me
     * @method getRootGroup
     * @return {flexiciousNmsp.FlexDataGridColumnGroup}
     *
     */
    getRootGroup() {
        return this.parentGroup?this.parentGroup.getRootGroup():this;
    }

    /**
     * Height of the column group row. Defaults to 25.
     * @method getHeight
     * @return {Number}
     */
    getHeight() {
        return this.level.getHeaderHeight();
    }

    /**
     * Support for columns and column groups as siblings. Prior to 2.7, you could not have a columngroup next to a column.
     * In other words, a column group could either have sub columngroups OR sub columns. If you use the new groupedColumns property
     * you can have column groups that contain both sub column groups as well as sub columns.
     * @method getGroupedColumns
     * @return {Array}
     */
    getGroupedColumns() {
        return this._groupedColumns.length==0?this.getColumns():this._groupedColumns;
    }

    /**
     * @method setGroupedColumns
     * @param value
     * 
     */
    setGroupedColumns(value) {
        this._groupedColumns=value.slice();
        const cols=[];
        const colgs=[];
        let i;
        for(i=0;i<value.length;i++){
            const val=value[i];
            if(val.implementsOrExtends('FlexDataGridColumnGroup')){
                const allCols=val.getAllColumns();

                for (const col of allCols) {
                    cols.push(col);
                }

                colgs.push(val)
            }else{
                val.columnGroup=this;
                cols.push(val);
            }
        }
        this.setColumns(cols);
        this.columnGroups=colgs;
    }

    /**
     * A copy of the child columns collection. Do not modify this, instead, use the setter to set a new value.
     * @method getColumns
     * @return {Array}
     *
     */
    getColumns() {
        return this._columns.slice();
    }

    /**
     * @method setColumns
     * @param {Array} value 
     */
    setColumns(value) {
        this.invalidateCalculations();
        this._columns=value;
        if(value.length>0){
            this.setStartColumn(value[0]);
            this.setEndColumn(value[value.length-1]);
            if(this.useLastColumnAsCollapseStateColumn)
                this.collapseStateColumn=this.getEndColumn();
        }

        for (const val of value) {
            val.columnGroup=this;
        }
    }

    /**
     * The FlexDataGridColumn object to end the column group at. Needs to be
     * at an index greater than the column specified in the start column. If child groups exist, goes through children to get the right most column
     * @method getEndColumn
     * @return {flexiciousNmsp.FlexDataGridColumn}
     */
    getEndColumn() {
        return this.calculatedEnd?this.calculatedEnd:this.getColumnAtExtremity(false);
    }

    /**
     * @method setEndColumn
     * @param {flexiciousNmsp.FlexDataGridColumn} value
     */
    setEndColumn(value) {
        this._endColumn = value;
    }

    /**
     * The FlexDataGridColumn object to start the column group at. Needs to be
     * at an index less than the column specified. If child groups exist, goes through children to get the left most column
     *
     * @method getStartColumn
     * @return {flexiciousNmsp.FlexDataGridColumn} value
     * 
     */
    getStartColumn() {
        return this.calculatedStart?this.calculatedStart:this.getColumnAtExtremity(true);
    }

    /**
     * @method setStartColumn
     * @param {flexiciousNmsp.FlexDataGridColumn} value
     */
    setStartColumn(value) {
        this._startColumn = value;
    }

    /**
     * Text for the header of this columnGroup
     * @method getHeaderText
     * @return {String} 
     */
    getHeaderText() {
        return this.headerText;
    }

    /**
     * Text for the header of this columnGroup
     * 
     * @method setHeaderText
     * @param {String} val
     */
    setHeaderText(val) {
        this.headerText=val;
    }

    kill() {
        flexiciousNmsp.EventDispatcher.prototype.kill.apply(this);
        if(this.columnGroups)
            this.killArray(this.columnGroups);
        if(this._columns)
            this.killArray(this._columns);
        this.parentGroup=null;
    }

    /**
     * @private
     */
    getConcatenatedHeaderText() {
        if(this.parentGroup == null)
        {
            return this.headerText;
        }
        else
        {
            return `${this.parentGroup.getConcatenatedHeaderText()}-${this.headerText}`;
        }
    }
}

flexiciousNmsp.FlexDataGridColumnGroup = FlexDataGridColumnGroup; //add to name space
FlexDataGridColumnGroup.prototype.typeName = FlexDataGridColumnGroup.typeName = 'FlexDataGridColumnGroup';//for quick inspection
FlexDataGridColumnGroup.static_FlexDataGridColumnGroupCell=new flexiciousNmsp.ClassFactory(flexiciousNmsp.FlexDataGridColumnGroupCell);