/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */


import Constants from '../../utils/Constants'
import UIUtils from '../../utils/UIUtils'
import EventDispatcher from '../../core/EventDispatcher'
import InsertionLocationInfo from '../valueobjects/InsertionLocationInfo'
import FlexDataGridPaddingCell from '../cells/FlexDataGridPaddingCell'
import PagerControl from '../toolbar/PagerControlAS'
import FilterSort from './filters/FilterSort'
import AdvancedFilter from './filters/AdvancedFilter'

/**
 * <p>
 * A class that contains information about a nest level of grid. This includes the columns at this
 * level, information about whether or not to enable paging, footers, filters, the row sizes of each,
 * the property of the dataprovider to be used as the key for selection, the property of the data provider
 * to be used as the children field, the renderers for each of the cells, etc.The Grid always contains
 * at least one level. This is the top level, and is accessible via the columnLevel property of the grid.
 * </p>
 * <p>
 * One of the most important concepts behind the Architecture of Flexicious Ultimate arose from the
 * fundamental requirement that the product was created for - that is display of Heterogeneous
 * Hierarchical Data.
 * </p>
 * <p>
 * The notion of nested levels is baked in to the grid via the "columnLevel" property.
 * This is a property of type "FlexDataGridColumnLevel". The grid always has at least one column level.
 * This is also referred to as the top level, or the root level. In flat grids (non hierarchical),
 * this is the only level. But in nested grids, you could have any number of nested levels.
 * </p>
 * <p>
 * The columns collection actually belongs to the columnLevel, and since there is one root level,
 * the columns collection of the grid basically points to the columns collection of this root level.
 * The FlexDataGridColumnLevel class has a "nextLevel" property, which is a pointer to another instance
 * of the same class, or a "nextLevelRenderer" property, which is a reference to a
 * ClassFactory the next level. Please note, currently, if you specify nextLevelRenderer, the nextLevel
 * is ignored. This means, at the same level, you cannot have both a nested sub grid as well as a
 * level renderer. Bottom line - use nextLevelRenderer only at the innermost level.
 * Our examples demonstrate this.
 * </p>
 * @constructor
 * @class FlexDataGridColumnLevel
 * @namespace flexiciousNmsp
 * 
 * */
export default class FlexDataGridColumnLevel extends EventDispatcher {
    constructor(grid) {
        super();
        /**
         * For inner level grids that have their own headers, when the header row scrolls out of view, there is no easy way to determine
         * columns for each row. This flag will make the header row "sticky". For the 3.2 release, this is a beta feature.
         * @type {boolean}
         * @property enableStickyHeaders
         * @default false
         */
        this.enableStickyHeaders = false;
        /**
         * The alpha to apply to sticky header row. See enableStickyHeaders.
         * @type {number}
         * @property stickyHeadersAlpha
         * @default 0.7
         */
        this.stickyHeadersAlpha = 0.7;
        /**
         * Setting this flag to true will resuse the columns from the previous level.
         *
         * Set this flag to true if you want the grid to behave like the advanced datagrid, where
         * there is one set of columns for hierarchical data.
         *
         * @property reusePreviousLevelColumns
         * @type {Boolean}
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/selectionui1
         */
        this.reusePreviousLevelColumns = false;
        /**
         * @private 
         */
        this._rowHeight = 25;

        /**
         * @private 
         */
        this._headerHeight = 25;

        /**
         * @private 
         */
        this._selectedObjects = [];

        /**
         * A property on the object that identifies the object uniquely.
         *
         * Similar to a surrogate key, or a business key, as long as it uniquely identifies the object.
         * When this property is set, the selected keys returns the ID values of the objects that were
         * selected.
         *
         * When a row is selected in the grid, we store the selectedKeyField property
         * of the selected object in this array collection. This allows for multiple
         * pages of data that comes down from the server and not maintained in memory
         * to still keep track of the ids that were selected on other pages.
         * 
         * @property selectedKeyField
         * @type {String}
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.selectedKeyField = "";
        /**
         * A property on the object that identifies whether the object is selectable.
         * @property selectableField
         * @type {String}
         * @default null
         * @see 
         */
        this.selectableField = "";
        /**
         * A property on the object that identifies whether the object is disabled.
         * @property disabledField
         * @type {String}
         * @default null
         */
        this.disabledField = "";
        /**
         * A property on the object that identifies if the object has children. Only needed in filterPageSortMode=server
         *
         * In lazy loaded (filterPageSortMode=server) hierarchical grids levels, the child
         * level items are loaded when the user actually expands this level for the first time.
         *
         * In scenarios where it is known initially that there are children, set this property to
         * the value of the object that identifies whether there are children. If this property is
         * set, the expand collapse icon will be drawn only when the value of this property on the object
         * returns an integer greater than zero. Otherwise, the level will always draw the expand collapse icon.
         * @property childrenCountField
         * @type {String}
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/multiplegroupingmanual
         */
        this.childrenCountField = "";
        /**
         * A property that points to the parent object
         * Used in conjunction with enableSelectionBubble property on the grid.
         * @property parentField
         * @type {String}
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/nested
         */
        this.parentField = "";
        /**
         * @property levelName
         * @type {String}
         * @default null
         */
        this.levelName = "";
        /**
         * Width to show the resize indicator between columns. Defaults to 4
         * @property headerSeperatorWidth
         * @type {Number}
         * @default 6
         */
        this.headerSeperatorWidth = 6;
        /**
         * The indentation to apply to each progressive nest level.
         * @property nestIndent
         * @type {Number}
         * @default 15
         * @see http://www.reactdatagrid.com/demo/#/nested
         */
        this.nestIndent = 15;
        /**
         * A List of FilterSort objects, for multi column sort support.
         * @property currentSorts
         * @type {Array}
         * @default [] empty array
         */
        this.currentSorts = [];
        /**
         * @private
         */
        this._nestDepth = null;
        /**
         * @private
         */
        this._filterArgs = [];

        /**
         * @private 
         */
        this._calculatedSelectedKeys = null;

        /**
         * @private
         */
        this._calculatedSelectedKeysCursor = null;
        /**
         * @private
         */
        this._parentLevel = null;
        /**
         * Childs groups, if any. FlexDataGridColumnGroup
         * @type {Array}
         * @property columnGroups
         * @default []
         */
        this.columnGroups = [];
        /**
         * Grouped columns - only supported at the top level.
         * @private
         */
        this._groupedColumns = [];
        /**
         * Flag to set if column groups exist
         * @type {boolean}
         * @property hasGroupedColumns
         * @default false
         */
        this.hasGroupedColumns = false;
        /**
         * @private 
         */
        this._columns = [];
        /**
         * The number of records to inspect to identify the max size of the data in the column when columnWidthMode=fitToContent.
         *
         * When you have very large datasets, it will be very slow if you inspect every row to identify the largest string to
         * come up with the column width. So we inspect only the first columnWidthModeFitToContentSampleSize rows.
         *
         * @default 100
         * @property columnWidthModeFitToContentSampleSize
         * @type {Number}
         */
        this.columnWidthModeFitToContentSampleSize = 100;
        /**
         * Declaration of the next level of columns
         * @property nextLevel
         * @type {flexiciousNmsp.FlexDataGridColumnLevel}
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/selectionoptions
         */
        this.nextLevel = null;
        /**
         * The property of the object on this level that should be used in the default sort.
         * @property initialSortField
         * @type {String}
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.initialSortField = null;
        /**
         * If true, initial sort direction for this level is ascending. Defaults to true.
         * @property initialSortAscending
         * @type {Boolean}
         * @default true
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.initialSortAscending = true;
        /**
         * Flag to indicate whether or not to enable paging functionality
         * @property enablePaging
         * @type {Boolean}
         * @default false
         * @see http://www.reactdatagrid.com/demo/#/outlook
         */
        this.enablePaging = false;
        /**
         * Use this flag to show the pager control even if the
         * enablePaging is set to false. This is used in a scenario
         * where you wish to show buttons other than the paging buttons
         * in the pager bar.
         * @property forcePagerRow
         * @type {Boolean}
         * @default false
         */
        this.forcePagerRow = false;

        /**
         * @private 
         */
        this._pagerRenderer = null;
        /**
         * The number of records to display per page
         * @property pageSize
         * @type {Number}
         * @default 5
         * @see http://www.reactdatagrid.com/demo/#/grouped
         */
        this.pageSize = 5;

        /**
         * @private 
         */
        this._pageIndex = 0;
        /**
         * @deprecated. This does not appply to Ultimate.
         *
         * The horizontal positioning of the contents of the pager. <br/>
         * Please use the displayOrder property to control the vertical positioning <br/>
         * Handled by default if you use the built in pager (com.flexicious.grids.columns.footers.PagerControl)
         * , but if you implement your own com.flexicious.controls.interfaces.IPager, then needs to be handled by you.
         * @see com.flexicious.grids.columns.footers.PagerControl.
         * @see implements com.flexicious.controls.interfaces.IPager.
         * @property pagerPosition
         * @type {String}
         * @default null
         */
        this.pagerPosition = null;

        /**
         * @private
         */
        this._enableFooters = false;
        /**
         * If you have enableFooters set to true but want the footer row to not
         * appear, you can set this value.
         * @property footerVisible
         * @type {Boolean}
         * @default false
         */
        this.footerVisible = true;

        /**
         * If you have enablePaging set to true but want the pager row to not
         * appear, you can set this value.
         * @property pagerVisible
         * @type {Boolean}
         * @default true
         */
        this.pagerVisible = true;

        /**
         * @private 
         */
        this._footerRowHeight = 25;

        /**
         * @private 
         */
        this._pagerRowHeight = 30;

        /**
         * @private 
         */
        this._filterRowHeight = 25;

        /**
         * @private 
         */
        this._enableFilters = false;
        /**
         * If you have enableFilters set to true but want the filter row to not
         * appear, you can set this value.
         * @property filterVisible
         * @type {Boolean}
         * @default true
         */
        this.filterVisible = true;
        /**
         * Whether or not to show the header row
         * @property headerVisible
         * @type {Boolean}
         * @default true
         */
        this.headerVisible = true;
        /**
         * The Filter/Page/Sort Mode. Can be either "server" or "client".
         * In client mode, the grid will take care of paging, sorting and
         * filtering once the dataprovider is set.
         * In server mode, the grid will fire a com.flexicious.nestedtreedatagrid.events.ExtendedFilterPageSortChangeEvent
         * named filterPageSortChange
         * that should be used to construct an appropriate query to be sent to the
         * backend.
         * @default "client"
         * @see com.flexicious.grids.events.FilterPageSortChangeEvent
         * @see http://www.reactdatagrid.com/demo/#/fulllazyload
         * @property filterPageSortMode
         * @type {String}
         */
        this.filterPageSortMode = "client";
        /**
         * The Item Load Mode. Can be either "server" or "client".
         * In client mode, the grid will assume that the children of items at this level
         * are prefreched.
         *
         * In server mode, the grid will dispatch a ITEM_LOAD event (itemLoad)
         * that should be used to construct an appropriate query to be sent to the backend, to
         * retrieve the child objects at this level.
         *
         * Once the results are retrieved, please call the "setChildData" method on the grid to
         * set the results at this level. Please note, the "childrenField" is still required at this level,
         * because that is where the setChildData method persists the loaded children.
         * Future itemOpen events do not result in itemLoads because the data for this particular entity has already been loaded and persisted.
         * @default "client"
         * @property itemLoadMode
         * @type {String}
         * @see http://www.reactdatagrid.com/demo/#/partiallazyload
         */
        this.itemLoadMode = "client";
        /**
         * A renderer that displays this level. If you specify a value for this, the columns are ignored,
         * and instead of drawing a nested data grid at this level, an instance of the levelRenderer is displayed.
         * All paging, footers, headers and other properties related to hierarchical display are also ignored.
         *
         * A level that has a levelRenderer cannot have subsequent levels. So, this is usually the innermost level.
         *
         * Please ensure that you explictly specify a levelRendererHeight so the vertical height of the grid can
         * be correctly calculated.
         * @property nextLevelRenderer
         * @type {flexiciousNmsp.ClassFactory}
         * @default null
         * @see http://www.reactdatagrid.com/demo/#/levelrenderer
         */
        this.nextLevelRenderer = null;
        /**
         * Height to assign to the renderer for each level. If not specified, a new instance of the level renderer is created, and the height
         * is defaulted to the height of that one instance. The same height is then reused for subsequent renderers of the same type.
         * @property levelRendererHeight
         * @type {Number}
         * @default 0
         * @see http://www.reactdatagrid.com/demo/#/levelrenderer
         */
        this.levelRendererHeight = 0;
        /**
         * Container for the expand collapse cell. Needs to inherit from FlexDataGridExpandCollapseCell.
         * Defaults to FlexDataGridExpandCollapseCell
         * @property expandCollapseCellRenderer
         * @type {*}
         * @default FlexDataGridColumnLevel.static_FlexDataGridExpandCollapseCell
         */
        this.expandCollapseCellRenderer = FlexDataGridColumnLevel.static_FlexDataGridExpandCollapseCell;
        /**
         * Container for the expand collapse cell. Needs to inherit from FlexDataGridExpandCollapseHeaderCell.
         * Defaults to FlexDataGridExpandCollapseHeaderCell
         * @property expandCollapseHeaderCellRenderer
         * @type {*}
         * @default FlexDataGridColumnLevel.static_FlexDataGridExpandCollapseHeaderCell
         */
        this.expandCollapseHeaderCellRenderer = FlexDataGridColumnLevel.static_FlexDataGridExpandCollapseHeaderCell;
        /**
         * Container for the nest indent padding cell. Needs to inherit from FlexDataGridExpandCollapseHeaderCell.
         * Defaults to FlexDataGridExpandCollapseCell
         * @property nestIndentPaddingCellRenderer
         * @type {*}
         * @default FlexDataGridColumnLevel.static_FlexDataGridPaddingCell
         */
        this.nestIndentPaddingCellRenderer = FlexDataGridColumnLevel.static_FlexDataGridPaddingCell;
        /**
         * The inner content of the padding cell. Defaults to a plain old UI component
         * @property nestIndentPaddingRenderer
         * @type {flexiciousNmsp.ClassFactory}
         * @default null
         */
        this.nestIndentPaddingRenderer = null;
        /**
         * The inner content of the scroll bar pad cell. Defaults to a plain old UI component
         * @property scrollbarPadRenderer
         * @type {flexiciousNmsp.ClassFactory}
         * @default null
         */
        this.scrollbarPadRenderer = null;
        /**
         * A function that can be used to control the enabled flag of each cell in this level.
         * If this function is null or returns null, the cell will be enabled, selectable, and rollovers
         * will work. If this function returns false, the cell will not be selectable, the item renderers will
         * be disabled and it will have no rollovers.
         * This function has two parameters:
         * 1) IFlexDataGridDataCell : Cell object which has a pointer to the row data (data) as well as other related information
         * found in the documentation of the FlexDataGridDataCell class. This parameter can be null, when the
         * function is called as a response to a higher level checkbox select. For example, when the user hits
         * select all. In this case, we do not always have a cell, and it will be passed in as null
         * 2) The actual data object bound to the row.
         * This function should true or false.
         *
         * For example:
         *
         * private function rowDisabledFunction(cell:FlexDataGridCell,data:Object):Boolean{
		 * 		return data.enabledl
		 * }
         *
         * @property rowDisabledFunction
         * @type {Function}
         * @default null
         * @example 
         *   getRowDisabled(cell, data) {
         *       if (data.legalName == 'Adobe Systems') {
         *       return true;
         *       }
         *       return false;//do not disable by default.
         *   }
         * @see http://www.reactdatagrid.com/demo/#/selectionoptions
         */
        this.rowDisabledFunction = null;

        /**
         * A function that can be used to control whether clicking on any cell in this level will select it.
         * If this function is null or returns false, clicking on the cell will not result in the row being selected.
         * If this function returns true, clicking on the cell will result in the row being selected in row selection modes only.
         * This function has two parameters:
         * 1) IFlexDataGridDataCell : Cell object which has a pointer to the row data (data) as well as other related information
         * found in the documentation of the IFlexDataGridDataCell class. This parameter can be null, when the
         * function is called as a response to a higher level checkbox select. For example, when the user hits
         * select all. In this case, we do not always have a cell, and it will be passed in as null
         * 2) The actual data object bound to the row.
         * This function should true or false.
         *
         * For example:
         *
         * private function rowSelectableFunction(cell:FlexDataGridCell,data:Object):Boolean{
		 * 		return data.enabled;
		 * }
         * @type {Function}
         * @property rowSelectableFunction
         * @type {null}
         * @example 
         *   returnFalse(cell, data) {
         *       return false;
         *   };
         * @see http://www.reactdatagrid.com/demo/#/grouped2
         */
        this.rowSelectableFunction = null;

        /**
         * @private 
         */
        this._rowBackgroundFunction = null;
        /**
         * A function that can be used to control the text color of each cell in this level.
         * If this function is null or returns null, the cell text will
         * use the alternatingTextColors style property. This function should take a IFlexDataGridDataCell
         * object, which has a pointer to the row data (data) as well as other related information
         * found in the documentation of the FlexDataGridDataCell. This function
         * should return a single color hex code (uint).
         *
         * @deprecated Use the grid.cellTextColorFunction
         * @property rowTextColorFunction
         * @type {Function}
         * @default null
         * @example 
         * getRowTextColor(cell) {
         *   if (cell.rowInfo.getData().headquarterAddress.state.name == "New York") {
         *       return 0xCC3300;
         *   }
         *   return null;
         *  };
         * @see http://www.reactdatagrid.com/demo/#/outlook
         */
        this.rowTextColorFunction = null;
        /**
         * A function that lets you control the border drawing mechanism for each cell.
         * By default, each cell will draw a 1px wide line to its right and bottom ends.
         *
         * This function offers you the opportunity to hook into this mechanism, and draw
         * your own border using the graphics property of the FlexDataGridCell object,
         * that is passed in as a parameter to this function. If this function returns false,
         * the default border is not drawn.
         * @property cellBorderFunction
         * @type {Function}
         * @default null
         * @example 
         * drawCellBorder(cell) {
         *       if (!cell.getIsDataCell()) {
         *           return true; //let the grid handle this
         *       }
         *       if (cell.level.getNestDepth() == 1) {
         *           cell.setStyleAttribute("borderBottom", "solid 1px #ececec");
         *           return false;
         *       }
         *       else if (cell.level.getNestDepth() == 2) {
         *           cell.setStyleAttribute("borderBottom", "solid 1px #afafaf");
         *           return false;
         *       }
         *       return true;//let the grid handle this
         *   }
         */
        this.cellBorderFunction = null;

        /**
         * @private 
         */
        this._cellCustomDrawFunction = null;
        /**
         * Container for the footer renderer. Needs to inherit from FlexDataGridPagerCell.
         * Defaults to FlexDataGridPagerCell
         * @property pagerCellRenderer
         * @type {*}
         * @default flexiciousNmsp.FlexDataGridColumnLevel.static_FlexDataGridPagerCell
         */
        this.pagerCellRenderer = flexiciousNmsp.FlexDataGridColumnLevel.static_FlexDataGridPagerCell;
        /**
         * In scenarios where there are detached filters, this function
         * may be used to add on to the list of arguments before the filter is
         * run. The function must take a column level as the only parameter,
         * and return an array collection of the filter expressions.
         * @property additionalFilterArgumentsFunction
         * @type {Function}
         * @default null
         */
        this.additionalFilterArgumentsFunction = null;
        /**
         * A comma separated list of the following strings:
         * filter,header,body,footer,pager.
         *
         * Changes the order in which the grid displays the
         * filter,header,body,footer,pager
         * The order in which the various sections of the grid
         * are laid out.
         *
         * Defaults to pager,filter,body,footer
         * @property displayOrder
         * @type {String}
         * @default "pager,header,filter,body,footer"
         */
        this.displayOrder = "pager,header,filter,body,footer";

        /**
         * @private 
         */
        this._cachedVisibleColumns = {};

        /**
         * @private 
         */
        this._lockCache = {};

        /**
         * @private 
         */
        this._flowColumns = null;

        /**
         * @private 
         */
        this._flowHeaderColumns = null;

        /**
         * @private 
         */
        this._wxInvalid = true;

        /**
         * @private 
         */
        this._enableRowNumbers = null;
        /**
         * In addition to the filter row, if you wish to apply external filters, or to apply filters at inner nested levels, you may use this property.
         * Works just like ListCollectionView.filterFunction
         * @property filterFunction
         * @type {Function}
         * @default null
         * @example 
         * filterDeviceTypes(item){
         *       if(! this.cbtimesheet1 && item.timeSheetTitle=='Time Sheet-1')
         *           return false;
         *       if(! this.cbtimesheet2 && item.timeSheetTitle=='Time Sheet-2')
         *           return false;
         *       return true;
         *   };
         * @see http://www.reactdatagrid.com/demo/#/multiplegroupingmanual
         */
        this.filterFunction = null;
        /**
         * A field on the dataprovider that indicates that this item should be selected
         * @type {String}
         * @property selectedField
         * @default null
         */
        this.selectedField = null;
        /**
         * Not Yet Implemented. Support for use case where inner level is single select. For use where inner level is single select and we need to cascade selection from the parent. When parent is chosen, one child needs to be the default selection. This field should point to a property on the model object that returns a true false determining which field to choose by default.
         * @type {string}
         * @property defaultSelectionField
         * @default null
         */
        this.defaultSelectionField = "";
        /**
         * 
         * Not Yet Implemented. Support for use case where inner level is single select. Sets this levels selection mode to be single select.
         * @type {boolean}
         * @property enableSingleSelect
         * @default false
         */
        this.enableSingleSelect = false;

        /**
         * @private 
         */
        this._unSelectedObjects = [];

        /**
         * @private 
         */
        this._pendingStyles = [];

        /**
         * The owner grid.
         * @property grid 
         * @type {FlexDataGrid}
         * @see http://www.reactdatagrid.com/demo/#/simple
         */
        this.grid = grid;
        /**
         * minimu header height
         * Not Yet Implemented.
         * @type {Number}
         * @property minHeaderHeight
         * @default 20
         */
        this.minHeaderHeight = 20;

        /**
         * This function will return total height for current level renderer
         * @property levelRendererHeightFunction
         * @default null
         * @example
         *  getLevelHeight(item) {
         *   let totalHeight = 0;
         *   // write logic here
         *   return totalHeight;
         *  }
         * 
         */
        this.levelRendererHeightFunction = null;

        /**
         * The property of the parent level object, that identifies the children that should be displayed on the next level.
         * This is only required if the collection is not a hierarchical view.
         *
         * Please note, in server mode, this property is also the "storage" for the lazy loaded children.
         *
         * See the Partially and Lazy Loaded Server Grid examples for more details
         * @type {String}
         * @property childrenField
         * @default "children"
         * @see http://www.reactdatagrid.com/demo/#/outlook
         */
        this.childrenField = "children";
        /**
         * [Style(name="backgroundColor",  type="Array", arrayType="uint", format="Color", inherit="no")]
         *  Background color, the color of the fill area.
		 *  @default null
         *  @type {Array}
		 *  @property backgroundColor
         */

        this.backgroundColor = null;
        /**
         * [Style(name="color", type="uint", format="Color", inherit="yes")]
         *  Color of text in the component, including the component label.
         *
         *  The default value for the Halo theme is <code>0x0B333C</code>.
         *  The default value for the Spark theme is <code>0x000000</code>.
		 *  @default null
         *  @type {uint}
		 *  @property  color
         */

        this.color = null;
        /**
         * [Style(name="disabledColor", type="uint", format="Color", inherit="yes")]
         *  Color of text in the component if it is disabled.
         *
         *  @default 0xAAB3B3
         *  @type {uint}
		 *  @property  disabledColor
         *
         */

        this.disabledColor = null;
        /**
         * [Style(name="textAlign", type="String", enumeration="left,center,right", inherit="yes")]
         *  Alignment of text within a container.
         *  Possible values are <code>"left"</code>, <code>"right"</code>,
         *  or <code>"center"</code>.
         *
         *  <p>The default value for most components is <code>"left"</code>.
         *  For the FormItem component,
         *  the default value is <code>"right"</code>.
         *  For the Button, LinkButton, and AccordionHeader components,
         *  the default value is <code>"center"</code>, and
         *  this property is only recognized when the
         *  <code>labelPlacement</code> property is set to <code>"left"</code> or
         *  <code>"right"</code>.
         *  If <code>labelPlacement</code> is set to <code>"top"</code> or
         *  <code>"bottom"</code>, the text and any icon are centered.</p>
         *
		 *  @default null
         *  @type {String}
		 *  @property  textAlign
         */

        this.textAlign = null;
        /**
         * [Style(name="textDecoration", type="String", enumeration="none,underline", inherit="yes")]
         *  Determines whether the text is underlined.
         *  Possible values are <code>"none"</code> and <code>"underline"</code>.
         *
         *  @default "none"
         *  @type {String}
		 *  @property  textDecoration
         */

        this.textDecoration = null;
        /**
         * [Style(name="multiColumnSortNumberWidth", type="Number", inherit="no")]
         *  The width of  the numeric value representing the order of the column sort.
         *  @default 10
         *  @type {Number}
		 *  @property  multiColumnSortNumberWidth
         *
         */

        this.multiColumnSortNumberWidth = null;
        /**
         * [Style(name="multiColumnSortNumberHeight", type="Number", inherit="no")]
         *  The height of  the numeric value representing the order of the column sort.
         *  @default 15
         *  @type {Number}
		 *  @property  multiColumnSortNumberHeight
         *
         */

        this.multiColumnSortNumberHeight = null;
        /**
         * [Style(name="multiColumnSortNumberStyleName", type="String", inherit="no")]
         *  The name of a CSS style declaration for controlling other aspects of
         *  the appearance of the numeric value representing the order of the column sort.
         *  @default "multiColumnSortNumberStyle"
         *  @type {String}
		 *  @property  multiColumnSortNumberStyleName
         *
         */

        this.multiColumnSortNumberStyleName = null;
        /**
         * [Style(name="columnGroupStyleName", type="String", inherit="no")]
         *  The name of a CSS style declaration for controlling other aspects of
         *  the appearance of the column groups.
         *  @default "columnGroupStyle"
         *
         *  @type {String}
		 *  @property  columnGroupStyleName
         */

        this.columnGroupStyleName = null;
        /**
         * [Style(name="headerStyleName", type="String", inherit="no")]
         *  The name of a CSS style declaration for controlling other aspects of
         *  the appearance of the column headers.
         *  @default "dataGridStyles"
         *  @type {String}
		 *  @property  headerStyleName
         *
         */

        this.headerStyleName = null;
        /**
         * [Style(name="footerStyleName", type="String", inherit="no")]
         *  The name of a CSS style declaration for controlling other aspects of
         *  the appearance of the column headers.
         *  @default "dataGridStyles"
         *  @type {String}
		 *  @property  footerStyleName
         */

        this.footerStyleName = null;
        /**
         * [Style(name="verticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type {Boolean}
		 *  @property  verticalGridLines
         */

        this.verticalGridLines = null;
        /**
         * [Style(name="horizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type {Boolean}
		 *  @property  horizontalGridLines
         */

        this.horizontalGridLines = null;
        /**
         * [Style(name="verticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
		 *  @property  verticalGridLineColor
         */

        this.verticalGridLineColor = null;
        /**
         * [Style(name="horizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
		 *  @default null
         *  @type uint
		 *  @property  horizontalGridLineColor
         *  @see http://www.reactdatagrid.com/demo/#/outlook
         */

        this.horizontalGridLineColor = null;
        /**
         * [Style(name="horizontalGridLineThickness", type="Number", format="Length", inherit="yes")]
         *  Thickness of the horizontal grid lines.
         *  @default 1
         *  @type Number
		 *  @property  horizontalGridLineThickness
         *  @see http://www.reactdatagrid.com/demo/#/outlook
         */

        this.horizontalGridLineThickness = null;
        /**
         * [Style(name="verticalGridLineThickness", type="Number", format="Length", inherit="yes")]
         *  Thickness of the vertical grid lines.
         *  @default 1
         *  @type Number
		 *  @property  verticalGridLineThickness
         */

        this.verticalGridLineThickness = null;
        /**
         * [Style(name="headerVerticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
		 *  @property  headerVerticalGridLines
         */

        this.headerVerticalGridLines = null;
        /**
         * [Style(name="headerHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
		 *  @property  headerHorizontalGridLines
         */

        this.headerHorizontalGridLines = null;
        /**
         * [Style(name="headerVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
		 *  @property  headerVerticalGridLineColor
         */

        this.headerVerticalGridLineColor = null;
        /**
         * [Style(name="headerHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
		 *  @default null
         *  @type uint
		 *  @property  headerHorizontalGridLineColor
         */

        this.headerHorizontalGridLineColor = null;
        /**
         * [Style(name="headerHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the header horizontal grid lines.
         *  @default 1
         *  @type Number
		 *  @property  headerHorizontalGridLineThickness
         */

        this.headerHorizontalGridLineThickness = null;
        /**
         * [Style(name="headerVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the header vertical grid lines.
         *  @default 1
         *  @type Number
		 *  @property  headerVerticalGridLineThickness
         *  @see http://www.reactdatagrid.com/demo/#/selectionui1
         */

        this.headerVerticalGridLineThickness = null;
        /**
         * [Style(name="headerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border
         *  @default false
         *  @type Boolean
		 *  @property  headerDrawTopBorder
         */

        this.headerDrawTopBorder = null;
        /**
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
		 *  @property  columnGroupVerticalGridLines
         * [Style(name="columnGroupVerticalGridLines", type="Boolean", inherit="no")]
         */

        this.columnGroupVerticalGridLines = null;
        /**
         * [Style(name="columnGroupHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
		 *  @property  columnGroupHorizontalGridLines
         */

        this.columnGroupHorizontalGridLines = null;
        /**
         * [Style(name="columnGroupVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
		 *  @property  columnGroupVerticalGridLineColor
         */

        this.columnGroupVerticalGridLineColor = null;
        /**
         * [Style(name="columnGroupHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
		 *  @default null
         *  @type uint
		 *  @property  columnGroupHorizontalGridLineColor
         */

        this.columnGroupHorizontalGridLineColor = null;
        /**
         * [Style(name="columnGroupHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the header horizontal grid lines.
         *  @default 1
         *  @type Number
		 *  @property  columnGroupHorizontalGridLineThickness
         */

        this.columnGroupHorizontalGridLineThickness = null;
        /**
         * [Style(name="columnGroupVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the header vertical grid lines.
         *  @default 1
         *  @type Number
		 *  @property  columnGroupVerticalGridLineThickness
         */

        this.columnGroupVerticalGridLineThickness = null;
        /**
         * [Style(name="columnGroupDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border
         *  @default false
		 *  @type Boolean
		 *  @property  columnGroupDrawTopBorder
         */

        this.columnGroupDrawTopBorder = null;
        /**
         * [Style(name="filterVerticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
		 *  @property  filterVerticalGridLines
         */

        this.filterVerticalGridLines = null;
        /**
         * [Style(name="filterHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
		 *  @property  filterHorizontalGridLines
         */

        this.filterHorizontalGridLines = null;
        /**
         * [Style(name="filterVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
		 *  @property  filterVerticalGridLineColor
         */

        this.filterVerticalGridLineColor = null;
        /**
         * [Style(name="filterHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
		 *	@default null
         *  @type uint
		 *  @property  filterHorizontalGridLineColor
         */

        this.filterHorizontalGridLineColor = null;
        /**
         * [Style(name="filterHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the filter horizontal grid lines.
         *  @default 1
         *  @type Number
		 *  @property  filterHorizontalGridLineThickness
         */

        this.filterHorizontalGridLineThickness = null;
        /**
         * [Style(name="filterVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the filter vertical grid lines.
         *  @default 1
         *  @type Number
		 *  @property  filterVerticalGridLineThickness
         */

        this.filterVerticalGridLineThickness = null;
        /**
         * [Style(name="filterDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
		 *  @property  filterDrawTopBorder
         */

        this.filterDrawTopBorder = null;
        /**
         * [Style(name="footerVerticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
		 *  @property  footerVerticalGridLines
         */

        this.footerVerticalGridLines = null;
        /**
         * [Style(name="footerHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
		 *  @property  footerHorizontalGridLines
         */

        this.footerHorizontalGridLines = null;
        /**
         * [Style(name="footerVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
		 *  @property  footerVerticalGridLineColor
         */

        this.footerVerticalGridLineColor = null;
        /**
         * [Style(name="footerHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
		 *  @default null
         *  @type uint
		 *  @property  footerHorizontalGridLineColor
         */

        this.footerHorizontalGridLineColor = null;
        /**
         * [Style(name="footerHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the footer horizontal grid lines.
         *  @default 1
         *  @type Number
		 *  @property  footerHorizontalGridLineThickness
         */

        this.footerHorizontalGridLineThickness = null;
        /**
         * [Style(name="footerVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the footer vertical grid lines.
         *  @default 1
         *  @type Number
		 *  @property  footerVerticalGridLineThickness
         */

        this.footerVerticalGridLineThickness = null;
        /**
         * [Style(name="footerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
		 *  @property  footerDrawTopBorder
         */

        this.footerDrawTopBorder = null;
        /**
         * [Style(name="pagerVerticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
		 *  @property  pagerVerticalGridLines
         */

        this.pagerVerticalGridLines = null;
        /**
         * [Style(name="pagerHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
		 *  @property  pagerHorizontalGridLines
         */

        this.pagerHorizontalGridLines = null;
        /**
         * [Style(name="pagerVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
		 *  @property  pagerVerticalGridLineColor
         */

        this.pagerVerticalGridLineColor = null;
        /**
         * [Style(name="pagerHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
		 *  @default null
         *  @type uint
		 *  @property  pagerHorizontalGridLineColor
         */

        this.pagerHorizontalGridLineColor = null;
        /**
         * [Style(name="pagerHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the pager horizontal grid lines.
         *  @default 1
         *  @type Number
		 *  @property  pagerHorizontalGridLineThickness
         */

        this.pagerHorizontalGridLineThickness = null;
        /**
         * [Style(name="pagerVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the pager vertical grid lines.
         *  @default 1
         *  @type Number
		 *  @property  pagerVerticalGridLineThickness
         */

        this.pagerVerticalGridLineThickness = null;
        /**
         * [Style(name="pagerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
		 *  @property  pagerDrawTopBorder
         */

        this.pagerDrawTopBorder = null;
        /**
         * [Style(name="rendererVerticalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show vertical grid lines between
         *  the columns.
         *  If <code>true</code>, shows vertical grid lines.
         *  If <code>false</code>, hides vertical grid lines.
         *  @default true
         *  @type Boolean
		 *  @property  rendererVerticalGridLines
         */

        this.rendererVerticalGridLines = null;
        /**
         * [Style(name="rendererHorizontalGridLines", type="Boolean", inherit="no")]
         *  Flag that indicates whether to show horizontal grid lines between
         *  the rows.
         *  If <code>true</code>, shows horizontal grid lines.
         *  If <code>false</code>, hides horizontal grid lines.
         *  @default false
         *  @type Boolean
		 *  @property  rendererHorizontalGridLines
         */

        this.rendererHorizontalGridLines = null;
        /**
         * [Style(name="rendererVerticalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the vertical grid lines.
         *  @default 0x666666
         *  @type uint
		 *  @property  rendererVerticalGridLineColor
         */

        this.rendererVerticalGridLineColor = null;
        /**
         * [Style(name="rendererHorizontalGridLineColor", type="uint", format="Color", inherit="yes")]
         *  The color of the horizontal grid lines.
		 *  @default null
         *  @type uint
		 *  @property  rendererHorizontalGridLineColor
         */

        this.rendererHorizontalGridLineColor = null;
        /**
         * [Style(name="rendererHorizontalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the renderer horizontal grid lines.
         *  @default 1
         *  @type Number
		 *  @property  rendererHorizontalGridLineThickness
         */

        this.rendererHorizontalGridLineThickness = null;
        /**
         * [Style(name="rendererVerticalGridLineThickness", type="Number", format="Length")]
         *  Thickness of the renderer vertical grid lines.
         *  @default 1
         *  @type Number
		 *  @property  rendererVerticalGridLineThickness
         */

        this.rendererVerticalGridLineThickness = null;
        /**
         * [Style(name="rendererDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
		 *  @property  rendererDrawTopBorder
         */

        this.rendererDrawTopBorder = null;
        /**
         * [Style(name="headerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
		 *  @property  headerDrawTopBorder
         */

        this.headerDrawTopBorder = null;
        /**
         * [Style(name="filterDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
		 *  @property  filterDrawTopBorder
         */

        this.filterDrawTopBorder = null;
        /**
         * [Style(name="footerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
		 *  @property  footerDrawTopBorder
         */

        this.footerDrawTopBorder = null;
        /**
         * [Style(name="pagerDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
		 *  @property  pagerDrawTopBorder
         */

        this.pagerDrawTopBorder = null;
        /**
         * [Style(name="rendererDrawTopBorder", type="Boolean", inherit="no")]
         *  Flag that indicates whether to force the top border, when horizontal gridlines are not drawn
         *  @default false
         *  @type Boolean
		 *  @property  rendererDrawTopBorder
         */

        this.rendererDrawTopBorder = null;
        /**
         * [Style(name="selectionColor", type="uint", format="Color", inherit="yes")]
         *  The color of the background for the row when the user selects
         *  an item renderer in the row.
         *
         *  The default value for the Halo theme is <code>0x7FCEFF</code>.
         *  The default value for the Spark theme is <code>0xA8C6EE</code>.
         *  @default null
         *  @type uint
		 *  @property  selectionColor
         */

        this.selectionColor = null;
        /**
         * [Style(name="color", type="uint", format="Color", inherit="yes")]
         *  Color of text in the component, including the component label.
         *
         *  The default value for the Halo theme is <code>0x0B333C</code>.
         *  The default value for the Spark theme is <code>0x000000</code>.
         *
         *  @default null
         *  @type uint
		 *  @property  color
         */

        this.color = null;
        /**
         * [Style(name="sortArrowSkin", type="Class", inherit="no")]
         *  The class to use as the skin for the arrow that indicates the column sort
         *  direction.
         *  The default value for the Halo theme is <code>mx.skins.halo.DataGridSortArrow</code>.
         *  The default value for the Spark theme is <code>mx.skins.spark.DataGridSortArrow</code>.
         *  @default null
         *  @type Class
		 *  @property  sortArrowSkin
         */

        this.sortArrowSkin = null;
        /**
         * [Style(name="paddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
         *  @type Number
		 *  @property  paddingLeft
         *  @see http://www.reactdatagrid.com/demo/#/selectionui2
         */

        this.paddingLeft = null;
        /**
         * [Style(name="paddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
		 *  @property  paddingRight
         *  @see http://www.reactdatagrid.com/demo/#/selectionui2
         */

        this.paddingRight = null;
        /**
         * [Style(name="paddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  paddingTop
         *  @see http://www.reactdatagrid.com/demo/#/outlook
         */

        this.paddingTop = null;
        /**
         * [Style(name="paddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  paddingBottom
         */

        this.paddingBottom = null;
        /**
         * [Style(name="headerPaddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
		 *  @type Number
		 *  @property  headerPaddingLeft
         */

        this.headerPaddingLeft = null;
        /**
         * [Style(name="headerPaddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
		 *  @property  headerPaddingRight
         */

        this.headerPaddingRight = null;
        /**
         * [Style(name="headerPaddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  headerPaddingTop
         *  @see http://www.reactdatagrid.com/demo/#/outlook
         */

        this.headerPaddingTop = null;
        /**
         * [Style(name="headerPaddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  headerPaddingBottom
         */

        this.headerPaddingBottom = null;
        /**
         * [Style(name="footerPaddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
         *  @type Number
		 *  @property  footerPaddingLeft
         */

        this.footerPaddingLeft = null;
        /**
         * [Style(name="footerPaddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
		 *  @property  footerPaddingRight
         */

        this.footerPaddingRight = null;
        /**
         * [Style(name="footerPaddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  footerPaddingTop
         */

        this.footerPaddingTop = null;
        /**
         * [Style(name="footerPaddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  footerPaddingBottom
         */

        this.footerPaddingBottom = null;
        /**
         * [Style(name="filterPaddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
         *  @type Number
		 *  @property  filterPaddingLeft
         */

        this.filterPaddingLeft = null;
        /**
         * [Style(name="filterPaddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
		 *  @property  filterPaddingRight
         */

        this.filterPaddingRight = null;
        /**
         * [Style(name="filterPaddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  filterPaddingTop
         */

        this.filterPaddingTop = null;
        /**
         * [Style(name="filterPaddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  filterPaddingBottom
         */

        this.filterPaddingBottom = null;
        /**
         * [Style(name="pagerPaddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
         *  @type Number
		 *  @property  pagerPaddingLeft
         */

        this.pagerPaddingLeft = null;
        /**
         * [Style(name="pagerPaddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
		 *  @property  pagerPaddingRight
         */

        this.pagerPaddingRight = null;
        /**
         * [Style(name="pagerPaddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  pagerPaddingTop
         */

        this.pagerPaddingTop = null;
        /**
         * [Style(name="pagerPaddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  pagerPaddingBottom
         */

        this.pagerPaddingBottom = null;
        /**
         * [Style(name="rendererPaddingLeft", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 1
         *  @type Number
		 *  @property  rendererPaddingLeft
         */

        this.rendererPaddingLeft = null;
        /**
         * [Style(name="rendererPaddingRight", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 0
         *  @type Number
		 *  @property  rendererPaddingRight
         */

        this.rendererPaddingRight = null;
        /**
         * [Style(name="rendererPaddingTop", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls left border
         *  and the left edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  rendererPaddingTop
         */

        this.rendererPaddingTop = null;
        /**
         * [Style(name="rendererPaddingBottom", type="Number", format="Length", inherit="no")]
         *  Number of pixels between the controls right border
         *  and the right edge of its content area.
         *
         *  @default 2
         *  @type Number
		 *  @property  rendererPaddingBottom
         */

        this.rendererPaddingBottom = null;
        /**
         * [Style(name="alternatingItemColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The colors to use for the backgrounds of the items in the grid.
         *  The value is an array of two or more colors.
         *  The backgrounds of the list items alternate among the colors in the array.
         *  @default undefined
         *  @type Array
		 *  @property  alternatingItemColors
         */

        this.alternatingItemColors = null;
        /**
         * [Style(name="alternatingTextColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The colors to use for the text of the items in the grid.
         *  The value is an array of two colors.
         *  The text color of the list items alternate among the colors in the array.
         *  @default [ #000000, #000000]
         *  @type Array
		 *  @property  alternatingTextColors
         */

        this.alternatingTextColors = null;
        /**
         * [Style(name="editItemColor", type="Array", arrayType="uint", format="Color")]
         *  The colors to use for the backgrounds of the items in the grid in the editable mode.
         *  The value is an array two colors.
         *  @default undefined
         *  @type Array
		 *  @property  editItemColor
         */

        this.editItemColor = null;
        /**
         * [Style(name="editTextColor", format="Color")]
         *  The colors to use for the text of the items in the editable grid.
         *  @default null
         *  @type null
		 *  @property  editItemColor
         */

        this.editTextColor = null;
        /**
         * [Style(name="selectionColor", type="uint", format="Color", inherit="yes")]
         *  The color of the background of a renderer when the user selects it.
         *
         *  The default value for the Halo theme is <code>0x7FCEFF</code>.
         *  The default value for the Spark theme is <code>0xA8C6EE</code>.
         *
         *  @default null
         *  @type uint
		 *  @property  selectionColor
         */

        this.selectionColor = null;
        /**
         * [Style(name="selectionDisabledColor", type="uint", format="Color", inherit="yes")]
         *  The color of the background of a renderer when the component is disabled.
         *  @default null
         *  @type uint
		 *  @property  selectionDisabledColor
         */

        this.selectionDisabledColor = null;
        /**
         * [Style(name="disclosureOpenIcon", type="Class", format="EmbeddedFile", inherit="no")]
         *  The icon that is displayed next to an open branch node of the navigation tree.
         *  The default value is <code>TreeDisclosureOpen</code> in the assets.swf file.
         *  @default null
         *  @type Class
		 *  @property  disclosureOpenIcon
         */

        this.disclosureOpenIcon = null;
        /**
         * [Style(name="disclosureClosedIcon", type="Class", format="EmbeddedFile", inherit="no")]
         *  The icon that is displayed next to a closed branch node of the navigation tree.
         *
         *  @default null
         *  @type Class
		 *  @property  disclosureClosedIcon
         *  The default value is <code>TreeDisclosureClosed</code> in the assets.swf file.
         */

        this.disclosureClosedIcon = null;
        /**
         * [Style(name="columnGroupOpenIcon", type="Class", format="EmbeddedFile", inherit="no")]
         *  The icon that is displayed next to an open column group.
         *
         *  The default value is <code>TreeDisclosureOpen</code> in the assets.swf file.
         *  @default null
         *  @type Class
		 *  @property  columnGroupOpenIcon
         */

        this.columnGroupOpenIcon = null;
        /**
         * [Style(name="columnGroupClosedIcon", type="Class", format="EmbeddedFile", inherit="no")]
         *  The icon that is displayed next to a closed column group.
         *
         *  The default value is <code>TreeDisclosureClosed</code> in the assets.swf file.
         *  @default null
         *  @type Class
		 *  @property  columnGroupClosedIcon
         */

        this.columnGroupClosedIcon = null;
        /**
         * [Style(name="rollOverColor", type="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the row.
         *
         *  The default value for the Halo theme is <code>0xB2E1FF</code>.
         *  The default value for the Spark theme is <code>0xCEDBEF</code>.
         *  @default null
         *  @type uint
		 *  @property  rollOverColor
         */

        this.rollOverColor = null;
        /**
         * [Style(name="activeCellColor", type="uint", format="Color", inherit="yes")]
         *  The color of the cell directly under the mouse or if using keyboard navigation, current keyboard seed.
         * @default #000000
         *  @type uint
		 *  @property   activeCellColor
         */

        this.activeCellColor = null;
        /**
         * [Style(name="footerColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  An array of two colors used to draw the footer background gradient.
         *  The first color is the top color.
         *  The second color is the bottom color.
         *  @default [0xCFCFCF, 0xCCCCCC]
         *  @type Array
		 *  @property  footerColors
         */

        this.footerColors = null;
        /**
         * [Style(name="footerRollOverColors",  type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the footer.
         *  The default value is [0xCCCCCC,0xCFCFCF]
         *  @default null
         *  @type Array
		 *  @property  footerRollOverColors
         */

        this.footerRollOverColors = null;
        /**
         * [Style(name="filterColors",  type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background for the filter.
         *  The default value is [0xCFCFCF,0xCFCFCF]
         *  @default null
         *  @type Array
		 *  @property  filterColors
         */

        this.filterColors = null;
        /**
         * [Style(name="filterRollOverColors",  type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the filter.
         *  The default value is [0xCFCFCF,0xCFCFCF]
         *  @default null
         *  @type Array
		 *  @property  filterRollOverColors
         */

        this.filterRollOverColors = null;
        /**
         * [Style(name="headerColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  An array of two colors used to draw the header background gradient.
         *  The first color is the top color.
         *  The second color is the bottom color.
         *  @default [0xFFFFFF, 0xE6E6E6]
         *  @type Array
		 *  @property  headerColors
         */

        this.headerColors = null;
        /**
         * [Style(name="headerRollOverColors",  type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the header.
         *
         *  The default value is [0xE6E6E6,0xFFFFFF]
         *  @default null
         *  @type Array
		 *  @property  headerRollOverColors
         */

        this.headerRollOverColors = null;
        /**
         * [Style(name="columnGroupColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  An array of two colors used to draw the Column Groups background gradient.
         *  The first color is the top color.
         *  The second color is the bottom color.
         *  @default [0xFFFFFF, 0xE6E6E6]
         *  @type Array
		 *  @property  columnGroupColors
         */

        this.columnGroupColors = null;
        /**
         * [Style(name="columnGroupRollOverColors",  type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the Column Groups.
         *
         *  The default value is [0xE6E6E6,0xFFFFFF]
         *  @type Array
		 *  @property  columnGroupRollOverColors
         */

        this.columnGroupRollOverColors = null;
        /**
         * [Style(name="rendererColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  An array of two colors used to draw the renderer background gradient.
         *  The first color is the top color.
         *  The second color is the bottom color.
         *  @default [0xFFFFFF, 0xFFFFFF]
         *  @type Array
		 *  @property  rendererColors
         */

        this.rendererColors = null;
        /**
         * * [Style(name="rendererRollOverColors", type="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over a level renderer.
         *  The default value is [0xFFFFFF,0xFFFFFF]
         *  @default null
         *  @type uint
		 *  @property  rendererRollOverColors
         */

        this.rendererRollOverColors = null;
        /**
         * [Style(name="pagerColors", type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  An array of two colors used to draw the pager background gradient.
         *  The first color is the top color.
         *  The second color is the bottom color.
         *  @default [0xCCCCCC, 0xCCCCCC]
         *  @type Array
		 *  @property  pagerColors
         */

        this.pagerColors = null;
        /**
         * [Style(name="pagerRollOverColors",type="Array", arrayType="uint", format="Color", inherit="yes")]
         *  The color of the row background when the user rolls over the pager.
         *  The default value is [0xE6E6E6,0xFFFFFF]
         *  @default null
         *  @type Array
		 *  @property  pagerRollOverColors
         */

        this.pagerRollOverColors = null;
        /**
         * [Style(name="textSelectedColor", type="uint", format="Color", inherit="yes")]
         *  Color of the text when the user selects a row.
         *  @default 0x000000
         *  @type uint
		 *  @property  textSelectedColor
         */

        this.textSelectedColor = null;
        /**
         * [Style(name="textRollOverColor", type="uint", format="Color", inherit="yes")]
         *  Color of the text when the user rolls over a row.
         *
         *  @default 0x000000
         *  @type uint
		 *  @property  textRollOverColor
         */

        this.textRollOverColor = null;
        /**
         * [Style(name="textDisabledColor", type="uint", format="Color", inherit="yes")]
         *  The color of the text of a renderer when the component is disabled.
         *  @default 0xDDDDDD
         *  @type uint
		 *  @property  textDisabledColor
         */

        this.textDisabledColor = null;

        /**
         * If set to true, this column wont be included in the
         * quick find mechanism by default.
         * @property excludeFromQuickFind
         */
        this.excludeFromQuickFind = false;


        //this._selectedObjects.addEventListener(this,flexiciousNmsp.ArrayCollection.EVENT_COLLECTION_CHANGE,this.onSelectedKeysChange);

    }

    getClassNames() {
        return ["FlexDataGridColumnLevel", "CSSStyleDeclaration"];
    }

    /**
     * Get the maximum depth or height of column groups.
	 * @method getMaxColumnGroupDepthOrHeight
     */
    getMaxColumnGroupDepthOrHeight(soFar, children, what) {
        if (typeof soFar == "undefined") soFar = 0;
        if (typeof children == "undefined") children = null;
        if (typeof what == "undefined") what = 'depth';

        if (children == null) {
            children = this.columnGroups;
        }
        if (what == 'depth') {
            soFar += 1;
        }
        const thisLevel = soFar;

        for (const colG of children) {
            if (what == 'height') {
                soFar += colG.getHeight();
            }
            const result = this.getMaxColumnGroupDepthOrHeight(thisLevel, colG.columnGroups, what);
            if (result > soFar) {
                soFar = result;
            }
            if (what == 'height') {
                soFar -= colG.getHeight();
            }
        }

        return soFar;
    }

    /**
     * Get the maximum  depth of column groups.
	 * @method getMaxColumnGroupDepth
     */
    getMaxColumnGroupDepth() {

        return this.getMaxColumnGroupDepthOrHeight(0, null, 'depth');

    }

    /**
     * Checks disabled flag.
	 * @method checkRowDisabled
     */
    checkRowDisabled(cell, item) {

        return item && item.hasOwnProperty(this.disabledField) && item[this.disabledField];

    }

    /**
     * Gets the index at which the given column appears in the sort list.
     * @return
	 * @method getSortIndex
     */
    getSortIndex(fld, return1, returnSortObject) {
        if (typeof return1 == "undefined") return1 = true;
        if (typeof returnSortObject == "undefined") returnSortObject = false;

        if (this.reusePreviousLevelColumns && this.getParentLevel())
            return this.getParentLevel().getSortIndex(fld);
        let index = 0;

        for (const srt of this.currentSorts) {
            index++;
            if (fld.getSortFieldName() == srt.sortColumn) {
                if (returnSortObject)
                    return srt;
                return index;
            }
        }

        return returnSortObject ? null : return1 ? 1 : -1;
    }

    /**
     * Returns true if the given column is already in the sort
     * @param fld
	 * @method hasSort
     *
     */
    hasSort(fld) {
        if (this.reusePreviousLevelColumns && this.getParentLevel())
            return this.getParentLevel().hasSort(fld);

        for (const srt of this.currentSorts) {
            if (fld.getSortFieldName() == srt.sortColumn) {
                return srt;
            }
        }

        return null;
    }

    /**
     * For multi column sort support.
	 * @method addSort
     */
    addSort(sort) {

        const col = this.getColumnByDataField(sort.sortColumn, 'sortField');
        let fs = new flexiciousNmsp.FilterSort();

        if (col) {
            const srt = this.getSortIndex(col, false, true);
            if (srt) {
                fs = srt
            }
            fs.sortNumeric = sort.sortNumeric;
            fs.sortCaseInsensitive = sort.sortCaseInsensitive;
            fs.sortColumn = sort.sortColumn;
            fs.sortCompareFunction = sort.sortCompareFunction != null ? sort.sortCompareFunction : col.sortCompareFunction != null ? col.sortCompareFunction : null;
            fs.isAscending = sort.isAscending;
            if (!srt)
                this.currentSorts.push(fs);

        } else {
            this.currentSorts.push(sort);
        }

    }

    /**
     * Removes all the sorts and calls doInvalidate.
	 * @method removeAllSorts
     */
    removeAllSorts() {

        flexiciousNmsp.UIUtils.removeAll(this.currentSorts);
        if (this.nextLevel)
            this.nextLevel.removeAllSorts();

    }

    /**
     * Clears out the current sort collection and sets it to the passed in values
     * @param fld
     * @param asc
	 * @method setCurrentSort
     *
     */
    setCurrentSort(fld, asc) {

        if (this.reusePreviousLevelColumns && this.getParentLevel())
            this.getParentLevel().setCurrentSort(fld, asc);
        flexiciousNmsp.UIUtils.removeAll(this.currentSorts);
        const fs = new flexiciousNmsp.FilterSort();

        fs.sortCaseInsensitive = fld.sortCaseInsensitive;
        fs.sortNumeric = fld.sortNumeric;
        fs.sortColumn = fld.getSortFieldName();
        fs.sortCompareFunction = fld.sortCompareFunction;
        fs.isAscending = asc;
        this.currentSorts.push(fs);

    }

    /**
     *
	 * @method createAscendingSortIcon
     */
    createAscendingSortIcon(icon) {

        return this.createSortArrow(true, icon);

    }

    /**
	 * @method createDescendingSortIcon
	 */
    createDescendingSortIcon(icon) {

        return this.createSortArrow(false, icon);

    }

    /**
	 * @method createSortArrow
	 */
    createSortArrow(d, existing) {

        const comp = existing?existing:new flexiciousNmsp.Image();
        const sortArrowClass = this.getStyleValue("sortArrowSkin");
        comp.setAttribute("src", this.grid.imagesRoot + sortArrowClass);
        comp.attachClass("sortIcon");
        comp.detachClass("descendingSort")
        if (!d) {
            comp.attachClass("descendingSort");
        }
        return comp;

    }

    /**
	 * @method onSelectedKeysChange
	 */
    onSelectedKeysChange() {
        this.grid.invalidateSelection();
        this._calculatedSelectedKeys = null;

    }

    /**
     * IF rowSelectableFunction is specified, returns result of that.
     * If selectableField is specified, returns value of that on the obj object
     * Else returns true
	 * @method checkRowSelectable
     */
    checkRowSelectable(cell, obj) {

        if (this.rowSelectableFunction != null) {
            return this.rowSelectableFunction(cell, obj);
        }
        let enabled = true;

        if (this.rowDisabledFunction != null) {
            enabled = !this.rowDisabledFunction(cell, obj);
        }
        if (this.selectableField && obj.hasOwnProperty(this.selectableField)) {
            return enabled && obj[this.selectableField];
        }
        return enabled;

    }

    /**
     *
	 * @method calculateChromeTopHeight
     */
    calculateChromeTopHeight() {

        return this.getHeaderHeight() * this.getMaxColumnGroupDepth();

    }

    /**
     *
	 * @method calculateChromeBottomHeight
     */
    calculateChromeBottomHeight() {

        let result = 0;
        result += this.getEnableFooters() ? this.getFooterRowHeight() : 0;
        result += this.enablePaging ? this.getPagerRowHeight() : 0;
        return result;

    }

    /**
     *
	 * @method setDepth
     */
    setDepth(nestLevel) {
        if (typeof nestLevel == "undefined") nestLevel = 1;

        this._nestDepth = nestLevel;
        if (this.nextLevel) {
            nestLevel++;
            this.nextLevel.setDepth(nestLevel);
        }
        // if (this.getNestDepth() > 1 && this.displayOrder.indexOf("pager") == 0) {
        //     this.displayOrder = "header,filter,body,footer,pager";
        // }

    }

    /**
	 * @method setParentLevel
     *
     */
    setParentLevel() {

        if (this.nextLevel) {
            this.nextLevel._parentLevel = this;
            this.nextLevel.setParentLevel(this);
            this.nextLevel._rowHeight = this._rowHeight;
            this.nextLevel._footerRowHeight = this._footerRowHeight;
            this.nextLevel._pagerRowHeight = this._pagerRowHeight;
            this.nextLevel._filterRowHeight = this._filterRowHeight;
            this.nextLevel._headerHeight = this._headerHeight;
        }

    }

    /**
	 * @method setGrid
     *
     */
    setGrid(grid) {

        this.grid = grid;
        if (this.nextLevel) {
            this.nextLevel.setGrid(grid);
        }


    }

    /**
	 * @method invalidateCache
     *
     */
    invalidateCache() {

        this._cachedVisibleColumns = {};
        this._lockCache = {};
        this._flowColumns = null;


    }

    /**
	 * @method setColumnLevel
     *
     */
    setColumnLevel() {

        this._wxInvalid = false;
        this.invalidateCache();
        let i;
        if (!this.reusePreviousLevelColumns || !this.getParentLevel()) {
            this.invalidateCalculationsDown();
            const loc = new InsertionLocationInfo();
            const cols = this.getVisibleColumns();
            for (i = 0; i < cols.length; i++) {
                const col = cols[i];
                if (col.getIsLeftLocked()) {
                    col.setX(loc.leftLockedContentX);
                    loc.leftLockedContentX += col.getWidth();
                }
                else if (col.getIsRightLocked()) {
                    col.setX(loc.rightLockedContentX);
                    loc.rightLockedContentX += col.getWidth();
                }
                else {
                    col.setX(loc.contentX);
                    loc.contentX += col.getWidth();
                }
                col.level = this;
            }
            const colgs = this.columnGroups;
            for (i = 0; i < colgs.length; i++) {
                const colG = colgs[i];
                colG.level = this;
                colG.initializeGroup();
                colG.initializeDepthY(); //this also initializes the first time.
            }
        }
        if (this.nextLevel) {
            this.nextLevel.setColumnLevel();
        }
    }

    /**
     *
	 * @method invalidateCalculationsDown
     */
    invalidateCalculationsDown() {
        if (this.columnGroups.length == 0) {
            return;
        }
        for (const cg of this.columnGroups) {
            cg.invalidateCalculationsDown();
        }
    }

    /**
     * Clears the selected rows and cells.
	 * @method clearSelection
     */
    clearSelection(dispatch) {
        if (typeof dispatch == "undefined") dispatch = true;


        if (this.getSelectedCells().length > 0)
            flexiciousNmsp.UIUtils.removeAll(this.getSelectedCells());
        if (this._selectedObjects.length > 0 || this._unSelectedObjects.length > 0) {
            flexiciousNmsp.UIUtils.removeAll(this._selectedObjects);
            flexiciousNmsp.UIUtils.removeAll(this._unSelectedObjects);
            this.onSelectedKeysChange();
        }
        else
            dispatch = false;

        if (this.nextLevel) {
            this.nextLevel.clearSelection(false);
        }

        if (dispatch) {
            const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.CHANGE,
                this.grid, this, null);
            this.dispatchEvent(evt);
        }

    }

    /**
	 * @method initializeLevel
     *
     * @param grid
     *
     */
    initializeLevel(grid) {


        this.setDepth();
        this.setGrid(grid);
        this.setParentLevel();
        this.setColumnLevel();
        this.ensureRowNumberColumn();


    }

    /**
     * Applies the passed in value to the provided property and cascades that value down to the next level, if one exists.
     * @param prp	The property to apply
     * @param value	The value to apply
	 * @method cascadeProperty
     */
    cascadeProperty(prp, value) {

        this[prp] = value;
        if (this.nextLevel) {
            this.nextLevel.cascadeProperty(prp, value);
        }

    }

    /**
     *
	 * @method transferProps
     */
    transferProps(from, to, checkForChanges) {
        if (typeof checkForChanges == "undefined") checkForChanges = false;

        const template = new FlexDataGridColumnLevel();
        template.grid = from.grid;
        for (const levelProp in from) {
            if (from.hasOwnProperty(levelProp) && UIUtils.isPrimitive(from[levelProp])) {
                if (to[levelProp] != from[levelProp]) {
                    if (!checkForChanges || from[levelProp] != template[levelProp])
                        to[levelProp] = from[levelProp];
                }
            }
        }
    }

    /**
     * Creates a new instance of this level, transfers all its primitive and factory
     * properties to the new instance, creates a clone collection of all its columns,
     * transfers the non primitive and factory properties of the columns to the cloned
     * columns and returns the new instance.
     * @return
	 * @method clone
     *
     */
    clone(transferCols) {
        if (typeof transferCols == "undefined") transferCols = true;

        const treeDataGridColumnLevel = new FlexDataGridColumnLevel();
        const cols = [];
        let i;
        let col;

        this.transferProps(this, treeDataGridColumnLevel);
        if (transferCols) {
            let hasColumnGroups = false;
            const groupedColumns = this.getGroupedColumns();
            for (i = 0; i < groupedColumns.length; i++) {
                col = groupedColumns[i];
                if (col.implementsOrExtends('FlexDataGridColumnGroup')) {
                    hasColumnGroups = true;
                }
            }

            for (i = 0; i < groupedColumns.length; i++) {
                col = groupedColumns[i];
                cols.push(this.cloneColumn(col));
            }
            if (hasColumnGroups)
                treeDataGridColumnLevel.setGroupedColumns(cols.slice());
            else
                treeDataGridColumnLevel.setColumns(cols.slice());//columns to print


            treeDataGridColumnLevel.currentSorts = this.currentSorts.slice();

            if (this.nextLevel) {
                treeDataGridColumnLevel.nextLevel = this.nextLevel.clone();
            }
        }

        return treeDataGridColumnLevel;
    }

    /**
	 * @method cloneColumn
	 */
    cloneColumn(col) {

        let newCol;

        if (col.implementsOrExtends('IFactory')) {//scenario when the concrete cla')){
            newCol = (col).newInstance();
        }
        else if (col.implementsOrExtends('FlexDataGridColumnGroup')) {
            newCol = new flexiciousNmsp.FlexDataGridColumnGroup();
        }
        else {
            newCol = new flexiciousNmsp.FlexDataGridColumn();
        }
        for (const colProp in col) {
            if (/*typeof col[colProp]=="function" ||*/ col[colProp] == null) continue; // TODO : commented by subash on 25/Feb/2016. label function like function silently ignored when clone.

            const isRendererOrFunctionOrFormatter = col[colProp]["implementsOrExtends"] && ((col[colProp].implementsOrExtends('IFactory'))
                || (col[colProp].implementsOrExtends('Function'))
                || (col[colProp].implementsOrExtends('Formatter)')));

            if (col.hasOwnProperty(colProp) && (
                UIUtils.isPrimitive(col[colProp]) || isRendererOrFunctionOrFormatter
            )) {
                newCol[colProp] = col[colProp];//we want to clone all primitive properties
            }
        }
        if (col.implementsOrExtends('FlexDataGridColumnGroup')) {
            const cgCols = [];
            const cgGroups = [];
            const children = col.columnGroups.length > 0 ? col.columnGroups : col.getChildren();

            for (const child of children) {
                const newChild = this.cloneColumn(child);
                if (newChild.implementsOrExtends('FlexDataGridColumnGroup')) {
                    cgGroups.push(newChild);
                } else {
                    cgCols.push(newChild)
                }
            }

            if (cgCols.length > 0)
                newCol.setColumns(cgCols);
            else if (cgGroups.length > 0)
                newCol.columnGroups = cgGroups;
        }
        return newCol;

    }

    /**
     * Gets the height of the header, footer, pager if visible.
	 * @method getRowHeightFromType
     */
    getRowHeightFromType(chromeLevel, item) {
        return chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_HEADER ||
            chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_COLUMN_GROUP ? this.getHeaderHeight() :
            chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_FOOTER ? this.getFooterRowHeight() :
                chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_PAGER ? this.getPagerRowHeight() :
                    chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_RENDERER ? this.getLevelRendererHeight(item) :
                        chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_FILTER ? this.getFilterRowHeight() : this.getRowHeight();

    }

    /**
     * Gets the sum of all column widths.
	 * @method getSumOfColumnWidths
     */
    getSumOfColumnWidths(modes) {
        if (typeof modes == "undefined") modes = null;

        if (!modes) modes = [flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE, flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_LEFT
            , flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_RIGHT];

        const arr = modes;

        let myWidth = 0;
        const cols = this.getVisibleColumns(arr);

        for (const col of cols) {
            myWidth += col.getWidth();
        }

        return myWidth;
    }

    /**
     * All columns where visible=true and lock modes match the specified lock modes.
	 * @method getVisibleColumns
     */
    getVisibleColumns(lockModes) {
        if (typeof lockModes == "undefined") lockModes = null;

        const key = lockModes ? lockModes.join("_") : "";
        if (this._cachedVisibleColumns[key]) return this._cachedVisibleColumns[key];
        const cols = [];
        const myCols = this.getColumns();

        for (const col of myCols) {
            if (col.getVisible()
                && (!lockModes ||
                    lockModes.includes(col.getColumnLockMode())))
                cols.push(col);
        }

        this._cachedVisibleColumns[key] = cols;
        return cols;
    }

    /**
     * All visible columns where excludeFromExport=false
	 * @method getExportableColumns
     * 
     * @return {Array}
     */
    getExportableColumns(lockModes, deep, options) {
        if (typeof lockModes == "undefined") lockModes = null;
        if (typeof deep == "undefined") deep = false;
        if (typeof options == "undefined") options = null;

        const cols = [];
        let i;
        const myCols = this.getColumns();
        for (i = 0; i < myCols.length; i++) {
            const col = myCols[i];
            if (!col.excludeFromExport && (!cols.includes(col)))
                cols.push(col);
        }
        if (deep && this.nextLevel) {
            const nextLevelCols = this.nextLevel.getExportableColumns(lockModes, deep, options);
            for (i = 0; i < nextLevelCols.length; i++) {
                const subCol = nextLevelCols[i];
                if (!cols.includes(subCol))
                    cols.push(subCol);
            }
        }
        const result = [];
        if (options) {
            for (i = 0; i < cols.length; i++) {
                const fdg = cols[i];
                if (!options.excludeHiddenColumns || fdg.getVisible())
                    result.push(fdg);
            }
        } else
            return cols;
        return result;
    }

    /**
     * All visible columns where sortable=true
	 * @method getSortableColumns
     * @return {Array} columns
     */
    getSortableColumns() {
        const cols = [];
        const myCols = this.getColumns();

        for (const col of myCols) {
            if (col.sortable && (!cols.includes(col)))
                cols.push(col);
        }

        return cols;
    }

    /**
     * Gets the list of filter arguments at this level.
	 * @method getFilterArguments
     * @return {ArrayCollection}
     */
    getFilterArguments() {

        return this._filterArgs;

    }

    /**
     * Removes the filter at this level, and all of this child levels, if specified.
	 * @method clearFilter
     * 
     * @param {boolean} recursive
     */
    clearFilter(recursive) {
        if (typeof recursive == "undefined") recursive = true;

        flexiciousNmsp.UIUtils.removeAll(this._filterArgs);
        if (recursive) {
            this.nextLevel.clearFilter(true);
        }

    }

    /**
     * All columns that can be "hidden"
	 * @method getShowableColumns
     * 
     * @param {Array} lockModes
     * @param {Boolean} deep
     * 
     * @return {Array}
     */
    getShowableColumns(lockModes, deep) {
        if (typeof lockModes == "undefined") lockModes = null;
        if (typeof deep == "undefined") deep = false;

        const cols = [];
        let i;
        const myCols = this.getColumns();
        for (i = 0; i < myCols.length; i++) {
            const col = myCols[i];
            if (!col.excludeFromSettings && (!cols.includes(col)))
                cols.push(col);
        }
        if (deep && this.nextLevel) {
            const nextLevelCols = this.nextLevel.getShowableColumns(lockModes, deep);
            for (i = 0; i < nextLevelCols.length; i++) {
                const subCol = nextLevelCols[i];
                if (!cols.includes(subCol))
                    cols.push(subCol);
            }
        }
        return cols;
    }

    /**
     * All visible columns where excludeFromPrint=false
	 * @method getPrintableColumns
     * 
     * @param {Array} lockModes
     * @param {boolean} deep
     * 
     * @return {Array}
     */
    getPrintableColumns(lockModes, deep) {
        if (typeof lockModes == "undefined") lockModes = null;
        if (typeof deep == "undefined") deep = true;

        const cols = [];
        let i;
        const myCols = this.getColumns();
        for (i = 0; i < myCols.length; i++) {
            const col = myCols[i];
            if (!col.excludeFromPrint && (!cols.includes(col)))
                cols.push(col);
        }

        if (deep && this.nextLevel) {
            const nextLevelCols = this.nextLevel.getPrintableColumns(lockModes, deep);
            for (i = 0; i < nextLevelCols.length; i++) {
                const subCol = nextLevelCols[i];
                if (!cols.includes(subCol))
                    cols.push(subCol);
            }
        }
        return cols;
    }

    /**
     * Returns the filter for this level
	 * @method createFilter
     * 
     * @param {object} parentObject
     * @param {flexiciousNmsp.Filter} inFilter
     * 
     * @return {flexiciousNmsp.Filter}
     */
    createFilter(parentObject, inFilter) {
        const filter = new flexiciousNmsp.AdvancedFilter();
        let i;
        let filterArg;
        filter.sorts = (this.currentSorts.slice());
        if (inFilter) {
            var items = inFilter.filterExpressions;
            for (i = 0; i < items.length; i++) {
                filterArg = items[i];
                filter.filterExpressions.push(filterArg);
            }
            filter.pageIndex = inFilter.pageIndex;
            filter.pageCount = inFilter.pageCount;
        }
        if (this.additionalFilterArgumentsFunction != null) {
            var items = this.additionalFilterArgumentsFunction(this);
            for (i = 0; i < items.length; i++) {
                filterArg = items[i];
                filter.filterExpressions.push(filterArg);
            }
        }
        if (this.enablePaging && filter.pageIndex == -1)
            filter.pageIndex = 0;
        filter.pageSize = this.pageSize;
        filter.parentObject = parentObject;
        filter.level = this;

        return filter;
    }

    /**
     * Returns the widest right locked width.
	 * @method getWidestRightLockedWidth
     * @param {number} widestRightLockedWidth
     * @return {number}
     */
    getWidestRightLockedWidth(widestRightLockedWidth) {
        if (typeof widestRightLockedWidth == "undefined") widestRightLockedWidth = -1;

        const myWidth = this.getRightLockedWidth();
        if (widestRightLockedWidth < myWidth)
            widestRightLockedWidth = myWidth;
        const nextLevelWidth = this.nextLevel ? this.nextLevel.getWidestRightLockedWidth(widestRightLockedWidth) : 0;
        if (widestRightLockedWidth < nextLevelWidth)
            widestRightLockedWidth = nextLevelWidth;
        return widestRightLockedWidth;

    }

    /**
	 * @method getWidestLeftLockedWidth
     * @param {number} widestLeftLockedWidth
     * @return {number}
	 */
    getWidestLeftLockedWidth(widestLeftLockedWidth) {
        if (typeof widestLeftLockedWidth == "undefined") widestLeftLockedWidth = -1;

        const myWidth = this.getLeftLockedWidth();
        if (widestLeftLockedWidth < myWidth)
            widestLeftLockedWidth = myWidth;
        const nextLevelWidth = this.nextLevel ? this.nextLevel.getWidestLeftLockedWidth(widestLeftLockedWidth) : 0;
        if (widestLeftLockedWidth < nextLevelWidth)
            widestLeftLockedWidth = nextLevelWidth;
        return widestLeftLockedWidth;

    }

    /**
     * List of visible columns of the specified lock modes.
	 * @method getColumnsByLockMode
     * @param {Array} types
     * @return {Array}
     */
    getColumnsByLockMode(types) {
        if (this._lockCache[types]) return this._lockCache[types];
        const result = [];
        const myCols = this.getVisibleColumns(types);

        for (const col of myCols) {
            result.push(col);
        }

        this._lockCache[types] = result;
        return result;
    }

    /**
     * At each level, get the sum of column widths, and return the width at the
     * level with the widest width. Only includes locked columns by default.
	 * @method getWidestWidth
     * 
     * @param {number} widestWidth
     * @param {boolean} deep
     * 
     * @return {number}
     */
    getWidestWidth(widestWidth, deep) {
        if (typeof deep == "undefined") deep = true;

        let myWidth = this.grid.lockDisclosureCell ? 0 : this.getNestDepth() * this.getNestIndent();
        myWidth += this.getSumOfColumnWidths([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE]);
        if (widestWidth < myWidth)
            widestWidth = myWidth;
        const nextLevelWidth = (this.nextLevel && deep) ? this.nextLevel.getWidestWidth(widestWidth) : 0;
        if (widestWidth < nextLevelWidth)
            widestWidth = nextLevelWidth;
        return widestWidth;

    }

    /**
	 * @method getColumnsByWidthMode
     * @param {Array} modes
     * @return {Array}
	 */
    getColumnsByWidthMode(modes) {
        const result = [];
        const cols = this.getVisibleColumns([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE]);

        for (const col of cols) {
            if (modes.includes(col.getColumnWidthMode())) {
                result.push(col);
            }
        }

        return result;
    }

    /**
     * Goes through the unlocked columns, figures out the widest width needed by each of the columns
     * by going through the data provider, and sets the width accordingly.
     *
     * If the equally parameter is set to true, all column settings are ignored, and the grid's width
     * is distributed equally across all columns
     *
     * Below is a description of the logic used:
     *
     * @copy FlexDataGridColumn._columnWidthMode
	 * @method setColumnWidthsUsingWidthMode
     * @param {boolean} equally
     */
    setColumnWidthsUsingWidthMode(equally) {
        if (typeof equally == "undefined") equally = false;

        let widthsAdded = 0;
        let cols = this.getVisibleColumns([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE]);
        let provider;
        let col;
        let i;
        let lstCol;

        if (!this.grid.getForceColumnsToFitVisibleArea()) {
            const myCols = this.getColumns();
            for (i = 0; i < myCols.length; i++) {
                lstCol = myCols[i];
                if (lstCol._naturalLastColWidth) {
                    lstCol.setWidth(lstCol._naturalLastColWidth);
                    lstCol._naturalLastColWidth = 0;
                }
            }
            if (cols.length > 0) {
                lstCol = cols[cols.length - 1];
            }
        }


        let av = this.getAvailableWidth();
        if (av <= 0) {
            return;
        }


        if (equally) {
            for (i = 0; i < cols.length; i++) {
                col = cols[i];
                col.setWidth((av - Constants.HORIZONTAL_SCROLLBAR_HEIGHT) / cols.length);
            }
            return;
        }
        const hasFitToContent = this.getColumnsByWidthMode([flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_FIT_TO_CONTENT]).length > 0;
        if (hasFitToContent) {
            provider = this.grid.flatten(this.getNestDepth(), false, true, true, true, this.columnWidthModeFitToContentSampleSize);
        }
        for (i = 0; i < cols.length; i++) {
            col = cols[i];
            if ((!this.grid.getForceColumnsToFitVisibleArea()) && !col.getIsLocked() && col.getIsLastUnLocked()) {
                continue;
            }
            if (col.getColumnWidthMode() == flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_FIXED
                || col.getColumnWidthMode() == flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_NONE) {
                //do nothing, we just use the width property user provided.
                //col.width=col.width;
                widthsAdded += col.getWidth();
            }
            else if (col.getColumnWidthMode() == flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_FIT_TO_CONTENT) {
                this.applyColumnWidthFromContent(col, provider);
                widthsAdded += col.getWidth();
            }
        }
        if (!this.grid.getForceColumnsToFitVisibleArea()) {
            //handle scenario where added column widths are less than the available,
            av=this.getAvailableWidth();
            if (widthsAdded < av) {
                if (lstCol) {
                    if (!lstCol._naturalLastColWidth)
                        lstCol._naturalLastColWidth = lstCol.getWidth();
                    lstCol.setWidth(av - widthsAdded);
                }
            }
            if (this.nextLevel)
                this.nextLevel.setColumnWidthsUsingWidthMode();
            return;
        }
        //now that we have allocated space to fixed Width and fitToContent columns, distribute the remaining space
        //based on percentages.
        let remaining = av - widthsAdded;
        if (remaining > 0) {
            cols = this.getColumnsByWidthMode([flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_PERCENT]);
            for (i = 0; i < cols.length; i++) {
                col = cols[i];
                const percent = col.percentWidth;
                if (percent) {
                    col.setWidth((percent * remaining) / 100);
                    widthsAdded += col.getWidth();
                }
            }
        }
        remaining = av - widthsAdded;
        //now that the columns are all allocated for, if we are required to fit within the visible area, then we need
        //to squish columns that are none width OR percent width.
        //if(remaining>1)
        {
            cols = this.getColumnsByWidthMode([flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_NONE,
            flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_PERCENT]);

            for (i = 0; i < cols.length; i++) {
                col = cols[i];
                col.setWidth(col.getWidth() + (remaining / (cols.length)));
                if (col.getWidth() < col.minWidth)
                    col.setWidth(col.minWidth);
            }
        }
        remaining = av - this.getSumOfColumnWidths([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE]);
        //this means we went over, because of some columns. we now HAVE to squish every non fixed on.
        cols = this.getColumnsByWidthMode([flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_NONE,
        flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_PERCENT, flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_FIT_TO_CONTENT]);
        let availableToTakeOff = 0;
        for (i = 0; i < cols.length; i++) {
            col = cols[i];
            if (col.getWidth() > col.minWidth) {
                availableToTakeOff += col.getWidth();
            }
        }
        if (availableToTakeOff > 0) {
            for (i = 0; i < cols.length; i++) {
                col = cols[i];
                if (col.getWidth() > col.minWidth) {
                    col.setWidth(col.getWidth() + ((remaining * col.getWidth()) / availableToTakeOff));
                }
            }
        } else {
            //cannot do anything - user asked to display too many columns in an area that is not enough!
        }


        cols = this.getVisibleColumns([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE]);
        let round = this.getSumOfColumnWidths([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE]) - av;
        if (cols.length > 0) {
            col = cols[cols.length - 1];
            col.setWidth(col.getWidth() - round);
        }
        cols = this.getVisibleColumns([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_RIGHT]);
        if (cols.length > 0) {
            round = this.getWidestRightLockedWidth() - this.grid.getRightLockedContent().getWidth();
            col = cols[cols.length - 1];
            col.setWidth(col.getWidth() - round);
        }
        cols = this.getVisibleColumns([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_LEFT]);
        if (cols.length > 0) {
            round = this.getWidestLeftLockedWidth() - this.grid.getLeftLockedContent().getWidth();
            col = cols[cols.length - 1];
            col.setWidth(col.getWidth() - round);
        }

        if (this.nextLevel)
            this.nextLevel.setColumnWidthsUsingWidthMode();

        this.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.COLUMNS_RESIZED, this.grid));
    }

    /**
     * Given a column and a dataprovider, measures the width required to render the text in the column, and if such width
     * is greater than the minwidth of the column, applies that width to the column
     *
     * @param col		The FlexDataGridColumn object
     * @param provider	The provider used to figure out the text width. You may use the grid.flatten(this.getNestDepth(),false,true,true,true,columnWidthModeFitToContentSampleSize);
	 * @method applyColumnWidthFromContent
     */
    applyColumnWidthFromContent(col, provider) {
        if (col.columnWidthModeFitToContentWidthCalculator) {
            col.setWidth(col.columnWidthModeFitToContentWidthCalculator(col, provider));
        } else {
            const offset = col.columnWidthOffset;
            let longestString = "";

            for (const obj of provider) {
                if (obj == null) continue;
                const str = col.itemToLabel(obj).toString();
                if (str && (longestString.length < str.length)) {
                    longestString = str;
                }
            }

            if (col.getHeaderText() && longestString.length < col.getHeaderText().length && !col.columnWidthModeFitToContentExcludeHeader)
                longestString = col.getHeaderText();
            if (longestString.length > 0) {
                const textLineMetrics = this.grid.measureText(longestString);
                col.setWidth(textLineMetrics.width + offset + parseInt(col.getStyleValue("paddingLeft")) +
                    parseInt(col.getStyleValue("paddingRight")));//need to account for default padding
            }

        }
        if (col.getWidth() < col.minWidth)
            col.setWidth(col.minWidth);
    }

    /**
     * If the grid's horizontal scroll policy is off, then returns the width of the
     * unlocked section minus the width of the scroll bar else returns the sum of all
     * column widths.
	 * @method getAvailableWidth
     * @return {number} availableWidth
     */
    getAvailableWidth() {

        let availableWidth;
        if (this.grid.getForceColumnsToFitVisibleArea())
            availableWidth = (this.grid.getBodyContainer().getWidth() - (!this.grid.implementsOrExtends("PrintFlexDataGrid") ? this.grid.isVScrollBarVisible : 0));
        else
            availableWidth = Math.max(this.grid.getColumnLevel().getWidestWidth(-1, false)
                , (this.grid.getBodyContainer().getWidth() - (!this.grid.implementsOrExtends("PrintFlexDataGrid") ? this.grid.isVScrollBarVisible : 0)));
        return availableWidth;

    }

    /**
     * Once we have the widest width, since we don't want the grid to appear jagged in multi levels
     * Adjust all columns where the widths are same so it looks like a contiguous rectangle.
     *
     * If the equally parameter is set to true, it will sum up the total column widths, and resize them
     * all equally.
	 * @method adjustColumnWidths
     * 
     * @param {number} widestWidth
     * @param {boolean} equally
     */
    adjustColumnWidths(widestWidth, equally) {
        if (typeof widestWidth == "undefined") widestWidth = -1;
        if (typeof equally == "undefined") equally = false;

        this.setColumnWidthsUsingWidthMode(equally);

    }

    /**
     * If a value exists for a style property at the  level return it, else return the value of the property at the grid level.
	 * @method getStyleValue
     * 
     * @param {String} styleProp
     * @return {*}
     */
    getStyleValue(styleProp) {

        let result = this.getStyle(styleProp);
        if (result == undefined)
            result = this.grid.getStyle(styleProp);
        return result;

    }

    /**
     *
	 * @method invalidateList
     */
    invalidateList() {

        if (this.grid)
            this.grid.reDraw();

    }

    /**
     * For header checkbox state, returns if all items in the grid are selected.
	 * @method areAllSelected
     * 
     * @param {Object} items
     * @param {Boolean} checkLength
     * 
     * @return {Boolean}
     */
    areAllSelected(items, checkLength) {
        if (typeof checkLength == "undefined") checkLength = true;

        //if(checkLength && (items.length != selectedKeys.length)) return false;

        for (const item of items) {
            if (!this.existsInCursor((item))) {
                if (this.checkRowSelectable(null, item))
                    return false;
            }
            if (this.nextLevel && this.nextLevel.reusePreviousLevelColumns) {
                if (!this.nextLevel.areAllSelected(this.getChildren(item), false)) {
                    return false;
                }

            }
        }

        return true;
    }

    /**
     * If there is a selected key field (usually a field that uniquely identifies an object)
     * returns the value of the selectedKeyField property of the item else returns the item itself.
     * If saveLevel is set to true, returns a string in the format key|levelNestDepth. Use this
     * in cases where you wish to refresh the grid with new data from the server and open items
     * should only refer to the keys so the open state can be persisted across server roundtrips.
	 * @method getItemKey
     * 
     * @param {object} item
     * @param {boolean} saveLevel
     * 
     * @return {object}
     */
    getItemKey(item, saveLevel) {
        if (typeof saveLevel == "undefined") saveLevel = false;

        if (saveLevel && this.selectedKeyField)
            return `${this.getValue(UIUtils.resolveExpression(item, this.selectedKeyField))}|${this.getNestDepth()}`;
        else
            return this.selectedKeyField ? this.getValue(UIUtils.resolveExpression(item, this.selectedKeyField)) : item;

    }

    /**
     * If value is primitive, does a to string, if XMLList, returns the text.
     * @param val
     * @return
	 * @method getValue
     *
     */
    getValue(val) {
        return val;

    }

    /**
     * Provided an item or a key, and a flag that indicates what it is,
     * returns true if the item matches the respective comparison.
	 * @method isItemEqual
     * 
     * @param {object} itemOrKeyToCompare
     * @param {object} rowData
     * @param {boolean} useItemKeyForCompare
     * 
     * @return {boolean}
     */
    isItemEqual(itemOrKeyToCompare, rowData, useItemKeyForCompare) {

        return (!useItemKeyForCompare && (rowData == itemOrKeyToCompare))
            || (useItemKeyForCompare && (this.getItemKey(rowData) == itemOrKeyToCompare));


    }

    /**
	 * @method areItemsEqual
     * 
     * @param {object} itemA
     * @param {object} itemB
     * 
     * @return {boolean}
	 */
    areItemsEqual(itemA, itemB) {

        if (this.selectedKeyField) {
            return ((UIUtils.resolveExpression(itemA, this.selectedKeyField)) == (UIUtils.resolveExpression(itemB, this.selectedKeyField)));

        } else {
            return itemA == itemB;
        }

    }

    /**
     * Provided an item key, loops through the data provider, and finds
     * the item with the provided key.
	 * @method getItemFromKey
     * 
     * @param {*} itemKey
     * @param {object} flat
     * 
     * @return {object}
     */
    getItemFromKey(itemKey, flat) {
        if (typeof flat == "undefined") flat = null;

        if (!flat) flat = this.grid.getRootFlat();

        for (const item of flat) {
            if (this.getItemKey(item) == itemKey) {
                return item;
            }
            if (this.nextLevel) {
                const result = this.nextLevel.getItemFromKey(itemKey, this.getChildren(item));
                if (result) {
                    return result;
                }
            }
        }

        return null;
    }

    /**
     * Returns true if the getItemKey for the specified item exists in the selectedKeys collection.
	 * @method isItemSelected
     * 
     * @param {object} item
     * @param {boolean} useExclusion
     * 
     * @return {boolean}
     */
    isItemSelected(item, useExclusion) {
        if (typeof useExclusion == "undefined") useExclusion = true;

        if (useExclusion && this.grid.enableSelectionExclusion) {
            return (this.getCheckBoxStateBasedOnExclusion(item) == flexiciousNmsp.TriStateCheckBox.STATE_CHECKED);
        }
        let selected = false;
        if (this.selectedKeyField) {
            const itemKey = UIUtils.resolveExpression(item, this.selectedKeyField);
            const v = this.getSelectedKeys();//just to ensure the cursor is read.
            selected = (v.includes(itemKey));
        }
        else
            selected = (this._selectedObjects.includes(item));
        return selected;

    }

    /**
     * Returns true if the CellInfo for the specified item and column exists in the selectedCells collection.
	 * @method isCellSelected
     * 
     * @param {object} item
     * @param {flexiciousNmsp.FlexDataGridColumn} col
     * 
     * @return {flexiciousNmsp.FlexDataGridCell} 
     */
    isCellSelected(item, col) {
        const items = this.getSelectedCells();

        for (const cell of items) {
            if (cell.rowData == item && cell.getColumn() == col)
                return cell;
        }

        return null;
    }

    /**
     *Selects or un selects a cell
	 * @method selectCell
     * 
     * @param {flexiciousNmsp.FlexDataGridColumn} cell
     * @param {boolean} selected
     */
    selectCell(cell, selected) {
        if (typeof selected == "undefined") selected = true;

        const cellInfo = this.isCellSelected(cell.rowInfo.getData(), cell.getColumn());
        if (!cell.getColumn())
            return;
        if (selected) {
            if (!cellInfo) {
                if (cell.level.grid.getSelectionMode() == flexiciousNmsp.NdgBase.SELECTION_MODE_SINGLE_CELL)
                flexiciousNmsp.UIUtils.removeAll(this.getSelectedCells());
                this.getSelectedCells().push(new flexiciousNmsp.CellInfo(cell.rowInfo.getData(), cell.getColumn()));
            }
        }
        else if (!selected) {
            if (cellInfo) {
                this.getSelectedCells().splice(this.getSelectedCells().indexOf(cellInfo), 1);
            }
        }
        const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.CHANGE,
            this.grid, this, null);
        this.dispatchEvent(evt);


    }

    /**
     * Inserts the column specified column before the specified column
	 * @method shiftColumns
     * 
     * @param {flexiciousNmsp.FlexDataGridColumn} columnToInsert
     * @param {flexiciousNmsp.FlexDataGridColumn} insertBefore
     * @param {boolean} movingCg
     */
    shiftColumns(columnToInsert, insertBefore, movingCg) {
        if (typeof movingCg == "undefined") movingCg = false;

        const cols = this.getColumns().slice();
        const arrayCollection = cols;
        arrayCollection.splice(arrayCollection.indexOf(columnToInsert), 1);
        arrayCollection.splice(arrayCollection.indexOf(insertBefore), 0, columnToInsert);
        this.setColumns(arrayCollection);

        if (columnToInsert.columnGroup && !movingCg) {
            const columnGroups = columnToInsert.level.getGroupedColumns();

            for (const cg1 of columnGroups) {
                if (cg1.implementsOrExtends('FlexDataGridColumnGroup')) {
                    this.swapColumns(cg1, columnToInsert, insertBefore);
                }
            }
        }

        this.setColumnLevel();
        this.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.COLUMNS_SHIFT, this.grid, this));

        const hScroll = this.grid.getHorizontalScrollPosition();
        const vScroll = this.grid.getVerticalScrollPosition();
        this.grid.reDraw();
        if (!this.grid.getForceColumnsToFitVisibleArea() && hScroll > 0) {
            this.grid.validateNow();
            const that = this;
            setTimeout(() => {
                that.grid.gotoHorizontalPosition(hScroll);
                that.grid.gotoVerticalPosition(vScroll);
            }, 10);
        }

    }

    /**
	 * @method swapColumns
     * 
     * @param {flexiciousNmsp.FlexDataGridColumnGroup} cg
     * @param {flexiciousNmsp.FlexDataGridColumn} columnToInsert
     * @param {flexiciousNmsp.FlexDataGridColumn} insertBefore
	 */
    swapColumns(cg, columnToInsert, insertBefore) {
        const arrayCollection = (cg.getColumns());
        if (arrayCollection.includes(columnToInsert) && arrayCollection.includes(insertBefore)) {
            arrayCollection.splice(arrayCollection.indexOf(columnToInsert), 1);
            arrayCollection.splice(arrayCollection.indexOf(insertBefore), 0, columnToInsert);
            cg.setColumns(arrayCollection);
        }

        for (const cg1 of cg.columnGroups) {
            if (cg1.implementsOrExtends('FlexDataGridColumnGroup')) {
                this.swapColumns(cg1, columnToInsert, insertBefore);
            }
        }
    }

    /**
     * Selects or un selects a row
	 * @method selectRow
     * 
     * @param {object} item
     * @param {boolean} selected
     * @param {boolean} dispatch
     * @param {boolean} recurse
     * @param {boolean} bubble
     * @param {object} parent
     * 
     */
    selectRow(item, selected, dispatch, recurse, bubble, parent) {
        if (typeof selected == "undefined") selected = true;
        if (typeof dispatch == "undefined") dispatch = true;
        if (typeof recurse == "undefined") recurse = true;
        if (typeof bubble == "undefined") bubble = true;
        if (typeof parent == "undefined") parent = null;
        let i;
        if (this.grid.enableSelectionExclusion) {
            this.selectRowExclusion(item, selected);
        }
        else {
            if (recurse) {
                if (this.nextLevel && (this.grid.enableSelectionCascade)) {
                    const items = this.getChildren(item, false, false, false);

                    if (this.nextLevel.enableSingleSelect) {
                        //Check on parent, at least 1 child row should be selected. Allow apps to provide which row will be default selected. Need data hooks/APIs for the same.
                        for (i = 0; i < items.length; i++) {
                            const child1 = items[i];
                            if (this.nextLevel.defaultSelectionField) {
                                if (UIUtils.resolveExpression(child1, this.nextLevel.defaultSelectionField)) {
                                    this.nextLevel.selectRow(child1, selected, false, true, false, item);
                                    break;
                                }
                            }
                            else {
                                //select the first child
                                this.nextLevel.selectRow(child1, selected, false, true, false, item);
                                break;
                            }
                        }
                    }
                    else {
                        for (const child of items) {
                            this.nextLevel.selectRow(child, selected, false, true, false, item);
                        }
                    }
                }
            }
            if (this.rowDisabledFunction != null) {
                const disable = this.rowDisabledFunction(null, item);
                if (disable) return;
            }
            const selectable = this.checkRowSelectable(null, item);
            if (!selectable) return;

            if (selected && this.enableSingleSelect && this.getNestDepth() > 1) {
                //need to un select all my siblings
                if (!parent)
                    parent = this.grid.getParent(item, this);
                const childList = this.grid.getChildren(parent, this.getParentLevel());
                for (i = 0; i < childList.length; i++) {
                    const sibling = childList[i];
                    if (this.getItemKey(sibling) != this.getItemKey(item)) {
                        if (this.isItemSelected(sibling)) {
                            this.selectRow(sibling, false, false, false, false, parent);
                        }
                    }
                }
            }


            if (this.grid.getSelectionMode() == flexiciousNmsp.NdgBase.SELECTION_MODE_SINGLE_ROW)
                this.grid.clearSelection();
            if (selected) {
                //if(!existsInCursor(item))
                {
                    this.addSelectedItem(item);
                }
            }
            else if (!selected) {
                //if(existsInCursor(item))
                {
                    this.removeSelectedItem(item);
                }
            }
            if (recurse && this.grid.enableSelectionBubble && bubble) {
                if (this.enableSingleSelect && selected) {
                    //just add parent to selection as well
                    if (this.getParentLevel()) {
                        const parent1 = this.grid.getParent(item, this);
                        this.getParentLevel().selectRow(parent1, true, false, false, false);
                    }
                }
                else
                    this.bubbleSelection(item);
            }
        }
        if (dispatch) {
            const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.CHANGE,
                this.grid, this, null);
            this.dispatchEvent(evt);
        }

    }

    /**
     *
	 * @method bubbleSelection
     * @param {Object} item
     */
    bubbleSelection(item) {

        if (this.getParentLevel()) {
            const parent = this.grid.getParent(item, this);
            if (parent) {
                const allChildren = this.grid.getChildren(parent, this.getParentLevel());
                if (this.areAllSelected(allChildren)) {
                    this.getParentLevel().selectRow(parent, true, false, false, false);
                } else if (this.getParentLevel().isItemSelected(parent, false)) {
                    this.getParentLevel().selectRow(parent, false, false, false, false);
                }
                this.getParentLevel().bubbleSelection(parent);
            }
        }

    }

    /**
	 * @method existsInCursor
     * @param {Object} item
     *
     */
    existsInCursor(item) {


        return this.isItemSelected(item);

    }

    /**
	 * @method addSelectedItem
     * @param {Object} item
     *
     */
    addSelectedItem(item) {
        if (!this._selectedObjects.includes(item)) {
            this._selectedObjects.push(item);
            this.onSelectedKeysChange();
            this.storeChain(item);
        }

    }

    /**
	 * @method setEnablePaging
     * @param val
     *
     */
    setEnablePaging(val) {
        this.enablePaging = val;
    }

    /*if(selectedKeyField && _calculatedSelectedKeysCursor){
     if(_calculatedSelectedKeysCursor.findAny(UIUtils.resolveExpression(item,selectedKeyField)))
     return;
     }*/
    /**
	 * @method storeChain
     * @param {Object} item
     *
     */
    storeChain(item) {

        if (this.grid.enableSelectionExclusion) {

            //when an item.implementsOrExtends('added to the selection (or exclusion), we will need to track its ascendants, because we ne')){
            //to mark them as middle state.
            if (!this.grid.selectionChain[this.getItemKey(item)]) {
                this.grid.selectionChain[this.getItemKey(item)] = [];
            }
            this.pushIntoChain(this.grid.selectionChain[this.getItemKey(item)], item);
            this.grid.selectionChain[this.getItemKey(item)].reverse();
        }

    }

    /**
	 * @method pushIntoChain
     * @param {Array} chain
     * @param {Object} item
     */
    pushIntoChain(chain, item) {

        chain.push(item);
        if (this.getParentLevel()) {
            const parent = this.grid.getParent(item, this);
            if (parent) {
                this.getParentLevel().pushIntoChain(chain, parent);
            }
        }

    }

    /**
     *
	 * @method removeSelectedItem
     * @param {object} item
     */
    removeSelectedItem(item) {


        if (this.selectedKeyField.length > 0 && !this._selectedObjects.includes(item)) {
            const itemKey = UIUtils.resolveExpression(item, this.selectedKeyField);

            for (const selectedItem of this._selectedObjects) {
                if (itemKey == UIUtils.resolveExpression(selectedItem, this.selectedKeyField)) {
                    item = selectedItem;
                    break;
                }
            }
        }
        const itemIndex = this._selectedObjects.indexOf(item);
        if (itemIndex >= 0)
            this._selectedObjects.splice(itemIndex, 1);
        this.onSelectedKeysChange();
    }

    /**
     * Sets the selected state at this level. If you specify checked, adds all rows
     * to the selected keys collection.
     * @param val
	 * @method setSelectedKeysState
     */
    setSelectedKeysState(val) {

        this.clearSelection(false);
        if (val == flexiciousNmsp.TriStateCheckBox.STATE_CHECKED) {
            const dp = this.grid.getDataProvider();

            for (const item of dp) {
                this.addSelectedItem((item));
            }
        }
        const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.CHANGE,
            this.grid, this, null);
        this.dispatchEvent(evt);


    }

    /**
     * For nested/grouped hierarchical datagrids, used to select records.
     * Iterates through the data provider, matching each object, and if matches, adds the object
     * to the selectedObjects for the level
	 * @method setSelectedObjects
     * 
     * @param {Array} objects
     * @param {boolean} openItems
     */
    setSelectedObjects(objects, openItems) {
        if (typeof openItems == "undefined") openItems = true;

        this.selectAll(true, true, null, objects, false, openItems);

    }

    /**
     * For nested/grouped hierarchical datagrids, used to select records.
     * Iterates through the data provider, matching each object, and if matches, adds the object
     * to the selectedObjects for the level
	 * @method setSelectedKeys
     * 
     * @param {Array} objects
     * @param {boolean} openItems
     */
    setSelectedKeys(objects, openItems) {
        if (typeof openItems == "undefined") openItems = true;

        this.selectAll(true, true, null, objects, true, openItems);

    }

    /**
     * @method selectAll
     * 
     * @param {boolean} select
     * @param {boolean} dispatch
     * @param {*} provider
     * @param {Array} subset
     * @param {boolean} useKeys
     * @param {boolean} openItems
     * 
     * @return 
     */
    selectAll(select, dispatch, provider, subset, useKeys, openItems) {
        if (typeof dispatch == "undefined") dispatch = true;
        if (typeof provider == "undefined") provider = null;
        if (typeof subset == "undefined") subset = null;

        if (this.getNestDepth() == 1) {
            this.grid._useSelectedSelectAllState = select ? flexiciousNmsp.TriStateCheckBox.STATE_CHECKED : flexiciousNmsp.TriStateCheckBox.STATE_UNCHECKED;
            //clear out everything before we start
            if (this.grid.enableSelectionExclusion) {
                this.grid.setSelectAllState(select ? flexiciousNmsp.TriStateCheckBox.STATE_CHECKED : flexiciousNmsp.TriStateCheckBox.STATE_UNCHECKED);
                return false;
            }
            else {
                if (this.grid.enableMaintainSelectionOnSelectAll) {
                    if (!select) {
                        const dp = this.grid.getRootFlat();
                        for (let idx = 0; idx < dp.length; idx++) {
                            this.removeSelectedItem(dp[idx]);
                        }
                    }
                }
                else {
                    this.clearSelection(false);
                }
            }
        }
        let returnValue = false;
        if (select) {
            if (!provider) {
                provider = this.grid.getRootFlat();
                if (this.getNestDepth() == 1 && this.grid.getIsClientFilterPageSortMode())
                    provider = UIUtils.filterArray(provider, this.grid.getRootFilter(), this.grid, this, false)
            }

            for (const item of provider) {
                if (this.enableSingleSelect && select) {
                    if (this.defaultSelectionField) {
                        if (!UIUtils.resolveExpression(item, this.defaultSelectionField)) {
                            continue;
                        }
                    }
                }

                let check = true;
                if (this.rowDisabledFunction != null) {
                    check = !this.rowDisabledFunction(null, item);
                }
                if (check) {
                    check = this.checkRowSelectable(null, item);

                }
                if (subset && check) {
                    check = subset.includes(useKeys ? this.getItemKey(item) : item);
                }
                if (check) {
                    this.addSelectedItem(item);
                    returnValue = true;
                }
                if (this.nextLevel && (this.grid.enableSelectionCascade || subset)) {
                    const nextVal = this.nextLevel.selectAll(select, false, this.getChildren(item), subset, useKeys, openItems);
                    if (nextVal && openItems) {
                        this.grid.getBodyContainer().addOpenItem(item, null, true, this);
                        returnValue = true;
                    }
                }
                if (this.enableSingleSelect && check) {
                    if (!this.defaultSelectionField) {
                        break;
                    }
                }
            }
        }

        if (dispatch) {
            const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.CHANGE,
                this.grid, this, null);
            this.dispatchEvent(evt);
        }
        //if(returnValue && this.getNestDepth()==1){
        //    this.grid.rebuildBody();
        //}
        return returnValue;

    }
    /**
     * For nested datagrids, used to get all records at all levels
	 * @method getSelectedObjectsDeep
     * 
     * @param {Array} soFar
     * @param {boolean} getKey
     * @param {boolean} unSelected
     */
    getSelectedObjectsDeep(soFar, getKey, unSelected) {
        if (typeof getKey == "undefined") getKey = false;
        if (typeof unSelected == "undefined") unSelected = false;

        const arr = unSelected ? this._unSelectedObjects : this._selectedObjects;

        for (const item of arr) {
            soFar.push(getKey ? this.getItemKey(item) : item);
        }

        if (this.nextLevel) {
            this.nextLevel.getSelectedObjectsDeep(soFar, getKey, unSelected);
        }
    }

    /**
     * If all items are selected, returns TriStateCheckBox.STATE_CHECKED.
     * If none are selected, returns TriStateCheckBox.STATE_UNCHECKED
     * else returns TriStateCheckBox.STATE_MIDDLE
	 * @method getSelectedKeysState
     * 
     * @param {object} allItems
     * @return {String}
     */
    getSelectedKeysState(allItems) {

        if (this.grid.getLength(allItems) == 0) {
            return flexiciousNmsp.TriStateCheckBox.STATE_UNCHECKED;
        }
        else if (this.areAllSelected(allItems)) {
            return flexiciousNmsp.TriStateCheckBox.STATE_CHECKED;
        }
        else if (!this.areAnySelected(allItems)) {
            return flexiciousNmsp.TriStateCheckBox.STATE_UNCHECKED;
        } else {
            return flexiciousNmsp.TriStateCheckBox.STATE_MIDDLE;
        }

    }

    /**
     * Returns true if there is at least one item in the selected keys collection
     * for this level, or if the next level uses this levels columns, then  the
     * selected keys collection for the next level.
	 * @method areAnySelected
     * 
     * @param {object} itemsToCheck
     * @param {boolean} recursive
     * 
     * @return {boolean}
     */
    areAnySelected(itemsToCheck, recursive) {
        if (typeof itemsToCheck == "undefined") itemsToCheck = null;
        if (typeof recursive == "undefined") recursive = false;

        for (const item of itemsToCheck) {
            for (const key of this._selectedObjects) {
                if (item == key) {
                    return true;
                }
            }

            if (recursive) {
                if (this.nextLevel && this.nextLevel.areAnySelected(this.getChildren(item))) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * If the dataprovider is IHierarchicalCollectionView, calls the getChildren method on the incoming object.
     * Else returns the value of the level.childrenField property on the object being passed in
	 * @method getChildren
     * 
     * @param {object} object
     * @param {boolean} filter
     * @param {boolean} page
     * @param {boolean} sort
     * 
     * @return {object}
     */
    getChildren(object, filter, page, sort) {
        if (typeof filter == "undefined") filter = false;
        if (typeof page == "undefined") page = false;
        if (typeof sort == "undefined") sort = false;

        return this.grid.getChildren(object, this, filter, page, sort)

    }

    /**
     * Calls getChildren, and if result is XML or XMLList, returns length() else returns length;
	 * @method getChildrenLength
     * 
     * @param {object} object
     * @param {boolean} filter
     * @param {boolean} page
     * @param {boolean} sort
     * 
     * @return {number}
     */
    getChildrenLength(object, filter, page, sort) {
        if (typeof filter == "undefined") filter = false;
        if (typeof page == "undefined") page = false;
        if (typeof sort == "undefined") sort = false;

        return this.grid.getChildrenLength(object, this, filter, page, sort)

    }

    /**
     * Whenever a tree datagrid event is dispatched at any level,
     * we also dispatch it from the top level grid so if someone wants
     * to listen for an event at all levels...
	 * @method dispatchEvent
     * @param event
     */
    dispatchEvent(event) {

        if (event.implementsOrExtends('FlexDataGridEvent') ||
            event.implementsOrExtends('ExtendedFilterPageSortChangeEvent')) {
            if (!this.grid.dispatchEvent(event)) {
                event.preventDefault();
            }
        }
        return flexiciousNmsp.EventDispatcher.prototype.dispatchEvent.apply(this, [event]);
    }

    /**
     * Sets the visible flag on all columns except the ones specified in the list to false.
	 * @method showColumns
     * @param {Array} colsToShow
     */
    showColumns(colsToShow) {
        const myCols = this.getColumns();

        for (const col of myCols) {
            col.setVisible(UIUtils.doesArrayContainObjectValue(colsToShow, "name", col.getUniqueIdentifier()));
        }

        if (this.nextLevel) this.nextLevel.showColumns(colsToShow);
    }

    /**
     * Sets the visible flag on all columns except the ones where excludeFromPrint=true.
	 * @method showPrintableColumns
     */
    showPrintableColumns() {
        const myCols = this.getColumns();

        for (const col of myCols) {
            col.setVisible(!col.excludeFromPrint);
        }

        if (this.nextLevel) this.nextLevel.showPrintableColumns();
    }

    /**
     *Returns the column with the specified datafield
	 * @method getColumnByDataField
     * 
     * @param {String} fld
     * @param {String} by
     * 
     * @return {flexiciousNmsp.FlexDataGridColumn}
     */
    getColumnByDataField(fld, by) {
        if (typeof by == "undefined") by = 'dataField';
        const myCols = this.getColumns();

        for (const col of myCols) {
            if (col[`get${UIUtils.doCap(by)}`]() == fld) {
                return col;
            }
        }

        if (this.nextLevel && !this.nextLevel.reusePreviousLevelColumns) {
            const retVal = this.nextLevel.getColumnByDataField(fld, by);
            if (retVal) {
                return retVal;
            }
        }
        return null;
    }

    /**
     *Returns the column with the specified UniqueIdentifier
	 * @method getColumnByUniqueIdentifier
     * @param {String} fld
     * @return {flexiciousNmsp.FlexDataGridColumnLevel}
     */
    getColumnByUniqueIdentifier(fld) {
        const myCols = this.getColumns();

        for (const col of myCols) {
            if (col.getUniqueIdentifier() == fld) {
                return col;
            }
        }

        return null;
    }

    /**
     * Clears out all the columns of the grid
     * If the parameter to rebuild is true, the grid will be rebuilt.
     * Otherwise, just the columns will be cleared, and the consumer code
     * should rebuild the grid.
	 * @method clearColumns
     * @param {boolean} rebuild
     */
    clearColumns(rebuild) {
        if (typeof rebuild == "undefined") rebuild = true;

        this._columns = [];

    }

    /**
     * Adds the column to the collection of columns at this level.
     * Deprecated. Use columns setter instead. <br/>
     * var cols:Array=level.columns;<br/>
     * //manipulate cols<br/>
     * level.columns=cols;<br/>
	 * @method addColumn
     * @param {flexiciousNmsp.FlexDataGridColumn} col
     */
    addColumn(col) {

        this._columns.push(col);
        col.level = this;
        this.setColumns(this._columns);
        this.adjustColumnWidths();

    }

    /**
     * Removes the column from the collection of columns at this level.
	 * @method removeColumn
     * @param {flexiciousNmsp.FlexDataGridColumn} col
     */
    removeColumn(col) {

        const colIndex = col ? col.getIndex() : -1;
        if (colIndex >= 0) {
            this._columns.splice(colIndex, 1);
        }

        this.setColumns(this._columns);

    }

    /**
     * Modifies the columns so the width is distributed equally.
     * Columns that are fixedWidth are not updated.
	 * @method distributeColumnWidthsEqually
     */
    distributeColumnWidthsEqually() {

        this.adjustColumnWidths(-1, true);

    }

    /**
     * Returns the column groups specified at the level cgLevel.
	 * @method getColumnGroupsAtLevel
     * 
     * @param {number} cgLevel
     * @param {Array} groups
     * @param {Array} result
     * @param {number} start
     * 
     * @return {Array} 
     */
    getColumnGroupsAtLevel(cgLevel, groups, result, start) {
        if (typeof groups == "undefined") groups = null;
        if (typeof result == "undefined") result = null;
        if (typeof start == "undefined") start = 0;

        start++;
        if (result == null)
            result = [];
        if (groups == null)
            groups = this.columnGroups;

        for (const grp of groups) {
            if (start == cgLevel) {
                result.push(grp);
            } else {
                this.getColumnGroupsAtLevel(cgLevel, grp.columnGroups, result, start);
            }
        }

        return result;
    }

    /**
	 * @method ensureRowNumberColumn
	 */
    ensureRowNumberColumn() {

    }

    /**
     * A list of ID values, based on the selectedKeyField property.
     * For hierarchical datagrids, used to get all open(expanded) records at all levels.
     * Please NOTE - this is a read only collection. Adding to this does nothing.
     *
     * If you want to programatically modify the collapse expand, use the setOpenKeys instead, or add directly to the open items.
	 * @method getOpenKeys
     * 
     * @param {Array} keys
     * @param {object} provider
     * 
     * */
    getOpenKeys(keys, provider) {
        if (typeof provider == "undefined") provider = null;

        if (!provider) {
            provider = this.grid.getRootFlat();
        }

        for (const item of provider) {
            if (this.isItemOpen(item))
                keys.push(this.getItemKey(item));
            if (this.nextLevel) {
                this.nextLevel.getOpenKeys(keys, this.getChildren(item));
            }
        }
    }

    /**
     * Sets the open keys at this level.
	 * @method setOpenKeys
     * 
     * @param {Array} keys
     * @param {object} provider
     * 
     * */
    setOpenKeys(keys, provider) {
        if (typeof provider == "undefined") provider = null;

        if (!provider) {
            provider = this.grid.getRootFlat();
        }

        for (const item of provider) {
            if (keys.includes(this.getItemKey(item)))
                this.grid.getBodyContainer().addOpenItem(item);

            if (this.nextLevel) {
                this.nextLevel.setOpenKeys(keys, this.getChildren(item));
            }
        }
    }

    /**
     * Iterates through items, checks to see if the selected bit is on, and if so, adds it to the open list.
	 * @method setSelectedItemsBasedOnSelectedField
     * 
     * @param {Array} items
     * @param {object} parent
     * @param {Array} openItems
     * 
     * @return {boolean}
     */
    setSelectedItemsBasedOnSelectedField(items, parent, openItems) {
        if (typeof parent == "undefined") parent = null;
        if (typeof openItems == "undefined") openItems = true;

        let result = false;

        for (const item of items) {
            if (item.hasOwnProperty(this.selectedField) && item[this.selectedField]) {
                if (parent) {
                    if (openItems) {
                        if (!this.grid.getOpenItems().indexOf(parent) >= 0) {
                            this.grid.getOpenItems().push(parent);
                        }
                    }
                }
                if (!this.isItemSelected(item, false)) {
                    this.addSelectedItem(item);
                    result = true;
                }
            }
            if (this.nextLevel) {
                const nlResult = this.nextLevel.setSelectedItemsBasedOnSelectedField(this.getChildren(item), item, openItems);
                result = nlResult || result;
            }
        }

        return result;
    }

    /**
	 * @method addUnSelectedItem
     * @param {object} item
	 */
    addUnSelectedItem(item) {


        this._unSelectedObjects.push(item);
        this.storeChain(item);

    }

    /**
     *
	 * @method removeUnSelectedItem
     * @param {object} item
     */
    removeUnSelectedItem(item) {


        if (this.selectedKeyField.length > 0 && !this._unSelectedObjects.includes(item)) {
            const itemKey = UIUtils.resolveExpression(item, this.selectedKeyField);

            for (const selectedItem of this._unSelectedObjects) {
                if (itemKey == UIUtils.resolveExpression(item, this.selectedKeyField)) {
                    item = selectedItem;
                    break;
                }
            }
        }
        const itemIndex = this._unSelectedObjects.indexOf(item);
        if (itemIndex >= 0)
            this._unSelectedObjects.splice(itemIndex, 1);

    }

    /**
     *
	 * @method selectRowExclusion
     * 
     * @param {object} item
     * @param {boolean} selected
     */
    selectRowExclusion(item, selected) {

        if (!this.checkRowSelectable(null, item)) return;
        if (!selected) {//user unchecked an item
            if (this.isItemSelected(item, false)) { //if it was explicitly selected previously
                this.removeSelectedItem(item); //remove it from the explicit selection
            } else if (!this.isItemUnSelected(item)) { //if it was not explicitly unselected previously
                this.addUnSelectedItem(item); //add it to explicit un selection
            }
        } else if (selected) {//user checked an item
            if (this.isItemUnSelected(item)) {//if it was explicitly unselected before
                this.removeUnSelectedItem(item);//remove from explicit un selection
            } else if (!this.isItemSelected(item, false)) {//if it was not explicitly selected before
                this.addSelectedItem(item); //add it to explicit selection.
                if (this.enableSingleSelect && this.getNestDepth() > 1) {
                    const myChain = this.grid.selectionChain[this.getItemKey(item)];

                    for (const sibling of this._selectedObjects) {
                        const siblingChain = this.grid.selectionChain[this.getItemKey(sibling)];
                        if (myChain && siblingChain && siblingChain.length > 1 && myChain.length > 1 && !this.areItemsEqual(sibling, item)) {
                            //this means that the
                            if (this.getParentLevel().areItemsEqual(siblingChain[this.getNestDepth() - 2], myChain[this.getNestDepth() - 2])) {
                                //this means the parents are same
                                this.removeSelectedItem(sibling);
                            }
                        }
                    }
                }
            }
        }
        if (this.grid.enableSelectionCascade && this.nextLevel) {
            this.nextLevel.getItemsInSelectionHierarchy(item, true, this.getNestDepth(), null, true);
        }

    }

    /**
	 * @method getCheckBoxStateBasedOnExclusion
     * 
     * @param {object} item
     * @param {boolean} useBubble
     * @param {boolean} checkDisabled
     * 
     * @return 
	 */
    getCheckBoxStateBasedOnExclusion(item, useBubble, checkDisabled) {
        if (typeof useBubble == "undefined") useBubble = true;
        if (typeof checkDisabled == "undefined") checkDisabled = true;

        //if explicitly selected, return checked
        //if explicitly unselected, return unchecked
        //if enableSelectionCascade, and.implementsOrExtends('parent Selected, return parent.sta')){
        //if enableSelectionBubble
        let state = this.grid.getSelectAllState();
        if (checkDisabled && this.rowDisabledFunction != null && this.rowDisabledFunction(null, item)) {
            return flexiciousNmsp.TriStateCheckBox.STATE_UNCHECKED;
        }
        if (checkDisabled && this.rowSelectableFunction != null && !this.rowSelectableFunction(null, item)) {
            return flexiciousNmsp.TriStateCheckBox.STATE_UNCHECKED;
        }
        if (this.isItemSelected(item, false)) {
            state = flexiciousNmsp.TriStateCheckBox.STATE_CHECKED;
        }
        else if (this.isItemUnSelected(item)) {
            state = flexiciousNmsp.TriStateCheckBox.STATE_UNCHECKED;
        }
        else if (this.grid.enableSelectionCascade && this.getParentLevel()) {
            const parent = this.grid.getParent(item, this);
            if (parent) {
                state = this.getParentLevel().getCheckBoxStateBasedOnExclusion(parent, false, false);
            }
        }
        if (this.grid.enableSelectionBubble && this.nextLevel && useBubble) {
            {

                //get next level children state
                const childrenCount = this.getNumChildren(item);
                if (childrenCount == 0) {
                    //whatever we inherited.
                } else {
                    let checkCount = 0;
                    let unCheckCount = 0;
                    const middleCount = 0;

                    checkCount = this.nextLevel.getItemsInSelectionHierarchy(item, false, this.getNestDepth(), null, false, false).length;//explicit check count
                    unCheckCount = this.nextLevel.getItemsInSelectionHierarchy(item, false, this.getNestDepth(), null, false, true).length;//explicit  un check count
                    if (this.nextLevel.enableSingleSelect) {
                        if (checkCount > 0)
                            state = flexiciousNmsp.TriStateCheckBox.STATE_CHECKED;
                        else
                            state = flexiciousNmsp.TriStateCheckBox.STATE_UNCHECKED;
                    }
                    else if (unCheckCount == childrenCount) {
                        state = flexiciousNmsp.TriStateCheckBox.STATE_UNCHECKED;
                    } else if (checkCount == childrenCount) {
                        state = flexiciousNmsp.TriStateCheckBox.STATE_CHECKED;
                    } else if (checkCount == 0 && unCheckCount == 0) {
                        //inherited
                    }
                    else if (this.grid.enableTriStateCheckbox && (checkCount || unCheckCount)) {
                        state = flexiciousNmsp.TriStateCheckBox.STATE_MIDDLE;
                    }
                }
            }
        }
        return state;

    }

    /**
     * Returns true if the getItemKey for the specified item exists in the selectedKeys collection.
	 * @method isItemUnSelected
     * @param {number} item
     * @return {boolean} 
     */
    isItemUnSelected(item) {

        if (this.selectedKeyField) {
            const itemKey = UIUtils.resolveExpression(item, this.selectedKeyField);

            for (const unselectedItem of this._unSelectedObjects) {
                const unselectedKey = UIUtils.resolveExpression(unselectedItem, this.selectedKeyField);
                if (unselectedKey == itemKey) {
                    return true;
                }
            }

            return false;
        }
        else
            return this._selectedObjects.includes(item);


    }

    /**
	 * @method getNumChildren
     * @param {object} item
     * @return {number}
	 */
    getNumChildren(item) {
        const numChildren = 0;
        if (this.childrenCountField) {
            return UIUtils.resolveExpression(item, this.childrenCountField);
        }
        const items = this.grid.getBodyContainer().getLoadedItems();

        for (const loadingItem of items) {
            if (loadingItem.isEqual(item, this, this.selectedKeyField.length > 0)) {
                return loadingItem.totalRecords;
            }
        }

        return this.grid.getLength(this.getChildren(item));
    }

    /**
     *
	 * @method getItemsInSelectionHierarchy
     * 
     * @param {object} parent
     * @param {boolean} recurse
     * @param {Number} inNestDepth
     * @param {Array} returnValue
     * @param {boolean} clear
     * @param {boolean} getUnSelected
     * 
     * @return {Array}
     */
    getItemsInSelectionHierarchy(parent, recurse, inNestDepth, returnValue, clear, getUnSelected) {
        if (!returnValue) {
            returnValue = [];
        }
        const toClearSelectedObjects = [];
        const toClearUnselectedObjects = [];
        let lvl;
        let i;
        let item;
        const coll = getUnSelected ? this._unSelectedObjects : this._selectedObjects;
        for (i = 0; i < coll.length; i++) {
            item = coll[i];
            var chain = this.grid.selectionChain[this.getItemKey(item)];
            if (chain == null) return returnValue;
            var itemAtMyLevel = chain[inNestDepth - 1];
            lvl = this.grid.getLevel(inNestDepth);
            if (lvl.areItemsEqual(itemAtMyLevel, parent)) {
                returnValue.push(item);
                toClearSelectedObjects.push(item);
            }
        }

        if (clear) {
            for (i = 0; i < this._unSelectedObjects.length; i++) {
                item = this._unSelectedObjects[i];
                chain = this.grid.selectionChain[this.getItemKey(item)];
                itemAtMyLevel = chain[inNestDepth - 1];
                lvl = this.grid.getLevel(inNestDepth);
                if (lvl.areItemsEqual(itemAtMyLevel, parent)) {
                    toClearUnselectedObjects.push(item);
                }
            }

            for (i = 0; i < toClearSelectedObjects.length; i++) {
                item = toClearSelectedObjects[i];
                this._selectedObjects.splice(this._selectedObjects.indexOf(item), 1);
            }
            for (i = 0; i < toClearUnselectedObjects.length; i++) {
                item = toClearUnselectedObjects[i];
                this._unSelectedObjects.splice(this._unSelectedObjects.indexOf(item), 1);
            }
            this.onSelectedKeysChange();
        }
        if (recurse && this.nextLevel) {
            this.nextLevel.getItemsInSelectionHierarchy(parent, recurse, inNestDepth, returnValue, clear, getUnSelected);
        }
        return returnValue;
    }

    /**
     * For columns with headerWordWrap, calculates the height required to display text in full.
	 * @method calculateHeaderHeights
     */
    calculateHeaderHeights() {

        if (this.headerVisible) {
            if (!this.grid.measurer) {
                this.grid.measurer = document.createElement("div");
                this.grid.measurer.style.position = "absolute";
                UIUtils.addChild(document.body, this.grid.measurer);
            }


            this.grid.getBodyContainer().variableRowHeightRenderers = new flexiciousNmsp.KeyValuePairCollection();
            let ht = this.minHeaderHeight;
            const cols = this.getFlowHeaderColumns();

            for (const col of cols) {
                const paddingLeft = col.getStyleValue("headerPaddingLeft");
                const paddingTop = col.getStyleValue("headerPaddingTop");
                const paddingRight = col.getStyleValue("headerPaddingRight");
                const paddingBottom = col.getStyleValue("headerPaddingBottom");
                //paddingLeft+=col.enableHierarchicalNestIndent?level.getMaxPaddingCellWidth():0;
                ht = this.grid.measureCellHeight(col, paddingLeft, paddingRight, paddingTop, paddingBottom,
                    col.headerRenderer, ht, col.getHeaderText(), null, this.getStyleValue("headerStyleName"), true);
            }

            ht += this.grid.variableHeaderHeightOffset;
            this.setHeaderHeight(ht);
            this.grid.getBodyContainer().clearVariableRowHeightRenderes();
            document.body.removeChild(this.grid.measurer);
            this.grid.measurer = null;
        }

    }

    /**
     * Columns where wordWrap=true || variableRowHeightUseRendererForCalculation and headerRenderer is not null
     * @return {Array}
	 * @method getFlowHeaderColumns
     */
    getFlowHeaderColumns() {
        if (this._flowHeaderColumns && this._flowHeaderColumns.length) return this._flowHeaderColumns;
        this._flowHeaderColumns = [];
        const cols = this.getVisibleColumns();

        for (const col of cols) {
            if (col.headerWordWrap) {
                this._flowHeaderColumns.push(col);
            } else if (this.grid.variableRowHeightUseRendererForCalculation && col.headerRenderer) {
                this._flowHeaderColumns.push(col);
            }
        }

        return this._flowHeaderColumns;
    }

    /**
	 * @method getRightLockedWidth
     * @return {number} column Widths
	 */
    getRightLockedWidth() {
        let colWidths = 0;
        const cols = this.getRightLockedColumns();

        for (const col of cols) {
            colWidths += col.getWidth();
        }

        return colWidths;
    }

    /**
     * Same as pagerRowHeight, except no built in logic to handle enableFilters and filterVisible.
     * @return {number}
	 * @method getPagerHeight
     */
    getPagerHeight() {
        return this._pagerRowHeight;
    }

    /**
	 * @method setPagerHeight
     * @param {number} value
	 */
    setPagerHeight(value) {
        this._pagerRowHeight = value;
    }

    /**
     * Height of the header for this level. Defaults to 25
	 * @method getHeaderHeight
     * @return {number}
     */
    getHeaderHeight() {
        return (this.headerVisible) ? this._headerHeight != 0 ? (this._headerHeight) : (this.getRowHeight()) : 0;
    }

    /**
     *
	 * @method setHeaderHeight
     * @param {number} o
     */
    setHeaderHeight(o) {
        if (this._headerHeight != o) {
            this.invalidateList();
        }
        this._headerHeight = o;
    }

    /**
     * A Class that implements com.flexicious.controls.interfaces.IPager.
     * Defaults to a built in implementation
     * @see com.flexicious.grids.columns.footers.PagerControl.
     * @see implements com.flexicious.controls.interfaces.IPager.
     * @return 
	 * @method getPagerRenderer
     */
    getPagerRenderer() {
        if (this._pagerRenderer == null) {
            this._pagerRenderer = FlexDataGridColumnLevel.static_FlexDataGridPager;
        }
        return this._pagerRenderer;
    }

    /**
     * A Class that implements com.flexicious.controls.interfaces.IPager.
     * Defaults to a built in implementation
     * @see com.flexicious.grids.columns.footers.PagerControl.
     * @see implements com.flexicious.controls.interfaces.IPager.
     * @param value
	 * @method setPagerRenderer
     *
     */
    setPagerRenderer(value) {
        this._pagerRenderer = value;
    }

    /**
     * Gets all unlocked columns
	 * @method getUnLockedColumns
     * @return {Array}
     */
    getUnLockedColumns() {
        return this.getColumnsByLockMode([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE])
    }

    /**
     * Grouped columns - only supported at the top level.
	 * @method getGroupedColumns
     * @return {Array}
     */
    getGroupedColumns() {
        return this._groupedColumns.length == 0 ? this.getColumns() : this._groupedColumns;
    }

    /**
     * Grouped columns - only supported at the top level.
	 * @method setGroupedColumns
     * @param {Array} value
     */
    setGroupedColumns(value) {
        this._groupedColumns = value.slice();
        this.hasGroupedColumns = this._groupedColumns.length >= 1;
        this.invalidateCache();
        const cols = [];
        const colgs = [];

        for (const val of value) {
            if (val.implementsOrExtends('FlexDataGridColumnGroup')) {
                for (const col of val.getAllColumns()) {
                    cols.push(col);
                }

                colgs.push(val)
            } else {
                cols.push(val);
            }
        }

        this.setColumns(cols);
        this.columnGroups = colgs;
    }

    /**
     * Returns true if filterPageSortMode==client.
	 * @method getIsClientFilterPageSortMode
     * @return {boolean} 
     */
    getIsClientFilterPageSortMode() {
        return this.filterPageSortMode == "client";
    }

    /**
	 * @method getTraceCols
     * @return {str}
	 */
    getTraceCols() {
        let str = "";
        const myCols = this.getColumns();

        for (const col of myCols) {
            str += `${col.getHeaderText()}:${col.getWidth()}`;
            str += "\n";
            //str +=row.getY() + ":" + (row.data.name?row.data.name:row.data.id) + ":" + row.rowPositionInfo.getRowType()+ "\n";
        }

        return str;
    }

    /**
     * Gets the width of the padding cell.
	 * @method getMaxPaddingCellWidth
     * @return {number}
     */
    getMaxPaddingCellWidth() {
        /*
        let lvl = this;
        let nd = 0;
        while (lvl != null) {
            nd += (lvl.getNestDepth() == 1 ? 0 : lvl.getNestIndent());
            lvl = lvl.getParentLevel();
        }
        
        return nd;
        */

        let totalParentNestIndent = (this.getNestIndentPlusParent() - this.getNestIndent());

        if (!this.nextLevel) {
            return this.grid.maxDisclosureCellWidth || totalParentNestIndent;
        } else {
            const extra = this.grid.maxDisclosureCellWidth ? (this.grid.maxDisclosureCellWidth - totalParentNestIndent) : 0;
            return (totalParentNestIndent - (this.nextLevel && !this.nextLevel.nextLevel ? extra : 0));
        }
    }

    /**
     * Height of the filter for this level. Defaults to 25
	 * @method getFilterRowHeight
     * @return {number} 
     */
    getFilterRowHeight() {
        return (this.getEnableFilters() && this.filterVisible) ? this._filterRowHeight != 0 ? this._filterRowHeight : this.getRowHeight() : 0;
    }

    /**
     *
	 * @method setFilterRowHeight
     * @param {number} val
     */
    setFilterRowHeight(val) {
        if (this._filterRowHeight != val) {
            this.invalidateList();
        }
        this._filterRowHeight = val;
    }

    /**
     * Columns where wordWrap=true || variableRowHeightUseRendererForCalculation and itemRenderer not null
	 * @method getFlowColumns
     * @return {Array}
     */
    getFlowColumns() {
        if (this._flowColumns) return this._flowColumns;
        this._flowColumns = [];
        const cols = this.getVisibleColumns();

        for (const col of cols) {
            if (col.wordWrap) {
                this._flowColumns.push(col);
            } else if (this.grid.variableRowHeightUseRendererForCalculation && col.itemRenderer) {
                this._flowColumns.push(col);
            }
        }

        return this._flowColumns;
    }

    /**
     * This array is a list of all items that are have been selected by the user or programmatically, at this level.
     *
     *
     * Please note, for hierarchical grids, each level has its own collection of selected keys.
     * The grid.selectedObjects returns the selectedKeys at the top level. To access the selected keys
     * at a lower level, you may navigate to that level using grid.getColumnLevel().nextLevel...nextLevel.selectedKeys
     * You can also call grid.getSelectedObjects to get selected objects at all levels.
	 * @method getSelectedObjects
     * @return {Array}
	 */
    getSelectedObjects() {
        return this._selectedObjects;
    }

    /**
     * Gets the sum of left lock column widths
	 * @method getUnLockedWidth
     * @return {number}
     */
    getUnLockedWidth() {
        let colWidth = 0;
        const cols = this.getUnLockedColumns();

        for (const col of cols) {
            colWidth += col.getWidth();
        }

        return colWidth;
    }

    /**
     * Gets all left locked columns
	 * @method getLeftLockedColumns
     * @return {Array}
     */
    getLeftLockedColumns() {
        return this.getColumnsByLockMode([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_LEFT])
    }

    /**
     * Same as filterRowHeight, except no built in logic to handle enableFilters and filterVisible.
	 * @method getFilterHeight
     * @return {number}
     */
    getFilterHeight() {
        return this._filterRowHeight;
    }

    /**
	 * @method setFilterHeight
	 */
    setFilterHeight(value) {
        this._filterRowHeight = value;
    }

    /**
     * Returns true if itemLoadMode==client.
	 * @method getIsClientItemLoadMode
     * @return {boolean}
     */
    getIsClientItemLoadMode() {
        return this.itemLoadMode == "client";
    }

    /**
     * Gets the max disclosure cell width.
	 * @method getMaxDisclosureCellWidth
     * @return {number}
     */
    getMaxDisclosureCellWidth() {
        return (this.grid.maxDisclosureCellWidth  || (this.getDeepNestIndent() - (this.getMaxPaddingCellWidth())))
    }

    /**
     * Gets the level that is the ultimate column owner.
	 * @method getColumnOwnerLevel
     * @return {flexiciousNmsp.FlexDataGridColumnLevel}
     */
    getColumnOwnerLevel() {
        if (this.reusePreviousLevelColumns && this.getParentLevel()) return this.getParentLevel().getColumnOwnerLevel();
        return this;
    }

    /**
     * A list of ID values, based on the selectedKeyField property.
     * When a row is selected in the grid, we store the selectedKeyField property
     * of the selected object in this array collection. This allows for multiple
     * pages of data that comes down from the server and not maintained in memory
     * to still keep track of the ids that were selected on other pages.
	 * @method getSelectedKeys
     * @return {Array}
     */
    getSelectedKeys() {
        if (this._calculatedSelectedKeys)
            return this._calculatedSelectedKeys;
        const sKeys = [];

        for (const item of this._selectedObjects) {
            sKeys.push(this.getItemKey(item));
        }

        this._calculatedSelectedKeys = sKeys;
        return this._calculatedSelectedKeys;
    }

    /**
     * Gets nestindent recursively
     * @return {number}
	 * @method getDeepNestIndent
     */
    getDeepNestIndent() {
        let lvl = this.grid.getColumnLevel();
        let nd = 0;
        while (lvl != null) {
            nd += lvl.getNestDepth() == 1 ? 0 : lvl.nestIndent;
            if (!lvl.nextLevel && lvl.nextLevelRenderer) {
                nd += lvl.nestIndent
            }

            lvl = lvl.nextLevel;

        }
        return nd;
    }

    /**
	 * @method getHasFilterFunction
     * @return {boolean}
	 */
    getHasFilterFunction() {
        return (this.filterFunction != null) || (this.nextLevel && this.nextLevel.getHasFilterFunction())
    }

    /**
     * Gets all right locked columns
	 * @method getRightLockedColumns
     * @return {Array}
     */
    getRightLockedColumns() {
        return this.getColumnsByLockMode([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_RIGHT])
    }

    /**
     * Height of the row for this level. Defaults to 25
	 * @method getRowHeight
     * @return {number}
     */
    getRowHeight() {
        return this._rowHeight;
    }

    /**
     * @return {boolean} o
	 * @method setRowHeight
     */
    setRowHeight(o) {
        if (this._rowHeight != o) {
            this.invalidateList();
        }
        this._rowHeight = o;
    }

    /**
     * Returns true if the nest depth is 1 and grid as the enableMultiColumnSort field enabled.
     * @return {boolean}
	 * @method getEnableMultiColumnSort
     *
     */
    getEnableMultiColumnSort() {
        return this.grid.enableMultiColumnSort;//false
    }

    /**
     * When grid.enableSelectionExclusion is set to true, a list of items that the user
     * explicitly unselected is stored here.
     * @return {Array}
	 * @method getUnSelectedObjects
     */
    getUnSelectedObjects() {
        return this._unSelectedObjects;
    }

    /**
     * Returns the first item from the selectedObjects collection, if one exists at this level. If there is none,
     * then returns the selectedItem collection of the next level.
	 * @method getSelectedItem
     * @return {Object}
     */
    getSelectedItem() {
        return this._selectedObjects.length > 0 ? this._selectedObjects[0] : this.nextLevel ? this.nextLevel.selectedItem : null;
    }

    /**
     * A function that lets you control the background drawing mechanism for each cell.
     * By default, each cell will draw a background on basis of the XXXXColors style property
     * and XXXXRollOverColors style property, where XXXX= header,filter,pager,footer,renderer or blank.
     *
     * This function offers you the opportunity to hook into this mechanism, and draw
     * your own background using the graphics property of the FlexDataGridCell object,
     * that is passed in as a parameter to this function. If this function returns false,
     * the default background is not drawn.
	 * @method getCellCustomBackgroundDrawFunction
     * @return {Function}
     */
    getCellCustomBackgroundDrawFunction() {
        return this._cellCustomDrawFunction;
    }

    /**
     *
	 * @method setCellCustomBackgroundDrawFunction
     * @param {Function} value
     */
    setCellCustomBackgroundDrawFunction(value) {
        this._cellCustomDrawFunction = value;
    }

    /**
     * Flag that indicates whether or not to show the filter
	 * @method getEnableFilters
     * @return {boolean}
     */
    getEnableFilters() {
        return this._enableFilters;
    }

    /**
     *
     * Sets whether or not to show the filter
     * @param value
	 * @method setEnableFilters
     */
    setEnableFilters(value) {
        this._enableFilters = value;
        this.invalidateList();
        this.dispatchEvent(new flexiciousNmsp.BaseEvent("enableFiltersChanged", undefined, undefined));
    }

    /**
     * Flag that indicates whether or not to show the footer
	 * @method getEnableFooters
     * @return {boolean}
     */
    getEnableFooters() {
        return this._enableFooters;
    }

    /**
     *
     * Sets whether or not to show the footer
	 * @method setEnableFooters
     * @param {boolean} value
     */
    setEnableFooters(value) {
        this._enableFooters = value;
        this.invalidateList();
        this.dispatchEvent(new flexiciousNmsp.BaseEvent("enableFootersChanged"));
    }

    /**
     * Height of the footer for this level. Defaults to 25
	 * @method getFooterRowHeight
     * @return {number} row height
     */
    getFooterRowHeight() {
        return (this.getEnableFooters() && this.footerVisible) ? this._footerRowHeight != 0 ? this._footerRowHeight : this.getRowHeight() : 0;
    }

    /**
     *
	 * @method setFooterRowHeight
     * @param {number} val
     */
    setFooterRowHeight(val) {
        if (this._footerRowHeight != val) {
            this.invalidateList();
        }
        this._footerRowHeight = val;
    }

    /**
     * A list of <code>CellInfo</code> objects. When grid.selectionMode=SINGLE_CELL, this collection contains only one item
	 * @method getSelectedCells
     * @return {Array}
     */
    getSelectedCells() {
        return this.grid ? this.grid.getSelectedCells() : [];
    }

    /**
     * How deep in the nesting hierarchy is this item.
     * Starts at 1 for the columnLevel that belongs to the grid.
	 * @method getNestDepth
     * @return
     */
    getNestDepth() {
        return this._nestDepth;
    }

    /**
     * FlexDataGrid.columns
	 * @method getColumns
     * @return {Array}
     */
    getColumns() {
        return (this.reusePreviousLevelColumns && this.getParentLevel()) ? this.getParentLevel().getColumns()
            : this._columns ? this._columns.slice() : [];
    }

    /**
     *
	 * @method setColumns
     * @param {Array} value
     */
    setColumns(value) {
        this._columns = value.slice(0);
        this.ensureRowNumberColumn();
        this.invalidateCache();

        for (const col of this._columns) {
            col.level = this;
            col.calcualtedMeasurements = {};
        }
    }

    /**
     * The parent level for this level. For the root level, this is null.
	 * @method getParentLevel
     * @return {flexiciousNmsp.FlexDataGridColumnLevel}
     */
    getParentLevel() {
        return this._parentLevel;
    }

    /**
     * The height of the pager row.
     * If not set, defaults to the value of the rowHeight property
	 * @method getPagerRowHeight
     * @return {number}
     */
    getPagerRowHeight() {
        return ((this.enablePaging && this.pagerVisible) || this.forcePagerRow) ? this._pagerRowHeight != 0 ? this._pagerRowHeight : this.getRowHeight() : 0;

    }

    /**
     *
	 * @method setPagerRowHeight
     * @param {number} val
     */
    setPagerRowHeight(val) {
        if (this._pagerRowHeight != val)
            this.invalidateList();
        this._pagerRowHeight = val;
    }

    getNestIndentPlusParent(){
        return this.nestIndent + (this._parentLevel? this._parentLevel.getNestIndentPlusParent() : 0)
    }

    /**
     * Gets the sum of left lock column widths
	 * @method getLeftLockedWidth
     * @return {number}
     */
    getLeftLockedWidth() {
        let colWidths = 0;
        const myCols = this.getLeftLockedColumns();

        for (const col of myCols) {
            colWidths += col.getWidth();
        }

        if (this.grid.lockDisclosureCell && (this.nextLevel || this.nextLevelRenderer || (this.itemLoadMode == "server")) &&
            this.grid.enableDefaultDisclosureIcon) {
                colWidths += (this.grid.maxDisclosureCellWidth || this.getNestIndentPlusParent())
        }
        return colWidths;
    }

    /**
	 * @method getNestIndent
     * @return {number}
	 */
    getNestIndent() {
        return this.nestIndent;
    }

    /**
	 * @method getLevelRendererHeight
     * @return {boolean}
	 */

    getLevelRendererHeight(item) {
        return item && this.levelRendererHeightFunction ? this.levelRendererHeightFunction(item) : this.levelRendererHeight;
    }

    /**
	 * @method kill
	 */
    kill() {
        flexiciousNmsp.EventDispatcher.prototype.kill.apply(this);
        this.invalidateCache();
        this._selectedObjects = [];
        this.openItems = [];
        this.currentSorts = [];
        if (this._columns)
            this.killArray(this._columns);
        this._columns = [];
        if (this.columnGroups)
            this.killArray(this.columnGroups);
        this.columnGroups = [];
        this.grid = null;
        if (this.nextLevel) {
            this.nextLevel.kill();
        }
    }
}

flexiciousNmsp.FlexDataGridColumnLevel = FlexDataGridColumnLevel; //add to name space
FlexDataGridColumnLevel.prototype.typeName = FlexDataGridColumnLevel.typeName = 'FlexDataGridColumnLevel';//for quick inspection

FlexDataGridColumnLevel.static_FlexDataGridExpandCollapseCell = new flexiciousNmsp.ClassFactory(flexiciousNmsp.FlexDataGridExpandCollapseCell, undefined);
FlexDataGridColumnLevel.static_FlexDataGridExpandCollapseHeaderCell = new flexiciousNmsp.ClassFactory(flexiciousNmsp.FlexDataGridExpandCollapseHeaderCell, undefined);
FlexDataGridColumnLevel.static_FlexDataGridPaddingCell = new flexiciousNmsp.ClassFactory(flexiciousNmsp.FlexDataGridPaddingCell, undefined);
FlexDataGridColumnLevel.static_FlexDataGridPagerCell = new flexiciousNmsp.ClassFactory(flexiciousNmsp.FlexDataGridPagerCell, undefined);
FlexDataGridColumnLevel.static_FlexDataGridPager = new flexiciousNmsp.ClassFactory(flexiciousNmsp.PagerControl, undefined);
/**
 * Dispatched when an item opens, both in server and client mode
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event itemClose
 */

FlexDataGridColumnLevel.EVENT_ITEMCLOSE = 'itemClose';
/**
 * Dispatched when an item opens, both in server and client mode
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event itemOpen
 */

FlexDataGridColumnLevel.EVENT_ITEMOPEN = 'itemOpen';
/**
 * Dispatched when an item opens, both in server and client mode
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event itemOpening
 */

FlexDataGridColumnLevel.EVENT_ITEMOPENING = 'itemOpening';
/**
 * Dispatched only in server mode, when an item opens for the first time.
 * Used to wire up an event handler to load the next level of data in lazy loaded grids.
 * Event Type:com.flexicious.nestedtreedatagrid.events.ExtendedFilterPageSortChangeEvent
 * @event itemLoad
 */

FlexDataGridColumnLevel.EVENT_ITEMLOAD = 'itemLoad';
/**
 * Dispatched when the grid's page, sort or filter state changes. Also Dispatched when
 * an item, that was not previously opened is opened. Used in lazy loaded (filterPageSortMode=server)
 * grids, to load a specific page of data from the backend.
 * Event Type:com.flexicious.nestedtreedatagrid.events.ExtendedFilterPageSortChangeEvent
 * @event filterPageSortChange
 */

FlexDataGridColumnLevel.EVENT_FILTERPAGESORTCHANGE = 'filterPageSortChange';
/**
 * Dispatched when any header cell is clicked
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event headerClicked
 */

FlexDataGridColumnLevel.EVENT_HEADERCLICKED = 'headerClicked';
/**
 *Dispatched when the columns at this level are resized
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event columnsResized
 */

FlexDataGridColumnLevel.EVENT_COLUMNSRESIZED = 'columnsResized';
/**
 *  Dispatched when the user attempts to edit an item. If you call preventDefault on this event, the edit session will not begin.
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridItemEditEvent
 * @event itemEditBeginning
 */

FlexDataGridColumnLevel.EVENT_ITEMEDITBEGINNING = 'itemEditBeginning';
/**
 *  Dispatched when the editor is instantiated.
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridItemEditEvent
 * @event itemEditBegin
 */

FlexDataGridColumnLevel.EVENT_ITEMEDITBEGIN = 'itemEditBegin';
/**
 *  Dispatched when the edit session ends.
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridItemEditEvent
 * @event itemEditEnd
 */

FlexDataGridColumnLevel.EVENT_ITEMEDITEND = 'itemEditEnd';
/**
 *  Dispatched when the edit session is cancelled.
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridItemEditEvent
 * @event itemEditCancel
 */

FlexDataGridColumnLevel.EVENT_ITEMEDITCANCEL = 'itemEditCancel';
/**
 * Dispatched just before committing the value of the editorDataField property of the editor
 * to the property specified by the datafield property of the column of the item being edited.
 * If you call preventDefault on the event, the value will not be committed.
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridItemEditEvent
 * @event itemEditValueCommit
 */

FlexDataGridColumnLevel.EVENT_ITEMEDITVALUECOMMIT = 'itemEditValueCommit';
/**
 *  Dispatched when row selection or cell selection changes.
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event change
 */

FlexDataGridColumnLevel.EVENT_CHANGE = 'change';
/**
 *  Dispatched when user rolls over a row in row selection mode or cell in cell selection mode
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event itemRollOver
 */

FlexDataGridColumnLevel.EVENT_ITEMROLLOVER = 'itemRollOver';
/**
 *  Dispatched when user rolls out of a row in row selection mode or cell in cell selection mode
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event itemRollOut
 */

FlexDataGridColumnLevel.EVENT_ITEMROLLOUT = 'itemRollOut';
/**
 *  Dispatched when user clicks on a row in row selection mode or cell in cell selection mode
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event itemClick
 */

FlexDataGridColumnLevel.EVENT_ITEMCLICK = 'itemClick';
/**
 *  Dispatched when user double clicks on a row in row selection mode or cell in cell selection mode
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 * @event itemDoubleClick
 */

FlexDataGridColumnLevel.EVENT_ITEMDOUBLECLICK = 'itemDoubleClick';
/**
 * Returns true if the getItemKey for the specified item exists in the selectedKeys collection.
 * @method isItemOpen
 * @return {boolean}
 */
flexiciousNmsp.FlexDataGridColumnLevel.prototype.isItemOpen = function (item) {
    if (this.selectedKeyField) {
        const itemKey = UIUtils.resolveExpression(item, this.selectedKeyField);
        const oi = this.grid.getOpenItems();

        for (const child of oi) {
            if (itemKey == UIUtils.resolveExpression(child, this.selectedKeyField))
                return true;
        }

        return false;
    }
    return this.grid.getOpenItems().includes(item);

}; 