/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */

import Constants from '../../utils/Constants'
import UIUtils from '../../utils/UIUtils'
import UIComponent from '../../core/UIComponent'
import FlexDataGridHeaderCell from '../cells/FlexDataGridHeaderCell'
import FlexDataGridFooterCell from '../cells/FlexDataGridFooterCell'
import FlexDataGridFilterCell from '../cells/FlexDataGridFilterCell'
import FlexDataGridPagerCell from '../cells/FlexDataGridPagerCell'
import FlexDataGridLevelRendererCell from '../cells/FlexDataGridLevelRendererCell'
import FlexDataGridExpandCollapseCell from '../cells/FlexDataGridExpandCollapseCell'
import FlexDataGridColumnGroupCell from '../cells/FlexDataGridColumnGroupCell'
import FlexDataGridDataCell from '../cells/FlexDataGridDataCell'
import FlexDataGridExpandCollapseHeaderCell from '../cells/FlexDataGridExpandCollapseHeaderCell'
import FlexDataGridHeaderSeperator from '../cells/FlexDataGridHeaderSeperator'
import ComponentAdditionResult from '../valueobjects/ComponentAdditionResult'
import CellInfo from '../valueobjects/CellInfo'
import SortInfo from '../valueobjects/SortInfo'
import RowInfo from '../valueobjects/RowInfo'
import ChangeInfo from '../valueobjects/ChangeInfo'
import FlexDataGridItemEditEvent from '../events/FlexDataGridItemEditEvent'
/**
 * FlexDataGridContainerBase is the base class for each of the containers of the grid. The Header, Footer, Pager and Filter Rows.
 * The grid has the following containers:
 * <ul>
 * <li>leftLockedHeader (com.flexicious.nestedtreedatagrid.cells.LockedContent) - The container for the left locked filter and header cells. This is an instance of the LockedContent class, which basically is an extended UIComponent that manages the filter and footer cell visibility, heights, and the y positions.</li>
 * <li>leftLockedContent (com.flexicious.nestedtreedatagrid.cells.ElasticContainer) - The container for the left locked data cells. This is an instance of the ElasticContainer class, which basically attaches to the owner component (which is the bodyContainer) and scrolls vertically along with it. The horizontal scroll of this component is set to off)</li>
 * <li>leftLockedFooter (com.flexicious.nestedtreedatagrid.cells.LockedContent)</li>
 * <li>filterContainer (com.flexicious.nestedtreedatagrid.cells.FlexDataGridChromeContainer)</li>
 * <li>headerContainer (com.flexicious.nestedtreedatagrid.cells.FlexDataGridChromeContainer)</li>
 * <li>bodyContainer(com.flexicious.nestedtreedatagrid.cells. FlexDataGridBodyContainer)</li>
 * <li>footerContainer (com.flexicious.nestedtreedatagrid.cells.FlexDataGridChromeContainer)</li>
 * <li>rightLockedHeader (com.flexicious.nestedtreedatagrid.cells.LockedContent)</li>
 * <li>rightLockedContent (com.flexicious.nestedtreedatagrid.cells.ElasticContainer)</li>
 * <li>rightLockedFooter (com.flexicious.nestedtreedatagrid.cells.LockedContent)</li>
 * <li>pagerContainer (com.flexicious.nestedtreedatagrid.cells.FlexDataGridChromeContainer) </li>
 * </ul>
 *
 * The filter, header, footer and pager containers are of type FlexDataGridChromeContainer that inherit from FlexDataGridContainerBase.
 * @constructor
 * @class FlexDataGridContainerBase
 * @namespace flexiciousNmsp
 * 
 */
export default class FlexDataGridContainerBase extends UIComponent {
    constructor(grid) {
        super({}, "div")
        /**
         * The grid that we belong to
         */
        this.grid = null;

        /**
         * @type {boolean}
         * @property verticalSpill
         * @default false
         */
        this.verticalSpill = false;
        /**
         * Collection of RowInfo objects currently being displayed. The Header,Footer and Filter containers
         * usually have only one row, while the body container has as many rows an be shown in the currently
         * visible area.
         * 
         * @type {Array}
         * @property rows
         * @default []
         */
        this.rows = [];

        /**
         * @private
         */
        this._sortInfo = {};

        /**
         * @private
         */
        this._loadedItems = [];

        /**
         * @private
         */
        this.itemVerticalPositions = [];

        /**
         * @private
         */
        this._horizontalScrollPending = -1;

        /**
         * Flag to turn on horizontal scroll recycle. Should be turned off in case of using ICustomMatchFilterControl, since you do not want these
         * controls to be recycled.
         * 
         * @type {boolean}
         * @property enableHorizontalRecycling
         * @default true
         */
        this.enableHorizontalRecycling = true;

        /**
         * @type {boolean}
         * @property columnDraggingToRight
         * @default false
         */
        this.columnDraggingToRight = false;

        /**
         * @private
         */
        this._keyboardhandled = false;

        /**
         * Last selected Row Index
         * 
         * @type {number}
         * @property lastSelectedRowIndex
         * @default -1
         */
        this.lastSelectedRowIndex = -1;

        /**
         * Timer to prevent double click from causing two item clicks.
         * 
         * @type {flexiciousNmsp.Timer}
         * @property itemClickTimer
         * @default null
         */
        this.itemClickTimer = null;

        /**
         * Duration, in milliseconds to wait until dispatching a new ITEM_CLICK event. Prevents two ITEM_CLICK events when user double clicks. Defaults to 250
         * 
         * @type {number}
         * @property itemClickTimerDuration
         * @default 250
         */
        this.itemClickTimerDuration = 250;

        /**
         * @private 
         */
        this._inEdit = false;

        /**
         * @type {number}
         * @property resizeCursorID
         * @default -1
         */
        this.resizeCursorID = -1;
        /**
         * When a column is being resized, this variable
         * holds the cell that initiated the resize operation. We render a
         * glyph that trails the users mouse pointer during the resize operation
         * This variable simply holds a reference to the
         * header cell associated with the column being resized.
         * */
        this.columnResizingCell = null;
        /**
         * When a column is being dragged and dropped into a different location, this variable
         * holds the cell that is being dragged. The cell actually does not drag, we render a
         * glyph that trails the users mouse pointer. This variable simply holds a reference to the
         * header cell associated with the column being dragged.
         */
        this.columnDraggingDragCell = null;

        /**
         * @property columnDraggingDropTargetCell
         * @default null
         */
        this.columnDraggingDropTargetCell = null;

        /**
         * @property startX
         * @default null
         */
        this.startX = null;
        /**
         * Once we are through calculating rowspans and colspans, we go through this array, popping cells and hiding cells
         * that are covered by the increased col and row spans of the cells in this array. Please note, this array will
         * only be populated when the snapToColumnWidths has just executed, and the hideSpannedCells has not.
         *
         * @type {Aray}
         * @property cellsWithColSpanOrRowSpan
         * @default []
         */
        this.cellsWithColSpanOrRowSpan = [];


        /**
         * 
         */
        this.grid = grid;


    }

    getClassNames() {
        return ["FlexDataGridContainerBase", "Container", "UIComponent"];
    }

    /**
     * @method createChildren
     */
    createChildren() {

        flexiciousNmsp.UIComponent.prototype.createChildren.apply(this, []);
        if (this == this.grid.getHeaderContainer()) {
            this.grid.addEventListener(this, Constants.EVENT_MOUSE_MOVE, this.onGridMouseMove);

        } else if (this == this.grid.getBodyContainer()) {
            this.grid.addEventListener(this, Constants.EVENT_FOCUS_CHANGE, this.onEditorKeyFocusChange);
        }

    }

    /**
     * @method onGridMouseMove
     * @param event
     */
    onGridMouseMove(event) {

        if (this.resizeCursorID != -1) {
            if (this.columnResizingCell) return;
            const cell = this.getCellFromMouseEventTarget(event.target);
            if (cell && cell.rowInfo && !cell.rowInfo.getIsHeaderRow()) {
                this.killResize();
            }
        }


    }

    /**
     * @method getCellFromMouseEventTarget
     * @param target
     * @return {flexiciousNmsp.FlexDataGridCell}
     */
    getCellFromMouseEventTarget(target) {

        if (!target) {
            return null;
        }
        else if (target.implementsOrExtends('IFlexDataGridCell')) {
            return target;
        } else if (target.hasOwnProperty("parent")) {
            return this.getCellFromMouseEventTarget(target.parent);
        } else
            return null;

    }

    /**
     * End the edit, if active, call removeComponent on each of the rows, and remove all the children.
     * 
     * @method removeAllComponents
     * @param {boolean} recycle
     */
    removeAllComponents(recycle) {
        this.pauseState();
        if (this._inEdit) this.endEdit(this.getEditor());
        while (this.rows.length > 0) {
            this.removeComponents(this.rows[0]);
        }
        UIUtils.removeAllChildren(this);
        this.resumeState();
    }

    /**
     * @method createComponents
     * 
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level
     * @param {number} currentScroll
     * 
     */
    createComponents(level, currentScroll) {
        if (typeof currentScroll == "undefined") currentScroll = 0;

        this.lastSelectedRowIndex = -1;
        this.removeAllComponents(false);
        this._sortInfo = {};
        this.grid.currentPoint.contentX = this.grid.currentPoint.leftLockedContentX = this.grid.currentPoint.rightLockedContentX = 0;
        this.grid.currentPoint.contentY = 0;

    }

    /**
     * Returns true if the passed in numbers are in the visible ViewPort
     * 
     * @method isInVisibleHorizontalRange
     * 
     * @param {number} x
     * @param {number} width
     * 
     * @return {boolean}
     */
    isInVisibleHorizontalRange(x, width) {
        //return true;
        if (this.grid.getForceColumnsToFitVisibleArea()) return true;
        if (!this.enableHorizontalRecycling) return true;
        const controlXEnd = x + width;
        const range = this.grid.getBodyContainer()._visibleRangeH;
        return (range.length == 0 || (

            (!(controlXEnd < range[0]) && !(x > range[1]))

        ));

    }

    /**
     * Removes the component from its parents hierarchy, and calls destroy method, if the component is a IFlexDataGridCell object
     * Hangs on to the component in the cache for further reuse.
     * 
     * @method removeComponent
     * @param comp component to remove
     */
    removeComponent(comp) {

        this.removeEventListeners(comp);
        if (this.grid.currentCell == comp.component)
            this.grid.currentCell = null;
        comp.row.removeCell(comp);

        if (comp.component == null || comp.component.parent == null) return;
        comp.component.parent.removeChild(comp.component);
        const cell = comp.component;
        if (cell) {
            cell.destroy();
        }

        this.grid.rendererCache.pushInstance(comp.component);


    }

    /**
     * Calls removeAllComponents, createComponents, validateNow, and snapToColumnWidths
     * @method reDraw
     */
    reDraw() {

        this.removeAllComponents(true);
        this.createComponents(this.grid.getColumnLevel());
        this.validateNow();
        this.snapToColumnWidths();

    }

    /**
     * @method recycleH
     * 
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level
     * @param {boolean} scrollRight
     * 
     */
    recycleH(level, scrollRight) {
        if (typeof scrollRight == "undefined") scrollRight = true;


        const dictionary = new flexiciousNmsp.KeyValuePairCollection();
        let i;
        let row;
        const rowsForRecycling = this.getRowsForRecycling();
        for (i = 0; i < rowsForRecycling.length; i++) {
            row = rowsForRecycling[i];
            dictionary.addItem(row, []);
            if (!row.getIsColumnBased()) continue;

            for (const cell of row.cells) {
                const tCell = cell.component;
                if (tCell && !this.isInVisibleHorizontalRange(cell.getX(), cell.component.getWidth()) && !tCell.getIsLocked()) {
                    if (tCell && tCell.colSpan > 1) {
                        //this means is cell that has a colspan, we should not take it out
                    } else
                        dictionary.getValue(row).push(cell);
                }
            }
        }
        const forRec = this.getRowsForRecycling();
        for (i = 0; i < forRec.length; i++) {
            row = forRec[i];
            if (!row.getIsColumnBased()) continue;
            const recyclable = dictionary.getValue(row);
            this.processRowPositionInfo(row.rowPositionInfo, false, row, recyclable);
            if(this.grid.enableHorizontalScrollOptimizationsBETA){
                
            }else{
                while (recyclable.length > 0) {
                    this.removeComponent(recyclable[recyclable.length - 1]);
                    recyclable.pop();
                }
            }

        }
        if (this.grid.getHasRowSpanOrColSpan()) {
            this.snapToColumnWidths();
        }


        //placeComponents();
    }

    /**
     * @method getRowsForRecycling
     * @return {Array}
     */
    getRowsForRecycling() {

        return this.rows;

    }

    /**
     * 
     */
    processRowPositionInfo(rowPositionInfo, scrollDown, existingRow, existingComponents) {
        if (typeof scrollDown == "undefined") scrollDown = true;
        if (typeof existingRow == "undefined") existingRow = null;
        if (typeof existingComponents == "undefined") existingComponents = null;

        if (rowPositionInfo.getRowType() == flexiciousNmsp.RowPositionInfo.ROW_TYPE_HEADER) {
            return this.processHeaderLevel(rowPositionInfo.getLevel(this.grid), rowPositionInfo, scrollDown, rowPositionInfo.rowData, flexiciousNmsp.RowPositionInfo.ROW_TYPE_HEADER, existingRow, existingComponents);
        }
        else if (rowPositionInfo.getRowType() == flexiciousNmsp.RowPositionInfo.ROW_TYPE_COLUMN_GROUP) {
            return this.processHeaderLevel(rowPositionInfo.getLevel(this.grid), rowPositionInfo, scrollDown, rowPositionInfo.rowData, flexiciousNmsp.RowPositionInfo.ROW_TYPE_COLUMN_GROUP, existingRow, existingComponents);
        }
        else if (rowPositionInfo.getRowType() == flexiciousNmsp.RowPositionInfo.ROW_TYPE_FILTER) {
            return this.processHeaderLevel(rowPositionInfo.getLevel(this.grid), rowPositionInfo, scrollDown, rowPositionInfo.rowData, flexiciousNmsp.RowPositionInfo.ROW_TYPE_FILTER, existingRow, existingComponents);
        }

        else if (rowPositionInfo.getRowType() == flexiciousNmsp.RowPositionInfo.ROW_TYPE_PAGER) {
            return this.processHeaderLevel(rowPositionInfo.getLevel(this.grid), rowPositionInfo, scrollDown, rowPositionInfo.rowData, flexiciousNmsp.RowPositionInfo.ROW_TYPE_PAGER, existingRow, existingComponents);
        }
        else if (rowPositionInfo.getRowType() == flexiciousNmsp.RowPositionInfo.ROW_TYPE_FOOTER) {
            return this.processHeaderLevel(rowPositionInfo.getLevel(this.grid), rowPositionInfo, scrollDown, rowPositionInfo.rowData, flexiciousNmsp.RowPositionInfo.ROW_TYPE_FOOTER, existingRow, existingComponents);
        }
        else if (rowPositionInfo.getRowType() == flexiciousNmsp.RowPositionInfo.ROW_TYPE_RENDERER) {
            return this.processRendererLevel(rowPositionInfo, scrollDown);
        }
        return null;
    }

    /**
     * Iterates through all rows and calls removeComponent on each of the cells.
     * Removes event listeners added to the RowInfo object
     * @param row
     *
     */
    removeComponents(row) {

        while (row.cells.length > 0) {
            this.removeComponent(row.cells[0]);
        }
        if (!row.getIsFillRow())
            this.rows.splice(this.rows.indexOf(row), 1);
        if (row.getIsFilterRow()) {
            row.removeEventListener(flexiciousNmsp.ExtendedFilterPageSortChangeEvent.FILTER_CHANGE, this.onFilterChange);
        }


    }

    /**
     * Sets the current highlight cell to the first available cell.
     */
    setCurrentCellToFirst() {

        const cell = this.getFirstAvailableCell(null, false);
        if (cell)
            this.grid.currentCell = cell;

    }

    /**
     * Gets the first cell of the first column.
     *
     */
    getFirstAvailableCell(cell, up) {
        if (typeof up == "undefined") up = true;

        if (this.rows.length > 0) {
            const row = this.rows[up ? this.rows.length - 1 : 0];
            if (cell && row.getIsColumnBased() && cell.rowInfo.getIsColumnBased() && cell.getColumn() && row.getCellForColumn(cell.getColumn())) {
                return row.getCellForColumn(cell.getColumn()).component
            }
            return this.getFirstHoverableCell(row);
        }
        return null;

    }

    getFirstHoverableCell(row, dataOnly, editableOnly) {
        if (typeof dataOnly == "undefined") dataOnly = false;
        if (typeof editableOnly == "undefined") editableOnly = false;

        if (row.cells.length > 0) {
            for (const cell of row.cells) {
                const fdgCell = cell.component;
                if (dataOnly && !(fdgCell.implementsOrExtends('IFlexDataGridDataCell'))) continue;
                if (editableOnly && (!fdgCell.getColumn() || !fdgCell.getColumn().getEditable())) continue;
                if (this.isHoverableCell(fdgCell))
                    return fdgCell;
            }
        }
        return null;

    }

    /**
     * Can this cell accept hover. Returns true if
     * (cell is IFlexDataGridDataCell
     * 		or cell is FlexDataGridHeaderCell
     *		or cell is FlexDataGridFooterCell
     *		or cell is FlexDataGridFilterCell
     *		or cell is FlexDataGridPagerCell
     *		or cell is FlexDataGridLevelRendererCell
     *		or cell is FlexDataGridExpandCollapseCell
     *		or cell is FlexDataGridColumnGroupCell) and cell.enabled and not cell is FlexDataGridExpandCollapseHeaderCell
     */
    isHoverableCell(cell) {

        return (cell.implementsOrExtends('IFlexDataGridDataCell')
            || cell.implementsOrExtends('FlexDataGridHeaderCell')
            || cell.implementsOrExtends('FlexDataGridFooterCell')
            || cell.implementsOrExtends('FlexDataGridFilterCell')
            || cell.implementsOrExtends('FlexDataGridPagerCell')
            || cell.implementsOrExtends('FlexDataGridLevelRendererCell')
            || cell.implementsOrExtends('FlexDataGridExpandCollapseCell')
            || cell.implementsOrExtends('FlexDataGridColumnGroupCell'))
            && cell.getEnabled() && !(cell.implementsOrExtends('FlexDataGridExpandCollapseHeaderCell'));
    }

    /**
    * Based upon the type of the cell, adds various event listeners to them
    * to respond to mouse overs, clicks, double clicks, mouse outs, keyboard input,
    * etc.
    */
    addEventListeners(comp) {
        const cell = comp.component;
        if (cell) {
            cell.addEventListener(this, Constants.EVENT_MOUSE_OVER, this.onCellMouseOver);
            cell.addEventListener(this, Constants.EVENT_KEY_UP, this.onCellKeyUp, false, -99);
            cell.addEventListener(this, Constants.EVENT_CLICK, this.onCellMouseClick);
            cell.addEventListener(this, Constants.EVENT_MOUSE_DOWN,this.onCellMouseDown);

            cell.addEventListener(this, Constants.EVENT_DOUBLE_CLICK, this.onCellDoubleClick);
            cell.addEventListener(this, Constants.EVENT_MOUSE_OUT, this.onCellMouseOut);
        }

        if (cell.implementsOrExtends('FlexDataGridHeaderCell')) {
            cell.addEventListener(this, Constants.EVENT_MOUSE_MOVE, this.onHeaderCellMouseMove);
            cell.addEventListener(this, Constants.EVENT_MOUSE_OUT, this.onHeaderCellMouseOut);
            cell.addEventListener(this, Constants.EVENT_MOUSE_DOWN, this.onHeaderCellMouseDown);

            const col = cell.getColumn();
            if (col && col.implementsOrExtends('FlexDataGridCheckBoxColumn'))
                cell.addEventListener(this, flexiciousNmsp.FlexDataGridEvent.SELECT_ALL_CHECKBOX_CHANGED, this.selectAllChangedHandler);
            cell.addEventListener(this, Constants.EVENT_MOVE, this.placeSortIcon);
            cell.addEventListener(this, Constants.EVENT_RESIZE, this.placeSortIcon);
        }
        else if (cell.implementsOrExtends('FlexDataGridColumnGroupCell')) {
            cell.addEventListener(this, Constants.EVENT_MOUSE_MOVE, this.onHeaderCellMouseMove);
            cell.addEventListener(this, Constants.EVENT_MOUSE_OUT, this.onHeaderCellMouseOut);
            cell.addEventListener(this, Constants.EVENT_MOUSE_DOWN, this.onHeaderCellMouseDown);
        }
        if (this.grid.enableDrop && cell) {
            cell.addEventListener(this, Constants.EVENT_MOUSE_MOVE, this.onCellDropMouseMove);
        }

    }

    /**
    * Handles Cell Mouse Over. If header cell is moving, draws the header drop indicator.
    */
    onCellMouseOver(event) {
        //calculateColumnDraggingDropTargetCell(event);
        //this has been moved to

        if (this.grid.contextMenuShown) return;
        const cell = (event.currentTarget);
        if (cell != this.grid.currentCell)
            this.handleMouseOver(cell, event);
        if ((this.resizeCursorID != -1) && !cell.implementsOrExtends('FlexDataGridHeaderCell')
            && !cell.implementsOrExtends('FlexDataGridColumnGroupCell')) {

            cell.setStyleAttribute("cursor", "");
            this.resizeCursorID = -1;
        }
    }

    /**
    * Given the mouse event, figures out which cell is the drop target.
    * In 2.9, code was added to drop on the right cell of the current if the mouse is to
    * the right of the middle of the cell, and on the left cell of the current if the
    * mouse is on the left hand side of the middle of the cell.
    *
    * This function checks to see if the mouse is over a header cell, and if the
    * cell being dragged is a header cell. If so, it will size and position a glyph
    * so that the user gets interactive feedback about where the column will be dropped
    * and which column will be dropped. The same applies to the column group cells for
    * drag and drop purposes.
    */
    calculateColumnDraggingDropTargetCell(event) {
        if (this.grid.contextMenuShown) return;
        const checkIconClass = this.grid.getStyle("checkIconClass");
        const crossIconClass = this.grid.getStyle("crossIconClass");

        const cell = this.getCellFromMouseEventTarget(event.target);
        if (!cell) return;
        const headerCell = cell.implementsOrExtends('FlexDataGridHeaderCell') ? cell : null;
        const movingHCell = this.columnDraggingDragCell;
        let pt = new flexiciousNmsp.Point(0, cell.getHeight());
        const cgCell = cell.implementsOrExtends('FlexDataGridColumnGroupCell') ? cell : null;
        const movingCgCell = this.columnDraggingDragCell;
        this.columnDraggingToRight = (event.localX > (cell.getWidth() / 2));
        const isRtl = (this.grid.getStyle("layoutDirection") == 'rtl');
        if (isRtl) this.columnDraggingToRight = !this.columnDraggingToRight;
        if (headerCell && movingHCell) {
            if (this.columnDraggingDragCell && this.columnResizingGlyph && (this.columnDraggingDragCell.level == headerCell.level)
                && (this.columnDraggingDragCell.getColumn().columnGroup == headerCell.getColumn().columnGroup)) {
                pt = this.grid.globalToLocal(headerCell.localToGlobal(pt));
                if (this.columnDraggingToRight) pt.setX(pt.getX() + headerCell.getWidth());
                this.columnResizingGlyph.move(pt.getX(), pt.getY());
                this.columnDraggingDropTargetCell = headerCell;
                if (this.columnResizingCellGlyph) {
                    this.columnResizingCellGlyph.colIcon.setSource(this.grid.imagesRoot + checkIconClass);
                }
            }
            else if (this.columnResizingCellGlyph) {
                this.columnResizingCellGlyph.colIcon.setSource(this.grid.imagesRoot + crossIconClass);
            }
        }
        else if (cgCell && movingCgCell) {
            if (movingCgCell && this.columnResizingGlyph && movingCgCell.columnGroup && (movingCgCell.columnGroup.parentGroup == cgCell.columnGroup.parentGroup) &&
                (movingCgCell.getColumn().getColumnLockMode() == cgCell.getColumn().getColumnLockMode())) {
                pt = this.grid.globalToLocal(cgCell.localToGlobal(pt));
                if (this.columnDraggingToRight) pt.setX(pt.getX() + cgCell.getWidth());
                this.columnResizingGlyph.move(pt.getX(), pt.getY());
                this.columnDraggingDropTargetCell = cgCell;
                if (this.columnResizingCellGlyph) {
                    this.columnResizingCellGlyph.colIcon.setSource(this.grid.imagesRoot + checkIconClass);
                }
            }
            else if (this.columnResizingCellGlyph) {
                this.columnResizingCellGlyph.colIcon.setSource(this.grid.imagesRoot + crossIconClass);
            }
        }
        else if (this.columnResizingCellGlyph) {
            this.columnResizingCellGlyph.colIcon.setSource(this.grid.imagesRoot + crossIconClass);
        }
    }

    /**
    *
    */
    onCellDropMouseMove(event) {



    }

    /**
    * Handles mouse over for data cells.
    * Calls highlightRow for item rollovers.
    */
    handleMouseOver(cell, triggerEvent) {

        if (!this.grid.allowInteractivity || !cell.getEnabled() || this.resizeCursorID != -1) return;
        const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.ITEM_ROLL_OVER, this.grid,
            cell.level, cell.getColumn(), cell, cell.rowInfo.getData(), triggerEvent);
        cell.level.dispatchEvent(evt);

        if (UIUtils.isMouseEvent(triggerEvent) && this._keyboardhandled) {
            this._keyboardhandled = false;
            return;
        }
        if (this.grid.currentCell) {
            this.highlightRow(this.grid.currentCell, this.grid.currentCell.rowInfo, false);
        }

        if (triggerEvent.type.toLowerCase().includes('key')) {
            this._keyboardhandled = true
        }
        this.grid.currentCell = cell;
        if (this.grid.useRollOver || !(this.grid.currentCell.rowInfo.getIsDataRow()))
            this.highlightRow(this.grid.currentCell, this.grid.currentCell.rowInfo, true);

    }

    /**
    *
    */
    onCellMouseOut(event) {
        const cell = event.currentTarget;
        this.handleMouseOut(cell, event)
    }

    /**
    * Handles mouse out. Dispatches the Item Roll Out event
    */
    handleMouseOut(cell, triggerEvent) {
        //if(!this.grid.currentCell || (this.grid.currentCell.rowInfo!=cell.rowInfo)){
        const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.ITEM_ROLL_OUT, this.grid,
            cell.level, cell.getColumn(), cell, cell.rowInfo.getData(), triggerEvent);
        cell.level.dispatchEvent(evt);
        //  }
    }

    /**
    * Handles mouse click. Ensures we're not on the resize area, and ends the active edit, if any, before calling the default click handler.
    */
    onCellMouseClick(event) {
        const cell = event.currentTarget;
        if (cell.implementsOrExtends('FlexDataGridHeaderCell')) {
            const onRightEdge = ((cell.getWidth() - 2 - event.localX) <= this.grid.getColumnLevel().headerSeperatorWidth);
            const onLeftEdge = (event.localX <= this.grid.getColumnLevel().headerSeperatorWidth);
            if (onRightEdge || onLeftEdge)
                return;
        }
        if (this._inEdit)
            this.endEdit(this.getEditor());
        this.handleMouseClick(cell, event);
    }

    /**
    *
    */
    onCellDoubleClick(event) {
        const cell = event.currentTarget;
        this.handleDoubleClick(cell, event);
    }

    /**
    * Handles the Double Click. If enableDoubleClickEdit and editable on the column is set to true, starts the edit session
    */
    handleDoubleClick(cell, triggerEvent) {
        if (!this.grid.allowInteractivity || !cell.enabled) return;

        if (cell.implementsOrExtends('IFlexDataGridDataCell') && cell.getColumn()) {
            if (cell.getColumn().getEditable() && this.grid.enableDoubleClickEdit)
                this.beginEdit(cell);
        }
    }

    /**
    * Used by accesibility
    *
    */
    emulateClick(cell) {
        const evt = new flexiciousNmsp.BaseEvent(Constants.EVENT_CLICK, true, false);
        this.handleMouseClick(cell, evt);
    }

    /**
    * Handles the Mouse click. Dispatches ITEM_CLICK. If the Cell is a header cell, calls
    * onHeaderCellClicked. If the cells is the checkbox selection cell, updates the grid selection.
    * Updates the Selected Cells, if this.grid.isCellSelectionMode.
    * If the column is editable, begins the edit session.
    */
    handleMouseClick(cell, triggerEvent) {
        if (!this.grid.allowInteractivity) return;

        if (!this.itemClickTimer || !this.itemClickTimer.running) {
            this.itemClickTimer = new flexiciousNmsp.Timer(this.itemClickTimerDuration, 1);
            this.itemClickTimer.addEventListener(this, Constants.EVENT_TIMER, this.onItemClickTimer);
            this.itemClickTimer.start();
        } else {
            return;
        }
        if (!cell.getEnabled() && cell.getColumn() && cell.getColumn().enableExpandCollapseIcon && cell.rowInfo && cell.rowInfo.getIsDataRow()) {
            cell.getIExpandCollapseComponent().doClick();
            return;
        }

        const keyboardEvent = triggerEvent.type.toLowerCase().includes('key') ? triggerEvent : null;
        const mouseEvent = UIUtils.isMouseEvent(triggerEvent) ? triggerEvent : null;
        const shiftKey = (mouseEvent && mouseEvent.shiftKey) || (keyboardEvent && keyboardEvent.shiftKey);
        let dispatchItemClick = false;
        this.grid.currentCell = cell;
        const headerCell = cell.implementsOrExtends('FlexDataGridHeaderCell') ? cell : null;
        const cgCell = cell.implementsOrExtends('FlexDataGridColumnGroupCell') ? cell : null;
        if (headerCell) {
            this.onHeaderCellClicked(headerCell, triggerEvent);
        }
        else if (cell.implementsOrExtends('FlexDataGridExpandCollapseCell')
            || (cell.getColumn() && cell.getColumn().enableExpandCollapseIcon && (triggerEvent && triggerEvent.triggerEvent && triggerEvent.triggerEvent.target.tagName == "IMG") && cell.getRowInfo().getIsDataRow())) {
            cell.getIExpandCollapseComponent().doClick();
        }
        else if (cgCell && cgCell.columnGroup.enableExpandCollapse) {
            if (cgCell.columnGroup.isClosed()) {
                cgCell.columnGroup.openColumns()
            } else {
                cgCell.columnGroup.closeColumns()
            }
            cgCell.invalidateDisplayList();
        }
        else if (this.grid.getIsRowSelectionMode() && cell.getColumn()) {
            if (cell.getColumn().enableCellClickRowSelect) {
                if (cell.implementsOrExtends("IFlexDataGridDataCell") && cell.getColumn()
                    && (!cell.getColumn().getEditable() || this.grid.enableDoubleClickEdit)
                    && (cell.level.checkRowSelectable(cell, cell.rowInfo.getData()))) {

                    let select = !cell.level.isItemSelected(cell.rowInfo.getData());
                    if (this.grid.isCtrlKeyDownOrSticky(triggerEvent))
                        cell.level.selectRow(cell.rowInfo.getData(), select);
                    else {
                        select = true;
                        this.grid.clearSelection();
                        cell.level.selectRow(cell.rowInfo.getData(), select);
                    }

                    this.highlightRow(cell, cell.rowInfo, true, select ? cell.getStyleValue("selectionColor") : cell.getRolloverColor());

                    if (shiftKey && (this.lastSelectedRowIndex > -1)) {
                        //selecting multiple rows.
                        let selectionStart = 0;
                        let selectionEnd = 0;
                        if (this.lastSelectedRowIndex > cell.rowInfo.rowPositionInfo.getRowIndex()) {
                            //shift up.
                            selectionStart = cell.rowInfo.rowPositionInfo.getRowIndex();
                            selectionEnd = this.lastSelectedRowIndex;
                        } else {
                            //shift down
                            selectionStart = this.lastSelectedRowIndex;
                            selectionEnd = cell.rowInfo.rowPositionInfo.getRowIndex();
                        }
                        for (let i = selectionStart; i <= selectionEnd; i++) {
                            const pos = this.itemVerticalPositions[i];
                            pos.getLevel(this.grid).selectRow(pos.rowData, true, false, true);
                        }

                    }
                    else if (!shiftKey)
                        this.lastSelectedRowIndex = cell.rowInfo.rowPositionInfo.getRowIndex();

                }
                this.grid.invalidateSelection();
            }
            dispatchItemClick = true;
        }
        else if (cell.implementsOrExtends("IFlexDataGridDataCell")
            && cell.getColumn() && (!cell.getColumn().getEditable() || this.grid.enableDoubleClickEdit)
            && !(cell.getColumn().implementsOrExtends("FlexDataGridCheckBoxColumn")) && this.grid.getIsCellSelectionMode()) {

            // this.lastSelectedRowIndex = cell.rowInfo.rowPositionInfo.getRowIndex();
            let selectCell = !cell.level.isCellSelected(cell.rowInfo.getData(), cell.getColumn());

            if (this.grid.isCtrlKeyDownOrSticky(triggerEvent))
                cell.level.selectCell(cell, selectCell);
            else {
                selectCell = true;
                this.grid.clearSelection();
              //  cell.level.selectCell(cell, selectCell);
            }
            this.highlightRow(cell, null, true, selectCell ? cell.getStyleValue("selectionColor") : cell.getRolloverColor());

            let nextCellDirection = null;

            if (shiftKey && (this.lastSelectedRowIndex > -1)) {
                //selecting multiple cells.
                let selectionStart = 0;
                let selectionEnd = 0;
                if (this.lastSelectedRowIndex > cell.rowInfo.rowPositionInfo.getRowIndex()) {
                    //shift up.
                    selectionStart = cell.rowInfo.rowPositionInfo.getRowIndex();
                    selectionEnd = this.lastSelectedRowIndex;
                    nextCellDirection = Constants.CELL_POSITION_BELOW;
                } else {
                    //shift down
                    selectionStart = this.lastSelectedRowIndex;
                    selectionEnd = cell.rowInfo.rowPositionInfo.getRowIndex();
                    nextCellDirection = Constants.CELL_POSITION_ABOVE;
                }

                let nextCell = null;

                for (let i = selectionStart; i <= selectionEnd; i++) {
                    const pos = this.itemVerticalPositions[i];
                    //nextCell = this.getCellInDirection(!nextCell?cell:nextCell, nextCellDirection, false, false, true, true);
                    var nextCell = new FlexDataGridDataCell();
                    nextCell.setColumn(cell.getColumn());
                    nextCell.rowInfo = new RowInfo(0,0,this.grid);
                    nextCell.level = pos.getLevel(this.grid);
                    nextCell.rowInfo.rowPositionInfo = pos;
                    if(nextCell) {
                        pos.getLevel(this.grid).selectCell(nextCell, true);
                    }
                }
            }
            if (selectCell) {
                cell.level.selectCell(cell, selectCell); 
            } 
            
            this.lastSelectedRowIndex = cell.rowInfo.rowPositionInfo.getRowIndex();

            this.grid.invalidateSelection();
            dispatchItemClick = true;
        }
        if (dispatchItemClick && cell.rowInfo && cell.rowInfo.getIsDataRow() && !cell.rowInfo.getIsFillRow()) {
            const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.ITEM_CLICK, this.grid,
                cell.level, cell.getColumn(), cell, cell.rowInfo.getData(), triggerEvent);
            evt.isItemSelected = (cell.level.isItemSelected(cell.rowInfo.getData()));
            cell.level.dispatchEvent(evt);

        }
        if ((cell.implementsOrExtends("IFlexDataGridDataCell")) && cell.getColumn()
            && cell.getColumn().getEditable() && (!this.grid.enableDoubleClickEdit || (keyboardEvent && (keyboardEvent.keyCode == Constants.KEYBOARD_F2)))
            && (cell.getColumn().isEditableFunction == null || cell.getColumn().isEditableFunction(cell))) {
            this.beginEdit(cell);
            triggerEvent.preventDefault();
        }

    }

    onCellMouseDown(event) {
        var cell = event.currentTarget;
        if(cell && event.triggerEvent && UIUtils.isRightClickEvent(event.triggerEvent) && cell.rowInfo && cell.rowInfo.getIsDataRow() && !cell.rowInfo.getIsFillRow()){
            var evt= new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.ITEM_RIGHT_CLICK,this.grid, cell.level,cell.getColumn(),cell,cell.rowInfo.getData(),event);
            cell.level.dispatchEvent(evt);
            return;
        }
    }

    /*&& !(cell.getColumn() is FlexDataGridCheckBoxColumn) */
    onItemClickTimer(event) {

        this.itemClickTimer.stop();
        this.itemClickTimer.removeEventListener(Constants.EVENT_TIMER, this.onItemClickTimer);
        this.itemClickTimer = null;

    }

    /**
     *
     */
    onCellKeyUp(event) {

    }

    /**
     * Handles cell key up. Handles KeyBoard.UP,KeyBoard.DOWN,KeyBoard.LEFT,KeyBoard.RIGHT
     */
    handleArrowKey(cell, keyCode, triggerEvent) {

        let nextCell;
        const keyboardEvent = triggerEvent.type.toLowerCase().includes('key') ? triggerEvent : null;
        let expCollCell;

        if ((keyboardEvent.ctrlKey && keyboardEvent.shiftKey && keyCode == Constants.KEYBOARD_DOWN)
            || (keyCode == Constants.KEYBOARD_NUMPAD_MULTIPLY)) {
            this.grid.expandChildrenOf(cell.rowInfo.getData(), true, cell.level);
        }
        else if ((keyboardEvent.ctrlKey && keyboardEvent.shiftKey && keyCode == Constants.KEYBOARD_UP)
            || (keyCode == Constants.KEYBOARD_NUMPAD_DIVIDE)) {
            this.grid.expandChildrenOf(cell.rowInfo.getData(), false, cell.level);
        }
        else if ((keyboardEvent.ctrlKey && keyboardEvent.shiftKey && keyCode == Constants.KEYBOARD_LEFT)
            || (keyCode == Constants.KEYBOARD_NUMPAD_SUBTRACT)) {
            expCollCell = this.getCellForRowColumn(cell.rowInfo.getData(), null, true);
            if (expCollCell && expCollCell.getIExpandCollapseComponent() && expCollCell.getIExpandCollapseComponent().open) {
                expCollCell.getIExpandCollapseComponent().doClick();
            }
        }
        else if ((keyboardEvent.ctrlKey && keyboardEvent.shiftKey && keyCode == Constants.KEYBOARD_RIGHT)
            || (keyCode == Constants.KEYBOARD_NUMPAD_ADD)) {
            expCollCell = this.getCellForRowColumn(cell.rowInfo.getData(), null, true);
            if (expCollCell && expCollCell.getIExpandCollapseComponent() && !expCollCell.getIExpandCollapseComponent().open) {
                expCollCell.getIExpandCollapseComponent().doClick();
            }
        }
        else if (keyCode == Constants.KEYBOARD_UP) {
            nextCell = this.getCellInDirection(cell, Constants.CELL_POSITION_ABOVE, false, false, true, true);
            if (!nextCell) return false;
            if (nextCell.rowInfo == cell.rowInfo) {
                //we're at the top row already.... so highlight the header
                const container = this.grid.getContainerInDirection(this.grid.getContainerName(this.grid.getContainerForCell(cell)), true);
                if (container) {
                    nextCell = container.getFirstAvailableCell(cell);
                }
            } else {
                //nextCell=nextCell;
            }
            if (triggerEvent)
                triggerEvent.preventDefault();
        }
        else if (keyCode == Constants.KEYBOARD_DOWN) {
            nextCell = this.getCellInDirection(cell, Constants.CELL_POSITION_BELOW, false, false, true, true);
            if (!nextCell) return false;
            if (nextCell.rowInfo == cell.rowInfo) {
                //we're at the last row already.... so highlight the header
                const container1 = this.grid.getContainerInDirection(this.grid.getContainerName(this.grid.getContainerForCell(cell)), false);
                if (container1) {
                    nextCell = container1.getFirstAvailableCell(cell, false);
                }
            }
            if (triggerEvent) {
                triggerEvent.preventDefault();
                // triggerEvent.stopPropagation();
            }
        }
        else if (keyCode == Constants.KEYBOARD_LEFT || (keyCode == Constants.KEYBOARD_TAB
            && triggerEvent.shiftKey)) {
            nextCell = this.getCellInDirection(cell, Constants.CELL_POSITION_LEFT, false, false, true, true);
            if (triggerEvent)
                triggerEvent.preventDefault();
        }
        else if (keyCode == Constants.KEYBOARD_RIGHT || keyCode == Constants.KEYBOARD_TAB) {
            nextCell = this.getCellInDirection(cell, Constants.CELL_POSITION_RIGHT, false, false, true, true);
            if (triggerEvent)
                triggerEvent.preventDefault();
        }
        else if (keyCode == Constants.KEYBOARD_PAGE_DOWN && this.rows.length > 2) {

            nextCell = this.getFirstHoverableCell(this.rows[this.rows.length -
                ((this.grid.getVerticalScrollPosition() == this.grid.getBodyContainer().getMaxVerticalScrollPosition()) ? 1 : 2)]);
            if (nextCell.rowInfo == cell.rowInfo) {
                const atMax = (this.grid.getVerticalScrollPosition() == this.grid.getBodyContainer().getMaxVerticalScrollPosition());

                this.grid.scrollToExistingRow(
                    Math.min(nextCell.rowInfo.rowPositionInfo.getVerticalPosition(),
                        this.grid.getBodyContainer().getMaxVerticalScrollPosition()),
                    true);
                this.grid.validateNow();
                this.grid.getBodyContainer().validateNow();
                if (!atMax) {
                    nextCell = this.getFirstHoverableCell(this.rows[this.rows.length > 1 ? 1 : 0]);
                }
            }
            if (triggerEvent)
                triggerEvent.preventDefault();
        }
        else if (keyCode == Constants.KEYBOARD_PAGE_UP && this.rows.length > 2) {
            nextCell = this.getFirstHoverableCell(this.rows[this.grid.getVerticalScrollPosition() == 0 ? 0 : 1]);
            if (nextCell.rowInfo == cell.rowInfo) {
                const atZero = (this.grid.getVerticalScrollPosition() == 0);

                this.grid.scrollToExistingRow(
                    Math.max(nextCell.rowInfo.rowPositionInfo.getVerticalPosition() -
                        this.grid.getBodyContainer().getHeight(),
                        0),
                    false);
                this.grid.validateNow();
                this.grid.getBodyContainer().validateNow();
                if (!atZero && this.rows.length > 2) {
                    nextCell = this.getFirstHoverableCell(this.rows[this.rows.length - 2]);
                }
            }
            if (triggerEvent)
                triggerEvent.preventDefault();
        }
        else if (keyCode == Constants.KEYBOARD_HOME) {
            if (this.grid.getVerticalScrollPosition() > 0) {
                this.grid.scrollToExistingRow(0, false);
                this.grid.validateNow();
            }
            nextCell = this.getFirstHoverableCell(this.rows[0]);
            if (triggerEvent)
                triggerEvent.preventDefault();
        }
        else if (keyCode == Constants.KEYBOARD_END) {
            if (this.grid.getVerticalScrollPosition() < this.grid.getBodyContainer().getMaxVerticalScrollPosition()) {
                this.grid.scrollToExistingRow(this.grid.getBodyContainer().getMaxVerticalScrollPosition(), true);
                this.grid.validateNow();
            }
            nextCell = this.getFirstHoverableCell(this.rows[this.rows.length - 1]);
            if (triggerEvent)
                triggerEvent.preventDefault();
        }

        else if (keyCode == Constants.KEYBOARD_F2 && cell.getColumn().getEditable()) {
            this.handleMouseClick(cell, triggerEvent);
        }
        else if (keyCode == Constants.KEYBOARD_SPACE) {
            this.handleMouseClick(cell, triggerEvent);
        }
        if (nextCell) {
            if (this.grid.getSelectionMode() == flexiciousNmsp.NdgBase.SELECTION_MODE_SINGLE_CELL && (nextCell.implementsOrExtends("IFlexDataGridDataCell"))
                && !this.grid.useRollOver) {
                //for excel like behavior, the keyboard will simply select the current cell as opposed to highlighting it, since useRollover is false.
                this.handleMouseClick(nextCell, triggerEvent);
            } else {
                this.handleMouseOver(nextCell, triggerEvent);
                if (nextCell.implementsOrExtends("FlexDataGridFilterCell")) {
                    nextCell.getRenderer().focus();
                }
            }


            if (this.grid.getContainerForCell(nextCell) == this.grid.getBodyContainer()) {
                let recycleNeeded = false;
                if ((keyCode == Constants.KEYBOARD_DOWN) && nextCell.rowInfo.getY() >= this.getVerticalScrollPosition() + this.getHeight()) {
                    this.setVerticalScrollPosition(nextCell.rowInfo.getY() + nextCell.rowInfo.getHeight() - this.getHeight());
                    recycleNeeded = true;
                }
                else if ((keyCode == Constants.KEYBOARD_UP) && nextCell.rowInfo.getY() <= this.getVerticalScrollPosition()) {
                    this.setVerticalScrollPosition(nextCell.rowInfo.getY());
                    recycleNeeded = true;
                }
                return recycleNeeded;
            }
        }
        return false;

    }

    /**
     * Handles cell key up. Handles the following keys:
     *
     * keyCode == Keyboard.UP
     * || keyCode == Keyboard.DOWN
     * || keyCode == Keyboard.LEFT
     * || keyCode == Keyboard.RIGHT
     * || keyCode == Keyboard.TAB
     * || keyCode == Keyboard.PAGE_DOWN
     * || keyCode == Keyboard.PAGE_UP
     * || keyCode == Keyboard.HOME
     * || keyCode == Keyboard.END
     * || keyCode == Keyboard.NUMPAD_SUBTRACT
     * || keyCode == Keyboard.NUMPAD_ADD
     * || keyCode == Keyboard.F2
     * || keyCode == Keyboard.NUMPAD_MULTIPLY
     */
    handleCellKeyUp(cell, keyCode, triggerEvent) {

        if (!this.grid.enableKeyboardNavigation || this.grid.getBodyContainer().getInEdit()) {
            return;
        }

        if (keyCode == Constants.KEYBOARD_UP || keyCode == Constants.KEYBOARD_DOWN
            || keyCode == Constants.KEYBOARD_LEFT || keyCode == Constants.KEYBOARD_RIGHT
            || keyCode == Constants.KEYBOARD_TAB
            || keyCode == Constants.KEYBOARD_PAGE_DOWN
            || keyCode == Constants.KEYBOARD_PAGE_UP
            || keyCode == Constants.KEYBOARD_HOME
            || keyCode == Constants.KEYBOARD_END
            || keyCode == Constants.KEYBOARD_NUMPAD_SUBTRACT
            || keyCode == Constants.KEYBOARD_NUMPAD_ADD
            || keyCode == Constants.KEYBOARD_F2
            || keyCode == Constants.KEYBOARD_NUMPAD_MULTIPLY) {

            if (this != this.grid.getBodyContainer() &&
                (keyCode == Constants.KEYBOARD_PAGE_DOWN
                    || keyCode == Constants.KEYBOARD_PAGE_UP
                    || keyCode == Constants.KEYBOARD_HOME
                    || keyCode == Constants.KEYBOARD_END)) {
                this.grid.getBodyContainer().handleCellKeyUp(this.grid.getBodyContainer().getFirstAvailableCell(null, false), keyCode, triggerEvent);
                return;
            }
            this.handleArrowKey(cell, keyCode, triggerEvent);
        }

        else if (keyCode == Constants.KEYBOARD_SPACE) {
            this.handleSpaceBar(cell, triggerEvent);
        }

    }

    handleSpaceBar(cell, triggerEvent) {

        this.handleMouseClick(cell, triggerEvent);

    }

    /**
     * Starts the Edit Session.
     */
    beginEdit(cell) {

        if (!cell.getColumn().getEditable()) return;
        this.grid._placementDirty = false;
        let evt = new flexiciousNmsp.FlexDataGridItemEditEvent(flexiciousNmsp.FlexDataGridEvent.ITEM_EDIT_BEGINNING, this.grid,
            cell.level, cell.getColumn(), cell, cell.rowInfo.getData());
        cell.getColumn().dispatchEvent(evt);
        if (evt.isDefaultPrevented()) return;

        if (this._inEdit)
            this.endEdit(this.getEditor());
        const prevEdit = this.getEditCell();
        this.setEditCell(cell);
        if (prevEdit)
            prevEdit.rowInfo.invalidateCells();
        this.getEditCell().rowInfo.invalidateCells();
        this._inEdit = true;
        this.setEditor(this.getEditCell().getColumn().itemEditor.newInstance({ cell: cell, row: cell.rowInfo, column: cell._column, level: cell.level, grid: cell.level.grid }, true));
        if (this.getEditor().implementsOrExtends("ISelectFilterControl")) {
            const iSelectEditor = this.getEditor().implementsOrExtends("ISelectFilterControl") ? this.getEditor() : null;
            if (this.getEditCell().getColumn().useFilterDataProviderForItemEditor) {
                const filterColumn = this.getEditCell().getColumn();
                this.getEditor().setAddAllItem(false);

                if (filterColumn.filterComboBoxBuildFromGrid)
                    this.getEditor().setDataProvider(filterColumn.getDistinctValues(this.grid.getRootFlat()));
                else if (filterColumn.filterComboBoxDataProvider)
                    this.getEditor().setDataProvider(filterColumn.filterComboBoxDataProvider);
                iSelectEditor.labelField = filterColumn.filterComboBoxLabelField;
                iSelectEditor.dataField = filterColumn.filterComboBoxDataField;
            }
        }

        this.getEditCell().attachUserClass(this.getEditCell().getColumn().getStyle("editorStyleName"));
        this.getEditCell().setText("");

        //editor.styleName = editCell.column;
        const pare = cell.getIsLeftLocked() ? this.grid.getLeftLockedContent() : cell.getIsLocked() ? this.grid.getRightLockedContent() : this;
        this.getEditor().grid = this.grid;
        this.getEditor().automationName = `${this.grid.automationName}_${this.getEditCell().getColumn().getUniqueIdentifier()}_editor`;

        const isTextInput = this.getEditor().implementsOrExtends("TextInput");
        const pt = this.getEditCell().placeComponent(this.getEditor(), this.getEditCell().getWidth() - (isTextInput ? 2 : 0),
            this.getEditCell().getHeight() - (isTextInput ? 4 : 0), false);
        const myPt = pare.globalToLocal(this.getEditCell().localToGlobal(pt));

        if (!cell.getColumn().itemEditorManagesPersistence) {
            this.grid.checkSetterAndApply(this.getEditor(), this.getEditCell().getColumn().editorDataField,
                UIUtils.resolveExpression(this.getEditCell().rowInfo.getData(), this.getEditCell().getColumn().getDataField()));
            this.getEditor().validateNow();
        }
        this.getEditor().setAttribute("className", "editor");
        //this.getEditor().setAttribute("id ", `${cell.getAutomationName()}_editor`);
        this.getEditor().move(myPt.getX(), myPt.getY());
        this.getEditor().setActualSize(this.getEditCell().getWidth(), this.getEditCell().getHeight());
        this.getEditor().addEventListener(this, Constants.EVENT_VALUE_COMMIT, this.onEditorValueCommit);
        this.getEditor().addEventListener(this, Constants.EVENT_KEY_DOWN, this.onEditorKeyDown);


        const sbRoot = flexiciousNmsp.DisplayList.instance().documentComponent;
        sbRoot.addEventListener(this, Constants.EVENT_MOUSE_UP, this.cancelEdit, true);
        flexiciousNmsp.DisplayList.instance().documentComponent.addEventListener(this, Constants.EVENT_RESIZE, this.onStageResize, true, 0);

        this.getEditor().focus();
        if (this.getEditor().implementsOrExtends("TextInput")) {
            const txt = this.getEditor().getText();
            if (txt.length > 0)
                this.getEditor().setSelection(txt.length, txt.length);
        }
        pare.addChild(this.getEditor());
        evt = new flexiciousNmsp.FlexDataGridItemEditEvent(flexiciousNmsp.FlexDataGridEvent.ITEM_EDIT_BEGIN, this.grid, cell.level, cell.getColumn(), cell, cell.rowInfo.getData());
        evt.itemEditor = this.getEditor();
        cell.getColumn().dispatchEvent(evt);

        evt = new flexiciousNmsp.FlexDataGridItemEditEvent(flexiciousNmsp.FlexDataGridEvent.ITEM_FOCUS_IN, this.grid, cell.level, cell.getColumn(), cell, cell.rowInfo.getData());
        evt.itemEditor = this.getEditor();
        this.grid.dispatchEvent(evt);


    }

    /**
     * Re-evaluates all the visible cell contents.
     */
    refreshCells() {
        for (const row of this.rows) {
            for (const cell of row.cells) {
                const cellComp = cell.component;
                if (cellComp)
                    cellComp.refreshCell();
            }
        }
    }

    /**
     *
     */
    onStageResize(event) {
        this.endEdit(this.getEditor());
        flexiciousNmsp.DisplayList.instance().documentComponent.removeEventListener(Constants.EVENT_RESIZE, this.onStageResize, true, 0);
    }

    /**
     * Handles the KeyDown on the Editor Component. If Escape, ends edit,
     * if Enter or Tab, commmits the value, and moves on to the next cell if Tab. Next Row if Enter.
     * If next Row or next cell belongs to a section that is not yet drawn, draws it and attempts
     * to focus into the next editor.
     */
    onEditorKeyDown(event) {
        if (event.keyCode == Constants.KEYBOARD_ESCAPE) {
            this.cancelEdit(event);
        }
        else if (event.keyCode == Constants.KEYBOARD_ENTER || event.keyCode == Constants.KEYBOARD_TAB) {
            let cell = this.getEditCell();

            const lhs = UIUtils.resolveExpression(cell.rowInfo.getData(), cell.getColumn().getDataField());
            const rhs = this.grid.checkGetterAndRetrieve(this.getEditor(), cell.getColumn().editorDataField);
            if (lhs == null && rhs == "") {
                //this commonly happens if the pojo has a null and the text input returns a blank string (when the user
                //does not make any change)
                this.endEdit(this.getEditor());
            }
            else if (lhs != rhs) {
                if (this.populateValue(event)) {
                    this.endEdit(this.getEditor());
                } else {
                    return;
                }
            } else {
                this.endEdit(this.getEditor());
            }
            //this.grid.refreshCells();
            cell.refreshCell();

            let nextCell;
            if (event.keyCode == Constants.KEYBOARD_ENTER) {
                nextCell = this.getCellInDirection(cell, Constants.CELL_POSITION_BELOW, true, true);
            }
            if (event.keyCode == Constants.KEYBOARD_TAB) {
                const ri = cell.rowInfo;
                const rowColumn = cell.getColumn();
                nextCell = this.getCellInDirection(cell, event.shiftKey ? Constants.CELL_POSITION_LEFT : Constants.CELL_POSITION_RIGHT, true, true);
                if (!nextCell || (nextCell == cell)) {//we're at the end...
                    //cell could have gotten recycled
                    cell = this.getCellForRowColumn(ri.getData(), rowColumn);
                    if (!cell) {
                        cell = this.getCellForRowColumn(ri.getData(), null, true);
                    }
                    nextCell = this.getCellInDirection(cell, event.shiftKey ? Constants.CELL_POSITION_ABOVE_LAST : Constants.CELL_POSITION_BELOW_FIRST, true, true);
                }
            }
            if (nextCell) {
                const cellData = new flexiciousNmsp.CellInfo(nextCell.rowInfo.getData(), nextCell.getColumn());
                if (nextCell) {

                    if (!nextCell.getIsLocked() && !this.grid.getForceColumnsToFitVisibleArea()) {
                        let pt = new flexiciousNmsp.Point(0, 0);
                        pt = nextCell.localToGlobal(pt);
                        pt = this.globalToLocal(pt);
                        let newHorizontal = 0;
                        newHorizontal = pt.getX();
                        if ((newHorizontal > (this.getHorizontalScrollPosition() + this.getWidth())) || (newHorizontal < (this.getHorizontalScrollPosition()))) {
                            this.grid.gotoHorizontalPosition(Math.min(newHorizontal, this.grid.getMaxHorizontalScrollPosition()));
                        }
                    }
                }
                nextCell = this.getCellForRowColumn(cellData.rowData, cellData.getColumn());
                const launchEdit = (nextCell.implementsOrExtends('IFlexDataGridDataCell'))
                    && nextCell.getColumn()
                    && nextCell.getColumn().getEditable();

                if (this.grid.enableDoubleClickEdit) {
                    if (launchEdit) {
                        this.beginEdit(nextCell);
                        event.preventDefault();
                    }
                }
                else {
                    if (launchEdit) {
                        event.preventDefault();
                        if (this.grid.getForceColumnsToFitVisibleArea()) {
                            this.beginEdit(nextCell);
                        }
                        else {
                            const that = this;
                            setTimeout(() => {
                                that.beginEdit(nextCell);
                            }, 100);
                        }

                    } else {
                        this.handleMouseClick(nextCell, event);
                    }
                }

            }
        }


    }

    /**
     *
     */
    doInvalidate() {

        this.invalidateDisplayList();

    }

    invalidateDisplayList() {

        flexiciousNmsp.UIComponent.prototype.invalidateDisplayList.apply(this, []);

    }

    /**
     *
     * @param hsp
     */
    gotoHorizontalPosition(hsp) {

        if (this.grid.getDataProvider() == null) {
            this._horizontalScrollPending = hsp;
            this.doInvalidate();
            return;
        }
        if (this.getHorizontalScrollPosition() != hsp) {
            const scrollRight = (this.getHorizontalScrollPosition() < hsp);
            this.setHorizontalScrollPosition(hsp);
            this.recycleH(this.grid.getColumnLevel(), scrollRight);
            this.grid.synchronizeHorizontalScroll();
        }


    }

    /**
     * Abstract
     */
    scrollToExistingRow(vsp, scrollDown) {



    }

    /**
     * Abstract
     **/
    isOutOfVisibleArea(row) {

        return false;

    }

    /**
     * Dispatches ITEM_EDIT_END.
     * Removes Event listeners
     * @method endEdit
     * @param editor
     */
    endEdit(editor) {

        if (!this._inEdit) return;
        this._inEdit = false;

        const evt = new flexiciousNmsp.FlexDataGridItemEditEvent(flexiciousNmsp.FlexDataGridEvent.ITEM_EDIT_END,
            this.grid, this.getEditCell().level, this.getEditCell().getColumn(), this.getEditCell(), this.getEditCell().rowInfo.getData());
        evt.itemEditor = this.getEditor();
        this.getEditCell().getColumn().dispatchEvent(evt);
        this.getEditCell().refreshCell();
        this.grid.highlightRow(this.getEditCell(), this.getEditCell().rowInfo, false);
        this.grid.invalidateCells();

        this.grid.currentCell = (null);
        this.setEditCell(null);

        const sbRoot = flexiciousNmsp.DisplayList.instance().documentComponent;
        sbRoot.removeEventListenerKeepDomListener(Constants.EVENT_MOUSE_UP, this.cancelEdit, true);
        this.getEditor().removeEventListener(Constants.EVENT_VALUE_COMMIT, this.onEditorValueCommit);
        this.getEditor().removeEventListener(Constants.EVENT_KEY_DOWN, this.onEditorKeyDown);
        editor.parent.removeChild(editor);

        editor.kill();

    }

    /**
     * @method onEditorKeyFocusChange
     * @param event
     */
    onEditorKeyFocusChange(event) {

        event.preventDefault();

    }

    /**
     * Cancels the current running edit. Dispatches ITEM_EDIT_CANCEL.
     * 
     * @method cancelEdit
     * @param event
     */
    cancelEdit(event) {
        if (!this._grid) {
            const sbRoot = flexiciousNmsp.DisplayList.instance().documentComponent;
            sbRoot.removeEventListenerKeepDomListener(Constants.EVENT_MOUSE_UP, this.cancelEdit, true);
            return;
        }
        if (UIUtils.isMouseEvent(event) &&
            (this.owns((event.target)) || (this.getEditor() && this.getEditor().owns((event.triggerEvent.target))))) {
            if ((event.target.implementsOrExtends("IFlexDataGridCell"))
                && (event.target).rowInfo.getIsFillRow()) {
                //end edit
            }
            else
                return;
        }
        if (UIUtils.isMouseEvent(event)) {
            if (event && !flexiciousNmsp.DisplayList.isMouseDown)
                return;
        }
        const evt = new flexiciousNmsp.FlexDataGridItemEditEvent(flexiciousNmsp.FlexDataGridEvent.ITEM_EDIT_CANCEL, this.grid, this.getEditCell().level,
            this.getEditCell().getColumn(), this.getEditCell(), this.getEditCell().rowInfo.getData(), event);
        evt.itemEditor = event.target;
        this.getEditCell().getColumn().dispatchEvent(evt);
        if (!evt.isDefaultPrevented())
            this.endEdit(this.getEditor());

    }

    /**
     *
     * @param event
     * @return
     *
     */
    onEditorValueCommit(event) {

        if (this.getEditCell().getColumn().itemEditorApplyOnValueCommit) {
            this.populateValue(event);
        }

    }

    /**
     * Applies the value from the editor back to the model object.
     * Dispatches ITEM_EDIT_VALUE_COMMIT
     * 
     * @method populateValue
     * @param event
     */
    populateValue(event) {

        const cell = this.getEditCell();
        if (cell.getColumn().itemEditorManagesPersistence) return true;
        //endEdit(event.currentTarget);
        const evt = new flexiciousNmsp.FlexDataGridItemEditEvent(
            flexiciousNmsp.FlexDataGridEvent.ITEM_EDIT_VALUE_COMMIT, this.grid, cell.level, cell.getColumn(), cell, cell.rowInfo.getData(), event);
        evt.itemEditor = event.currentTarget;
        evt.item = this.getEditCell().rowInfo.getData();
        evt.newValue = this.grid.checkGetterAndRetrieve(this.getEditor(), cell.getColumn().editorDataField);
        cell.getColumn().dispatchEvent(evt);
        if (!evt.isDefaultPrevented()) {
            if (cell.getColumn().itemEditorValidatorFunction == null ||
                cell.getColumn().itemEditorValidatorFunction(this.getEditor())) {

                if (this.grid.enableTrackChanges) {
                    this.grid.trackChange(cell.rowInfo.getData(), flexiciousNmsp.ChangeInfo.CHANGE_TYPE_UPDATE, cell.level,
                        cell.getColumn().getDataField(), UIUtils.resolveExpression(cell.rowInfo.getData(), cell.getColumn().getDataField())
                        , evt.newValue)
                }
                UIUtils.resolveExpression(cell.rowInfo.getData(), cell.getColumn().getDataField(),
                    this.grid.checkGetterAndRetrieve(this.getEditor(), cell.getColumn().editorDataField), false, true, this.grid);
                cell.refreshCell();
                return true;
            }
            else if (cell.getColumn().itemEditorValidatorFunction != null)
                return false;
        }
        else
            return false;
        return true;

    }

    /**
     * Calls invalidateBackground() on each of the cells.
     * 
     * @method invalidateCells
     */
    invalidateCells() {
        const allRows = this.getAllRows();

        for (const row1 of allRows) {
            for (const cell1 of row1.cells) {
                if (cell1.component.implementsOrExtends("IFlexDataGridCell")) {
                    cell1.component.invalidateBackground();
                }
            }
        }
    }

    /**
     * Gets the Rows Collection
     * 
     * @method getAllRows
     * @return {Array}
     *
     */
    getAllRows() {

        return this.rows;

    }

    /**
     * @private
     * @param {flexiciousNmsp.FlexDataGridCell} cell
     * @param {flexiciousNmsp.RowInfo} row
     * @param {Boolean} highLight
     * @param {*} highLightColor
     */
    highlightRow(cell, row, highLight, highLightColor) {
        if (typeof highLightColor == "undefined") highLightColor = 0;

        this.grid.highlightRow(cell, row, highLight, highLightColor);

    }

    /**
     * Gets the cell in the specified direction of the provided cell
     * @param thisCell		The cell to start searching at.
     * @param direction	The direction to search. Should be one of FlexDataGrid.CELL_POSITION_XXXX constants
     * @param dataOnly	Boolean flag to look only for data cells (IFlexDataGridDataCell)
     * @param editableOnly	Boolean flag to look only for editable cells
     * @param scrollIfNecessary	If next visible cell is beyond scrollable area, recycle and then search again.
     * @param hoverableOnly	Boolean flag to look only for cells where isHoverable(cell) is true.
     * @return The cell, if one matches the provided criteria, or null
     */
    getCellInDirection(
        thisCell,
        direction,
        dataOnly,
        editableOnly,
        scrollIfNecessary,
        hoverableOnly) {
        if (typeof dataOnly == "undefined") dataOnly = false;
        if (typeof editableOnly == "undefined") editableOnly = false;
        if (typeof scrollIfNecessary == "undefined") scrollIfNecessary = true;
        if (typeof hoverableOnly == "undefined") hoverableOnly = false;

        let row;
        let i;
        let adjacentRow;
        const thisRow = thisCell.rowInfo;
        const thiCells = thisRow.getInnerCells();
        thiCells.sort((a, b) => a.getPerceivedX() - b.getPerceivedX());
        const thisIndex = thiCells.indexOf(thisCell);
        const above = direction == Constants.CELL_POSITION_ABOVE || direction == Constants.CELL_POSITION_ABOVE_LAST;
        const toLeft = direction == Constants.CELL_POSITION_LEFT;
        const vertical = direction == Constants.CELL_POSITION_ABOVE ||
            direction == Constants.CELL_POSITION_BELOW || direction == Constants.CELL_POSITION_BELOW_FIRST
            || direction == Constants.CELL_POSITION_ABOVE_LAST;
        let adjacentCell;
        if (direction == Constants.CELL_POSITION_LEFT || direction == Constants.CELL_POSITION_RIGHT)
            adjacentRow = thisCell.rowInfo;
        else {
            for (i = 0; i < this.rows.length; i++) {
                row = this.rows[i];
                if (!adjacentRow)
                    adjacentRow = above ? row : this.rows[this.rows.length - 1];
                if (above && (row != thisRow) && (row.getY() > adjacentRow.getY()) && (row.getY() < thisRow.getY()) && (!dataOnly || row.getIsDataRow())) {
                    adjacentRow = row;
                }
                else if (!above && (row != thisRow) && (row.getY() < adjacentRow.getY()) && (row.getY() > thisRow.getY()) && (!dataOnly || row.getIsDataRow())) {
                    adjacentRow = row;
                }
            }
        }
        if (direction == Constants.CELL_POSITION_ABOVE_LAST && !this.grid.getForceColumnsToFitVisibleArea() &&
            scrollIfNecessary) {
            //we will need to scroll
            this.grid.gotoHorizontalPosition(this.grid.getMaxHorizontalScrollPosition());
        }
        else if (direction == Constants.CELL_POSITION_BELOW_FIRST &&
            !this.grid.getForceColumnsToFitVisibleArea() && scrollIfNecessary) {
            //we will need to scroll
            this.grid.gotoHorizontalPosition(0);
        }
        if (scrollIfNecessary && vertical && this.isOutOfVisibleArea(adjacentRow)) {
            if (!above) {
                //scroll down
                this.scrollToExistingRow(Math.min(this.getMaxVerticalScrollPosition(), this.getVerticalScrollPosition() + adjacentRow.getHeight()), true);
            }
            else {
                //scroll up
                this.scrollToExistingRow(Math.max(0, this.getVerticalScrollPosition() - adjacentRow.getHeight()), false);
            }
        }

        const cells = adjacentRow.cells.slice();
        cells.sort((a, b) => a.getPerceivedX() - b.getPerceivedX());
        if (direction == Constants.CELL_POSITION_ABOVE_LAST || direction == Constants.CELL_POSITION_RIGHT)
            cells.reverse();


        if (direction == Constants.CELL_POSITION_ABOVE ||
            direction == Constants.CELL_POSITION_BELOW) {
            if (thisRow.getIsColumnBased() && adjacentRow.getIsColumnBased() && thisCell.getColumn() && adjacentRow.getCellForColumn(thisCell.getColumn())) {
                return this.checkRowSpanColSpan((adjacentRow.getCellForColumn(thisCell.getColumn()).component), thisCell, direction, dataOnly, editableOnly, scrollIfNecessary, hoverableOnly);
            } else {
                return this.checkRowSpanColSpan(this.getFirstHoverableCell(adjacentRow, dataOnly, editableOnly), thisCell, direction, dataOnly, editableOnly, scrollIfNecessary, hoverableOnly);
            }
        }

        for (i = 0; i < cells.length; i++) {
            const adjCellComp = cells[i];
            const adjCell = adjCellComp.component;
            if (hoverableOnly && !this.isHoverableCell(adjCell)) continue;
            if (dataOnly && !(adjCell.implementsOrExtends("IFlexDataGridDataCell"))) continue;
            if (editableOnly && (!adjCell.getColumn() || !adjCell.getColumn().getEditable())) continue;
            {
                if (direction == Constants.CELL_POSITION_BELOW_FIRST || direction == Constants.CELL_POSITION_ABOVE_LAST)
                    return this.checkRowSpanColSpan(adjCell, thisCell, direction, dataOnly, editableOnly, scrollIfNecessary, hoverableOnly);
                else if (direction == Constants.CELL_POSITION_LEFT || direction == Constants.CELL_POSITION_RIGHT) {
                    if (direction == Constants.CELL_POSITION_LEFT) {
                        if (adjCell.getPerceivedX() < thisCell.getPerceivedX())
                            adjacentCell = adjCell;
                    } else {
                        if (adjCell.getPerceivedX() > thisCell.getPerceivedX())
                            adjacentCell = adjCell;
                    }
                }
                else {
                    throw new Error(`Invalid cell direction${direction}`);
                }
            }
        }
        const cellData = new flexiciousNmsp.CellInfo(thisCell.rowInfo.getData(), thisCell.getColumn());
        let diffParentAdjacentCell = null;
        if (scrollIfNecessary && !this.grid.getForceColumnsToFitVisibleArea() &&

            (
                (!toLeft && this.grid.getHorizontalScrollPosition() < this.grid.getMaxHorizontalScrollPosition()) ||
                (toLeft && this.grid.getHorizontalScrollPosition() > 0)
            )

        ) {

            if (adjacentCell) {
                if (thisCell.getIsLeftLocked() && !adjacentCell.getIsLocked() && !toLeft)
                    return this.checkRowSpanColSpan(adjacentCell, thisCell, direction, dataOnly, editableOnly, scrollIfNecessary, hoverableOnly);
                else if (thisCell.getIsRightLocked() && !adjacentCell.getIsLocked() && toLeft)
                    return this.checkRowSpanColSpan(adjacentCell, thisCell, direction, dataOnly, editableOnly, scrollIfNecessary, hoverableOnly);
                else if (thisCell.parent == adjacentCell.parent) {
                    if (adjacentCell.parent == this.grid.getBodyContainer()) {
                        //left and right scroll over the columns that are just outside the viewport causes them to half appear. so we set hsp
                        const halfVisibleRowdata = adjacentCell.rowInfo.getData();
                        const halfVisibleColumn = adjacentCell.getColumn();
                        if (halfVisibleRowdata && halfVisibleColumn) {
                            if (!this.grid.getForceColumnsToFitVisibleArea()) {
                                if (((adjacentCell.getX() + adjacentCell.getWidth()) >= this.getHorizontalScrollPosition()) &&
                                    (adjacentCell.getX() < this.getHorizontalScrollPosition())) {
                                    this.grid.gotoHorizontalPosition(adjacentCell.getX());
                                    return this.checkRowSpanColSpan(this.getCellForRowColumn(halfVisibleRowdata, halfVisibleColumn), thisCell, direction, dataOnly, editableOnly, scrollIfNecessary, hoverableOnly);
                                } else if (((adjacentCell.getX() + adjacentCell.getWidth()) > (this.getHorizontalScrollPosition() + this.getWidth())) &&
                                    (adjacentCell.getX() < (this.getHorizontalScrollPosition() + this.getWidth()))) {
                                    this.grid.gotoHorizontalPosition(Math.min(this.getHorizontalScrollPosition() + (adjacentCell.getWidth()), this.grid.getMaxHorizontalScrollPosition()));
                                    return this.checkRowSpanColSpan(this.getCellForRowColumn(halfVisibleRowdata, halfVisibleColumn), thisCell, direction, dataOnly, editableOnly, scrollIfNecessary, hoverableOnly);
                                }
                            }
                        }
                    }
                    return this.checkRowSpanColSpan(adjacentCell, thisCell, direction, dataOnly, editableOnly, scrollIfNecessary, hoverableOnly);
                }
                else
                    diffParentAdjacentCell = adjacentCell;
            }
            this.grid.gotoHorizontalPosition(!toLeft ?
                Math.min(this.grid.getHorizontalScrollPosition() + thisCell.getWidth(), this.grid.getMaxHorizontalScrollPosition()) :
                Math.max(this.grid.getHorizontalScrollPosition() - thisCell.getWidth(), 0)
            );

            const adjCol = cellData.getColumn() ?
                cellData.getColumn().getAdjacentColumn(toLeft ? -1 : 1) : null;
            if (!adjCol)
                return this.checkRowSpanColSpan(diffParentAdjacentCell, thisCell, direction, dataOnly, editableOnly, scrollIfNecessary, hoverableOnly);
            thisCell = this.getCellForRowColumn(cellData.rowData, adjCol);

            if (thisCell == null) {
                //this means our anchor column actually scrolled out of view, so we move on to the next visible.
                let cols = adjCol.level.getVisibleColumns([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE]);
                if (toLeft) cols = cols.reverse();

                for (i = 0; i < cols.length; i++) {
                    const col = cols[i];
                    if (((col.getColIndex() < adjCol.getColIndex) && toLeft)
                        || ((col.getColIndex > adjCol.getColIndex) && !toLeft)) {
                        const oCell = this.getCellForRowColumn(cellData.rowData, col);
                        if (oCell) {
                            thisCell = oCell;
                            break;
                        }
                    }
                }
            }

            const result =
                this.getCellInDirection(thisCell, direction,
                    dataOnly, editableOnly, scrollIfNecessary);
            if (result) {
                adjacentCell = result;
            } else
                adjacentCell = diffParentAdjacentCell;
        }
        return this.checkRowSpanColSpan(adjacentCell, thisCell, direction, dataOnly, editableOnly, scrollIfNecessary, hoverableOnly);
    }

    checkRowSpanColSpan(
        retCell,
        thisCell,
        direction,
        dataOnly,
        editableOnly,
        scrollIfNecessary,
        hoverableOnly) {

        if (!this.grid.getHasRowSpanOrColSpan()) {
            return retCell;
        }
        else if (retCell == thisCell) {
            return retCell;
        }
        else if (retCell && this.isCoveredByRowSpanOrColspan(retCell)) {
            const cells = retCell.rowInfo.cells.slice().reverse();

            for (const row of this.rows) {
                for (const cell of row.cells) {
                    if ((cell.component.implementsOrExtends('IFlexDataGridDataCell'))) {
                        if ((cell.component).colSpan > 1) {
                            return cell.component;
                        }
                    }
                }
            }

            return this.getCellInDirection(retCell, direction, dataOnly, editableOnly, scrollIfNecessary, hoverableOnly);
        } else
            return retCell;


    }

    isCoveredByRowSpanOrColspan(cell) {

        return this.isCoveredByRowSpan(cell) || this.isCoveredByColSpan(cell);

    }

    /**
     * Given a cell, returns true if the cell is "hidden" by another cell that has a row span which will cover this cell.
     * @param cell	The cell to check
     */
    isCoveredByRowSpan(cell) {
        if (!this.grid.getHasRowSpanOrColSpan()) return false;

        for (const row of this.rows) {
            if (row.rowPositionInfo.getVerticalPosition() < cell.rowInfo.rowPositionInfo.getVerticalPosition()) {
                //this row is above us.
                const maxCellHt = row.getMaxCellHeight(cell.getColumn());
                if ((maxCellHt > row.getHeight()) && ((maxCellHt + row.rowPositionInfo.getVerticalPosition()) > cell.rowInfo.rowPositionInfo.getVerticalPosition())) {
                    return true;
                }
            } else
                return false;
        }

        return false;
    }

    /**
     * Given a cell, returns true if the cell is "hidden" by another cell that has a col span which will cover this cell.
     * @param cell	The cell to check
     */
    isCoveredByColSpan(cell) {
        if (!this.grid.getHasRowSpanOrColSpan()) return false;

        for (const comp of cell.rowInfo.cells) {
            const cellComp = comp.component;
            if (cellComp.getPerceivedX() > cell.getPerceivedX()) return false;//we're ahead of cell.
            if (cellComp && cellComp.getColumn() && (cellComp.getWidth() > cellComp.getColumn().getWidth())) {
                if ((cellComp.getPerceivedX() + cellComp.getWidth()) > cell.getPerceivedX())
                    return true;
            }
        }

        return false;
    }

    /**
     * Given a column and a data object, returns the cell associated with the data object. If the column is null, will
     * return the first unlocked cell of the row associated with the data object, if includeExp is set to true, returns the
     * expand collapse cell. Please note, this only checks the rows that are currently drawn, not the entire dataprovider.
     * Only the data objects that have currently drawn rows (the visible viewport) will return a cell object, others will reuturn null
     * @param dataObject	An item in the dataprovider.
     * @param col			A column to match. Can be null.
     * @param includeExp	Whether or not to include the expand collapse cell.
     * @return IFlexDataGridCell
     */
    getCellForRowColumn(dataObject, col, includeExp) {
        if (typeof includeExp == "undefined") includeExp = false;

        for (const row of this.rows) {
            if (row.rowPositionInfo.rowData == dataObject) {
                for (const cell of row.cells) {
                    const fCell = cell.component;
                    if (includeExp && (fCell.getIsExpandCollapseCell())) {
                        return fCell;
                    }
                    else if (fCell && fCell.getColumn() == col && col != null) {
                        return fCell;
                    }
                    else if (col == null && !fCell.getIsLocked() && (fCell.implementsOrExtends("IFlexDataGridDataCell"))) {
                        return fCell;
                    }
                }
            }
        }

        return null;
    }

    /**
     *
     */
    removeEventListeners(comp) {

        const cell = comp.component;
        if (cell.implementsOrExtends("FlexDataGridPagerCell")) {
            const iPager = cell.getRenderer();
            if (iPager.level.getNestDepth() == 1)
                iPager.removeEventListener(flexiciousNmsp.ExtendedFilterPageSortChangeEvent.PAGE_CHANGE, this.rootPageChange);
            else
                iPager.removeEventListener(flexiciousNmsp.ExtendedFilterPageSortChangeEvent.PAGE_CHANGE, this.onPageChange);
        }
        if (cell.implementsOrExtends("FlexDataGridHeaderCell")) {
            cell.removeEventListener(Constants.EVENT_MOUSE_MOVE, this.onHeaderCellMouseMove);
            cell.removeEventListener(Constants.EVENT_MOUSE_OUT, this.onHeaderCellMouseOut);
            cell.removeEventListener(Constants.EVENT_MOUSE_DOWN, this.onHeaderCellMouseDown);
            if (cell.getColumn().implementsOrExtends("FlexDataGridCheckBoxColumn"))
                cell.removeEventListener(flexiciousNmsp.FlexDataGridEvent.SELECT_ALL_CHECKBOX_CHANGED, this.selectAllChangedHandler);
            cell.removeEventListener(Constants.EVENT_MOVE, this.placeSortIcon);
            cell.removeEventListener(Constants.EVENT_RESIZE, this.placeSortIcon);

        }
        if (cell.implementsOrExtends("FlexDataGridColumnGroupCell")) {
            cell.removeEventListener(Constants.EVENT_MOUSE_MOVE, this.onHeaderCellMouseMove);
            cell.removeEventListener(Constants.EVENT_MOUSE_OUT, this.onHeaderCellMouseOut);
            cell.removeEventListener(Constants.EVENT_MOUSE_DOWN, this.onHeaderCellMouseDown);


        }
        if (cell.implementsOrExtends("FlexDataGridFilterCell") && cell.getRenderer().implementsOrExtends && cell.getRenderer().implementsOrExtends("IFilterControl")) {
            cell.rowInfo.unRegisterIFilterControl(cell.getRenderer());
        }
        if (this.grid.enableDrop && cell) {
            cell.removeEventListener(Constants.EVENT_MOUSE_MOVE, this.onCellDropMouseMove);
        }

        cell.removeEventListener(Constants.EVENT_MOUSE_OVER, this.onCellMouseOver);
        cell.removeEventListener(Constants.EVENT_KEY_UP, this.onCellKeyUp);
        cell.removeEventListener(Constants.EVENT_CLICK, this.onCellMouseClick);
        cell.removeEventListener(Constants.EVENT_DOUBLE_CLICK, this.onCellDoubleClick);
        cell.removeEventListener(Constants.EVENT_MOUSE_OUT, this.onCellMouseOut);

    }

    /**
     *
     */
    addRow(ht, scrollDown, rowPositionInfo) {


        const rowInfo = new flexiciousNmsp.RowInfo(ht, this.implementsOrExtends("FlexDataGridBodyContainer") ?
            rowPositionInfo.getVerticalPosition() : this.grid.currentPoint.contentY, this.grid);
        rowInfo.rowPositionInfo = rowPositionInfo;

        this.grid.currentPoint.contentY += ht;
        this.grid.currentPoint.contentX = this.grid.currentPoint.leftLockedContentX = this.grid.currentPoint.rightLockedContentX = 0;
        if (!rowInfo.getIsFillRow()) {
            if (scrollDown)
                this.rows.push(rowInfo);
            else {
                //scrolling up...
                let insertionPoint = 0;
                while (insertionPoint < this.rows.length && rowInfo.getY() > this.rows[insertionPoint].getY()) {
                    insertionPoint++;
                }
                this.rows.splice(insertionPoint, 0, rowInfo);
            }
            //trace("Added "+rowInfo.getY())

            if (rowInfo.getIsFilterRow()) {
                rowInfo.addEventListener(this, flexiciousNmsp.ExtendedFilterPageSortChangeEvent.FILTER_CHANGE, this.onFilterChange);
            }
        }
        return rowInfo;

    }

    getBorderWidth(cell) {
        return 0;//cell.hasVerticalGridLines()?cell.getVerticalGridLineThickness():0;
    }

    getBorderHeight(cell) {
        return 0;//cell.hasHorizontalGridLines()?cell.getVerticalGridLineThickness():0;
    }

    /**
     *
     */
    addCell(component, row, existingComponent) {
        if (typeof existingComponent == "undefined") existingComponent = null;

        const result = new flexiciousNmsp.ComponentAdditionResult();
        let comp = existingComponent;
        let cell;
        if (component.implementsOrExtends("IFlexDataGridCell")) {
            cell = component;

            if (cell.getIsLeftLocked()) {
                var remaining = 0;
                if (cell.getIsContentArea()) {
                    comp = this.addToSection(cell, row, "left", existingComponent);
                }
                else if (cell.rowInfo.getIsFooterRow() || cell.rowInfo.getIsPagerRow()) {
                    if (!existingComponent) {
                        this.grid.getLeftLockedFooter().addChild(component);
                        comp = row.addCell(component, this.grid.currentPoint.leftLockedFooterX);
                    }
                    comp.setX(this.grid.currentPoint.leftLockedFooterX);

                    if (cell.getColumn() && cell.getColumn().getIsLastLeftLocked() || (cell.level.getNestDepth() == 1 && cell.level.getLeftLockedColumns().length == 0)) {
                        remaining = this.grid.getLeftLockedContent().getWidth() - this.grid.currentPoint.leftLockedFooterX;
                        cell.setActualSize(remaining - this.getBorderWidth(cell), cell.getHeight() - this.getBorderHeight(cell));
                    }
                    this.grid.currentPoint.leftLockedFooterX += component.getWidth();
                    component.setY(this.grid.currentPoint.leftLockedFooterY);
                    comp.inCornerAreas = true;
                }
                else if (cell.rowInfo.getIsHeaderRow() || cell.rowInfo.getIsFilterRow() || cell.rowInfo.getIsColumnGroupRow()) {
                    if (!existingComponent) {
                        this.grid.getLeftLockedHeader().addChild(component);
                        comp
                            = row.addCell(component, this.grid.currentPoint.leftLockedHeaderX);
                    }
                    comp.setX(this.grid.currentPoint.leftLockedHeaderX);

                    if (cell.getColumn() && cell.getColumn().getIsLastLeftLocked() || (cell.level.getNestDepth() == 1 && cell.level.getLeftLockedColumns().length == 0)) {

                        remaining = this.grid.getLeftLockedContent().getWidth() - this.grid.currentPoint.leftLockedHeaderX;
                        cell.setActualSize(remaining - this.getBorderWidth(cell), cell.getHeight() - this.getBorderHeight(cell));
                    }
                    this.grid.currentPoint.leftLockedHeaderX += component.getWidth();
                    component.setY(this.grid.currentPoint.leftLockedHeaderY);
                    comp.inCornerAreas = true;
                }

            }
            else if (cell.getIsRightLocked()) {
                if (cell.getIsContentArea()) {
                    comp = this.addToSection(cell, row, "right");
                }
                else if (cell.rowInfo.getIsFooterRow() || cell.rowInfo.getIsPagerRow()) {
                    if (!existingComponent) {
                        this.grid.getRightLockedFooter().addChild(component);
                        comp = row.addCell(component, this.grid.currentPoint.rightLockedFooterX);
                    }
                    comp.setX(this.grid.currentPoint.rightLockedFooterX);
                    if (cell.getColumn() && cell.getColumn().getIsLastrightLocked()) {
                        remaining = this.grid.getRightLockedContent().getWidth() - this.grid.currentPoint.rightLockedFooterX;
                        cell.setActualSize(remaining - this.getBorderWidth(cell), cell.getHeight() - this.getBorderHeight(cell));
                    }
                    this.grid.currentPoint.rightLockedFooterX += component.getWidth();
                    component.setY(this.grid.currentPoint.rightLockedFooterY);
                    comp.inCornerAreas = true;
                }
                else if (cell.rowInfo.getIsHeaderRow() || cell.rowInfo.getIsFilterRow() || cell.rowInfo.getIsColumnGroupRow()) {
                    if (!existingComponent) {
                        this.grid.getRightLockedHeader().addChild(component);
                        comp = row.addCell(component, this.grid.currentPoint.rightLockedHeaderX);
                    }
                    comp.setX(this.grid.currentPoint.rightLockedHeaderX);
                    if (cell.getColumn() && cell.getColumn().getIsLastrightLocked()) {
                        remaining = this.grid.getRightLockedContent().getWidth() - this.grid.currentPoint.rightLockedHeaderX;
                        cell.setActualSize(remaining - this.getBorderWidth(cell), cell.getHeight() - this.getBorderHeight(cell));
                    }
                    this.grid.currentPoint.rightLockedHeaderX += component.getWidth();
                    component.setY(this.grid.currentPoint.rightLockedHeaderY);
                    comp.inCornerAreas = true;
                }
            }
            else {

                if (!existingComponent) {
                    this.addChild(component);
                    comp = row.addCell(component, this.grid.currentPoint.contentX);
                }
                comp.setX(this.grid.currentPoint.contentX);
                this.grid.currentPoint.contentX += component.getWidth();
            }
        }
        else {
            if (!existingComponent) {
                this.addChild(component);
                comp = row.addCell(component, this.grid.currentPoint.contentX);
            }
            comp.setX(this.grid.currentPoint.contentX);
            this.grid.currentPoint.contentX += component.getWidth();
        }

        result.componentInfo = comp;
        if (comp.component.implementsOrExtends("IFlexDataGridCell")) {
            (comp.component).componentInfo = comp;
        }

        result.horizontalSpill = (this.grid.currentPoint.contentX > this.getWidth());
        result.verticalSpill = (this.grid.currentPoint.contentY > this.getHeight() + 1);
        return result;
    }

    /**
     *
     */
    addToSection(cell, row, lockDir, existingComponent) {
        if (typeof existingComponent == "undefined") existingComponent = null;

        let remaining = 0;
        if (!existingComponent)
            this.grid[`get${UIUtils.doCap(lockDir)}LockedContent`]().addChild(cell);
        //if this level has a renderer, the first lockDir locked cell is also the last
        //if this level does not have a renderer and there are no lockDir locked columns,
        //the disclosure cell, which is the second cell added is going to be the last
        //else if the level does have lockDir locked columns, then we ask the column if
        //it is the last lockDir locked column.
        const isLast = (cell.rowInfo.getIsRendererRow() || cell.rowInfo.getIsPagerRow())
            || ((cell.rowInfo.cells.length > 0 || cell.level.getNestDepth() == 1) && cell.level[`get${UIUtils.doCap(lockDir)}LockedColumns`]().length == 0)
            || cell.getColumn() && cell.getColumn()[`getIsLast${lockDir}Locked`]();

        const comp = existingComponent ? existingComponent : row.addCell(cell, this.grid.currentPoint[`${lockDir}LockedContentX`]);
        comp.setX(this.grid.currentPoint[`${lockDir}LockedContentX`]);
        if (isLast) {
            remaining = this.grid[`get${UIUtils.doCap(lockDir)}LockedContent`]().getWidth() - this.grid.currentPoint[`${lockDir}LockedContentX`];
            cell.setActualSize(remaining - this.getBorderWidth(cell), cell.getHeight() - this.getBorderHeight(cell))
        }
        this.grid.currentPoint[`${lockDir}LockedContentX`] += cell.getWidth();
        return comp;

    }

    /**
     * In lazy loaded grids, gets the loaded info of the provided object.
     */
    findLoadingInfo(item, level, useSelectedKeyField) {
        if (typeof useSelectedKeyField == "undefined") useSelectedKeyField = false;

        for (const loadingItem of this._loadedItems) {
            if (loadingItem.isEqual(item, level, useSelectedKeyField)) {
                return loadingItem;
            }
        }

        return null;
    }

    /**
     * @method processRendererLevel
     * 
     * @param {flexiciousNmsp.RowPositionInfo} rowPositionInfo
     * @param {boolean} scrollDown
     * 
     * @return {flexiciousNmsp.RowIndo} row
     */
    processRendererLevel(rowPositionInfo, scrollDown) {
        if (typeof scrollDown == "undefined") scrollDown = false;


        const level = rowPositionInfo.getLevel(this.grid);
        const levelRendererHeight = level.getLevelRendererHeight(rowPositionInfo.rowData);
        const paddingY = level.getStyleValue("horizontalGridLines") ? 1 : 0;
        const row = this.addRow(levelRendererHeight, scrollDown, rowPositionInfo);
        if (!row.paddingExists())
            this.addPadding(level.getNestDepth(), row, levelRendererHeight - paddingY, level);
        else {

        }
        const dataCell = this.grid.rendererCache.popInstance(FlexDataGridContainerBase._levelRendererFactory);
        dataCell.setRendererFactory(level.nextLevelRenderer);
        dataCell.level = level;
        dataCell.rowInfo = row;

        this.addEventListeners(this.addCell(dataCell, row).componentInfo);
        dataCell.refreshCell();
        dataCell.setActualSize(rowPositionInfo.getLevel(this.grid).getSumOfColumnWidths([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE]) - this.getBorderWidth(dataCell),
            levelRendererHeight - this.getBorderHeight(dataCell));
        this.grid.placeComponents(row);
        return row;

    }

    /**
     * @method placeComponents
     */
    placeComponents() {
        for (const row of this.rows) {
            this.grid.placeComponents(row);
        }
    }

    /**
     * @method processHeaderLevel
     * 
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level
     * @param {flexiciousNmsp.RowPositionInfo} rowPositionInfo
     * @param {boolean} scrollDown
     * @param {object} item
     * @param {number} chromeLevel
     * @param {flexiciousNmsp.RowInfo} existingRow
     * @param {Array} existingComponents
     */
    processHeaderLevel(
        level,
        rowPositionInfo,
        scrollDown,
        item,
        chromeLevel,
        existingRow,
        existingComponents) {


        const paddingY = level.getStyleValue("horizontalGridLines") ? 1 : 0;
        const rowHeight = level.getRowHeightFromType(chromeLevel);

        let row;
        if (!existingRow) {
            row = this.addRow(rowHeight, scrollDown, rowPositionInfo);
        }
        else {
            this.grid.currentPoint.contentX = this.grid.currentPoint.rightLockedContentX = this.grid.currentPoint.leftLockedContentX = 0;
            row = existingRow;
        }
        const nestLevel = rowPositionInfo ? rowPositionInfo.getRowNestlevel() : 0;
        if (this.grid.enableDefaultDisclosureIcon) {
            if ((this.isInVisibleHorizontalRange
                (this.grid.currentPoint.contentX, level.nestIndent * (nestLevel - 1)) || this.grid.lockDisclosureCell)
                && (!existingRow || !row.paddingExists()))
                this.addPadding(rowPositionInfo.getRowNestlevel(), row, rowHeight, level);
            else if (!this.grid.lockDisclosureCell)
                this.grid.currentPoint.contentX += level.nestIndent * (nestLevel - 1);

            if (row.paddingCell) {
                row.paddingCell.refreshCell();
            }


            if (!row.getIsPagerRow()) {

                if ((level.nextLevel || level.nextLevelRenderer)) {
                    if ((this.isInVisibleHorizontalRange(this.grid.currentPoint.contentX,
                        level.nestIndent) || this.grid.lockDisclosureCell)
                        && (!existingRow || !row.disclosureExists())) {
                        if (!row.getIsColumnGroupRow()) {
                            const disclosureCell = this.grid.rendererCache.popInstance(level.expandCollapseHeaderCellRenderer);
                            disclosureCell.rowInfo = row;
                            disclosureCell.level = level;
                            disclosureCell.setActualSize(level.getMaxDisclosureCellWidth() - this.getBorderWidth(disclosureCell),
                                (level.getMaxColumnGroupDepth() > 1 && chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_HEADER ?
                                    this.grid.getHeaderContainer().getHeight() : rowHeight) - this.getBorderHeight(disclosureCell));
                            this.addEventListeners(this.addCell(disclosureCell, row).componentInfo)
                        }
                    }
                    else if (!this.grid.lockDisclosureCell)
                        this.grid.currentPoint.contentX += level.getMaxDisclosureCellWidth();

                    if (row.disclosureCell) {
                        row.disclosureCell.refreshCell();
                    }
                }
            }
        }

        if (chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_PAGER) {

            const pagerCell = this.grid.rendererCache.popInstance(level.pagerCellRenderer, level.getPagerRenderer());
            pagerCell.setRendererFactory(level.getPagerRenderer());

            pagerCell.level = level;
            pagerCell.rowInfo = row;
            const iPager = pagerCell.getRenderer();
            iPager.grid = null;
            iPager.grid = this.grid;
            let pagerWidth;
            if (level.getNestDepth() == 1) {
                pagerWidth = this.getPagerWidth();
            }
            else {
                pagerWidth = level.getSumOfColumnWidths([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE]);
            }
            if (this.grid.getStyle("pagerStyleName"))
                pagerCell.attachUserClass(this.grid.getStyle("pagerStyleName"));
            pagerCell.setActualSize(pagerWidth - this.getBorderWidth(pagerCell), rowHeight - this.getBorderHeight(pagerCell));


            pagerCell.invalidateDisplayList();
            if (level.getNestDepth() == 1) {
                pagerCell.setVisible(this.grid.getPagerRowHeight() > 0);
            }
            this.addEventListeners(this.addCell(pagerCell, row).componentInfo);
            iPager.rowInfo = row;
            let destroyed = false;
            if (iPager.level != null && iPager.level != level) {
                iPager.destroy();
                destroyed = true;
            }
            iPager.level = level;
            if (destroyed) {
                iPager.initialize();
            }
            pagerCell.refreshCell();
            if (level.getNestDepth() == 1) {
                iPager.addEventListener(this, flexiciousNmsp.ExtendedFilterPageSortChangeEvent.PAGE_CHANGE, this.rootPageChange);
            }
            else
                iPager.addEventListener(this, flexiciousNmsp.ExtendedFilterPageSortChangeEvent.PAGE_CHANGE, this.onPageChange);

        }
        else {
            let cols = level.getVisibleColumns([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE, flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_LEFT]);
            let flatDp;
            if (chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_FILTER) {
                if (level.getNestDepth() == 1) {
                    flatDp = this.grid.getRootFlat();
                } else {
                    flatDp = level.getChildren(item);
                }
            }

            if (chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_COLUMN_GROUP) {
                let currentCGCell;
                const cgLevel = row.getColumnGroupDepth(this.grid);
                const cgs = level.getColumnGroupsAtLevel(cgLevel);

                for (const cg of cgs) {
                    const cgWx = cg.getWX();
                    if (cg.getStartColumn() && (this.isInVisibleHorizontalRange(cgWx[1], cgWx[0])
                        || cg.getStartColumn().getIsLocked())
                        && (!existingRow || !row.columnGroupCellExists(cg))) {
                        currentCGCell =
                            this.processColumnGroupCell(level, cg.columnGroupCellRenderer, row,
                                item, rowHeight, cg, existingComponents);
                        currentCGCell.setText(cg.getHeaderText(currentCGCell));
                        currentCGCell.refreshCell();
                    }
                }
            }
            else {
                this.grid.currentPoint.headerX = 0;
                for (var colIndex = 0; colIndex < cols.length; colIndex++) {
                    var col = cols[colIndex];
                    if (!col.getWidth()) col.setWidth(100);
                    if ((this.isInVisibleHorizontalRange(this.grid.currentPoint.headerX, col.getWidth())
                        || col.getIsLocked())
                        && (!existingRow || !row.columnCellExists(col))) {
                        if (chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_HEADER) {
                            this.processHeaderCell(level, cols, colIndex, row, item, rowHeight, existingComponents);
                        }
                        else if (chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_FILTER) {
                            this.processFilterCell(level, cols, colIndex, row, item, flatDp, existingComponents);
                        }
                        else {
                            this.processFooterCell(level, cols, colIndex, row, item, existingComponents);
                        }

                    }
                    else {
                        if (!col.getIsLocked())
                            this.grid.currentPoint.contentX += col.getWidth();
                    }
                    if (!col.getIsLocked())
                        this.grid.currentPoint.headerX += col.getWidth();

                }
                if (level.getNestDepth() == 1) {
                    if ((this.grid.getForceColumnsToFitVisibleArea()
                        || (this.getWidth() >= level.getSumOfColumnWidths([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE]))
                        || (this.getMaxHorizontalScrollPosition() > 0 &&
                            (this.getMaxHorizontalScrollPosition() - this.getHorizontalScrollPosition() < 20)))
                        && !row.scrollbarPadExists()
                        && !this.grid.implementsOrExtends("PrintFlexDataGrid"))
                        this.addPadding(2, row, this.getHeight(), level, false, true, Constants.VERTICAL_SCROLLBAR_WIDTH);
                }
                //now that we're done with the left locked and unlocked cols, do the right locked ones..
                cols = level.getVisibleColumns([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_RIGHT]);
                let wSoFar = 0;
                for (colIndex = 0; colIndex < cols.length; colIndex++) {
                    col = cols[colIndex];
                    this.grid.currentPoint.contentX = this.getWidth() - wSoFar - col.getWidth() + 1;
                    if ((!existingRow || !row.columnCellExists(col))) {
                        if (chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_HEADER) {
                            this.processHeaderCell(level, cols, colIndex, row, item, rowHeight, existingComponents);
                        }
                        else if (chromeLevel == flexiciousNmsp.RowPositionInfo.ROW_TYPE_FILTER) {
                            this.processFilterCell(level, cols, colIndex, row, item, flatDp, existingComponents);
                        }
                        else {
                            this.processFooterCell(level, cols, colIndex, row, item, existingComponents);
                        }
                    }
                    wSoFar += col.getWidth();
                }

            }
        }
        //do we have something right locked? if not, add something so we dont show a hole
        //        if(!row.rightLockedExists() && level.getNestDepth()>1)
        //            this.addPadding(rowPositionInfo.getRowNestlevel(),row,rowHeight,level,true);

        this.grid.placeComponents(row);
        return row;

    }

    /**
     * @method getPagerWidth
     */
    getPagerWidth() {

        const borderThickness = this.grid.getStyle("borderThickness");
        return this.grid.getWidth() - (this.grid.hasBorderSide("left") ? borderThickness : 0)
            - (this.grid.hasBorderSide("right") ? borderThickness : 0);

    }

    /**
     * @method initializeRendererFromColumn
     * 
     * @param filterRenderer
     * @param filterColumn
     */
    initializeRendererFromColumn(filterRenderer, filterColumn) {
        if (filterRenderer.implementsOrExtends('IFilterControl')) {
            const txt = filterRenderer.implementsOrExtends("TextInput") ? filterRenderer : null;
            if (txt) {
                if (filterColumn.hasOwnProperty("filterWaterMark"))
                    txt.watermark = filterColumn.filterWaterMark;
                if (filterColumn.getStyle("filterIcon"))
                    txt.setStyle("insideIcon", filterColumn.getStyle("filterIcon"));
                if (filterColumn.hasOwnProperty("clearFilterOnIconClick"))
                    txt.clearTextOnIconClick = filterColumn.clearFilterOnIconClick;
                if (filterColumn.hasOwnProperty("enableFilterAutoComplete")
                    && filterColumn.enableFilterAutoComplete) {
                    txt.enableAutoComplete = true;
                }
                if (filterColumn.hasOwnProperty("showClearIconWhenHasText"))
                    txt.showIconWhenHasText = filterColumn.showClearIconWhenHasText;
            }
            const mscb = filterRenderer.implementsOrExtends("MultiSelectComboBox") ? filterRenderer : null;
            if (mscb) {
                if (filterColumn.hasOwnProperty("filterWaterMark"))
                    mscb.searchBoxWatermark = filterColumn.filterWaterMark;
                if (filterColumn.getStyle("filterIcon"))
                    mscb.setStyle("filterIcon", filterColumn.getStyle("filterIcon"));
            }
            filterRenderer.grid = filterColumn.getOwner();
            filterRenderer.searchField = filterColumn.getSearchField();
            filterRenderer.filterOperation = filterColumn.filterOperation;
            filterRenderer.filterTriggerEvent = filterColumn.filterTriggerEvent;
            if (filterColumn.filterComparisionType != flexiciousNmsp.FilterExpression.FILTER_COMPARISION_TYPE_AUTO)
                filterRenderer.filterComparisionType = filterColumn.filterComparisionType;
            filterRenderer.gridColumn = (filterColumn);
        }
        //here we need to do the following
        //1) If filterRenderer.implementsOrExtends('ISelectFilterControl, set its dataProvider, dataFie')){
        const iSelectFilterControl = filterRenderer.implementsOrExtends("ISelectFilterControl") ? filterRenderer : null;
        if (iSelectFilterControl != null) {
            if (filterColumn.filterComboBoxDataProvider != null)
                iSelectFilterControl.setDataProvider(filterColumn.filterComboBoxDataProvider);
            if (filterColumn.filterComboBoxDataField != "data")
                //the user overrode it
                iSelectFilterControl.dataField = filterColumn.filterComboBoxDataField;
            if (filterColumn.filterComboBoxLabelField != "label")
                //the user overrode it
                iSelectFilterControl.labelField = filterColumn.filterComboBoxLabelField;

            iSelectFilterControl.setAddAllItem(true);
            if (filterColumn.filterComboBoxWidth)
                iSelectFilterControl.dropdownWidth = filterColumn.filterComboBoxWidth;
        }
        const iTriStateCheckBoxFilterControl = filterRenderer.implementsOrExtends("ITriStateCheckBoxFilterControl") ? filterRenderer : null;
        if (iTriStateCheckBoxFilterControl != null) {
            iTriStateCheckBoxFilterControl.allowUserToSelectMiddle = true;
            iTriStateCheckBoxFilterControl.setSelectedState("middle");
        }
        const iDateFilterControl = filterRenderer.implementsOrExtends("IDateComboBox") ? filterRenderer : null;
        if (iDateFilterControl && filterColumn.filterDateRangeOptions.length > 0)
            iDateFilterControl.setDateRangeOptions(filterColumn.filterDateRangeOptions);
        //(filterRenderer).percentWidth=100;
        filterRenderer.invalidateDisplayList();
    }

    /**
     * @method processFilterCell
     * 
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level
     * @param {Array} cols
     * @param {number} colIndex
     * @param {flexiciousNmsp.RowInfo} row
     * @param {object} item
     * @param {object} flatDp
     * @param {Array} existingComponents
     */
    processFilterCell(level, cols, colIndex, row, item, flatDp, existingComponents) {

        const paddingY = level.getStyleValue("horizontalGridLines") ? 1 : 0;
        const col = cols[colIndex];
        const rowHeight = level.getFilterRowHeight();
        const filterCell = this.grid.rendererCache.popInstance(col.filterCellRenderer, col.getFilterRenderer());

        filterCell.setRendererFactory(col.getFilterRenderer());
        filterCell.rowInfo = row;
        filterCell.level = level;
        filterCell.setColumn(col);
        filterCell.setActualSize(col.getWidth() - this.getBorderWidth(filterCell), rowHeight - this.getBorderHeight(filterCell));
        this.addEventListeners(this.addCell(filterCell, row).componentInfo);
        filterCell.setActualSize(col.getWidth() - this.getBorderWidth(filterCell), rowHeight - this.getBorderHeight(filterCell));
        if (level.getNestDepth() == 1) {
            filterCell.setVisible(level.getFilterRowHeight() > 0);
        }

        filterCell.attachUserClass(col.getStyleValue("filterStyleName"));

        const filterRenderer = filterCell.getRenderer();
        if (filterRenderer && filterRenderer.implementsOrExtends && filterRenderer.implementsOrExtends('IFilterControl')) {
            this.initializeRendererFromColumn(filterRenderer, col);
            this.initializeFilterRenderer(filterRenderer, col, item, flatDp, level);
            row.registerIFilterControl(filterRenderer);
        }

        filterCell.refreshCell();
    }

    /**
     * Given a filter control, initializes it from the provided column
     * @param filterRenderer	The filter control
     * @param filterColumn		The column to initialize the control with
     * @param item				The item (only required for inner level filters)
     * @param flatDp			The dataprovider
     * @param level				The level against which to initialize the filter.
     */
    initializeFilterRenderer(filterRenderer, filterColumn, item, flatDp, level) {

        const iSelectFilterControl = (filterRenderer);
        if (iSelectFilterControl != null) {
            if (filterColumn.filterComboBoxBuildFromGrid)
                this.buildFilter(iSelectFilterControl, filterColumn, level.getNestDepth() == 1 ? null : item, flatDp);
            else if (filterColumn.filterComboBoxDataProvider) {
                iSelectFilterControl.setDataProvider(filterColumn.filterComboBoxDataProvider);
                if (filter) {
                    for (const fexp of filter.filterExpressions) {
                        if (fexp.columnName == filterColumn.getSearchField()) {
                            iSelectFilterControl.setValue(fexp.expression);
                        }
                    }
                }
            }
        }
        if (filterColumn.hasOwnProperty("enableFilterAutoComplete")
            && filterColumn.enableFilterAutoComplete) {
            this.buildFilter(iSelectFilterControl, filterColumn, level.getNestDepth() == 1 ? null : item, flatDp);
            if (filterColumn.filterComboBoxDataProvider)
                iSelectFilterControl.setDataProvider(filterColumn.filterComboBoxDataProvider);
        }

        var filter = this.grid.itemFilters.getValue(level.getNestDepth() == 1 ? Constants.TOP_LEVEL_FILTER : item);
        if (filter && filterColumn.getSearchField()) {
            const filterValue = filter.getFilterValue(filterColumn.getSearchField());
            const filterControl = filterRenderer.implementsOrExtends("IFilterControl") ? filterRenderer : null;
            if (filterControl && UIUtils.hasMethod(filterControl, "isMatch")) {
                const filterExpression = filter.getFilterExpression(filterColumn.getSearchField(), true);
                if(filterExpression && filterExpression.filterControl) {
                    filterExpression.filterControl = filterControl;
                }
            }
            if (filterControl) {
                filterControl.clear();
                if (filterValue)
                    filterControl.setValue(filterValue)
            }

        }

    }

    /**
     * @method buildFilter
     * 
     * @param {Object} iSelectFilterControl
     * @param {flexiciousNmsp.FlexDataGridColumn} column
     * @param {Object} parentObject
     * @param {Object} flatValues
     */
    buildFilter(iSelectFilterControl, column, parentObject, flatValues) {
        if(typeof  parentObject=="undefined") parentObject=null;
        if(typeof flatValues=="undefined")flatValues=null;
        var valueSet = false;
        
        var collection= column.getDistinctValues(flatValues);
        iSelectFilterControl.setDataProvider(collection);
        var filter=this.grid.itemFilters.getValue(parentObject==null?Constants.TOP_LEVEL_FILTER:parentObject);
        if(filter)
        {
            for(var i=0;i<filter.filterExpressions.length;i++)
            {

                var filterExpression=filter.filterExpressions[i];
                if(filterExpression.columnName == column.getSearchField()){
                    valueSet = true;
                    iSelectFilterControl.setValue(filterExpression.expression);
                }
            }

            if(!valueSet) {
                if(this.grid.filterComboBoxDataProviderEnableReduce && filter.filterExpressions.length > 0) {
                    flatValues = this.grid.getFilteredPagedSortedData({}, true);
                    collection= column.getDistinctValues(flatValues);
                    iSelectFilterControl.setDataProvider(collection);
                }
            }
        }

    }

    /**
     * @method processFooterCell
     * 
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level
     * @param {Array} cols
     * @param {number} colIndex
     * @param {flexiciousNmsp.RowInfo} row
     * @param {object} item
     * @param {Array} existingComponents
     */
    processFooterCell(level, cols, colIndex, row, item, existingComponents) {

        const paddingY = level.getStyleValue("horizontalGridLines") ? 1 : 0;
        const col = cols[colIndex];
        const rowHeight = level.getFooterRowHeight();
        const rendererFactory = col.deriveRenderer(flexiciousNmsp.RowPositionInfo.ROW_TYPE_FOOTER);
        const existingComponent = this.getExistingCell(existingComponents, rendererFactory, col);
        const footerCell = existingComponent ? existingComponent.component : this.grid.rendererCache.popInstance(col.footerCellRenderer, rendererFactory);
        footerCell.setRendererFactory(rendererFactory);
        footerCell.rowInfo = row;
        footerCell.level = level;
        footerCell.setColumn(col);
        footerCell.setActualSize(col.getWidth() - this.getBorderWidth(footerCell), rowHeight - this.getBorderHeight(footerCell));
        footerCell.setWordWrap(col.footerWordWrap);
        if (existingComponent)
            this.addCell(footerCell, row, existingComponent);
        else
            this.addEventListeners(this.addCell(footerCell, row, existingComponent).componentInfo);
        //footerCell.setActualSize(col.getWidth(),rowHeight);
        footerCell.getRenderer().setStyleAttribute("textAlign", (col.footerAlign || col.getStyleValue("textAlign")));
        footerCell.attachClass(col.getStyleValue("footerStyleName"));
        //if(item!=null){
        footerCell.refreshCell();
        //}
        if (level.getNestDepth() == 1) {
            footerCell.setVisible(level.getFooterRowHeight() > 0);
        }
        footerCell.invalidateDisplayList();

    }

    /**
     * @method processColumnGroupCell
     * 
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level
     * @param rendererFactory
     * @param {flexiciousNmsp.RowInfo} row
     * @param {object} item
     * @param {number} rowHeight
     * @param {flexiciousNmsp.FlexDataGridColumnGroup} columnGroup
     * @param {Array} existingComponents
     */
    processColumnGroupCell(
        level,
        rendererFactory,
        row,
        item,
        rowHeight,
        columnGroup,
        existingComponents) {

        const paddingY = level.getStyleValue("horizontalGridLines") ? 1 : 0;
        let cgRenderer = columnGroup.columnGroupRenderer;
        if (!cgRenderer) cgRenderer = UIUtils.UIComponentFactory;
        const existingComponent = this.getExistingCell(existingComponents, rendererFactory, columnGroup.getStartColumn());
        const columnGroupCell = existingComponent ? existingComponent.component : this.grid.rendererCache.popInstance
            (rendererFactory, cgRenderer);
        columnGroupCell.setRendererFactory(cgRenderer);

        columnGroupCell.rowInfo = row;
        columnGroupCell.level = level;
        columnGroupCell.columnGroup = columnGroup;
        if (existingComponent)
            this.addCell(columnGroupCell, row, existingComponent);
        else
            this.addEventListeners(this.addCell(columnGroupCell, row, existingComponent).componentInfo);
        if (columnGroupCell.getHeight() == 0)
            columnGroupCell.setActualSize(columnGroup.getWX()[0] - this.getBorderWidth(columnGroupCell), rowHeight - this.getBorderHeight(columnGroupCell));

        columnGroupCell.attachUserClass(level.getStyleValue("columnGroupStyleName"));
        return columnGroupCell;

    }

    /**
     * @method processHeaderCell
     * 
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level
     * @param {Array} cols
     * @param {number} colIndex
     * @param {flexiciousNmsp.RowInfo} row
     * @param {object} item
     * @param {number} rowHeight
     * @param {Array} existingComponents
     */
    processHeaderCell(level, cols, colIndex, row, item, rowHeight, existingComponents) {

        const paddingY = level.getStyleValue("horizontalGridLines") ? 1 : 0;
        const col = cols[colIndex];
        const rendererFactory = col.deriveRenderer(flexiciousNmsp.RowPositionInfo.ROW_TYPE_HEADER);
        const existingComponent = this.getExistingCell(existingComponents, rendererFactory, col);
        //if you use enableHorizontalScrollOptimizationsBETA you HAVE to use FlexDataGridHeaderCellHS
        const headerCellRenderer  = this.grid.enableHorizontalScrollOptimizationsBETA && col.headerCellRenderer == flexiciousNmsp.FlexDataGridColumn.static_FlexDataGridHeaderCell ?
         flexiciousNmsp.FlexDataGridColumn.static_FlexDataGridHeaderCellHS : col.headerCellRenderer;
        const headerLabel = existingComponent ? existingComponent.component : this.grid.rendererCache.popInstance(headerCellRenderer, rendererFactory);
        headerLabel.setRendererFactory(rendererFactory);
        headerLabel.rowInfo = row;
        headerLabel.level = level;
        headerLabel.setColumn(col);
        this.grid.currentCell = headerLabel;
        headerLabel.setText(col.getHeaderText(headerLabel));
        if (headerLabel.getRenderer().typeName) {
            headerLabel.getRenderer().setStyleAttribute("textAlign", (col.headerAlign || col.getStyleValue("textAlign")));
            headerLabel.setWordWrap(col.headerWordWrap);
        }
        if (level.getMaxColumnGroupDepth() > 1) {
            if (col.columnGroup == null) {
                //we need to be as tall as the entire header.
                headerLabel.setActualSize(col.getWidth() - this.getBorderWidth(headerLabel), this.grid.getHeaderContainer().getHeight() - this.getBorderHeight(headerLabel));
            } else {

                headerLabel.setActualSize(col.getWidth() - this.getBorderWidth(headerLabel), this.grid.getHeaderContainer().getHeight() -
                    (col.columnGroup.getHeight() + col.columnGroup.getY()) - this.getBorderHeight(headerLabel));
            }
        }
        else
            headerLabel.setActualSize(col.getWidth() - this.getBorderWidth(headerLabel), rowHeight - this.getBorderHeight(headerLabel));
        if (existingComponent)
            this.addCell(headerLabel, row, existingComponent);
        else
            this.addEventListeners(this.addCell(headerLabel, row, existingComponent).componentInfo);

        headerLabel.attachUserClass(col.getStyleValue("headerStyleName"));

        if (level.currentSorts.length == 0 && level.initialSortField && (level.initialSortField == col.getDataField())) {
            headerLabel.sortAscending = level.initialSortAscending;
            this.sortByCell(headerLabel);
        }
        //if(item!=null){
        headerLabel.refreshCell();
        //}

    }

    /**
     * @method getExistingCell
     * 
     * @param {Array} existingComponents
     * @param rendererFactory
     * @param col
     */
    getExistingCell(existingComponents, rendererFactory, col) {
        if(this.grid.enableHorizontalScrollOptimizationsBETA){
            if(existingComponents){
                for(var i=0;i<existingComponents.length;i++){
                    var comp=existingComponents[i];
                    if( comp.component._rendererFactory == rendererFactory){
                        existingComponents.splice(existingComponents.indexOf(comp),1);
                        comp.component.setColumn(col);
                        comp.setX(col.getX());
                        comp.component.setX(col.getX());
                        return comp;
                    }
                }
            }    
        }
        return null;
    }

    /**
     * @private
     * @method onFilterChange
     * @param event
     */
    onFilterChange(event) {

        let rowInfo;
        if (event.triggerEvent) {
            const trigger = event.triggerEvent.currentTarget;
            const treeCell = trigger.parent;
            rowInfo = treeCell.rowInfo;
        }
        const filter = event.filter;
        filter.pageIndex = 0;
        const level = rowInfo ? rowInfo.rowPositionInfo.getLevel(this.grid) : this.grid.getColumnLevel();
        filter.pageSize = level.pageSize;
        this.grid.itemFilters.addItem(level.getNestDepth() == 1 ? Constants.TOP_LEVEL_FILTER : rowInfo.getData(), flexiciousNmsp.AdvancedFilter.from(filter));
        const nested =
            new flexiciousNmsp.ExtendedFilterPageSortChangeEvent(flexiciousNmsp.ExtendedFilterPageSortChangeEvent.FILTER_PAGE_SORT_CHANGE,
                this.createFilter(level, level.getNestDepth() == 1 ? null : rowInfo.getData()));
        nested.filter.pageIndex = filter.pageIndex;
        nested.filter.sorts = level.currentSorts;
        nested.triggerEvent = event;
        nested.cause = flexiciousNmsp.ExtendedFilterPageSortChangeEvent.FILTER_CHANGE;
        if (level.getNestDepth() != 1)//the grid takes care of this scenario
            level.dispatchEvent(nested);


    }

    /**
     * @private
     * @method onPageChange
     * @param event
     */
    onPageChange(event) {

        this.dispatchPageChange(event);

    }

    /**
     * @method rootPageChange
     * @param event
     */
    rootPageChange(event) {

        this.dispatchPageChange(event);

    }

    /**
     * @method dispatchPageChange
     * @param event
     */
    dispatchPageChange(event) {

        const iPager = event.target;

        if (iPager && !iPager.level.getIsClientFilterPageSortMode() && iPager.level.getNestDepth() > 1) {
            const loadInfo = this.findLoadingInfo(iPager.rowInfo.getData(), iPager.level.getParentLevel(), iPager.level.childrenField.length > 0);
            if (loadInfo) {
                loadInfo.totalRecords = iPager.getTotalRecords();
                loadInfo.pageIndex = iPager.getPageIndex();
            }
        }
        const filter = this.grid.itemFilters.getValue(iPager.level.getNestDepth() == 1 ? Constants.TOP_LEVEL_FILTER : iPager.rowInfo.getData());
        if (filter)
            filter.pageIndex = iPager.getPageIndex();
        const nested =
            new flexiciousNmsp.ExtendedFilterPageSortChangeEvent(flexiciousNmsp.ExtendedFilterPageSortChangeEvent.FILTER_PAGE_SORT_CHANGE,
                this.createFilter(iPager.level, iPager.level.getNestDepth() == 1 ? null : iPager.rowInfo.getData()));
        nested.filter.pageIndex = iPager.getPageIndex();
        nested.triggerEvent = event;
        nested.cause = flexiciousNmsp.ExtendedFilterPageSortChangeEvent.PAGE_CHANGE;
        iPager.level.dispatchEvent(nested);


    }

    /**
     * @method createFilter
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level
     * @param {Object} parentObject
     * 
     * @return {AdvancedFilter} 
     */
    createFilter(level, parentObject) {

        return level.createFilter(parentObject, this.grid.itemFilters.getValue(parentObject == null ? Constants.TOP_LEVEL_FILTER : parentObject));

    }

    /**
     * @method killResize
     */
    killResize() {
        //        if(this.resizeCursorID!=-1){
        //            var cell=this.grid.currentCell;
        //            if(cell){
        //                cell.setStyleAttribute("cursor="";
        //                this.resizeCursorID=-1;
        //            }
        //        }
    }

    /**
     * @method onHeaderCellMouseMove
     * @param event
     */
    onHeaderCellMouseMove(event) {


        const cell = event.currentTarget;
        if (this.columnDraggingDragCell && flexiciousNmsp.DisplayList.isMouseDown) {
            if (cell.getStyleAttribute("cursor") === "w-resize") {
                cell.setStyleAttribute("cursor", "");
                if ("typeName" in cell.getRenderer())
                    cell.getRenderer().setStyleAttribute("cursor", "");
                this.resizeCursorID = -1;
            }
            this.calculateColumnDraggingDropTargetCell(event);
            return;
        }


        if (this.columnResizingCell || this.columnResizingGlyph || !((event.target.implementsOrExtends("FlexDataGridHeaderCell")))) {
            this.killResize();
            return; //we're resizing or we are over something else
        }
        let ptx = new flexiciousNmsp.Point(event.localX, event.localY);
        if (!cell) return;
        if (event.target != cell && (event.target.implementsOrExtends("UIComponent"))) {
            ptx = cell.globalToLocal((event.target).localToGlobal(ptx))
        }
        const onRightEdge = ((cell.getWidth() - ptx.getX()) <= this.grid.getColumnLevel().headerSeperatorWidth);
        const onLeftEdge = (ptx.getX() <= this.grid.getColumnLevel().headerSeperatorWidth);


        let prevCel;

        if (onLeftEdge || onRightEdge) {
            if (this.resizeCursorID == -1) {

                //if we're over a cell edge on the left side, we're resizing the column before it.
                if (onLeftEdge) {
                    prevCel = this.getCellInDirection(cell, Constants.CELL_POSITION_LEFT);
                    if (!prevCel || !prevCel.getColumn() || !prevCel.getColumn().resizable) {
                        return;
                    }
                }
                else if (onRightEdge) {
                    if (!cell.getColumn().resizable || (cell.getColumn().getIsLastUnLocked())
                        || cell.getColumn().getIsLastrightLocked()) {
                        return;
                    }
                }
                if ("typeName" in cell.getRenderer())
                    cell.getRenderer().setStyleAttribute("cursor", "w-resize");
                cell.setStyleAttribute("cursor", "w-resize");
                //cell.getRenderer().setText(event.localX +":"+prevCel.getColumn().resizable);
                this.resizeCursorID = 1;
            }
        }
        else {
            if (this.resizeCursorID != -1) {
                if ("typeName" in cell.getRenderer())
                    cell.getRenderer().setStyleAttribute("cursor", "");
                cell.setStyleAttribute("cursor", "");
                this.resizeCursorID = -1;
            }

        }

    }

    /**
     * @method onHeaderCellMouseOut
     * @param event
     */
    onHeaderCellMouseOut(event) {



    }

    /**
     * @method columnDraggingMouseMoveHandler
     * @param event
     */
    columnDraggingMouseMoveHandler(event) {

        if (!flexiciousNmsp.DisplayList.isMouseDown || !this.columnDraggingDragCell) {
            this.columnDraggingMouseUpHandler(event);
            return;
        }

        if (!this.columnResizingGlyph) {
            const that = this;
            if (UIUtils.isWebKit()) {
                setTimeout(function () {
                    if (flexiciousNmsp.DisplayList.isMouseDown) {
                        if (!that.columnResizingGlyph) {
                            that.startColumnDrag();
                        }
                    }
                }, 300);

            }
            else {
                that.startColumnDrag();
            }
        }
        if (this.columnResizingCellGlyph && this.columnResizingCellGlyph.parent != null) {
            let pt = new flexiciousNmsp.Point(event.pageX, event.pageY);
            pt = this.grid.globalToLocal(pt);
            this.columnResizingCellGlyph.move(pt.x, pt.y + 16);
        }
    }

    /**
     * @method startColumnDrag
     */
    startColumnDrag() {
        let pt = new flexiciousNmsp.Point(0, 0);

        let pt1 = new flexiciousNmsp.Point(0, 0);
        pt1 = this.columnDraggingDragCell.localToGlobal(pt1);
        pt1 = this.grid.globalToLocal(pt1);
        this.columnDraggingDragCell.moving = true;
        this.columnDraggingDropTargetCell = null;

        this.columnResizingGlyph = new flexiciousNmsp.FlexDataGridHeaderSeperator();
        pt = this.grid.globalToLocal(this.columnDraggingDragCell.localToGlobal(pt));
        this.grid.addChild((this.columnResizingGlyph));
        this.columnResizingGlyph.key = new Date().getTime();
        this.columnResizingGlyph.propertyBag.key = new Date().getTime();
        this.columnResizingGlyph.attachClass("columnResizingGlyph");
        this.columnResizingGlyph.attachClass("flexiciousUnSelectableText");




        const activeCellColor = this.grid.getStyle("activeCellColor");
        const movingcellht = this.columnDraggingDragCell.getHeight();
        this.columnDraggingDragCell.currentBackgroundColors = [activeCellColor, activeCellColor];
        this.columnDraggingDragCell.setHeight(this.grid.getHeight() - pt1.getY());
        this.columnDraggingDragCell.invalidateDisplayList();
        this.columnDraggingDragCell.validateNow();
        this.columnDraggingDragCell.setHeight(movingcellht);

        this.columnResizingGlyph.move(pt.getX(), pt1.getY());
        this.columnResizingGlyph.setActualSize(1, this.grid.getHeight() - pt1.getY() - this.grid.isHScrollBarVisible - this.columnDraggingDragCell.getHeight());
        this.startX = this.columnResizingGlyph.getX();
        this.startY = this.columnResizingGlyph.getY();


        this.columnResizingCellGlyph = new flexiciousNmsp.FlexDataGridHeaderCell();
        this.columnResizingCellGlyph.setColumn(this.columnDraggingDragCell.getColumn());
        this.columnResizingCellGlyph.level = this.columnDraggingDragCell.level;
        this.columnResizingCellGlyph.rowInfo = this.columnDraggingDragCell.rowInfo;
        this.columnResizingCellGlyph.setText(this.columnDraggingDragCell.getText());
        this.columnResizingCellGlyph.attachClass("flexDataGridHeaderCell");
        this.columnResizingCellGlyph.attachClass("columnResizingCellGlyph");
        this.columnResizingCellGlyph.attachClass("flexiciousUnSelectableText");
        UIUtils.attachClass(document.body, "flexiciousUnSelectableText");

        this.grid.addChild((this.columnResizingCellGlyph));

        this.columnResizingCellGlyph.refreshCell();
        this.columnResizingCellGlyph.createColumnIcon();
        this.columnResizingCellGlyph.move(pt.getX(), pt1.getY() + 16);
        this.columnResizingCellGlyph.setActualSize(this.columnDraggingDragCell.getWidth(), this.columnDraggingDragCell.getHeight());
        this.columnResizingCellGlyph.iconRight = 8;
        this.columnResizingCellGlyph.iconTop = 4;
    }

    /**
     * @method columnDraggingMouseUpHandler
     * @param event
     */
    columnDraggingMouseUpHandler(event) {

        const sbRoot = flexiciousNmsp.DisplayList.instance().documentComponent;
        sbRoot.removeEventListener(Constants.EVENT_MOUSE_MOVE, this.columnDraggingMouseMoveHandler, true);
        sbRoot.removeEventListener(Constants.EVENT_MOUSE_UP, this.columnDraggingMouseUpHandler, true);
        this.grid.detachClass("flexiciousUnSelectableText");
        UIUtils.detachClass(document.body, "flexiciousUnSelectableText");
        if (!this.columnDraggingDragCell)
            return;
        if (this.columnResizingGlyph && this.columnResizingGlyph.parent != null) {
            this.grid.removeChild((this.columnResizingGlyph));
        }
        if (this.columnResizingCellGlyph && this.columnResizingCellGlyph.parent != null) {
            this.columnResizingCellGlyph.destroy();
            this.grid.removeChild((this.columnResizingCellGlyph));
        }
        if (this.columnDraggingDropTargetCell && (this.columnDraggingDropTargetCell != this.columnDraggingDragCell) && (this.columnDraggingDropTargetCell.implementsOrExtends("FlexDataGridHeaderCell"))) {
            const fromCol = this.columnDraggingDragCell.getColumn();
            const toCol = this.columnDraggingDropTargetCell.getColumn();
            const lvl = this.columnDraggingDropTargetCell.level;
            this.grid.shiftColumns(fromCol, toCol, lvl);
            if (this.columnDraggingToRight) {
                this.grid.shiftColumns(toCol, fromCol, lvl);
            }
        }
        else if (this.columnDraggingDropTargetCell && (this.columnDraggingDropTargetCell != this.columnDraggingDragCell)
            && (this.columnDraggingDropTargetCell.implementsOrExtends("FlexDataGridColumnGroupCell"))
            && (this.columnDraggingDragCell.implementsOrExtends("FlexDataGridColumnGroupCell"))) {
            const movingCGCell = (this.columnDraggingDragCell);
            let col;
            let i;
            let allCols;
            if (this.columnDraggingToRight) {
                allCols = this.columnDraggingDropTargetCell.getColumn().columnGroup.getAllColumns();
                for (i = 0; i < allCols.length; i++) {
                    col = allCols[i];
                    this.grid.shiftColumns(col, movingCGCell.getColumn(), this.columnDraggingDropTargetCell.level, true);
                }

            } else {
                allCols = movingCGCell.columnGroup.getAllColumns();
                for (i = 0; i < allCols.length; i++) {
                    col = allCols[i];
                    this.grid.shiftColumns(col, this.columnDraggingDropTargetCell.getColumn(), this.columnDraggingDropTargetCell.level, true);
                }
            }


            flexiciousNmsp.DisplayList.isMouseDown = 0;
        }
        this.columnDraggingDragCell.moving = false;
        this.columnDraggingDragCell = null;
        this.columnResizingGlyph = null;
        this.columnDraggingDropTargetCell = null;
        this.columnResizingCellGlyph = null;


    }

    /**
     * @method onHeaderCellMouseDown
     * @param event
     */
    onHeaderCellMouseDown(event) {

        const cell = event.currentTarget;
        const headerCell = event.currentTarget;
        const sbRoot = flexiciousNmsp.DisplayList.instance().documentComponent;

        const eventLocalX = cell.globalToLocal(new flexiciousNmsp.Point(event.pageX, event.pageY)).getX();

        const onRightEdge = ((cell.getWidth() - 4 - eventLocalX) <= this.grid.getColumnLevel().headerSeperatorWidth);
        const onLeftEdge = (eventLocalX <= this.grid.getColumnLevel().headerSeperatorWidth);

        if (this.resizeCursorID == -1 && !onRightEdge && !onLeftEdge) {
            if (this.columnDraggingDragCell || this.columnResizingCell) return;
            if (!cell.getColumn().draggable) return;
            sbRoot.addEventListener(this, Constants.EVENT_MOUSE_MOVE, this.columnDraggingMouseMoveHandler, true);
            sbRoot.addEventListener(this, Constants.EVENT_MOUSE_UP, this.columnDraggingMouseUpHandler, true);
            this.columnDraggingDragCell = event.currentTarget;
        }
        else if (headerCell) {
            if (this.columnDraggingDragCell || this.columnResizingCell) return;
            if (!cell.getColumn().resizable) return;
            if (onLeftEdge) {
                var prevCel = this.getCellInDirection(headerCell, Constants.CELL_POSITION_LEFT);
                if (!prevCel || !prevCel.getColumn() || !prevCel.getColumn().resizable) {
                    return;
                }
            }
            headerCell.resizing = true;
            //headerCell.resizingPrevious=onLeftEdge;
            this.columnResizingCell = headerCell;
            this.grid.attachClass("flexiciousUnSelectableText");
            if (!onRightEdge) {
                this.columnResizingCell = this.getCellInDirection(this.columnResizingCell, Constants.CELL_POSITION_LEFT);
            }
            flexiciousNmsp.DisplayList.isMouseDown = 1;
            sbRoot.addEventListener(this, Constants.EVENT_MOUSE_MOVE, this.columnResizingHandler, true, 0, true);
            sbRoot.addEventListener(this, Constants.EVENT_MOUSE_UP, this.columnResizeMouseUpHandler, true, 0, true);

            this.columnResizingGlyph = new flexiciousNmsp.FlexDataGridHeaderSeperator();
            this.columnResizingGlyph.mouseEnabled = false;
            let pt = new flexiciousNmsp.Point(onLeftEdge ? 0 : cell.getWidth(), 0);
            pt = this.grid.globalToLocal(cell.localToGlobal(pt));
            this.grid.addChild((this.columnResizingGlyph));
            this.columnResizingGlyph.key = new Date().getTime();
            this.columnResizingGlyph.attachClass("columnResizingGlyph");
            this.columnResizingGlyph.attachClass("flexiciousUnSelectableText");
            this.columnResizingGlyph.move(pt.getX(), 0);
            this.columnResizingGlyph.setActualSize(1, this.grid.getHeight() - this.grid.isHScrollBarVisible);
            this.startX = this.columnResizingGlyph.getX();



        }

    }

    /**
     *  @method columnResizingHandler
     *  @param event
     */
    columnResizingHandler(event) {

        if (!flexiciousNmsp.DisplayList.isMouseDown || !this.columnResizingCell) {
            this.columnResizeMouseUpHandler(event);
            return;
        }
        let pt = new flexiciousNmsp.Point(event.pageX, event.pageY);
        pt = this.grid.globalToLocal(pt);
        const minWidth = this.columnResizingCell.getColumn() ? this.columnResizingCell.getColumn().minWidth + 1 : 20;
        const min = this.startX - this.columnResizingCell.getWidth() + minWidth;
        const max = this.grid.getWidth() - minWidth;
        this.columnResizingGlyph.move(Math.min(Math.max(min, pt.getX()), max), 0);

    }

    /**
     * The IFlexDataGridCell under edit currently. Please note, this is the IFlexDataGridCell component, not the editor component.
     * This will be populated only after an edit session has started. To access the actual editor, please use the
     * getCurrentEditor property
     * 
     * @method getCurrentEditCell
     * @return
     */
    getCurrentEditCell() {

        return this.getEditCell();

    }

    /**
     * The editor being used as the current component to edit. This is an instance of the itemEditor class factory.
     * This will be populated only after an edit session has started.
     * 
     * @method getCurrentEditor
     * @return
     */
    getCurrentEditor() {

        return this.getEditor();

    }

    /**
     * @method columnResizeMouseUpHandler
     * @param event
     */
    columnResizeMouseUpHandler(event) {
        this.grid.detachClass("flexiciousUnSelectableText");
        flexiciousNmsp.DisplayList.isMouseDown = 0;
        if (!this.columnResizingCell) return;
        const change = this.columnResizingGlyph.getX() - this.startX;
        let colResizing = this.columnResizingCell.getColumn();
        if (this.columnResizingCell.resizingPrevious) {
            const prevCel = this.getCellInDirection(this.columnResizingCell, Constants.CELL_POSITION_LEFT);
            if (!prevCel || !prevCel.getColumn() || !prevCel.getColumn().resizable) {
                return;
            }
            else {
                colResizing = prevCel.getColumn();
            }
        }
        const hPos = this.grid.getHorizontalScrollPosition();
        const vPos = this.grid.getVerticalScrollPosition();

        if (this.grid.enableColumnWidthUserOverride) {
            colResizing.setColumnWidthMode(flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_FIXED); //so we do not recalculate widths in the future
        }

        if ((!colResizing.getIsLocked()) && (this.grid.getForceColumnsToFitVisibleArea() || (colResizing.level.getNestDepth() != 1))) {
            const filter = colResizing.getIsLeftLocked() ? [flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_LEFT] :
                colResizing.getIsRightLocked() ? [flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_RIGHT] : [flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE];

            const cols = colResizing.level.getVisibleColumns(filter);
            const colsAffected = [];
            let elasticCount = 0;
            let colHit = false;
            let i;
            let col;
            for (i = 0; i < cols.length; i++) {
                col = cols[i];
                if (colHit) {
                    if (col.getIsElastic()) {
                        elasticCount++;
                        colsAffected.push(col);
                    }
                }
                if (!colHit) {
                    colHit = (col == colResizing);
                }
            }
            const share = change / elasticCount;
            let removed = 0;
            for (i = 0; i < colsAffected.length; i++) {
                col = colsAffected[i];
                if ((col != colResizing) && col.getIsElastic()) {
                    let toRemove = share;
                    if (col.getWidth() - share < col.minWidth) {
                        toRemove = col.getWidth() - col.minWidth;
                    }

                    col.setWidth(col.getWidth() - toRemove);
                    removed += toRemove;
                    if (this.grid.enableColumnWidthUserOverride) {
                        col.setColumnWidthMode(flexiciousNmsp.FlexDataGridColumn.COLUMN_WIDTH_MODE_FIXED); //so we do not recalculate widths in the future
                    }
                }
            }
            colResizing.setWidth(colResizing.getWidth() + removed);
        }
        else {
            colResizing.setWidth(colResizing.getWidth() + change);
            this.grid.getColumnLevel().adjustColumnWidths();


        }

        if (this.grid.variableRowHeight) {
            this.grid.rebuildBody();
            if (hPos > 0 || vPos > 0) {
                this.grid.validateNow();
                if (vPos) {
                    this.grid.gotoVerticalPosition(vPos);
                }
                if (hPos) {
                    this.grid.gotoHorizontalPosition(hPos);
                }
            }
        }
        if (colResizing.getIsLocked()) {
            this.grid.reDraw();
            if (hPos > 0 || vPos > 0) {
                this.grid.validateNow();
                if (vPos) {
                    this.grid.gotoVerticalPosition(vPos);
                }
                if (hPos) {
                    this.grid.gotoHorizontalPosition(hPos);
                }
            }
        }


        const sbRoot = flexiciousNmsp.DisplayList.instance().documentComponent;
        sbRoot.removeEventListener(Constants.EVENT_MOUSE_MOVE, this.columnResizingHandler, true);
        sbRoot.removeEventListener(Constants.EVENT_MOUSE_UP, this.columnResizeMouseUpHandler, true);
        this.grid.removeChild((this.columnResizingGlyph));
        this.columnResizingCell.resizing = false;
        this.columnResizingCell = null;
        this.columnResizingGlyph = null;

        this.resizeCursorID = -1;

        const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.COLUMNS_RESIZED, colResizing.level.grid, colResizing.level, colResizing, null, event);
        this.dispatchEvent(evt);
        colResizing.level.dispatchEvent(evt);
        this.grid.invalidateCells();
        this.grid.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.COLUMN_STRETCH, colResizing.level.grid, colResizing.level, colResizing, null, event));

    }

    /**
     * @method snapToColumnWidths
     */
    snapToColumnWidths() {
        this.cellsWithColSpanOrRowSpan = [];
        const rowsForSnapping = this.getRowsForSnapping();
        let i;
        let row;
        let j;
        for (i = 0; i < rowsForSnapping.length; i++) {
            row = rowsForSnapping[i];
            if (row.getIsRendererRow() || row.getIsPagerRow()) {
                for (j = 0; j < row.cells.length; j++) {
                    const cell = row.cells[j];
                    if (cell.component.implementsOrExtends("FlexDataGridPagerCell")
                        || cell.component.implementsOrExtends("FlexDataGridLevelRendererCell")) {
                        cell.component.setWidth(row.rowPositionInfo.getLevel(this.grid).getSumOfColumnWidths([flexiciousNmsp.FlexDataGridColumn.LOCK_MODE_NONE]));
                    }
                }
            }
            else if (row.getIsColumnGroupRow()) {
                for (j = 0; j < row.cells.length; j++) {
                    const cgCellComp = row.cells[j];
                    const cgCell = cgCellComp.component;
                    if (cgCell) {
                        cgCell.onColumnsResized(null);
                    }
                }
            }
            else
                this.snapRowToColumnWidth(row);
        }
        this.hideSpannedCells();
        this.cellsWithColSpanOrRowSpan = [];
    }

    /**
     * Given a cell, gets the colspan associated with that cell by calling the this.grid.colSpanFunction.
     * 
     * @method getColSpan
     * @param {flexiciousNmsp.FlexDataGridDataCell} cell
     * 
     * @return {number}
     */
    getColSpan(cell) {

        if (this.grid.getColSpanFunction() == null) return 1;
        let colSp = 1;
        const fdgCell = cell.implementsOrExtends("IFlexDataGridDataCell") ? cell : null;
        if (this.grid.getColSpanFunction() != null) {
            colSp = this.grid.getColSpanFunction()(cell);
        }
        if (fdgCell)
            fdgCell.colSpan = colSp;
        return colSp;

    }

    /**
     * Given a cell, gets the rowSpan associated with that cell by calling the this.grid.rowSpanFunction.
     * 
     * @method getRowSpan
     * @param {flexiciousNmsp.FlexDataGridDataCell} cell
     * 
     * @return {number}
     */
    getRowSpan(cell) {

        if (this.grid.getRowSpanFunction() == null) return 1;
        let rowSp = 1;
        const fdgCell = cell.implementsOrExtends("IFlexDataGridDataCell") ? cell : null;
        if (this.grid.getRowSpanFunction() != null) {
            rowSp = this.grid.getRowSpanFunction()(cell);
            if (rowSp > 1)
                cell.rowInfo.rowPositionInfo.rowSpan = Math.max(rowSp, cell.rowInfo.rowPositionInfo.rowSpan);
        }
        if (fdgCell)
            fdgCell.rowSpan = rowSp;
        return rowSp;

    }

    /**
     * @method getCellWidth
     * @param {flexiciousNmsp.FlexDataGridDataCell} cell
     * 
     * @return {number}
     */
    getCellWidth(cell) {

        if (this.grid.getColSpanFunction() == null
            || (cell.implementsOrExtends("FlexDataGridHeaderCell"))
            || (cell.implementsOrExtends("FlexDataGridColumnGroupCell"))) return cell.getColumn() ? cell.getColumn().getWidth() : cell.getWidth();

        const colSpan = this.getColSpan(cell);
        if (!cell.getVisible()) return 0;
        if (colSpan == 1) {
            return cell.getColumn() ? cell.getColumn().getWidth() : cell.getWidth();
        }
        this.cellsWithColSpanOrRowSpan.push(cell);
        const cols = cell.level.getColumnsByLockMode([cell.getColumn().getColumnLockMode()]);
        if (colSpan == 0) {
            //this means we need to span all columns to the right.
            if (!cell.getIsLocked()) {
                return this.grid.getBodyContainer().getWidth() + this.grid.getBodyContainer().getMaxHorizontalScrollPosition() - cell.getColumn().getX() - this.grid.isVScrollBarVisible;
            } else if (cell.getIsLeftLocked()) {
                return this.grid.getLeftLockedContent().getWidth() - cell.getColumn().getX();
            } else {
                //right locked
                return this.grid.getRightLockedContent().getWidth() - cell.getColumn().getX();
            }
        } else {
            //this means we need to span the specified number of columns
            let w = cell.getColumn() ? cell.getColumn().getWidth() : cell.getWidth();
            const meIdx = cols.indexOf(cell.getColumn());
            for (let idx = meIdx + 1; idx < (Math.min(cols.length, meIdx + colSpan)); idx++) {
                const nextCol = cols[idx];
                w += nextCol.getWidth();
            }
            return w;
        }
        //return cell.getColumn()?cell.getColumn().getWidth():cell.getWidth();
    }

    /**
     * @method getCellHeight
     * @param {flexiciousNmsp.FlexDataGridDataCell} cell
     * 
     * @return {number}
     */
    getCellHeight(cell) {

        if (this.grid.getRowSpanFunction() == null
            || !(cell.rowInfo.getIsDataRow())) return cell.getHeight();
        const row = cell.rowInfo;
        let rowSpan = this.getRowSpan(cell);
        if (!cell.getVisible()) return 0;
        if (rowSpan == 1) {
            return row.getHeight();
        }
        if (rowSpan == 0) {
            rowSpan = this.rows.length;
        }
        this.cellsWithColSpanOrRowSpan.push(cell);
        //this means we need to span all rows to the bottom.
        const meIdx = this.itemVerticalPositions.indexOf(row.rowPositionInfo);
        let ht = row.getHeight();
        for (let idx = meIdx + 1; idx < (Math.min(this.itemVerticalPositions.length, meIdx + rowSpan)); idx++) {
            const nextRow = this.itemVerticalPositions[idx];
            ht += nextRow.getRowHeight();
        }
        return ht;

    }

    /**
     * Once the snapToColumnWidths finishes, the cellsWithColSpanOrRowSpan array contains all cells
     * that have a row or col span. This method will bring such cells to the top so they hide the ones that are covered.
     * 
     * @method hideSpannedCells
     */
    hideSpannedCells() {
        for (const largerCell of this.cellsWithColSpanOrRowSpan) {
            largerCell.parent.setChildIndex(largerCell, largerCell.parent.numChildren() - 1);
        }
    }

    /**
     * @method getRowsForSnapping
     * @return {Object}
     */
    getRowsForSnapping() {

        return this.getAllRows();

    }

    /**
     * @method snapRowToColumnWidth
     * @param {flexiciousNmsp.RowInfo} row
     */
    snapRowToColumnWidth(row) {
        const xSoFar = 0;
        const loc = new flexiciousNmsp.InsertionLocationInfo();
        if (!this.grid.getForceColumnsToFitVisibleArea()) {
            row.cells.sort((a, b) => a.getColIndex() - b.getColIndex());
        }

        for (const cell of row.cells) {
            if (cell.component) {
                const tdgc = cell.component;
                const column = tdgc.getColumn();
                if (loc.contentX == 0 && column && !column.getIsLocked())
                    loc.contentX = column.getX();
                let toSet = this.getCellWidth(tdgc);//column?column.getWidth():cell.component.getWidth();
                const origWidth = column ? column.getWidth() : cell.component.getWidth();
                if (!tdgc.getIsLocked()) {
                    if(this.grid.enableHorizontalScrollOptimizationsBETA && column)
                        cell.setX(column.getX());
                    else
                        cell.setX(loc.contentX);
                    
                    loc.contentX += origWidth;

                } else if (tdgc.getIsLeftLocked()) {
                    if ((tdgc.getColumn() && tdgc.getColumn().getIsLastLeftLocked())) {
                        toSet = this.grid.getLeftLockedContent().getWidth() - loc.leftLockedContentX;
                    }
                    cell.setX(loc.leftLockedContentX);
                    loc.leftLockedContentX += origWidth;
                } else if (tdgc.getIsRightLocked()) {
                    if (tdgc.getColumn() && tdgc.getColumn().getIsLastrightLocked()) {
                        toSet = this.grid.getRightLockedContent().getWidth() - loc.rightLockedContentX;
                    }
                    cell.setX(loc.rightLockedContentX);
                    loc.rightLockedContentX += origWidth;
                }
                //if(tdgc.visible)
                tdgc.setActualSize(toSet - this.getBorderWidth(tdgc), this.getCellHeight(tdgc) - this.getBorderHeight(tdgc));

            }
        }

        this.grid.placeComponents(row);
    }

    /**
     * @method placeSortIcon
     * @param event
     */
    placeSortIcon(event) {

        const headerCell = event.target;
        if (headerCell.level.hasSort(headerCell.getColumn())) {
            headerCell.placeSortIcon();
        }

    }

    /**
     * @method selectAllChangedHandler
     * @param event
     */
    selectAllChangedHandler(event) {

        this.onSelectAllChanged(event);

    }

    /**
     * @method onSelectAllChanged
     * @param event
     */
    onSelectAllChanged(event) {

        const cell = event.target;
        this.dispatchEvent(new flexiciousNmsp.FlexDataGridEvent(event.type, cell.level.grid, cell.level, cell.getColumn(), cell, null, event));

    }

    /**
     * @method sortByColumn
     * @param col
     */
    sortByColumn(col) {

        const sort = new flexiciousNmsp.FilterSort(col.getSortFieldName());
        const existingsort = col.level.getSortIndex(col, false, true);
        if (!existingsort) {
            sort.isAscending = true;
        } else {
            sort.isAscending = !existingsort.isAscending;
        }
        col.level.addSort(sort);
        if (col.level.getNestDepth() == 1) {
            this.grid.rebuildHeader();
        }

    }

    /**
     * 
     * @method onHeaderCellClicked
     * 
     * @param {flexiciousNmsp.FlexDataGridDataCell} cell
     * @param triggerEvent
     * @param {boolean} isMsc
     * 
     */
    onHeaderCellClicked(cell, triggerEvent, isMsc) {
        if (typeof isMsc == "undefined") isMsc = false;

        if (!this.grid.allowInteractivity) return;

        const lvl = cell.level;
        const cellData = cell.rowInfo.getData();

        const evt = new flexiciousNmsp.FlexDataGridEvent(flexiciousNmsp.FlexDataGridEvent.HEADER_CLICKED, lvl.grid, lvl, cell.getColumn(), cell, null, triggerEvent, false, true);
        this.dispatchEvent(evt);
        lvl.dispatchEvent(evt);

        //        if(evt.isDefaultPrevented()){
        //            return;
        //        }

        if (cell.getColumn().sortable) {
            const mouseEvent = UIUtils.isMouseEvent(triggerEvent);
            if (isMsc||(this.grid.getEnableSplitHeader() && mouseEvent && ((mouseEvent.ctrlKey || mouseEvent.shiftKey) && this.grid.enableControlClickMultiSort) || (!this.grid.enableControlClickMultiSort && mouseEvent.localX>(mouseEvent.currentTarget.getWidth()-this.grid.headerSortSeparatorRight)))) {
                //MCS
                this.sortByColumn(cell.getColumn());
            } else {
                if (cell.getColumn().level.hasSort(cell.getColumn())) {
                    cell.sortAscending = !cell.sortAscending;
                } else {
                    cell.sortAscending = true;
                }

                for (const oCell of cell.rowInfo.cells) {
                    const othercell = oCell.component.implementsOrExtends("FlexDataGridHeaderCell") ? oCell.component : null;
                    if (othercell) {
                        othercell.destroySortIcon();
                    }
                }

                this.storeSort(cell.rowInfo.getData(), cell.getColumn(), cell.sortAscending);
                this.sortByCell(cell);

                cell.placeComponent(cell.getRenderer(), cell.getWidth(), cell.getHeight(), undefined);
            }
            const nested =
                new flexiciousNmsp.ExtendedFilterPageSortChangeEvent(flexiciousNmsp.ExtendedFilterPageSortChangeEvent.FILTER_PAGE_SORT_CHANGE,
                    this.createFilter(lvl, lvl.getNestDepth() == 1 ? null : cellData));
            nested.cause = flexiciousNmsp.ExtendedFilterPageSortChangeEvent.SORT_CHANGE;
            if (lvl.getNestDepth() == 1)
                nested.filter.parentObject = null;
            lvl.dispatchEvent(nested);

            if (lvl.getNestDepth() == 1) {
                if (lvl.getIsClientFilterPageSortMode())
                    this.grid.rebuildBody();
            }
        }



    }

    /**
     * @method storeSort
     * 
     * @param {object} item
     * @param {flexiciousNmsp.FlexDataGridColumn} column
     * @param {boolean} ascending
     */
    storeSort(item, column, ascending) {

        if (!this._sortInfo[item]) {
            this._sortInfo[item] = new flexiciousNmsp.SortInfo(column, ascending);
        } else {
            this._sortInfo[item].sortCol = column;
            this._sortInfo[item].sortAscending = ascending;
        }

    }

    /**
     * @method sortByCell
     * @param {flexiciousNmsp.FlexDataGridDataCell} cell
     */
    sortByCell(cell) {

        const level = cell.level;
        level.setCurrentSort(cell.getColumn(), cell.sortAscending);
        cell.createSortIcon(this);


    }

    /**
     * @method filterPageSort
     * 
     * @param {Object} flat
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level
     * @param {Object} parentObj
     * @param {boolean} applyFilter
     * @param {boolean} applyPaging
     * @param {boolean} applySort
     * @param {Array} pages
     * @param {boolean} updatePager
     * 
     * @return {Object}
     * 
     */
    filterPageSort(
        flat,
        level,
        parentObj,
        applyFilter,
        applyPaging,
        applySort,
        pages,
        updatePager) {

        const columnOwner = level.getColumnOwnerLevel();
        if ((columnOwner.currentSorts.length == 0) && columnOwner.initialSortField) {
            const items = columnOwner.getVisibleColumns();

            for (const col of items) {
                if (col.getDataField() == columnOwner.initialSortField) {
                    columnOwner.setCurrentSort(col, columnOwner.initialSortAscending);
                }
            }
        }
        if (columnOwner.currentSorts.length > 0 && applySort) {
            if (this.grid.getLength(flat) > 0) {
                flat = UIUtils.sortArray(flat, columnOwner.currentSorts);
            }
        }

        let filter = this.grid.itemFilters.getValue(parentObj == null || level.reusePreviousLevelColumns ||
            (level.nextLevel && level.nextLevel.reusePreviousLevelColumns) ? Constants.TOP_LEVEL_FILTER : parentObj);
        if (this.grid.getHasFilterFunction()) {
            filter = this.grid.createFilter();
        }
        if (filter && applyFilter) {
            flat = UIUtils.filterArray(flat, filter, this.grid, level, this.grid.enableHideIfNoChildren);
        }
        if (updatePager && this.grid.updateTotalRecords) {
            this.grid.setTotalRecords(this.grid.getLength(flat));
        }
        if (level.enablePaging && applyPaging) {
            filter = this.grid.itemFilters.getValue(parentObj == null ? Constants.TOP_LEVEL_FILTER : parentObj);
            const pageIndex = filter ? filter.pageIndex : 0;
            const totalRecords = flat.length;
            if (pages && pages.length == 2)
                flat = UIUtils.pageArrayByPageNumbers(flat, pages, level.pageSize);
            else
                flat = UIUtils.pageArray(flat, pageIndex, level.pageSize);
        }
        return flat;

    }

    /**
     * @method addPadding
     * 
     * @param {number} nestLevel
     * @param {flexiciousNmsp.RowInfo} row
     * @param {number} paddingHeight
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level
     * @param {boolean} forceRightLock
     * @param {boolean} scrollPad
     * @param {number} width
     */
    addPadding(nestLevel, row, paddingHeight, level, forceRightLock, scrollPad, width) {

        if (typeof forceRightLock == "undefined") forceRightLock = false;
        if (typeof scrollPad == "undefined")
            scrollPad = false;
        else
            scrollPad = scrollPad;
        if (typeof width == "undefined") width = -1;
        if (nestLevel > 1) {
            const pad = this.grid.rendererCache.popInstance(level.nestIndentPaddingCellRenderer
                , scrollPad ? level.scrollbarPadRenderer : level.nestIndentPaddingRenderer);
            pad.rowInfo = row;
            pad.level = level;
            pad.forceRightLock = forceRightLock;
            pad.scrollBarPad = scrollPad;
            pad.setActualSize(level.grid.forceZeroPadding && level.nextLevel ?0:width == -1 ? level.getMaxPaddingCellWidth() - this.getBorderWidth(pad) : width - this.getBorderWidth(pad), paddingHeight - this.getBorderHeight(pad));
            pad.invalidateBackground();
            this.addEventListeners(this.addCell(pad, row).componentInfo);
            return pad;
        }
        return null;

    }

    /**
     * @method updateDisplayList
     * 
     * @param {number} unscaledWidth
     * @param {number} unscaledHeight
     */
    updateDisplayList(unscaledWidth, unscaledHeight) {

        if (this._horizontalScrollPending != -1) {
            this.gotoHorizontalPosition(this._horizontalScrollPending);
            this._horizontalScrollPending = -1;
        }
        flexiciousNmsp.UIComponent.prototype.updateDisplayList.apply(this, [unscaledWidth, unscaledHeight]);

    }

    /**
     * Used by accesibility
     * 
     * @method getChildIds
     * @param {Array} arr
     */
    getChildIds(arr) {
        for (const row of this.rows) {
            for (const comp of row.cells) {
                arr.push(arr.length + 1);
            }
        }
    }

    /**
     * Used by accesibility
     * @method getSelectedIds
     * @param bodyStart
     * @return {Array}
     */
    getSelectedIds(bodyStart) {
        const arr = [];
        const arr1 = [];

        for (const row of this.rows) {
            const level = row.rowPositionInfo.getLevel(this.grid);

            const selected = (level.getSelectedKeys().includes(level.getItemKey(row.getData())));

            for (const comp of row.cells) {
                arr.push(arr.length);
                const cell = comp.component;
                if (selected) {
                    arr1.push(arr.length + bodyStart);
                }
            }
        }

        return arr1;
    }

    /**
     * Used by accesibility
     * @method getChildId
     * @param {Array} arr
     * @param {flexiciousNmsp.FlexDataGridDataCell} cell
     * 
     * @return {number}
     */
    getChildId(arr, cell) {
        for (const row of this.rows) {
            for (const comp of row.cells) {
                arr.push(arr.length);
                if (comp.component == cell) {
                    return arr.length + 1;
                }
            }
        }

        return 0;
    }

    /**
     * Used by accesibility
     * @method getChildAtId
     * 
     * @param {Array} arr
     * @param {number} id
     * 
     * @return {Object}
     */
    getChildAtId(arr, id) {
        for (const row of this.rows) {
            for (const comp of row.cells) {
                arr.push(arr.length);
                if (arr.length == id - 2) {
                    return comp.component;
                }
            }
        }

        return null;
    }

    /**
     * @method refreshCheckBoxes
     */
    refreshCheckBoxes() {
        for (const row of this.rows) {
            if (row.getIsDataRow() || row.getIsHeaderRow()) {
                for (const cell of row.cells) {
                    const comp = cell.component;

                    if (comp.implementsOrExtends("FlexDataGridCell")) {
                        if (comp.getColumn() && comp.getColumn().implementsOrExtends("FlexDataGridCheckBoxColumn")) {
                            comp.initializeCheckBoxRenderer(comp._renderer);
                        }
                    }
                }
            }
        }
    }

    /**
     * Row positions
     * @method getItemVerticalPositions
     * @return {number}
     */
    getItemVerticalPositions() {
        return this.itemVerticalPositions;
    }

    /**
     * Returns true if grid is in edit mode
     * @method getInEdit
     * @return
     *
     */
    getInEdit() {
        return this._inEdit;
    }

    /**
     * @method getTraceRows
     * @return {String}
     */
    getTraceRows() {
        let str = "";

        for (const row of this.rows) {
            str += `${row.rowPositionInfo.getRowIndex()}:${row.getY()}:${row.rowPositionInfo.getVerticalPosition()}:${row.rowPositionInfo.getRowType()}:${row.cells.length}:`;

            for (const cell of row.cells) {
                const comp = cell.component;
                const col = comp ? comp.getColumn() : null;
                str += `|${comp.text}|${cell.getX()}|${col ? col.getHeaderText() : " "}`;
            }

            str += "\n";
            //str +=row.getY() + ":" + (row.data.name?row.data.name:row.data.id) + ":" + row.rowPositionInfo.getRowType()+ "\n";
        }

        return str;
    }

    /**
     * The editor being used as the current component to edit. This is an instance of the itemEditor class factory.
     * This will be populated only after an edit session has started. 
     * @method getEditor
     * @return 
     */
    getEditor() {
        return this.grid._editor;
    }

    /**
     * @method setEditor
     * @param val
     */
    setEditor(val) {
        this.grid._editor = val;
    }

    /**
     * The IFlexDataGridCell under edit currently. Please note, this is the IFlexDataGridCell component, not the editor component.
	 * This will be populated only after an edit session has started. To access the actual editor, please use the 
	 * getCurrentEditor property
     * 
     * @method getEditCell
     * @return
     */
    getEditCell() {
        return this.grid._editCell;
    }

    /**
     * @method setEditCell
     * @param val
     */
    setEditCell(val) {
        this.grid._editCell = val;
    }

    /**
     * @method getTraceCells
     * @return {String}
     */
    getTraceCells() {
        let str = "";
        const c = this.getChildren();

        for (const obj of c) {
            str += (`${obj.getX()}:${obj.getY()}:${obj.getWidth()}:${obj.getText()}\n`)
        }

        return str;
    }

    /**
     * @method kill
     */
    kill() {
        this.grid = null;

        for (const row of this.rows) {
            row.kill();
        }

        flexiciousNmsp.UIComponent.prototype.kill.apply(this);
    }
}

flexiciousNmsp.FlexDataGridContainerBase = FlexDataGridContainerBase; //add to name space
FlexDataGridContainerBase._levelRendererFactory = new flexiciousNmsp.ClassFactory(flexiciousNmsp.FlexDataGridLevelRendererCell, undefined);
FlexDataGridContainerBase.prototype.typeName = FlexDataGridContainerBase.typeName = 'FlexDataGridContainerBase';//for quick inspection
/**
 * Dispatched when the header cells are clicked
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 */

FlexDataGridContainerBase.EVENT_HEADERCLICKED = 'headerClicked';
/**
 * Dispatched when the columns at any level are resized.
 * Event Type:com.flexicious.nestedtreedatagrid.events.FlexDataGridEvent
 */

FlexDataGridContainerBase.EVENT_COLUMNSRESIZED = 'columnsResized';