/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */
const uiUtil = flexiciousNmsp.UIUtils;
const flxConstants = flexiciousNmsp.Constants;
import TypedObject from '../../../core/TypedObject'
/**
 * A container object that encapsulates all the
 * individual property filter settings, page settings
 * and the sort state.
 * @constructor
 * @class Filter
 * @namespace flexiciousNmsp
 * 
 */
export default class Filter extends TypedObject {
    constructor() {
        super();
        /**
         * Size of the page
         * @type {number}
         * @property  pageSize
         * @default 20
         */
        this.pageSize=20;
        /**
         * The current page index.
         * @type {number}
         * @property pageIndex
         * @default -1
         */
        this.pageIndex=-1;
        /**
         * The number of pages
         * @type {number}
         * @property pageCount
         * @default 1
         */
        this.pageCount=1;
        /**
         * The total number of records
         * @type {number}
         * @property recordCount
         * @default null
         */
        this.recordCount=null;
        /**
         * A collection of com.flexicious.grids.filters.FilterExpression
         * objects.
         * @type {Array}
         * @property filterExpressions
         * @default []
         */
        this.filterExpressions=[];
        /**
         * A collection of com.flexicious.grids.filters.FilterSort
         * objects. Must be just 1 for Basic DataGrid, but can be
         * more than one for advanced datagrid.
         * @type {Array}
         * @property sorts
         * @default []
         */
        this.sorts=[];
        /**
         * For the filter dropdown, show the label for the pre built filters.
         * @type {string}
         * @property filterDescrption
         * @default null
         */
        this.filterDescrption=null;
        /**
         * List of records
         * @type {Object}
         * @property records
         * @default null
         */
        this.records=null; 
    }

    getClassNames() {
        return ["Filter ","TypedObject"];
    }

    /**
     * Takes a filter and populates the current filter's
     * properties with the provided filter
     * @method copyFrom
     * @param {Object} filter to copy
     */
    copyFrom(filter) {

        this.filterDescrption = filter.filterDescrption;
        this.pageSize = filter.pageSize;
        this.pageIndex = filter.pageIndex;
        this.pageCount = filter.pageCount;
        this.recordCount = filter.recordCount;
        const clonedArguments = [];
        let i=0;
        for( i=0;i<filter.filterExpressions.length;i++){
            const arg=filter.filterExpressions[i];
            const filterExpression=new flexiciousNmsp.FilterExpression();
            filterExpression.copyFrom(arg);
            if(filterExpression.expression!=null || filterExpression.filterControlValue)
                clonedArguments.push(filterExpression);
        }
        this.filterExpressions=clonedArguments;

        const sortsCollection = filter.sorts;
        if(sortsCollection)
        {
            const clonedSorts = [];
            for( i=0;i<sortsCollection.length;i++){
                const argSort=sortsCollection[i];
                const filterSort=new flexiciousNmsp.FilterSort();
                filterSort.copyFrom(argSort);
                clonedSorts.push(filterSort);
            }
            this.sorts=clonedSorts;
        }

    }

    /**
     * Adds a com.flexicious.grids.filters.FilterSort
     * object to the list. Must be just 1 for Basic DataGrid, but can be
     * more than one for advanced datagrid.
     * @method addSort
     * 
     * @param {String} sortColumn
     * @param {Boolean} isAscending
     * @param {String} sortComparisionType default value is "auto"
     * @param {Function} sortCompareFunction 
     * 
     */
    addSort(sortColumn, isAscending, sortComparisionType, sortCompareFunction) {
        if(typeof  sortComparisionType=="undefined") sortComparisionType='auto';
        if(typeof sortCompareFunction=="undefined")sortCompareFunction=null;

        for (const filterSort of this.sorts) {
            if(filterSort.sortColumn == sortColumn)
            {
                filterSort.isAscending = isAscending;
                filterSort.sortComparisionType = sortComparisionType;
                filterSort.sortCompareFunction = sortCompareFunction;
                return;
            }
        }

        const newSort = new flexiciousNmsp.FilterSort();
        newSort.sortColumn = sortColumn;
        newSort.isAscending = isAscending;
        newSort.sortComparisionType=sortComparisionType;
        newSort.sortCompareFunction = sortCompareFunction;
        this.sorts.push(newSort);
    }

    /**
     * Adds a FilterExpression to the current list
     * Uses the default equality operator.
     * @method addCriteria
     * @param {String} columnName
     * @param {Object} expression
     *
     */
    addCriteria(columnName, expression) {

        this.addOperatorCriteria(columnName, flexiciousNmsp.FilterExpression.FILTER_OPERATION_TYPE_EQUALS, expression);

    }

    /**
     * Adds a com.flexicious.grids.filters.FilterExpression to the current list. If an Filter Expression
     * was added with the same column name previously, overwrites the operation and the compareValue
     * @method addOperatorCriteria
     * @param {String} columnName 	The column to filter on
     * @param {String} operation  	The filter operation to apply. See com.flexicious.grids.filters.FilterExpression for a list of available operations.
     * @param {Object} compareValue	The expression to filter with. A single value in case of Binary operators like equals, greater than, less than, and an array for rante operators like between, in list, etc.
     * @param {Boolean} wasContains
     */
    addOperatorCriteria(columnName, operation, compareValue, wasContains) {
        if(typeof  wasContains=="undefined") wasContains=false;

        for (const filterExpression of this.filterExpressions) {
            if(filterExpression.columnName == columnName)
            {
                filterExpression.expression = compareValue;
                filterExpression.filterOperation = operation;
                return;
            }
        }

        this.filterExpressions.push(flexiciousNmsp.FilterExpression.createFilterExpression(this, columnName,operation, compareValue,wasContains));
    }

    /**
     * Expression to add
     * @method addFilterExpression
     * @param {flexiciousNmsp.FilterExpression} filterExpression
     *
     */
    addFilterExpression(filterExpression) {

        this.addOperatorCriteria(filterExpression.columnName, filterExpression.filterOperation,filterExpression.expression,filterExpression.wasContains);

    }

    /**
     * Removes a previously added FilterExpression from the current list
     * that matches the specified search field.
     * @method removeCriteria
     * @param {String} searchField
     */
    removeCriteria(searchField) {

        const argument= UIUtils.doesArrayContainValue(this.filterExpressions,"ColumnName",searchField);
        if(argument!=null)
        {
            this.filterExpressions.splice(this.filterExpressions.indexOf(argument),1);
        }

    }

    /**
     * Given a filter column, returns the corresponding value of the filter expression of that column.
     * 
     * @method getFilterValue
     * @param {String} fld
     * @return {*}
     * 
     */
    getFilterValue(fld) {
        for (const exp of this.filterExpressions) {
            if(exp.columnName==fld){
                return exp.filterControlValue?exp.filterControlValue:exp.expression;
            }
        }

        return null;
    }

    /**
     * Given a filter column, returns the corresponding filter expression.
     * @method getFilterExpression
     * @param {String} fld
     * @return {*} 
     */
    getFilterExpression(fld,orig) {
        for (const exp of this.filterExpressions) {
            if(exp.columnName==fld){
                return orig?exp:exp.clone();
            }
        }

        return null;
    }

    /**
     * @method getFilterExpressions
     * @return {Object}
     */
    getFilterExpressions() {
        return this.filterExpressions;
    }

    /**
     * @method setFilterExpressions
     * @param {*} val 
     */
    setFilterExpressions(val) {
        return this.filterExpressions=val;
    }
}

flexiciousNmsp.Filter = Filter; //add to name space
Filter.prototype.typeName = Filter.typeName = "Filter";//for quick inspection
Filter.ALL_ITEM= "All";
