/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */


import Constants from '../../../utils/Constants'
import UIUtils from '../../../utils/UIUtils' 
import TypedObject from '../../../core/TypedObject'


/**
 * A class that represents an individual filter expression
 * A filter expression consists of a columnName (this is the
 * property of the object to search), an operation type (one of the
 * FILTER_OPERATION_TYPE constants) and an expression (the value to search on).
 * @constructor
 * @class FilterExpression
 * @namespace flexiciousNmsp
 * 
 */
export default class FilterExpression extends TypedObject {
    constructor(columnName, filterOperation, expression) {
        super();
        if (typeof columnName == "undefined") columnName = "";
        if (typeof filterOperation == "undefined") filterOperation = "Equals";
        if (typeof expression == "undefined") expression = null;

        this.wasContains = false;
        /**
         * The property of the object to search on
         * @type {String}
         * @property columnName
         * @default null
         */
        this.columnName = null;
        /**
         * The filter operation to apply - one of the
         * FILTER_OPERATION_TYPE constant values
         * @type {String}
         * @property filterOperation
         * @default FilterExpression.FILTER_OPERATION_TYPE_EQUALS
         */
        this.filterOperation = FilterExpression.FILTER_OPERATION_TYPE_EQUALS;
        /**
         * This is ususally automatically set, you dont have to manually set it,
         * unless you're sending strings as Date objects. When set, will attempt
         * to first convert the current value to the type you specified and then
         * do the conversion.
         * @type {String}
         * @property filterComparisionType
         * @default FilterExpression.FILTER_COMPARISION_TYPE_AUTO
         */
        this.filterComparisionType = FilterExpression.FILTER_COMPARISION_TYPE_AUTO;
        /**
         * The expression to compare against
         * @type {String}
         * @property expression
         * @default null
         */
        this.expression = null;
        /**
         * For some filter controls, the expression is different than the selected value. We store the actual
         * value of the filter control, so it can be rebuilt.
         * @property filterControlValue
         * @default null
         */
        this.filterControlValue = null;
        /**
         * @type {flexiciousNmsp.Filter}
         * @property filter
         * @default null
         */
        this.filter = null;
        /**
         * An filterControl that was responsible for the creation
         * of this argument. Used in case of IConverterFilterControl,
         * which in addition to being a filter control, also encapsulates
         * the logic of performing the comparisions and actual matching
         * of the filter criteria.
         * @type {*}
         * @property filterControl
         * @default null
         */
        this.filterControl = null;
        /**
         * Applies to Ultimate only.
         * When set to true, will check to see if either the current item, or any of it's children or grandchildren match the criteria.
         * Used in hierarchical searches, when the parent does not match, but one of the children matches so we have to show the parent as well.
         * @type {boolean}
         * @property recurse
         * @default false
         */
        this.recurse = false;



        /**
         * The property of the object to search on
         * @property
         */
        this.columnName = columnName;
        /**
         * The filter operation to apply - one of the
         * FILTER_OPERATION_TYPE constant values
         * @type {Function}
         * @property filterOperation
         */
        this.filterOperation = filterOperation;
        /**
         * The expression to match
         * @type {Object}
         * @property expression
         */
        this.expression = expression;
    }

    getClassNames() {
        return ["FilterExpression ", "TypedObject"];
    }

    /**
     * @method copyFrom
     * @param {Object} filterExpression 
     * @return {flexiciousNmsp.FilterExpression} 
     */
    copyFrom(filterExpression) {
        this.columnName = filterExpression.columnName;
        this.filterOperation = filterExpression.filterOperation;
        this.filterComparisionType = filterExpression.filterComparisionType;
        this.filterControlValue = filterExpression.filterControlValue;
        this.recurse = filterExpression.recurse;
        this.expression = !filterExpression.expression ? null :
            (filterExpression.expression.hasOwnProperty("item")) ? filterExpression.expression.item :
                (filterExpression.expression.hasOwnProperty("object")) ? filterExpression.expression.object :
                    filterExpression.expression;
        return this;

    }

    /**
     * @method clone
     * @return {flexiciousNmsp.FilterExpression} return clone of this filter expresssion
     */
    clone() {
        const filterExpression = new FilterExpression();
        filterExpression.columnName = this.columnName;
        filterExpression.filterOperation = this.filterOperation;
        filterExpression.filterComparisionType = this.filterComparisionType;
        filterExpression.expression = this.expression;
        filterExpression.filterControlValue = this.filterControlValue;
        filterExpression.recurse = this.recurse;
        return filterExpression;
    }

    /**
     * Creates a filter expression
     * @param filter            The filter create the expression for
     * @param columnName        The column to search
     * @param filterOperation    The filter operation to apply. See the list of FILTER_OPERATION_TYPE constants
     * @param expression        The expression to compare against
     * @param wasContains
     * @method createFilterExpression
     */
    static createFilterExpression(filter, columnName, filterOperation, expression, wasContains) {
        if (typeof wasContains == "undefined") wasContains = false;
        const filterExpression = new FilterExpression();
        filterExpression.filter = filter;
        filterExpression.columnName = columnName;
        filterExpression.expression = expression;
        filterExpression.filterOperation = filterOperation;
        filterExpression.wasContains = wasContains;
        return filterExpression;

    }

    /**
     * convert object to other datatype using filterComparisionType parameter
     * 
     * @method convert
     * @param {String} filterComparisionType
     * @param {Object} object
     * 
     * @return {*}
     */
    static convert(filterComparisionType, object) {

        if (object == null) return object;

        if (filterComparisionType == FilterExpression.FILTER_COMPARISION_TYPE_NUMBER && !(typeof object == "number")) {
            return parseFloat(object.toString());
        }
        else if (filterComparisionType == FilterExpression.FILTER_COMPARISION_TYPE_DATE && !(typeof object == "object")) {
            const str = object.toString();
            return str.length > 0 ? new Date(Date.parse(str)) : null;
        }
        else if (filterComparisionType == FilterExpression.FILTER_COMPARISION_TYPE_BOOLEAN && !(typeof object == "boolean")) {
            return FilterExpression.parseBoolean(object.toString());
        }
        else if (filterComparisionType == FilterExpression.FILTER_COMPARISION_TYPE_STRING && !(typeof object == "string")) {
            return object.toString();
        }
        return object;

    }

    /**
     * Given an object checks to see if it matches the criteria specified by this filter expression
     * @method isMatch
     * @param src
     * @param grid
     * 
     * @return {Boolean}
     * @throws {Error} Invalid expression for Filter expression
     */
    isMatch(src, grid) {
        if (this.filterControl && UIUtils.hasMethod(this.filterControl, "isMatch")) {
            return this.filterControl.isMatch(src, this);
        }

        if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_NONE ||
            this.expression == null || src == null || this.columnName == null) return true;

        let object = null;

        const col = grid && grid.getFilterColumn(this.columnName);
        if ((col && UIUtils.hasMethodOrProperty(col, "useLabelFunctionForFilterCompare")
            && col.useLabelFunctionForFilterCompare)) {
            object = col.itemToLabel(src);
        }
        else if ((col
            && (col.filterCompareFunction != null))) {
            return col.filterCompareFunction(src, this);
        }
        else if ((col && UIUtils.hasMethodOrProperty(col, "filterConverterFunction")
            && (col.filterConverterFunction != null))) {
            object = col.filterConverterFunction(src, col);
        }
        else if (this.columnName.indexOf(".") > 0) {
            //object = UIUtils.resolveExpression(src, this.columnName,null, true);
            const dotobject = UIUtils.resolveExpression(src, this.columnName, null, true);
            if (dotobject == undefined)
                return true;//this means the item does not have our field/
            else if (dotobject == null)
                return false;
            else
                object = dotobject;
        }
        else if (src.hasOwnProperty(this.columnName))
            object = src[this.columnName];
        else
            return true;


        if (object == null) return false;
        if (this.filterControl && UIUtils.hasMethod(this.filterControl, "convert")) {
            object = this.filterControl.convert(object.toString());
        }
        else if (this.filterComparisionType != FilterExpression.FILTER_COMPARISION_TYPE_AUTO) {
            object = FilterExpression.convert(this.filterComparisionType, object);
        }

        //begins with operation - comapre text
        if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_BEGINS_WITH)
            return (object.toString().toLocaleLowerCase().indexOf(this.expression.toString().toLocaleLowerCase().toString()) == 0);
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_CONTAINS)
            return (object.toString().toLocaleLowerCase().includes(this.expression.toString().toLocaleLowerCase().toString()));
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_DOES_NOT_CONTAIN)
            return (!object.toString().toLocaleLowerCase().includes(this.expression.toString().toLocaleLowerCase().toString()));
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_BETWEEN) {
            if (this.expression == null) throw new Error("Expression is not an array collection for between filter operation");
            if (this.expression.length != 2) throw new Error("Invalid expression for between filter operation. Between filter operation requires array collection with exactly 2 items.");
            return (this.expression[0] <= object) && (object <= this.expression[1]);
        }
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_ENDS_WITH)
            return (object.toString().toLowerCase().lastIndexOf(this.expression.toString().toLowerCase()) == object.toString().length - this.expression.toString().length);
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_EQUALS)
            return object == this.expression;
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_GREATER_THAN)
            return object > this.expression;
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_GREATERTHANEQUALS)
            return object >= this.expression;
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_IN_LIST) {
            if (this.expression == null) throw new Error("expression is not an array collection for between filter operation");
            for (i = 0; i < this.expression.length; i++) {
                obj = this.expression[i];
                if (this.wasContains && obj && object) {
                    if (object.toString().includes(obj.toString())) {
                        return true;
                    }
                }
                else if (obj == object)
                    return true;
            }
            return false;
        }
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_IS_NOT_NULL) {
            return object != null
        }
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_LESS_THAN)
            return object < this.expression;
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_LESS_THAN_EQUALS)
            return object <= this.expression;
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_NOT_EQUALS)
            return object != this.expression;
        else if (this.filterOperation == FilterExpression.FILTER_OPERATION_TYPE_NOT_IN_LIST) {
            if (this.expression == null) throw new Error("expression is not an array collection for between filter operation");
            for (var i = 0; i < this.expression.length; i++) {
                var obj = this.expression[i];
                if (obj == object)
                    return false;
            }
            return true;
        }
        else
            throw new Error("Invalid expression for Filter expression");

    }

    /**
     * It parse "1" or "true" or "yes" or "y" as true and
     * "0" or "false" or "no" or "n" as false otherwise handled by default boolean function
     * @param {String} str 
     * @return {Boolean} description
     */
    static parseBoolean(str) {

        switch (str.toLowerCase()) {
            case "1":
            case "true":
            case "yes":
            case "y":
                return true;
            case "0":
            case "false":
            case "no":
            case "n":
                return false;
            default:
                return Boolean(str);
        }

    }
}

flexiciousNmsp.FilterExpression = FilterExpression; //add to name space
FilterExpression.prototype.typeName = FilterExpression.typeName = "FilterExpression";//for quick inspection
FilterExpression.FILTER_OPERATION_TYPE_NONE = "None";
FilterExpression.FILTER_OPERATION_TYPE_EQUALS = "Equals";
FilterExpression.FILTER_OPERATION_TYPE_NOT_EQUALS = "NotEquals";
FilterExpression.FILTER_OPERATION_TYPE_BEGINS_WITH = "BeginsWith";
FilterExpression.FILTER_OPERATION_TYPE_ENDS_WITH = "EndsWith";
FilterExpression.FILTER_OPERATION_TYPE_CONTAINS = "Contains";
FilterExpression.FILTER_OPERATION_TYPE_DOES_NOT_CONTAIN = "DoesNotContain";
FilterExpression.FILTER_OPERATION_TYPE_GREATER_THAN = "GreaterThan";
FilterExpression.FILTER_OPERATION_TYPE_LESS_THAN = "LessThan";
FilterExpression.FILTER_OPERATION_TYPE_GREATERTHANEQUALS = "GreaterThanEquals";
FilterExpression.FILTER_OPERATION_TYPE_LESS_THAN_EQUALS = "LessThanEquals";
FilterExpression.FILTER_OPERATION_TYPE_IN_LIST = "InList";
FilterExpression.FILTER_OPERATION_TYPE_NOT_IN_LIST = "NotInList";
FilterExpression.FILTER_OPERATION_TYPE_BETWEEN = "Between";
FilterExpression.FILTER_OPERATION_TYPE_IS_NOT_NULL = "IsNotNull";
FilterExpression.FILTER_OPERATION_TYPE_IS_NULL = "IsNull";
FilterExpression.FILTER_COMPARISION_TYPE_AUTO = "auto";
FilterExpression.FILTER_COMPARISION_TYPE_STRING = "string";
FilterExpression.FILTER_COMPARISION_TYPE_NUMBER = "number";
FilterExpression.FILTER_COMPARISION_TYPE_DATE = "date";
FilterExpression.FILTER_COMPARISION_TYPE_BOOLEAN = "boolean";