/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */


import Constants from '../../utils/Constants'
import UIUtils from '../../utils/UIUtils'
import TypedObject from '../../core/TypedObject'
/**
 * A set of utility functions that are shared by FlexDataGridDataCell, FlexDataGridDataCell2 and FlexDataGridDataCell3
 * @constructor
 * @class CellUtils
 * @namespace flexiciousNmsp
 * 
 */
export default class CellUtils extends TypedObject {
    getClassNames() {
        return ["CellUtils", "TypedObject"];
    }

    /**
     * Draws the background as well as calls drawBorder on the given cell
     *
     * Each cell in Ultimate will draw its own background
     * and border. This method will use the evaluated results of a number of properties on the cell
     * object to calculate the color, width, and visibility of the background associated with this cell.
     * <ul>
     * <li>First, this method will check to see if you have defined a cellCustomBackgroundDrawFunction for the
     * associated column or level. If so, it will call this method, with the cell as a parameter. You can draw
     * a custom background in this function. If you return false, from this function, this method will not
     * execute any other code, and simply call the drawBorders method</li>
     * <li>First, this method will check to see if you have defined a cellCustomBackgroundDrawFunction for the
     * associated column or level. If so, it will call this method, with the cell as a parameter. You can draw
     * a custom background in this function. If you return false, from this function, this method will not
     * execute any other code, and simply call the drawBorders method</li>
     * </ul>
     */
    static drawBackground(cell) {

        if (!cell.backgroundDirty) return;
        cell.backgroundDirty = false;
        //cell.graphics.clear();
        let result;
        const column = cell.getColumn();
        const level = cell.level;
        if (column && column.cellCustomDrawFunction != null) {
            result = column.cellCustomDrawFunction(cell);
            //if(result==false)
            //return;
        }
        if (level.cellCustomBackgroundDrawFunction != null) {
            result = level.cellCustomBackgroundDrawFunction(cell);
            //if(result==false)
            //return;
        }
        if (result != false) {

            const paddingX = cell.hasVerticalGridLines() ? cell.getVerticalGridLineThickness() : 0;
            const paddingY = cell.hasHorizontalGridLines() ? cell.getHorizontalGridLineThickness() : 0;
            const colors = cell.getBackgroundColors();

            if (colors instanceof Array) {
                cell.setStyleAttribute("backgroundColor", "");
                UIUtils.gradientFill(cell, colors);
            }
            else {
                UIUtils.gradientFill(cell, null);
                cell.setStyleAttribute("backgroundColor", UIUtils.decimalToColor(colors));
            }
            const res = cell.getTextColors();
            if (res == 0) {
                cell.setStyleAttribute("color", "#000000");
            }
            else if (res != null) {
                cell.setStyleAttribute("color", UIUtils.decimalToColor(res));
            }
        }
        CellUtils.drawBorders(cell);

    }

    /**
     * If this is a fillRow, returns the value of the color property.
     * Else, returns the value of the "prop" property which is the value of one of the following style properties:
     * <ul>
     * <li>rollOverColors</li>
     * <li>headerRollOverColors</li>
     * <li>footerRollOverColors</li>
     * <li>filterRollOverColors</li>
     * <li>pagerRollOverColors</li>
     * <li>columnGroupRollOverColors</li>
     * </ul>
     */
    static getRolloverColor(cell, prop) {

        if (cell.rowInfo && cell.rowInfo.getIsFillRow()) return cell.getStyleValue("color");
        if (cell.getColumn()) return cell.getColumn().getStyleValue(prop);
        return cell.getStyleValue(prop);

    }

    /**
     * The getTextColors method is responsible for evaluating the text color of the cell.
     * <ul>
     * <li>First, it checks to see if the cell is disabled. If so, it will return the value of the textDisabledColor style property</li>
     * <li>Next, it checks to see if there is a value for the currentTextColors property (based upon rollover) and if so returns that. </li>
     * <li>Next, it checks to see if there is a custom cellTextColorFunction on the column and if so, it returns the result of this function</li>
     * <li>Next, it checks to see if there is a custom rowTextColorFunction on the level and if so, it returns the result of this function</li>
     * <li>Next, it checks to see if there enableEditRowHighlight is true amd this row is in edit mode, then it returns the value of the editTextColor style property.</li>
     * <li>Next, it checks to see if the row is selected (row selection mode) or cell is selected (cell selection mode) and if so, returns textSelectedColor property.</li>
     * <li>Next, it checks to see if a columnTextColor property is specified on the associated column and if so, returns the value of this property.</li>
     * <li>Finally, it uses the alternatingTextColors array and return its 0th or 1st index based upton the rowInfo.rowPositionInfo.rowIndex.</li>
     * </ul>
     */
    static getTextColors(cell) {

        const rowInfo = cell.rowInfo;
        const level = cell.level;
        const column = cell.getColumn();
        if (rowInfo && rowInfo.getIsFillRow()) return cell.getStyleValue("color");
        if (cell.currentTextColors) return cell.currentTextColors;
        let result;
        if (!cell.getEnabled()) {
            result = cell.getStyleValue("textDisabledColor");
            if (result != null) return result;
        }
        //this will be evetually all removed in favor of grid.cellTextColorFunction
        if (column && column.cellTextColorFunction != null) {
            result = column.cellTextColorFunction(cell);
            if (result != null) return result;
        }
        if (level.rowTextColorFunction != null) {
            result = level.rowTextColorFunction(cell);
            if (result != null) return result;
        }
        //end
        if (level.grid.getBodyContainer().getCurrentEditCell && level.grid.getBodyContainer().getCurrentEditCell() &&
            level.grid.getBodyContainer().getCurrentEditCell().rowInfo == cell.rowInfo) {
            const editColors = cell.getStyleValue("editTextColor");
            if (editColors) {
                return editColors;
            }
        }
        const odd = (rowInfo && rowInfo.rowPositionInfo ? (rowInfo.rowPositionInfo.getRowIndex() % 2 == 0) : false);
        if (cell.getStyleValue("alternatingTextColors")[odd ? 0 : 1] != cell.getStyleValue("textSelectedColor")) {
            if (level.isItemSelected(rowInfo.getData())) {
                return cell.getStyleValue("textSelectedColor");
            }
            else if (column && level.isCellSelected(rowInfo.getData(), column)) {
                return cell.getStyleValue("textSelectedColor");
            }
        }
        if (column && column.getStyle("columnTextColor")) {
            return cell.getStyleValue("columnTextColor");
        }
        return cell.getStyleValue("alternatingTextColors")[odd ? 0 : 1];

    }

    /**
     * If a value is defined for grid.cellBackgroundColorFunction, returns the result of that function
     * else returns null.
     */
    static getBackgroundColorFromGrid(cell) {

        if (cell.level.grid.cellBackgroundColorFunction != null) {
            return (cell.level.grid.cellBackgroundColorFunction(cell))
        }
        return null;

    }

    /**
     * If a value is defined for grid.cellTextColorFunction, returns the result of that function
     * else returns null.
     */
    static getTextColorFromGrid(cell) {

        if (cell.level.grid.cellTextColorFunction != null) {
            return (cell.level.grid.cellTextColorFunction(cell))
        }
        return null;

    }

    /**
     * The getBackgroundColors method is responsible for evaluating the background color of the cell.
     * <ul>
     * <li>First, it checks to see if this is a filler row. If so, it will use the alternatingItemColors array and return its 0th or 1st index based upton the rowInfo.rowPositionInfo.rowIndex. </li>
     * <li>Next, it checks to see if the cell is disabled. If so, it will return the value of the selectionDisabledColor style property</li>
     * <li>Next, it checks to see if there is a custom cellBackgroundColorFunction on the column and if so, it returns the result of this function</li>
     * <li>Next, it checks to see if there is a custom rowBackgroundColorFunction on the level and if so, it returns the result of this function</li>
     * <li>Next, it checks to see if there rowInfo's data object is in the errors collection of the grid, and if so, it return the value of the errorBackgroundColor property.</li>
     * <li>Next, it checks to see if there enableEditRowHighlight is true amd this row is in edit mode, then it returns the value of the editItemColors style property.</li>
     * <li>Next, it checks to see if there is a value for the currentBackgroundColors property (based upon rollover) and if so returns that.
     * The currentBackgroundColors property is a result of the grid.highlightRow method.
     * <ul>
     * <li>If enableActiveCellHighlight=true uses the value of the activeCellColor style.</li>
     * <li>Else, uses the value of the rollOverColor style property.</li>
     * </ul>
     * </li>
     * <li>Next, it checks to see if the row is selected (row selection mode) or cell is selected (cell selection mode) and if so, returns selectionColor property.</li>
     * <li>Next, it checks to see if a backgroundColor property is specified on the associated column and if so, returns the value of this property.</li>
     * <li>Finally, it uses the alternatingItemColors array and return its 0th or 1st index based upton the rowInfo.rowPositionInfo.rowIndex.</li>
     * </ul>
     */
    static getBackgroundColors(cell) {

        const rowInfo = cell.rowInfo;
        const level = cell.level;
        const column = cell.getColumn();
        const odd = (rowInfo && rowInfo.rowPositionInfo ? (rowInfo.rowPositionInfo.getRowIndex() % 2 == 0) : false);
        if (rowInfo && rowInfo.getIsFillRow()) {
            return level.grid.enableFillerRows ? cell.getStyleValue("alternatingItemColors")[odd ? 0 : 1] : level.grid.getStyle("backgroundColor");
        }
        let result;

        if (!cell.getEnabled()) {
            result = cell.getStyleValue("selectionDisabledColor");
            if (result != "null" && result != null)
                return result;
        }

        if (column && column.cellBackgroundColorFunction != null) {
            result = column.cellBackgroundColorFunction(cell);
            if (result != null) return result;
        }
        if (level.rowBackgroundColorFunction != null) {
            result = level.rowBackgroundColorFunction(cell);
            if (result != null) return result;
        }
        if (level.grid.hasErrors && level.grid.getError(rowInfo.getData())) {
            return cell.getStyleValue("errorBackgroundColor");
        }
        if (level.grid.enableEditRowHighlight && level.grid.getBodyContainer().getCurrentEditCell() &&
            level.grid.getBodyContainer().getCurrentEditCell().rowInfo == cell.rowInfo) {
            const editColors = cell.getStyleValue("editItemColors");
            if (editColors) {
                return editColors;
            }
        }
        if (cell.currentBackgroundColors) {
            return cell.currentBackgroundColors;
        }
        if (level.grid.hasErrors && level.grid.getError(rowInfo.getData())) {
            return cell.getStyleValue("errorBackgroundColor");
        }
        else if (level.isItemSelected(rowInfo.getData()) && level.grid.getIsRowSelectionMode()) {
            return cell.getStyleValue("selectionColor");
        }
        else if (column && level.isCellSelected(rowInfo.getData(), column)) {
            return cell.getStyleValue("selectionColor");
        }
        else if (column && column.isValidStyleValue(column.getStyle("backgroundColor"))) {
            return column.getStyle("backgroundColor")
        }

        return cell.getStyleValue("alternatingItemColors")[odd ? 0 : 1];

    }

    /**
     * Draws the borders of the given cell. Each cell in Ultimate will draw its own background
     * and border. This method will use the evaluated results of a number of properties on the cell
     * object to calculate the color, width, and visibility of the border associated with this cell.
     * <ul>
     * <li>First, it checks to see if the grid.hasErrors evaluates to true. If it does, the nit checks to
     * see if the cell.rowInfo.getData() object associated with this cell is within the error collection.
     * If so, it will use the result of the errorBorderColor style property to figure out the color
     * of the border for this cell. </li>
     * <li>Next, it checks to see if there is a cellBorderFunction defined. If there is one, it will just
     * pass the cell to the given cellBorderFunction and if this function returns false, this method
     * performs no additional processing and returns.</li>
     * <li>Next, if the cell's hasHorizontalGridLines evaulates to true, this method then uses the results
     * of horizontalGridLineThickness, horizontalGridLineColor property to draw a bottomn border</li>
     * <li>Next, if the cell's drawTopBorder evaulates to true, this method then uses the results
     * of horizontalGridLineThickness, horizontalGridLineColor property to draw a top border.</li>
     * <li>Finally, it calls the drawRightBorder method on the cell.</li>
     * </ul>
     */
    static drawBorders(cell) {


        const rowInfo = cell.rowInfo;
        const level = cell.level;
        const column = cell.getColumn();
        const paddingY = cell.hasHorizontalGridLines() ? 1 : 0;

        let result;
        const cellComp = cell;
        const errStr = cellComp ? cellComp.getErrorString() : "";
        if (errStr) {
            cellComp.setErrorString("");
        }
        if (level.grid.hasErrors) {
            if (column) {
                const err = level.grid.getError(rowInfo.getData());
                if (err && err[column.getDataField()]) {
                    (cell).setErrorString(err[column.getDataField()], cell.level.grid.getStyle('errorBorderColor'));
                    return;
                }
            }
        }

        if (column && column.cellBorderFunction != null) {
            result = column.cellBorderFunction(cell);
            if (result == false)
                return;
        }
        if (level.cellBorderFunction != null) {
            result = level.cellBorderFunction(cell);
            if (result == false)
                return;
        }
        if (cell.hasHorizontalGridLines()) {
            cell.setStyleAttribute("borderBottom", `solid ${cell.getHorizontalGridLineThickness()}px ${UIUtils.decimalToColor(cell.getHorizontalGridLineColor())}`);
        } else {
            cell.setStyleAttribute("borderBottom", "");
        }
        if (cell.getDrawTopBorder()) {
            cell.setStyleAttribute("borderTop" , `solid ${cell.getHorizontalGridLineThickness()}px ${UIUtils.decimalToColor(cell.getHorizontalGridLineColor())}`);
        } else {
            cell.setStyleAttribute("borderTop" , "");
        }
        cell.drawRightBorder(cell.width, cell.height);
    }

    /**
     * Returns the value of the given style property based on the following logic:
     * <ul>
     * <li> If there is a column associated with this cell, and there is a property in the styleOverrides  collection of the column with styleProp
     * name, returns the value of that property from styleOverrides</li>
     * <li> If there is a column, calls the getStyleValue method on that column (which checkes to see if a there is a value for this style property
     * in the column, return that, else if in the level, return that, else if in the grid, return that). The idea being that
     * you can define a property at the highest possible container (the grid) or one below (at the level) or one below (at the column)</li>
     * </ul>
     */
    static getStyleValue(cell, styleProp) {

        const result = cell.getColumn() ? cell.getColumn().getStyleValue(styleProp) : cell.level.getStyleValue(styleProp);
        return result;

    }

    /**
     * Draws the right border for this cell.
     * There are a few rules on basis of which a right (or a left) border is drawn on a cell.
     * <ul>
     * <li>First, if the cell is a Pager cell, no right border is drawn, since the pager shares its left and right border with the grid.</li>
     * <li>Second, if the cell has no vertical gridlines, no border is drawn.</li>
     * <li>Else, if the cell has a column and its right locked, then a left border is drawn.</li>
     * </ul>
     */
    static drawRightBorder(cell) {
 
        if (cell.implementsOrExtends("FlexDataGridPagerCell")) {
            cell.setStyleAttribute("borderRight", "");
            return;
        }
        else if (cell.implementsOrExtends("FlexDataGridPaddingCell") && cell.level.getNestDepth() != cell.level.grid.getMaxDepth()) {
            cell.setStyleAttribute("borderRight", "");
            return;
        }
        else if (cell.getColumn() && cell.getColumn().getIsLastleftLocked()) {
            cell.setStyleAttribute("borderRight" , "");
            return;
        }
        else if (cell.scrollBarPad || cell.getColumn()) {
            if (cell.scrollBarPad || cell.getColumn().getIsLastUnLocked() || cell.getColumn().getIsLastrightLocked()) {
                if (cell.scrollBarPad)
                    cell.setStyleAttribute("borderLeft", `solid ${cell.getVerticalGridLineThickness()}px ${UIUtils.decimalToColor(cell.getVerticalGridLineColor())}`);
                cell.setStyleAttribute("borderRight" , "");
                return;
            }
        }
        const paddingX = cell.hasVerticalGridLines() ? cell.getVerticalGridLineThickness() : 0;
        const paddingY = cell.hasHorizontalGridLines() ? cell.getHorizontalGridLineThickness() : 0;
        const cgc = (cell.implementsOrExtends("FlexDataGridColumnGroupCell") ? cell : null);
        const sbp = (cell.implementsOrExtends("FlexDataGridPaddingCell") ? cell : null);
        if (cgc && cgc.background) return;
        if (cell.hasVerticalGridLines()) {
            if (sbp) {
                if (sbp.scrollBarPad) {
                    cell.setStyleAttribute("borderLeft" , `solid ${cell.getVerticalGridLineThickness()}px ${UIUtils.decimalToColor(cell.getVerticalGridLineColor())}`);
                }
            }
            else if (cgc) {
                let colBeforeMe;
                const drawLeftBorder = false;

                if (drawLeftBorder) {
                    cell.setStyleAttribute("borderLeft" , `solid ${cell.getVerticalGridLineThickness()}px ${UIUtils.decimalToColor(cell.getVerticalGridLineColor())}`);
                }
                if (cgc.columnGroup.getEndColumn().getIsLastUnLocked() || cgc.columnGroup.getEndColumn().getIsLastrightLocked()) {
                    return;
                }
            }

            cell.setStyleAttribute("borderRight" , `solid ${cell.getVerticalGridLineThickness()}px ${UIUtils.decimalToColor(cell.getVerticalGridLineColor())}`);
        }

    }

    /**
     * @copy com.flexicious.nestedtreedatagrid.interfaces.IFlexDataGridCell#refreshCell
     */
    static refreshCell(cell) {



        cell.backgroundDirty = true;
        const rowInfo = cell.rowInfo;
        if (rowInfo.getIsFillRow()) return;
        const level = cell.level;
        const column = cell.getColumn();

        let enabled = true;
        if (column) {
            cell.setText(column.itemToLabel(rowInfo.getData(), cell));
            if (column.cellDisabledFunction != null)
                enabled = (!column.cellDisabledFunction(cell));
        }
        if (level.rowDisabledFunction != null) {
            enabled = (!level.rowDisabledFunction(cell, rowInfo.getData()));
        }
        const renderer = cell.getRenderer();
        if (renderer.setData) {
            renderer.setData(rowInfo.getData());
        }
        //if(renderer.hasOwnProperty("enabled") && renderer["enabled"]!=enabled)renderer["enabled"]=enabled;
        if (renderer.setEnabled) {
            renderer.setEnabled(enabled)
            if (cell.getEnabled() != enabled)
                cell.setEnabled(enabled);

        }

    }

    /**
     * Given a renderer, sets the size on basis of whether or not there are vertical and horizotnal gridlines .
     * If there are vertical gridlines, sets the width to width minus verticalGridLineThickness
     * If there are vertical gridlines, sets the height to width minus horizontalGridLineThickness
     */
    static setRendererSize(cellRenderer, w, h) {
        const cell = cellRenderer.parent;
        const borderW = cell && cell.implementsOrExtends('IFlexDataGridCell') && cell.hasVerticalGridLines() ? cell.getVerticalGridLineThickness() : 0;
        const borderH = cell && cell.implementsOrExtends('IFlexDataGridCell') && cell.hasHorizontalGridLines() ? cell.getHorizontalGridLineThickness() : 0;
        (cellRenderer).setActualSize(w - borderW, h - borderH);


    }

    /**
     * If there is a value for the prefix parameter, capitalizes the first word of the val parameter and returns it
     */
    static capitalizeFirstLetterIfPrefix(prefix, val) {

        return (prefix && prefix.length) > 0 ? CellUtils.doCap(val) : val;

    }

    /**
     * Capitalizes the first character of the passed in string.
     */
    static doCap(val) {

        return val.substr(0, 1).toUpperCase() + val.substr(1);

    }

    /**
     * This method is responsible for initializing the checkbox renderer for both the header and data cells.
     * It performns the following tasks
     * <ul>
     * <li>First, it will set the radio button mode of the associated TriStateCheckBox to true if col.radioButtonMode or level.enableSingleSelect is set to true </li>
     * <li>Next, if the grid.enableSelectionExclusion flag is set to true, it will set the selected state of the associated TriStateCheckBox to the result of the getCheckBoxStateBasedOnExclusion method of the level</li>
     * <li>Next, if the grid.enableSelectionExclusion flag is set to false (default),
     * it will set the selected state of the associated TriStateCheckBox based upon the following logic (if enableTriStateCheckBox=true):
     * <ul>
     * 	<li>If the associated rowData object is in the selected objects collection of the level, sets it to checked</li>
     * 	<li>If any of the children the associated rowData object are selected, sets it to middle</li>
     * 	<li>Else sets it to unchecked</li>
     * </ul>
     * </li>
     * <li>Finally, if the enableLabelAndCheckBox flag is set to true on the column, this method will apply the result of the columns itemToLabel function
     * to the label property of the checkbox</li>
     * </ul>
     */
    static initializeCheckBoxRenderer(cell, level) {

        const renderer = (cell.getRenderer().implementsOrExtends("TriStateCheckBox") ? cell.getRenderer() : null);
        const col = (cell.getColumn() && cell.getColumn().implementsOrExtends("FlexDataGridCheckBoxColumn") ? cell.getColumn() : null);

        if (!renderer || !col) return;
        renderer.enableDelayChange = false;//we dont want the delay timer
        renderer.radioButtonMode = col.radioButtonMode || level.enableSingleSelect;

        if (level.grid.enableSelectionExclusion) {
            renderer.setSelectedState(level.getCheckBoxStateBasedOnExclusion(cell.rowInfo.getData()));
        }
        else {
            if (level.grid.enableTriStateCheckbox) {
                if (level.isItemSelected(cell.rowInfo.getData())) {
                    renderer.setSelectedState(flexiciousNmsp.TriStateCheckBox.STATE_CHECKED);
                } else if (level.nextLevel && level.nextLevel.areAnySelected(level.getChildren(cell.rowInfo.getData()), true)) {
                    if (level.nextLevel && level.nextLevel.enableSingleSelect)
                        renderer.setSelectedState(flexiciousNmsp.TriStateCheckBox.STATE_CHECKED);
                    else
                        renderer.setSelectedState(flexiciousNmsp.TriStateCheckBox.STATE_MIDDLE);
                } else {
                    renderer.setSelectedState(flexiciousNmsp.TriStateCheckBox.STATE_UNCHECKED);
                }
            }
            else
                renderer.setSelected(level.isItemSelected(cell.rowInfo.getData()));

        }
        if (col.enableLabelAndCheckBox) {
            if (cell.implementsOrExtends("IFlexDataGridDataCell")) {
                renderer.label = col.itemToLabel(cell.rowInfo.getData(), cell);
            }
        }

    }

    /**
     * If the cell.backgroundColors is an array, returns a string join using a comma.
     * Else, returns the result of ExtendedUIUtils.getColorName on it.
     */
    static getBackGroundColorString(cell) {

        const colors = cell.getBackgroundColors();
        if (colors instanceof Array) {
            return (colors).join(",");
        }
        else {
            return UIUtils.getColorName(colors).toString()
        }

    }
}

flexiciousNmsp.CellUtils = CellUtils; //add to name space
CellUtils.prototype.typeName = CellUtils.typeName = 'CellUtils';//for quick inspection