/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */
import React from 'react';
import FlexDataGrid from '../flexgrid/grids/FlexDataGrid'
import ReactDataGridColumnLevel from './ReactDataGridColumnLevel'
import ReactDataGridColumn from './ReactDataGridColumn'

import UIUtils from '../utils/UIUtils'
import Constants from '../utils/Constants'


/**
 * The ReactDataGrid class is the main component intended to be consumed by application developers to embed the 
 * Flexicious React DataGrid component in their React Apps. It ultimately inherits from React.Component base class. The 
 * immediate super class of this class is the {@link FlexDataGrid} class, which contains most of the grid API.   
 * 
 * In order to make the Flexicious React DataGrid (base class of this class) build purely from JSX markup,
 * This class provides a hook into JSX parsing (making its children not visual DOM children, but actually
 * just configuration)
 * 
 * This class will also be the actual DOM node that renders the REACT datagrid - it will contain all the 
 * scrolling containers (pager, header, filter, body and footer) - which in turn will contain instances of
 * {@link ReactDataGridCellContainer} classes which will in turn contain the actual renderers (div's)
 * 
 * The children of this class in markup will be {@link ReactDataGridColumn} or {@link ReactDataGridColumnGroup} or {@link ReactDataGridColumnLevel}
 * classes. They are config classes, and are ignored for rendering purposes - their only use is to pass the
 * configuration of the grid into the base class for this class ({@link FlexDataGrid})
 * 
 * For a complete list of properties and methods, please refer to the following classes:
 * {@link FlexDataGrid} (base class of this class, which eventually inherits from React.Component)
 * {@link FlexDataGridColumnLevel} (Please read the Introduction section of the user guide to understand what a level means)
 * {@link FlexDataGridColumnGroup} (For grouped columns)
 * {@link FlexDataGridColumn} (The actual column object within the grid)    .
 * 
 * For a list of examples, review http://reactdatagrid.com/demo/ - Each example has a "Source Code" section where you 
 * can review sample configurations.
 * 
 * Here is the source code for all running examples:
 * https://github.com/flexicious/react-datagrid
 * 
 * Here is a video that demonstrates how to quickly get up and running with the React DataGrid:
 * https://www.youtube.com/watch?v=sO3o0zG93fM 
 * 
 * For a simple example:
 *  https://github.com/flexicious/react-datagrid-getting-started
 * 
 * Running simple example:
 *  https://www.webpackbin.com/bins/-Kj9JhCTXC3CghFUfRtQ
 * 
 * For a example with the base adapter [CSS Styles, Popups, etc included] included see 
 *  https://github.com/flexicious/react-datagrid-with-styles
 * 
 * React-Redux example code: 
 *  https://github.com/flexicious/redux-react-datagrid
 * 
 * React-Redux example running: 
 *  https://www.webpackbin.com/bins/-Kl_rVyde2vHvB1s4gmf
 *
 * Typical markup looks like:
 *  @example
 * <ReactDataGrid dataProvider={this.state.dataProvider} width={"100%"} height={300} ... [See {FlexDataGrid} class for full list of properties]>
 *      <ReactDataGridColumnLevel [See { FlexDataGridColumnLevel} for full list of properties - note that this is not required, you can directly 
 *                                      list { ReactDataGridColumn} nodes here, a level is only needed for hierarchical grids where each level
 *                                      of hierarchy may potentially have different sets of columns.]> 
 *            <ReactDataGridColumnGroup ... [See { FlexDataGridColumnGroup} class for full list of properties]>
 *                 <ReactDataGridColumn  ... [See { FlexDataGridColumn} class for full list of properties]/>
 *                 <ReactDataGridColumn  ... [See { FlexDataGridColumn} class for full list of properties]/>
 *                 <ReactDataGridColumn  ... [See { FlexDataGridColumn} class for full list of properties]/>
 *                 <ReactDataGridColumn  ... [See { FlexDataGridColumn} class for full list of properties]/>
 *            </ReactDataGridColumnGroup>
 *      </ReactDataGridColumnLevel>
 * </ReactDataGrid>
 * 
 * Complete Example:
 * @example
 *      <ReactDataGrid  width={"100%"} ref={(grid) => { this.grid = grid; }} enablePrint enablePreferencePersistence enableExport enableCopy enableFilters enableFooters initialSortField="title" preferencePersistenceKey="variableRowHeight" initialSortAscending forcePagerRow variableRowHeight
 *          horizontalScrollPolicy="off">
 *            <ReactDataGridColumn dataField="id" headerText="ID" filterControl="TextInput" filterOperation="Contains" columnWidthMode="fitToContent" />
 *            <ReactDataGridColumn dataField="title" headerText="Title" filterControl="TextInput" columnWidthMode="fitToContent" footerLabel="Count:" footerOperation="count" footerAlign="center" filterOperation="Contains" />
 *            <ReactDataGridColumn dataField="description" headerText="Description" wordWrap />
 *            <ReactDataGridColumn dataField="genre" headerText="Genre" filterControl="MultiSelectComboBox" filterComboBoxBuildFromGrid columnWidthMode="fitToContent" />
 *            <ReactDataGridColumn dataField="price" headerText="Price" filterControl="NumericRangeBox" columnWidthMode="fixed" width="100" footerLabel="Avg:" footerOperation="average" footerAlign="center" footerOperationPrecision={2} footerFormatter={flexiciousNmsp.CurrencyFormatter} labelFunction={UIUtils.dataGridFormatCurrencyLabelFunction} />
 *              <ReactDataGridColumn dataField="publish_date" headerText="Publish Date" filterControl="DateComboBox" columnWidthMode="fitToContent" filterConverterFunction={this.convertDate} labelFunction={this.dataGridFormatDateLabelFunction} />
 *      </ReactDataGrid>
 */
export default class ReactDataGrid extends FlexDataGrid {
    constructor(props, arg1, arg2) {
        super(props, arg1, arg2)
        const that = this;
        let directChildren = false;
        let hasGroups = false;
        var childrenWithProps = React.Children.map(this.props.children, (child) => {
            if (child.type.typeName == "ReactDataGridColumn") {
                ReactDataGridColumn.process(that, that.getColumnLevel(), child.props);
                directChildren = true;
            } else if (child.type.typeName == "ReactDataGridColumnGroup") {
                ReactDataGridColumnGroup.process(that, that.getColumnLevel(), child.props);
                hasGroups = true;
            } else {
                ReactDataGridColumnLevel.process(that, null, child.props);
            }
        });
        if (props.styles) {
            for (var prop in props.styles) {
                this.applyAttribute(this, prop, props.styles[prop], true);
            }
        }
        if (directChildren) {
            let level = that.getColumnLevel();
            if (level._tempCols) {
                if (hasGroups) {
                    level.setGroupedColumns(level._tempCols);
                } else {
                    level.setColumns(level._tempCols);
                }

            }
        }
    }

    getAllChildren() {
        //all my props.children are just configuration.
        return this.children;
    }
    getClassNames() {
        return ["ReactDataGrid", "FlexDataGrid", "NdgBase", "UIComponent", "IExtendedFlexDataGrid", " ISpinnerOwner"
            , "IFocusManagerComponent", "IExtendedDataGrid"];
    }


}

flexiciousNmsp.ReactDataGrid = ReactDataGrid; //add to name space
ReactDataGrid.prototype.typeName = ReactDataGrid.typeName = 'ReactDataGrid';//for quick inspection

ReactDataGrid.SELECTION_MODE_SINGLE_ROW = "singleRow";
ReactDataGrid.SELECTION_MODE_MULTIPLE_ROWS = "multipleRows";
ReactDataGrid.SELECTION_MODE_SINGLE_CELL = "singleCell";
ReactDataGrid.SELECTION_MODE_MULTIPLE_CELLS = "multipleCells";
ReactDataGrid.SELECTION_MODE_NONE = "none";
