/**
 * Copyright (c) 2008 Gareth Arch
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/**
 *
 * http://code.google.com/p/flexdateutils/
 * Code license:       MIT License
 * This utility class packages together various date and time functions.  Most of them duplicate the functionality
 * found in the corresponding ColdFusion date/time functions, and argument-wise are set up the same way.
 */

/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */

/**
 * Given a date range, sets the startDate and endDate values to the appropraite values.
 * @constructor
 * @class DateUtils
 */


import Constants from './Constants'
import UIUtils from './UIUtils' 
import TypedObject from '../core/TypedObject'
export default class DateUtils extends TypedObject {
    getClassNames() {
        return "DateUtils";
    }

    /**
     * Formats a date into a certain date/time format
     *
     * @param date    The date to format
     * @param mask    How the date should be formatted
     *
     * @return        A formatted date
     */
    static dateTimeFormat(date, mask) {
        if (typeof  mask == "undefined") mask = Constants.DEFAULT_DATE_FORMAT;

        return buildDateTime(date, mask, "(Y|M|D|E|A|J|H|K|L|N|S|TZD|\\W)+", `${Constants.SHORT_DATE_MASK} ${Constants.SHORT_TIME_MASK}`);

    }

    /**
     * Formats a time into a certain time format
     *
     * @param date    The date to format
     * @param mask    How the date should be formatted
     *
     * @return        A formatted time
     */
    static getTimeFormat(date, mask) {
        if (typeof  mask == "undefined") mask = Constants.SHORT_TIME_MASK;

        return buildDateTime(date, mask, "(A|:|J|H|K|L|N|S|TZD|\\s)+", Constants.SHORT_TIME_MASK);

    }

    /**
     * Formats a date into a certain date format
     *
     * @param date    The date to format
     * @param mask    How the date should be formatted
     *
     * @return        A formatted date
     */
    static formatString(date, mask) {
        if (typeof  mask == "undefined") mask = Constants.SHORT_DATE_MASK;

        return buildDateTime(date, mask, "(Y|M|D|E|\\W)+", Constants.SHORT_DATE_MASK);

    }

    /**
     * Adds the specified number of "date parts" to a date, e.g. 6 days
     *
     * @param datePart    The part of the date that will be added
     * @param number    The total number of "dateParts" to add to the date
     * @param date        The date on which to add
     *
     * @return            The new date
     */
    static dateAdd(datePart, number, date) {

        const _returnDate = new Date(date);

        switch (datePart) {
            case DateUtils.YEAR:
            case DateUtils.MONTH:
            case DateUtils.DAY_OF_MONTH:
            case DateUtils.HOURS:
            case DateUtils.MINUTES:
            case DateUtils.SECONDS:
            case DateUtils.MILLISECONDS:
                _returnDate[ `set${UIUtils.doCap(datePart)}` ]((_returnDate["get" + UIUtils.doCap(datePart)]()) + number);
                break;
            case DateUtils.WEEK:
                _returnDate.setTime(((number * 7 * DateUtils.millisecondsPerDay ) + _returnDate.getTime()));
                _returnDate[ DateUtils.DAY_OF_MONTH ] += number * 7;
                break;
            default:
                break;
        }
        return _returnDate;
    }

    /**
     * Gets the day of the week
     *
     * @param date    The date for which to get the day of the week
     *
     * @return        A number representing the day of the week, 0 to 6
     */
    static dayOfWeek(date) {
        return date.getDay();

    }

    /**
     * Gets the ordinal value or day of the year
     *
     * @param date    The date for which to get the day of the year
     *
     * @return        A number representing the day of the year, 1 to 365 or 366 for a leap year
     */
    static dayOfYear(date) {
        // add one as it has to include first of year
        return DateUtils.dateDiff(DateUtils.DAY_OF_MONTH, new Date(date.getFullYear(), DateUtils.monthAsNumber(DateUtils.JANUARY), 1), date) + 1;

    }

    /**
     * Gets the week of the year
     *
     * @param date    The date for which to get the week of the year
     *
     * @return        A number representing the week of the year, 1 to 53 ( as there are slightly more than 52 weeks of days in a year)
     */
    static weekOfYear(date) {
        return Math.ceil(DateUtils.dayOfYear(date) / 7);

    }

    /**
     * Converts the day of the week to a Flex day of the week
     *
     * @param {Number} localDayOfWeek    The human readable day of week
     *
     * @return {Number}       The Flex converted day of week or 0 aka Sunday
     */
    static toFlexDayOfWeek(localDayOfWeek) {
        return ( localDayOfWeek > 0 && localDayOfWeek < 8 ) ? localDayOfWeek - 1 : 0;

    }

    /**
     * Gets the Xth day of the month.
     * e.g. get the 3rd Wednesday of the month
     *
     * @param iteration        The iteration of the month to get e.g. 4th or Last
     * @param strDayOfWeek    The day of the week as a string
     * @param date            The date containing the month and year
     *
     * @return                The date of the xth dayOfWeek of the month
     */
    static dayOfWeekIterationOfMonth(iteration, strDayOfWeek, date) {
        // get the numeric day of the week for the requested day
        const _dayOfWeek = DateUtils.dayOfWeekAsNumber(strDayOfWeek);
        // get the date for the first of the month
        const _firstOfMonth = new Date(date.getFullYear(), date.getMonth(), 1);
        // calculate how many days to add to get to the requested day from the first of the month
        let _daysToAdd = _dayOfWeek - DateUtils.dayOfWeek(_firstOfMonth);
        // if dayOfWeek is before the first of the month, get the dayOfWeek for the following week
        if (_daysToAdd < 0) {
            _daysToAdd += 7;
        }
        // set the date to the first day of the week for the requested date
        const _firstDayOfWeekOfMonth = DateUtils.dateAdd(DateUtils.DAY_OF_MONTH, _daysToAdd, _firstOfMonth);
        // return the date if iteration is 1
        if (iteration == 1) {
            return _firstDayOfWeekOfMonth;
        } else {
            // if requesting an iteration that is more than is in that month or requesting the last day of week of month
            // return last date for that day of week of month
            if (( DateUtils.totalDayOfWeekInMonth(strDayOfWeek, date) < iteration ) || ( iteration == DateUtils.LAST )) {
                iteration = DateUtils.totalDayOfWeekInMonth(strDayOfWeek, date);
            }
            // subtract 1 as it starts from the first dayOfWeek of month
            return DateUtils.dateAdd(DateUtils.WEEK, iteration - 1, _firstDayOfWeekOfMonth);
        }

    }

    /**
     * Gets the days in the month
     *
     * @param date    The date to check
     *
     * @return        The number of days in the month
     */
    static daysInMonth(date) {
        // get the first day of the next month
        const _localDate = new Date(date.getFullYear(), DateUtils.dateAdd(DateUtils.MONTH, 1, date).getMonth(), 1);
        // subtract 1 day to get the last day of the requested month
        return DateUtils.dateAdd(DateUtils.DAY_OF_MONTH, -1, _localDate).getDate();

    }

    /**
     * Gets the total number of dayOfWeek in the month
     *
     * @param strDayOfWeek    The day of week to check
     * @param date            The date containing the month and year
     *
     * @return {Number}               The number of <code>strDayOfWeek</code> in that month and year
     */
    static totalDayOfWeekInMonth(strDayOfWeek, date) {
        const _startDate = DateUtils.dayOfWeekIterationOfMonth(1, strDayOfWeek, date);
        const _totalDays = DateUtils.dateDiff(DateUtils.DAY_OF_MONTH, _startDate, new Date(date.getFullYear(), date.getMonth(), DateUtils.daysInMonth(date)));
        // have to add 1 because have to include first day that is found i.e. if wed is on 2nd of 31 day month, would total 5, of if wed on 6th, would total 4
        return Math.floor(_totalDays / 7) + 1;

    }

    /**
     * Converts the month to a Flex month
     * @method toFlexMonth
     * @param {Number} localMonth  The human readable month
     *
     * @return {Number}       The Flex converted month or 0 aka January
     */
    static toFlexMonth(localMonth) {
        return ( localMonth > 0 && localMonth < 13 ) ? localMonth - 1 : 0;

    }

    /**
     * Determines whether a value is actually a valid date
     *
     * @method isDate
     * @param {String} value    The date value
     *
     * @return        <code>true</code> means this is a valid date, <code>false</code> means it is not a valid date
     */
    static isDate(value) {
        return Date.parse(value) > 0;

    }

    /**
     * Formats a date to the string version of the day of the week
     *
     * @param date    The date to format
     *
     * @return        A formatted day of week
     */
    static dayOfWeekAsString(date) {
        return DateUtils.formatString(date, "EEEE");

    }

    /**
     * Formats a date to the numeric version of the day of the week
     *
     * @param strDayOfWeek    The day of week to convert
     *
     * @return                A formatted day of week or -1 if day not found
     */
    static dayOfWeekAsNumber(strDayOfWeek) {
        return ( getObjDaysOfWeek()[ strDayOfWeek ] >= 0 ) ? getObjDaysOfWeek()[ strDayOfWeek ] : -1;

    }

    /**
     * Formats a date to the string version of the month
     *
     * @param date    The date to format
     *
     * @return        A formatted month
     */
    static monthAsString(date) {
        return DateUtils.formatString(date, "MMMM");

    }

    /**
     * Formats a month to the numeric version of the month
     *
     * @param strMonth    The month to convert
     *
     * @return            A formatted month or -1 if month not found
     */
    static monthAsNumber(strMonth) {
        return ( getObjMonth()[ strMonth ] >= 0 ) ? getObjMonth()[ strMonth ] : -1;

    }

    /**
     * Gets the number of days in the year
     *
     * @param date    The date to check
     *
     * @return        The total number of days in the year
     */
    static daysInYear(date) {
        return DateUtils.dateDiff(
            DateUtils.DAY_OF_MONTH,
            new Date(date.getFullYear(), DateUtils.monthAsNumber(DateUtils.JANUARY), 1),
            DateUtils.dateAdd(DateUtils.YEAR, 1, new Date(date.getFullYear(), DateUtils.monthAsNumber(DateUtils.JANUARY), 1)));

    }

    /**
     * Determines whether the year is a leap year or not
     *
     * @param date    The date to check
     *
     * @return        <code>true</code> means it is a leap year, <code>false</code> means it is not a leap year.
     */
    static isLeapYear(date) {
        return  DateUtils.daysInYear(date) > 365;

    }

    /**
     * Determines the number of "dateParts" difference between 2 dates
     *
     * @param datePart    The part of the date that will be checked
     * @param startDate    The starting date
     * @param endDate    The ending date
     *
     * @return            The number of "dateParts" difference
     */
    static dateDiff(datePart, startDate, endDate) {
        let _returnValue = 0;

        switch (datePart) {
            case DateUtils.MILLISECONDS:
                _returnValue = endDate.getTime() - startDate.getTime();
                break;
            case DateUtils.SECONDS:
                _returnValue = Math.floor(DateUtils.dateDiff(DateUtils.MILLISECONDS, startDate, endDate) / DateUtils.SECOND_VALUE);
                break;
            case DateUtils.MINUTES:
                _returnValue = Math.floor(DateUtils.dateDiff(DateUtils.MILLISECONDS, startDate, endDate) / DateUtils.MINUTE_VALUE);
                break;
            case DateUtils.HOURS:
                _returnValue = Math.floor(DateUtils.dateDiff(DateUtils.MILLISECONDS, startDate, endDate) / DateUtils.HOUR_VALUE);
                break;
            case DateUtils.DAY_OF_MONTH:
                // Math.floor causes rounding down error with DST start at dayOfYear
                _returnValue = Math.floor(DateUtils.dateDiff(DateUtils.MILLISECONDS, startDate, endDate) / DateUtils.DAY_VALUE);
                break;
            case DateUtils.WEEK:
                _returnValue = Math.floor(DateUtils.dateDiff(DateUtils.MILLISECONDS, startDate, endDate) / DateUtils.WEEK_VALUE);
                break;
            case DateUtils.MONTH:
                // if start date is after end date, switch values and take inverse of return value
                if (DateUtils.dateDiff(DateUtils.MILLISECONDS, startDate, endDate) < 0) {
                    _returnValue -= DateUtils.dateDiff(DateUtils.MONTH, endDate, startDate);
                } else {
                    // get number of months based upon years
                    _returnValue = DateUtils.dateDiff(DateUtils.YEAR, startDate, endDate) * 12;
                    // subtract months then perform test to verify whether to subtract one month or not
                    // the check below gets the correct starting number of months (but may need to have one month removed after check)
                    if (endDate.getMonth() != startDate.getMonth()) {
                        _returnValue += ( endDate.getMonth() <= startDate.getMonth() ) ? 12 - startDate.getMonth() + endDate.getMonth() : endDate.getMonth() - startDate.getMonth();
                    }
                    // have to perform same checks as YEAR
                    // i.e. if end date day is <= start date day, and end date milliseconds < start date milliseconds
                    if (( endDate[ DateUtils.DAY_OF_MONTH ] < startDate[ DateUtils.DAY_OF_MONTH ] ) ||
                        ( endDate[ DateUtils.DAY_OF_MONTH ] == startDate[ DateUtils.DAY_OF_MONTH ] &&
                            endDate[ DateUtils.MILLISECONDS ] < startDate[ DateUtils.MILLISECONDS ] )) {
                        _returnValue -= 1;
                    }
                }
                break;
            case DateUtils.YEAR:
                _returnValue = endDate.getFullYear() - startDate.getFullYear();
                // this fixes the previous problem with dates that ran into 2 calendar years
                // previously, if 2 dates were in separate calendar years, but the months were not > 1 year apart, then it was returning too many years
                // e.g. 11/2008 to 2/2009 was returning 1, but should have been returning 0 (zero)
                // if start date before end date and months are less than 1 year apart, add 1 to year to fix offset issue
                // if end date before start date and months are less than 1 year apart, subtract 1 year to fix offset issue
                // added month and milliseconds check to make sure that a date that was e.g. 9/11/07 9:15AM would not return 1 year if the end date was 9/11/08 9:14AM
                if (_returnValue != 0) {
                    // if start date is after end date
                    if (_returnValue < 0) {
                        // if end date month is >= start date month, and end date day is >= start date day, and end date milliseconds > start date milliseconds
                        if (( endDate[ DateUtils.MONTH ] > startDate[ DateUtils.MONTH ] ) ||
                            ( endDate[ DateUtils.MONTH ] == startDate[ DateUtils.MONTH ] && endDate[ DateUtils.DAY_OF_MONTH ] > startDate[ DateUtils.DAY_OF_MONTH ] ) ||
                            ( endDate[ DateUtils.MONTH ] == startDate[ DateUtils.MONTH ] && endDate[ DateUtils.DAY_OF_MONTH ] == startDate[ DateUtils.DAY_OF_MONTH ] &&
                                endDate[ DateUtils.MILLISECONDS ] > startDate[ DateUtils.MILLISECONDS ] )) {
                            _returnValue += 1;
                        }
                    } else {
                        // if end date month is <= start date month, and end date day is <= start date day, and end date milliseconds < start date milliseconds
                        if (( endDate[ DateUtils.MONTH ] < startDate[ DateUtils.MONTH ] ) ||
                            ( endDate[ DateUtils.MONTH ] == startDate[ DateUtils.MONTH ] && endDate[ DateUtils.DAY_OF_MONTH ] < startDate[ DateUtils.DAY_OF_MONTH ] ) ||
                            ( endDate[ DateUtils.MONTH ] == startDate[ DateUtils.MONTH ] && endDate[ DateUtils.DAY_OF_MONTH ] == startDate[ DateUtils.DAY_OF_MONTH ] &&
                                endDate[ DateUtils.MILLISECONDS ] < startDate[ DateUtils.MILLISECONDS ] )) {
                            _returnValue -= 1;
                        }
                    }
                }
                break;

        }

        return _returnValue;


    }
}


DateUtils.MONDAY = "monday";
DateUtils.TUESDAY = "tuesday";
DateUtils.WEDNESDAY = "wednesday";
DateUtils.THURSDAY = "thursday";
DateUtils.FRIDAY = "friday";
DateUtils.SATURDAY = "saturday";
DateUtils.SUNDAY = "sunday";
DateUtils.JANUARY = "january";
DateUtils.FEBRUARY = "february";
DateUtils.MARCH = "march";
DateUtils.APRIL = "april";
DateUtils.MAY = "may";
DateUtils.JUNE = "june";
DateUtils.JULY = "july";
DateUtils.AUGUST = "august";
DateUtils.SEPTEMBER = "september";
DateUtils.OCTOBER = "october";
DateUtils.NOVEMBER = "november";
DateUtils.DECEMBER = "december";
DateUtils.YEAR = "fullYear";
DateUtils.MONTH = "month";
DateUtils.WEEK = "week";
DateUtils.DAY_OF_MONTH = "date";
DateUtils.HOURS = "hours";
DateUtils.MINUTES = "minutes";
DateUtils.SECONDS = "seconds";
DateUtils.MILLISECONDS = "milliseconds";
DateUtils.DAY_OF_WEEK = "day";
DateUtils.LAST = -1;

DateUtils.SECOND_VALUE = 1000;
DateUtils.MINUTE_VALUE = DateUtils.SECOND_VALUE * 60;
DateUtils.HOUR_VALUE = DateUtils.MINUTE_VALUE * 60;
DateUtils.DAY_VALUE = DateUtils.HOUR_VALUE * 24;
DateUtils.WEEK_VALUE = DateUtils.DAY_VALUE * 7;


let _objDaysOfWeek = null;
let _objMonth = null;
/**
 *
 */
var getObjDaysOfWeek = () => {
    if (!_objDaysOfWeek) {
        _objDaysOfWeek = {};
        _objDaysOfWeek[ DateUtils.SUNDAY ] = 0;
        _objDaysOfWeek[ DateUtils.MONDAY ] = 1;
        _objDaysOfWeek[ DateUtils.TUESDAY ] = 2;
        _objDaysOfWeek[ DateUtils.WEDNESDAY ] = 3;
        _objDaysOfWeek[ DateUtils.THURSDAY ] = 4;
        _objDaysOfWeek[ DateUtils.FRIDAY ] = 5;
        _objDaysOfWeek[ DateUtils.SATURDAY ] = 6;
    }
    return _objDaysOfWeek;
};

var getObjMonth = () => {
    if (!_objMonth) {
        _objMonth = {};
        _objMonth[ DateUtils.JANUARY ] = 0;
        _objMonth[ DateUtils.FEBRUARY ] = 1;
        _objMonth[ DateUtils.MARCH ] = 2;
        _objMonth[ DateUtils.APRIL ] = 3;
        _objMonth[ DateUtils.MAY ] = 4;
        _objMonth[ DateUtils.JUNE ] = 5;
        _objMonth[ DateUtils.JULY ] = 6;
        _objMonth[ DateUtils.AUGUST ] = 7;
        _objMonth[ DateUtils.SEPTEMBER ] = 8;
        _objMonth[ DateUtils.OCTOBER ] = 9;
        _objMonth[ DateUtils.NOVEMBER ] = 10;
        _objMonth[ DateUtils.DECEMBER ] = 11;
    }
    return _objMonth;

};

function removeInvalidDateTimeCharacters(mask, pattern, defaultValue) {
    // test for invalid date and time characters
    if (mask.replace(new RegExp(pattern, "ig"), "").length > 0) {
        // if user is passing an invalid mask, default to defaultValue
        mask = defaultValue;
    }
    // temporarily replace TZD with lowercase tzd for replacing later
    return mask.replace(new RegExp("TZD", "i"), "tzd");

}


DateUtils.millisecondsPerDay = 1000 * 60 * 60 * 24;

function buildDateTime(date, mask, pattern, defaultValue) {
    const df = new flexiciousNmsp.DateFormatter();
    df.formatString = removeInvalidDateTimeCharacters(mask, pattern, defaultValue);
    return df.format(date).replace(new RegExp("TZD", "i"), buildTimeZoneDesignation(date));
}


function buildTimeZoneDesignation(date) {

    if (!date) {
        return "";
    }
    let timeZoneAsString = "GMT ";
    // timezoneoffset is the number that needs to be added to the local time to get to GMT, so
    // a positive number would actually be GMT -X hours
    if (date.getTimezoneOffset() / 60 > 0 && date.getTimezoneOffset() / 60 < 10) {
        timeZoneAsString += `-0${( date.getTimezoneOffset() / 60 ).toString()}`;
    } else if (date.getTimezoneOffset() < 0 && date.getTimezoneOffset() / 60 > -10) {
        timeZoneAsString += `0${( -1 * date.getTimezoneOffset() / 60 ).toString()}`;
    }
    // add zeros to match standard format
    timeZoneAsString += "00";

    return timeZoneAsString;

}

flexiciousNmsp.DateUtils = DateUtils;