/*eslint-disable */
/**
 * Flexicious
 * Copyright 2011, Flexicious LLC
 */
import TypedObject from '../core/TypedObject'
import CurrencyFormatter from './CurrencyFormatter'

import DateFormatter from './DateFormatter'
/**
 * A Utility class that contains generic functions needed in multiple
 * places.
 * @namespace flexiciousNmsp
 * @class UIUtils
 * @name UIUtils
 * 
 */
export default class UIUtils extends TypedObject {

    getClassNames() {
        return ["TypedObject", "UIUtils"];
    }

    /**
     * Returns property or subproperty (dot delimited) of an object <br/>
     * @param data The source object
     * @param property The name of the property with the path if dot-delimited
     */
    static getObjectProperty(data, property) {
        if (data) {
            if (property) {
                const names = property.split(".");
                let obj = data;
                let parent = data;

                for (const name of names) {
                    obj = parent[name];
                    if (obj) {
                        parent = obj;
                    }
                    else {
                        // "null" may be a valid value. In any event no need to search futher
                        return null;
                    }
                }

                return obj;
            }
            else {
                return data;
            }
        }
        else {
            return null;
        }

    }

    /**
     * Parses the value of a complex property and return its String equivalent
     * @param item the source object
     * @param column the DataGridColumn object. Its dataField property gets the property path with name
     */
    static labelObjectProperty(item, column) {
        const obj = UIUtils.getObjectProperty(item, column.getDataField());
        return obj != null ? String(obj) : "";
    }

    /**
     * Generates single string from collection of items, delimited by delimiter
     * @param list array of objects
     * @param propertyName name of a property to be included, may be a complex "." delimited property name
     * @param delimiter a delimiter between elements in the string, by default it's ", "
     */
    static concatenateList(list, propertyName, delimiter) {
        if (typeof delimiter == "undefined") delimiter = ", ";

        let result = "";
        if (list) {
            for (const obj of list) {
                const item = (UIUtils.getObjectProperty(obj, propertyName));
                if (item) {
                    result += result.length > 0 ? delimiter + item : item;
                }
            }
        }
        return result;

    }

    /**
     * Parses the value of a complex list property, concatenating string values of all elements.
     * To specify field name and display field name use dataField="dataField/displayPropertyName/delimiter",
     * where "dataField" is the field name, and "displayPropertyName" is the display field of elements in the list,
     * and "delimiter" is the delimiter string to be used (optional).
     * @param item the source object
     * @param column the DataGridColumn object. Its dataField property gets the property path with name
     */
    static labelObjectPropertyList(item, column) {
        if (column.getDataField()) {
            const names = column.getDataField().split("/");
            const fieldName = names[0];
            const propName = names[1];
            const delimiter = names.length > 2 ? names[2] : null;
            return UIUtils.concatenateList(UIUtils.getObjectProperty(item, fieldName), propName, delimiter);
        }
        else {
            return "";
        }

    }

    /**
     * Datagrid function to format the yes no column in a grid.
     * @return A formatted boolean string
     *
     */
    static dataGridFormatYesNoLabelFunction(item, dgColumn) {
        const val = UIUtils.resolveExpression(item, dgColumn.getDataField());
        return UIUtils.formatBoolean(val);

    }

    /**
     * Datagrid function to format the date column in a grid.
     * @return A formatted date string
     *
     */
    static dataGridFormatDateLabelFunction(item, dgColumn) {
        const num = UIUtils.resolveExpression(item, dgColumn.getDataField());
        const date = typeof num == 'string' ? new Date(Date.parse(num.toString())) : num;
        if (date)
            return UIUtils.formatDate(date);
        return null;

    }

    /**
     * Datagrid function to format the currency column in a grid.
     * @return A formatted currency string
     *
     */
    static dataGridFormatCurrencyLabelFunction(item, dgColumn) {
        const num = UIUtils.resolveExpression(item, dgColumn.getDataField());
        const number = typeof num == 'string' ? parseFloat(num.toString()) : num;
        if (!isNaN(number))
            return UIUtils.formatCurrency(number, ' ');
        else
            return "";

    }

    /**
     * Takes in a boolean, returns a Yes or No
     * @param val  Boolean indicator to format
     * @return     The formatted yes or no string
     *
     */
    static formatBoolean(val) {
        return val ? "Yes" : "No";

    }

    /**
     * Takes in a date, and returns it formatted as
     * MM/DD/YY
     * @param {Object} date The date to format
     * @param {String} formatString
     * @return     The formatted date string
     *
     */
    static formatDate(date, formatString) {
        if (typeof formatString == "undefined") formatString = flexiciousNmsp.Constants.DEFAULT_DATE_FORMAT;
        const f = new flexiciousNmsp.DateFormatter();
        f.formatString = formatString;
        return f.format(date);

    }
    static getDateValue(dateStr, formatString) {
        if (typeof formatString == "undefined") formatString = flexiciousNmsp.Constants.DEFAULT_DATE_FORMAT;
        const f = new flexiciousNmsp.DateFormatter();
        f.formatString = formatString;
        return f.parse(dateStr);
    }
    /**
     * Takes in a number, and returns it formatted as
     * Currency
     * @param val  The number to format
     * @param currencySymbol
     * @return     The formatted currency string
     *
     */
    static formatCurrency(val, currencySymbol) {
        if (typeof currencySymbol == "undefined") currencySymbol = '';
        const f = new flexiciousNmsp.CurrencyFormatter();
        f.currencySymbol = currencySymbol;
        return f.format(val);
    }

    /**
     * Calls add popup.
     * @param pop
     * @param parent
     * @param modal
     * @param triggerEvent - If you specify this, it will move the popup right below the mouse click
     * @param title
     * @param actions
     * callbackFunctionDetail - a callback wrapper for the popup getting close.
     *                                 it should be
     *                                                {
     *                                                     context : this,
     *                                                     method : onPopupClosed. // which means this.onPopupClosed.
     *                                                }
     * @return
     *
     */
    static addPopUp(pop, parent, modal, triggerEvent, title, actions) {
        return UIUtils.adapter.showDialog(pop, parent, modal, pop.width || 800, pop.height || 600, title, actions);
    }
    static removePopUp(popup) {
        //todo wire up removal??
    }
    static createContextMenuPopup() {
        return UIUtils.adapter.createContextMenuPopup()
    }
    static createSettingsPopup() {
        return UIUtils.adapter.createSettingsPopup()
    }
    static createSaveSettingsPopup() {
        return UIUtils.adapter.createSaveSettingsPopup()
    }
    static createExportPopup() {
        return UIUtils.adapter.createExportPopup()
    }

    static getDomElementIndex(parent, child) {
        for (let i = 0; i < parent.children.length; i++) {
            if (parent.children[i] == child) {
                return i;
            }
        }
        return -1;
    }

    /**
     * Returns true if passed in object is a boolean, number or string
     */
    static isPrimitive(input) {
        const typeOfA = typeof (input);
        return (typeOfA == "boolean" || typeOfA == "number" || typeOfA == "string" || (input instanceof Date) || (typeOfA == "function"));

    }

    /**
     * Returns true if passed in object is a undefined or null data types.
     */
    static isSpecialDataType(input) {
        const typeOfA = typeof (input);
        return (typeOfA == "undefined" || (typeOfA == "object" && input == null));
    }

    /**
     * Returns true if passed in object is Array or Object data type
     */
    static isCompositeDataType(input) {
        return (input.constructor === Array || typeof (input) == "object");
    }

    /**
     * Returns true if passed in object is Array
     */

    static isArray(input) {
        return (input.constructor === Array);
    }

    /**
     * Returns true if passed in object is clonable
     */

    static isCloneable(input) {
        return (typeof (input["clone"]) == "function");
    }

    static ensureWithinView(popup, parent) {
        if ((popup.y + popup.height) > parent.height) {
            popup.y -= popup.height;
        }
        if ((popup.getX() + popup.width) > parent.width) {
            popup.setX(
                popuup.getX() - popup.getWidth()
            );
        }

    }

    /**
     * Returns the average of all items the specified data provider
     * @param dataProvider
     * @param fld
     * @return
     *
     */
    static average(dataProvider, fld) {
        if (typeof fld == "undefined") fld = "";


        if (dataProvider && UIUtils.getLength(dataProvider) > 0) {
            return (UIUtils.sum(dataProvider, fld) / UIUtils.getLength(dataProvider));
        }
        else {
            return 0;
        }

    }

    /**
     * If arr is XML or XMLList, returns length() else returns length;
     */
    static getLength(arr) {
        if (!arr) return 0;
        //        if(arr is XML){
        //            return (arr as XML).length();
        //        }
        //        else if (arr is XMLList){
        //            return (arr as XMLList).length();
        //        }
        //        else{
        return arr.length;
        //}
        //return -1;

    }

    /**
     * Returns the sum of all items the specified data provider
     * @param dataProvider
     * @param fld
     * @return
     *
     */
    static sum(dataProvider, fld) {
        if (typeof fld == "undefined") fld = "";

        let total = 0;

        for (const row of dataProvider) {
            const num = parseFloat(UIUtils.toString(UIUtils.resolveExpression(row, fld)));
            if (!isNaN(num))
                total += num;
        }

        return total;
    }

    /**
     * Returns the minimum value in the specified data provider
     * @param dataProvider
     * @param fld
     * @param comparisionType
     * @return
     *
     */
    static min(dataProvider, fld, comparisionType) {
        if (typeof fld == "undefined") fld = "";
        if (typeof comparisionType == "undefined") comparisionType = "auto";

        let min;

        for (const row of dataProvider) {
            const value = flexiciousNmsp.FilterExpression.convert(comparisionType, UIUtils.resolveExpression(row, fld));

            if (value != null) {

                if ((value) < min || !min) {
                    min = (value);
                }
            }
        }

        return min;
    }

    /**
     * Returns the maximum value in the specified data provider
     * @param dataProvider
     * @param fld
     * @param comparisionType
     * @return
     *
     */
    static max(dataProvider, fld, comparisionType) {
        if (typeof fld == "undefined") fld = "";
        if (typeof comparisionType == "undefined") comparisionType = "auto";

        let max;

        for (const row of dataProvider) {
            const value = flexiciousNmsp.FilterExpression.convert(comparisionType, UIUtils.resolveExpression(row, fld));
            if (value != null) {

                if ((value) > max || !max) {
                    max = (value);
                }
            }
        }

        return max;
    }

    /**
     * Pastes the passed in data to the clipboard.
     * @param strToPaste
     * @return
     *
     */
    static pasteToClipBoard(strToPaste) {
        if (!navigator.clipboard) {
           UIUtils.fallbackCopyTextToClipboard(strToPaste);
            return;
          }

          navigator.clipboard.writeText(strToPaste).then(function() {
          }, function(err) {
          });
       
    }

    static fallbackCopyTextToClipboard(strToPaste) {

         // window.prompt("Press the Ctrl+C key to copy to Clipboard", strToPaste)

        // saving whole string to actual clipboard
        try{
            window.clipboardData.setData('text', strToPaste);
          } catch(err) {
        const el = document.createElement('textarea');  // Create a <textarea> element
        el.value = strToPaste;  
        el.id = 'pasteArea'                        // Set its value to the string that you want copied
        el.setAttribute('readonly', '');                // Make it readonly to be tamper-proof
        el.style.position = 'absolute';                 
        el.style.left = '-9999px';                      // Move outside the screen to make it invisible
        document.body.appendChild(el);                  // Append the <textarea> element to the HTML document
        const selected =            
            document.getSelection().rangeCount > 0      // Check if there is any content selected previously
            ? document.getSelection().getRangeAt(0)     // Store selection if found
            : false;                                    // Mark as false to know no selection existed before
        el.select();                                    // Select the <textarea> content
        try {
            var successful = document.execCommand('copy');
            var msg = successful ? 'successful' : 'unsuccessful';
          } catch (err) {
          }                   // Copy - only works as a result of a user action (e.g. click events)
      //  document.body.removeChild(el);                  // Remove the <textarea> element
        if (selected) {                                 // If a selection existed before copying
            document.getSelection().removeAllRanges();  // Unselect everything on the HTML document
            document.getSelection().addRange(selected); // Restore the original selection
        }
    }
    }

    /**
     * Checks to see if the passed in value is a string representation of a numeric value
     * @param str The string representation to check
     * @return True if str is a string representation of a number value
     *
     */
    static isStringNumeric(str) {
        //this should be renamed.
        if (str == "") return false;
        for (let i = 0; i < str.length; i++) {
            const obj = str.charAt(i);
            if ((obj < '0' || obj > '9') && obj != '.' && obj != '-') {
                return false
            }
        }
        return true;
    }

    /**
     * A utility function that encapsulates the task of showing a confirmation popup.
     * @param msg     Text string that appears in the Alert control.
     * @param parent  Object upon which the confirm dialog centers itself.
     * @param modal   Flag represents if this is a modal window or not
     * @param title   Text string that appears in the title bar.
     * @param features An array of objects having the details about the flag, class(style) and its callback function
     *
     */
    static showConfirm(msg, parent, modal, title, features) {
        UIUtils.adapter.showConfirm(msg, parent, modal, title, features);
    }

    /**
     * Shows a popup with an error message
     * @param errorMessage     Text string that appears in the Alert control.
     * @param [errorTitle]    Text string that appears in the title bar.
     */
    static showError(errorMessage, errorTitle) {
        if (typeof errorTitle == "undefined") errorTitle = "Error Occurred";
        UIUtils.showToaster(errorMessage, errorTitle, 'error');
    }

    /**
     * Shows a popup with an informational message
     * @param message     Text string that appears in the Alert control.
     * @param title    Text string that appears in the title bar.
     * @param flag   Test string which represents the type of message. e.g error
     */
    static showMessage(message, title, flag) {
        if (UIUtils.adapter && typeof UIUtils.adapter.showMessage == "function")
            UIUtils.adapter.showMessage(message, title, flag);
        else
            UIUtils.showToaster(message, title);
    }

    /**
     *
     * @param obj
     *
     */
    static traceData(obj) {
        Console.log(obj);
    }

    /**
     *
     */
    static handleError(event) {
        UIUtils.showError(`Error occurred : ${event.toString()}`, "Error occured");
    }

    /**
     * Removes the supplied item from the supplied array,
     * and returns that array
     * @param array            The array to remove the item from
     * @param itemToRemove    The item to remove
     *
     */
    static removeFromArray(array, itemToRemove) {
        let index = 0;
        let indexToRemove = -1;

        for (const item of array) {
            if (item == itemToRemove) {
                indexToRemove = index;
                break;
            }
            index++;
        }

        if (indexToRemove >= 0) {
            array.splice(indexToRemove, 1);
        }
        return array;
    }

    /**
     * Iterates through an array to see if the toString of the "valFld" property of any of the <br/>
     * items in the array array equals "compareVal"
     * @param array            Array to search
     * @param valFld        Property of the objects in the array
     * @param compareVal    The value to compare against.
     * @return
     *
     */
    static doesArrayContainValue(array, valFld, compareVal) {
        for (const o1 of array) {
            if (o1.hasOwnProperty(valFld) && o1[valFld] != null && o1[valFld].toString().toLocaleLowerCase() == compareVal.toLowerCase()) {
                return o1;
            }
        }

        return null;
    }

    /**
     *
     * Iterates through an array to see if the "toString" methods of any of the <br/>
     * items in the a1 array equals "compareVal"
     * @param array            Array to search
     * @param compareVal    The value to compare against.
     * @return The first object that matches the search criteria, null if no match
     *
     */
    static doesArrayContainStringValue(array, compareVal) {
        for (const o1 of array) {
            if (o1.toString().toLocaleLowerCase() == compareVal.toLowerCase()) {
                return o1;
            }
        }

        return null;
    }

    /**
     * Iterates through an array to see if the "valFld" property of any of the <br/>
     * items in the a1 array equals "compareVal"
     * @param array            Array to search
     * @param valFld        Property of the objects in the array
     * @param compareVal    The value to compare against.
     * @return The first object that matches the search criteria, null if no match
     *
     */
    static doesArrayContainObjectValue(array, valFld, compareVal) {
        for (const o1 of array) {
            if (o1[valFld] != null &&
                o1[valFld] == compareVal) {
                return o1;
            }
        }

        return null;
    }

    /**
     * Opens a browser popup
     * @param url        The URL to open
     * @param [options]    Options string, defaults to width=800,height=600,lef t=0,top=0, toolbar=No,location=No,scrollbars=Yes,status=No,resizable=No,fullscreen=No
     * @param [useNew]    If true, creates a new window, else reuses the existing one. Defaults to true
     *
     */
    static openBrowserPopup(url, options, useNew) {
        if (typeof useNew == 'undefined') useNew = true;
        if (typeof options == 'undefined') options = 'width=800,height=600,left=0,top=0,toolbar=No,location=No,scrollbars=Yes,status=No,resizable=No,fullscreen=No';
        return window.open(url, '', options);
    }

    /**
     * Iterates through each array and returns false if either contains
     * an element not contained within the other.
     * @param a1
     * @param a2
     * @return
     *
     */
    static areArraysEqual(a1, a2) {
        const areArraysEqual = true;
        if (a1 == null && a2 == null)
            return true;
        else if (a1 != null && a2 != null) {
            if (a1.length != a2.length)
                return false;

            for (let index = 0; index < a1.length; index++)
                if (a1[index] != a2[index])
                    return false;
        }
        else
            return false;
        return true;

    }

    /**
     * Used to create a filter renderer along with initialization properties
     * @param renderer         Class factory to be used as the generator
     * @param properties    Initialization properties.
     * @return The created renderer factory
     *
     */
    static createRenderer(renderer, properties) {
        const obj = eval(renderer);
        for (const prop in properties) {
            obj[prop] = properties[prop]
        }
    }

    /**
     * Returns the top level application.
     * @return
     */
    static getTopLevelApplication() {
        return window.document.body;
    }

    /**
     * Adds a child to the provided display object.
     * Takes into account the scenario where the parent is a Spark application.
     * @param {flexiciousNmsp.UIComponent} parent
     * @param {flexiciousNmsp.UIComponent} child
     * @param {flexiciousNmsp.UIComponent} before
     * @return
     */
    static addChild(parent, child, before) {
        if (child.hasOwnProperty("$$typeof")) {
            parent.children.push(child);
            return;
        }
        let origParentComponent;
        let origChildComponent;
        if (typeof parent["implementsOrExtends"] == "function") {
            if (parent.implementsOrExtends("UIComponent")) {
                origParentComponent = parent;
                if (parent.hasOwnProperty('contentGroup')) {
                    parent = parent.contentGroup;
                }
                else {
                    parent = parent.domElement;
                }

            }
        }
        if (typeof child["implementsOrExtends"] == "function") {
            if (child.implementsOrExtends("UIComponent")) {
                origChildComponent = child;
                child.invalidateDisplayList();
                child = child.domElement;
            }
        }

        if (typeof before !== "undefined" && before !== null) {
            if (typeof before["implementsOrExtends"] == "function") {
                if (before.implementsOrExtends("UIComponent")) {
                    before.invalidateDisplayList();
                    before = before.domElement;
                }
            }
        }

        if (origParentComponent && origChildComponent) {
            //both are UIComponents.
            origChildComponent.propertyBag.key = origParentComponent.children.length;
            origParentComponent.children.push(origChildComponent);
            origChildComponent.parent = origParentComponent;

            return;
        }
        //console.log(parent.constructor.toString()+":"+child.constructor.toString());
        //console.log(parent.id+":"+child.id)
        //UIUtils.adapter.addChild(parent,child);
        if (before)
            parent.insertBefore(child, before);
        else
            parent.appendChild(child);

        //        if(origChildComponent&&(origChildComponent.getWidth()==-1 || origChildComponent.getHeight()==-1)){
        //            origChildComponent.setActualSizeFromDomElement();
        //        }
    }

    /**
     * remove child from parent element
     * @param parent
     * @param child
     *
     */
    static removeChild(parent, child) {

        if (child.hasOwnProperty("$$typeof")) {
            parent.children.splice(parent.children.indexOf(child), 1);
            return;
        }
        let origParentComponent;
        let origChildComponent;
        if (typeof parent["implementsOrExtends"] == "function") {
            if (parent.implementsOrExtends("UIComponent")) {
                origParentComponent = parent;
                parent = parent.domElement;
            }
        }
        if (typeof child["implementsOrExtends"] == "function") {
            if (child.implementsOrExtends("UIComponent")) {
                origChildComponent = child;
                child = child.domElement;
            }
        }
        if (origParentComponent && origChildComponent) {
            origParentComponent.children.splice(origParentComponent.children.indexOf(origChildComponent), 1);
            origChildComponent.parent = null;
            origParentComponent.invalidateDisplayList();
        }
        //UIUtils.adapter.removeChild(parent, child);

    }

    /**
     * Given an array collection of complex objects, and a property,
     * returns an array collection of the propertyToExtract values
     * of each object within the array collection
     * @param source    A list of objects that contain the property to extract
     * @param propertyToExtract    The name of the property to extract
     * @return An array collection of the extracted property values
     */
    static extractPropertyValues(source, propertyToExtract) {
        const result = [];

        for (const obj of source) {
            if (obj)
                result.push(UIUtils.checkGetterAndRetrieve(obj, propertyToExtract));
        }

        return result;
    }

    /**
     * Returns the string value of the passed in object, or if null, returns an empty string.
     */
    static toString(host) {
        if (host) return host.toString();
        return "";

    }

    /**
     * Returns a specially formatted string that stores information about the
     * type. Used by preference persistence serialization, can handle array,
     * arraycollection and primitives.
     */
    static toPersistenceString(host) {
        let result = "";
        if ((host["getClassNames"] && host.getClassNames() == "ArrayCollection") || host instanceof Array) {
            result += (host["getClassNames"] && host.getClassNames() == "ArrayCollection") ? "ac+=" : "ar+=";

            for (const obj of host) {
                result += `${obj.toString()}+=`;
            }
        }
        else if (typeof host == "boolean") {
            result += "b+=";
            result += host ? "y" : "n";
        } else if (typeof host == "string") {
            result += "s+=";
            result += host;
        } else if (typeof host["getTime"] == "function") {
            result += "d+=";
            result += host.getTime().toString();
        } else if (typeof host == "number") {
            result += "n+=";
            result += host.toString();
        }
        return result;
    }

    /**
     * Takes in a specifically formatted strings and spits out an object
     * that was used to make that string.
     */
    static fromPersistenceString(host) {
        let result;
        const parts = host.split("+=");
        if (parts[0] == "ac" || parts[0] == "ar") {
            result = parts[0] == "ac" ? [] : [];
            let partIndex = 0;

            for (const part of parts) {
                if (partIndex != 0) {
                    if (part != null && part.length > 0)
                        (host["getClassNames"] && host.getClassNames() == "ArrayCollection") ? result.push(part) : result.push(part);
                }
                partIndex++;
            }
        } else if (parts[0] == "b") {
            return parts[1] == "y" ? true : false;
        } else if (parts[0] == "s") {
            return parts[1] ? parts[1].toString() : "";
        } else if (parts[0] == "d") {
            return Date.parse(parts[1].toString());
        } else if (parts[0] == "n") {
            return parseFloat(parts[1].toString());
        }
        return result;

    }

    /**
     * For prebuilt filters, short cut method to create date filter
     * @param description    Name of the filter (to show in filter dropdown)
     * @param fld            DataField of the column to apply filter to
     * @param dateRange        One of the DateRange constants (See the DateRange object)
     * @param start            If you specify DateRange=Custom, the start date for the custom range
     * @param end            If you specify DateRange=Custom, the end date for the custom range
     * @return A Filter object that contains the passed in fld and date range in the arguments collection.
     */
    static createDateFilter(description, fld, dateRange, start, end) {
        if (typeof start == "undefined") start = null;
        if (typeof end == "undefined") end = null;

        const filter = new flexiciousNmsp.Filter();
        filter.filterDescrption = description;
        filter.setFilterExpressions([
            {
                columnName: fld, filterOperation: flexiciousNmsp.FilterExpression.FILTER_OPERATION_TYPE_BETWEEN, filterControlValue: dateRange != flexiciousNmsp.DateRange.DATE_RANGE_CUSTOM ? dateRange :
                    `Custom__${start.toString()}__${end.toString()}`, expression: dateRange != flexiciousNmsp.DateRange.DATE_RANGE_CUSTOM ? getDateRange(dateRange) :
                        ([start, end])
            }
        ]);
        return filter;

    }

    /**
     * For prebuilt filters, short cut method to create list based filters (ComboBox, MSCB)
     * @param description    Name of the filter (to show in filter dropdown)
     * @param fld            DataField of the column to apply filter to
     * @param values        An array of strings that represents the dataField of the filter list control
     * @return A Filter object that contains the passed in fld and values in the arguments collection.
     */
    static createListFilter(description, fld, values) {
        const filter = new flexiciousNmsp.Filter();
        filter.filterDescrption = description;
        filter.setFilterExpressions([
            {
                columnName: fld, filterOperation: values.length > 1 ?
                    flexiciousNmsp.FilterExpression.FILTER_OPERATION_TYPE_IN_LIST : flexiciousNmsp.FilterExpression.FILTER_OPERATION_TYPE_EQUALS, filterControlValue: (values), expression: (values)
            }
        ]);
        return filter;

    }

    /**
     * Goes through columns of the passed in grid (or level for Ultimate) and checks to see if the given
     * column is in a columnGroup. If yes, returns the top level Column Group for it.
     * Added for Printing support of column groups
     *
     * @param grid    EADG, FDG or FDGCL
     * @param col    EADGC or FDGC
     * @return         ADGCG or FDGCG
     */
    static getGroupForColumn(grid, col) {
        //these should be interfaces
        if (!grid.hasOwnProperty("groupedColumns"))
            return null;
        const items = grid.getGroupedColumns();

        for (const grp of items) {
            if (grp.hasOwnProperty("children")) {//this means this is a Grouped column (ADGCG or FDGCG)
                if (UIUtils.isChildRecursive(grp, col)) {
                    return grp;
                }
            }
        }

        return null;
    }

    /**
     * Is col a child of grp or any of its children?
     * Added for Printing support of column groups
     * @param grp
     * @param col
     * @return
     */
    static isChildRecursive(grp, col) {
        let result = false;

        for (const childGrp of grp.children) {
            if (childGrp.hasOwnProperty("children")) {
                result = UIUtils.isChildRecursive(childGrp, col);
                if (result)
                    return true;
            }
            else if (childGrp == col) {
                return true;
            }
        }

        return false;
    }

    /**
     * check whether value is null or not
     * @param {Object} val 
     */
    static emptyIfNull(val) {
        if (!val) return "";
        return val.toString();

    }

    /**
     * capitalize first letter if prefix exist
     * 
     * @param prefix 
     * @param val 
     */
    static capitalizeFirstLetterIfPrefix(prefix, val) {
        if (!UIUtils.capitalizeFirstLetterIfPrefixCache[prefix + val]) {
            UIUtils.capitalizeFirstLetterIfPrefixCache[prefix + val] = (prefix && prefix.length) > 0 ? UIUtils.doCap(val) : val;
        }
        return UIUtils.capitalizeFirstLetterIfPrefixCache[prefix + val];
    }

    /**
     * capitalize first letter
     * @param {String} val 
     */
    static doCap(val) {
        return val.substr(0, 1).toUpperCase() + val.substr(1);
    }

    /**
     * change first letter to lower case
     * @param {String} val 
     */
    static doLower(val) {
        return val.substr(0, 1).toLowerCase() + val.substr(1);
    }

    /**
     * fetch style properties of cell
     * @param {flexiciousNmsp.FlexDataGridCell} cell 
     * @param {Object} styleProp 
     */
    static getStyleValue(cell, styleProp) {
        const result = cell.getColumn() ? cell.getColumn().getStyleValue(styleProp) : cell.level.getStyleValue(styleProp);
        return result;
    }

    /**
     * get wheel delta value on occuring mouse event
     * @param evt 
     * @return wheel data value
     */
    static getMouseWheelDelta(evt) {
        if (!evt) evt = event;
        const w = evt.wheelDelta;
        const d = evt.detail;
        const mutiplier = UIUtils.wheelDeltaMultiplier;
        if (d) {
            if (w) return mutiplier * w / d / 40 * d > 0 ? 1 : -1; // Opera
            else return -mutiplier * d / 3;              // Firefox;
        } else return mutiplier * w / 120;             // IE/Safari/Chrome


        //
        //        if (!event)
        //            event = window.event;
        //        return (event.wheelDelta) ? event.wheelDelta : (event.detail > 0) ? 5 : -5;
    }

    /**
     * remove all children of the parent element
     * @param {*} parent 
     */
    static removeAllChildren(parent) {
        parent.innerHTML = "";
    }

    static removeAll(arr) {
        return arr.splice(0, arr.length);       
    }

    /**
     * A utility function that resolves expressions like x.y.z (for complex object dataField Support)
     * Also used to apply values. For example, if dataField=x.y.z, we can apply x.y.z=valueToApply.
     *
     * @param host
     * @param expression
     * @param valueToApply
     * @param returnUndefinedIfPropertyNotFound
     * @param applyNullValues
     * @param valueApplier
     * @return {*}
     */
    static resolveExpression(
        host,
        expression,
        valueToApply,
        returnUndefinedIfPropertyNotFound,
        applyNullValues,
        valueApplier) {
        if (typeof valueToApply == "undefined") valueToApply = null;
        if (typeof returnUndefinedIfPropertyNotFound == "undefined") returnUndefinedIfPropertyNotFound = false;
        if (typeof applyNullValues == "undefined") applyNullValues = false;


        if (null == expression || expression == "") {
            return host;
        }
        if (!UIUtils.expressionCache[expression]) {
            UIUtils.expressionCache[expression] = {};
            UIUtils.expressionCache[expression].isSimple = !expression.includes(".") && !expression.includes("[");
        }

        if (UIUtils.expressionCache[expression].isSimple) {

            if (valueToApply != null || applyNullValues) {
                if (valueApplier)
                    valueApplier.checkSetterAndApply(host, expression, valueToApply);
                else
                    host[expression] = valueToApply;
            }
            const result = UIUtils.checkGetterAndRetrieve(host, expression);
            return (typeof result !== 'undefined' && result !== null) ? result : returnUndefinedIfPropertyNotFound ? undefined : null;
        }

        const fields = expression.split(".");
        let endpointData = host;
        var i = 0;

        for (var i = 0; i < fields.length; i++) {
            const field = fields[i];

            if (field.includes("[")) {
                const indexString = field.substring(field.indexOf("[") + 1, field.indexOf("]"));
                endpointData = endpointData[field.substring(0, field.indexOf("["))];

                if (parseInt(indexString) <= (endpointData).length - 1) {
                    endpointData = (endpointData)[parseInt(indexString)];
                }
                else {
                    return "";
                }
            }
            else if (endpointData != null && i <= fields.length) {

                if (endpointData.hasOwnProperty(field)) {
                    if (valueToApply && (i == (fields.length - 1)))
                        endpointData[field] = valueToApply;
                    endpointData = endpointData[field];
                }
                else {
                    return returnUndefinedIfPropertyNotFound ? undefined : null;
                }
            }
        }

        return endpointData;

    }

    /**
     * Sorts the given array
     */
    static sortArray(arrayIn, sorts) {
        const array = arrayIn;
        if (array.length == 0) return array;
        const funcs = [];
        const modelItem = array[0];

        for (const srt of sorts) {
            if (srt.sortColumn || srt.sortCompareFunction != null) {

                if (srt.sortCompareFunction != null) {
                    funcs.push(srt.sortCompareFunction)
                }
                else if (srt.sortCaseInsensitive) {
                    funcs.push(srt.isAscending ? (a, b, srt) => {
                        const aVal = (UIUtils.toString(UIUtils.resolveExpression(a, srt.sortColumn))).toLowerCase();
                        const bVal = (UIUtils.toString(UIUtils.resolveExpression(b, srt.sortColumn))).toLowerCase();

                        if(!aVal && !bVal) return 0;
                        if(!aVal) return -1;
                        if(!bVal) return 1;

                        if (aVal < bVal)
                            return -1;
                        if (aVal > bVal)
                            return 1;
                        return 0;
                    } :
                        (a, b, srt) => {
                            const aVal = (UIUtils.toString(UIUtils.resolveExpression(a, srt.sortColumn))).toLowerCase();
                            const bVal = (UIUtils.toString(UIUtils.resolveExpression(b, srt.sortColumn))).toLowerCase();
                            
                            if(!aVal && !bVal) return 0;
                            if(!bVal) return -1;
                            if(!aVal) return 1;
                            
                            if (aVal > bVal)
                                return -1;
                            if (aVal < bVal)
                                return 1;
                            return 0;
                        });
                } else if (srt.sortNumeric) {
                    funcs.push(srt.isAscending ? (a, b, srt) => {
                        const aVal = parseFloat(UIUtils.resolveExpression(a, srt.sortColumn));
                        const bVal = parseFloat(UIUtils.resolveExpression(b, srt.sortColumn));
                        
                        if(!aVal && !bVal) return 0;
                        if(!aVal) return -1;
                        if(!bVal) return 1;

                        if (aVal < bVal)
                            return -1;
                        if (aVal > bVal)
                            return 1;
                        return 0;
                    } :
                        (a, b, srt) => {
                            const aVal = parseFloat(UIUtils.resolveExpression(a, srt.sortColumn));
                            const bVal = parseFloat(UIUtils.resolveExpression(b, srt.sortColumn));
                            
                            if(!aVal && !bVal) return 0;
                            if(!bVal) return -1;
                            if(!aVal) return 1;

                            if (aVal > bVal)
                                return -1;
                            if (aVal < bVal)
                                return 1;
                            return 0;
                        });
                }
                else {
                    funcs.push(
                        srt.isAscending ? (a, b, srt) => {
                            const aVal = (UIUtils.resolveExpression(a, srt.sortColumn));
                            const bVal = (UIUtils.resolveExpression(b, srt.sortColumn));
                            
                            if(!aVal && !bVal) return 0;
                            if(!aVal) return -1;
                            if(!bVal) return 1;

                            if (aVal < bVal)
                                return -1;
                            if (aVal > bVal)
                                return 1;
                            return 0;
                        } :
                            (a, b, srt) => {
                                const aVal = (UIUtils.resolveExpression(a, srt.sortColumn));
                                const bVal = (UIUtils.resolveExpression(b, srt.sortColumn));

                                if(!aVal && !bVal) return 0;
                                if(!bVal) return -1;
                                if(!aVal) return 1;

                                if (aVal > bVal)
                                    return -1;
                                if (aVal < bVal)
                                    return 1;
                                return 0;
                            }
                    );
                }
            }
        }

        const finalFunc = (a, b) => {
            for (let j = 0; j < funcs.length; j++) {
                //for (var i=funcs.length-1;i>=0;i--){
                const result = sorts[j].sortCompareFunction ? sorts[j].isAscending ? funcs[j](a, b, sorts[j]) : funcs[j](b, a, sorts[j]) : funcs[j](a, b, sorts[j]);
                if (result != 0) {
                    return result;
                }
            }
            return 0;
        };
        array.sort(finalFunc);
        return array;
    }

    /**
     * Returns array of pages
     * @param {Array} arrayIn 
     * @param {Number} pageIndex 
     * @param {Number} pageSize 
     * @return {Array}
     */
    static pageArray(arrayIn, pageIndex, pageSize) {
        const totalRecords = arrayIn.length;
        const pageStart = ((pageIndex) * pageSize);
        const pageEnd = (pageStart + pageSize) > totalRecords - 1 ? totalRecords - 1 : pageStart + pageSize - 1;
        const newFlat = [];
        for (let i = pageStart; i <= pageEnd; i++)
            newFlat.push(arrayIn[i]);
        return newFlat;

    }

    /**
     * returns array of pages by page number
     * @param {Array} arrayIn 
     * @param {Array} pageIndexes
     * @param {Number} pageSize 
     * @return {Array}
     */
    static pageArrayByPageNumbers(arrayIn, pageIndexes, pageSize) {
        const totalRecords = arrayIn.length;
        const pageIndex = pageIndexes[0] > 0 ? pageIndexes[0] - 1 : 0;
        if (pageIndexes[0] >= pageIndexes[1])
            return UIUtils.pageArray(arrayIn, pageIndex, pageSize);

        const pageStart = ((pageIndex) * pageSize);
        const spanSize = pageSize * (pageIndexes[1] - pageIndex);

        const pageEnd = (pageStart + spanSize) > totalRecords - 1 ? totalRecords - 1 : pageStart + spanSize - 1;
        const newFlat = [];
        for (let i = pageStart; i <= pageEnd; i++)
            newFlat.push(arrayIn[i]);
        return newFlat;

    }

    /**
     * Provided an array, filters it on basis of the filter arguments. Returns a new array that only contains the values that match filters.
     * 
     * @param {Array} arrayIn 
     * @param {flexiciousNmsp.Filter} filter 
     * @param {flexiciousNmsp.FlexDataGrid} grid 
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level 
     * @param {Boolean} hideIfNoChildren 
     * 
     * @return {Array} returns filtered array.
     */
    static filterArray(arrayIn, filter, grid, level, hideIfNoChildren) {
        const retVal = [];

        for (const obj of arrayIn) {
            if (UIUtils.filterRecursive(obj, filter, level, hideIfNoChildren)) {
                retVal.push(obj);
            }
        }

        return retVal;
    }

    /**
     * if object do not have children, or none of them matched or object did not match and none of my children did either or there are are no recursing expressions its return false.
     * if <code>hideIfNoChildren=false</code> and object matched or object didn't match but atleast one of its children matched or there have nothing to filter its return true.
     * Filter data of all level of data grid recursively using defind filter function
     * @param {*} obj 
     * @param {flexiciousNmsp.Filter} filter 
     * @param {flexiciousNmsp.FlexDataGridColumnLevel} level 
     * @param {Boolean} hideIfNoChildren 
     * @return {Boolean} 
     */
    static filterRecursive(obj, filter, level, hideIfNoChildren) {
        if (level == null) {
            level = filter.level;
        }
        if (level != null && level.grid.globalFilterMatchFunction != null) {
            return level.grid.globalFilterMatchFunction(obj);
        }
        if ((filter.getFilterExpressions().length > 0 && (level.getNestDepth() == 1 || level.reusePreviousLevelColumns || level.getEnableFilters())) || (level.getHasFilterFunction())) {
            let match = true;
            if (filter.getFilterExpressions().length > 0 && (level.getNestDepth() == 1 || level.reusePreviousLevelColumns || level.getEnableFilters())) {
                //match filter
                var items = filter.getFilterExpressions();
                for (var i = 0; i < items.length; i++) {
                    const fExp = items[i];
                    if (level.grid.filterExcludeObjectsWithoutMatchField) {
                        if (fExp.columnName && obj && !obj.hasOwnProperty(fExp.columnName)) {
                            match = false;
                            break;
                        }
                    }
                    if (!fExp.isMatch(obj, level.grid)) {
                        match = false;
                        break;
                    }
                }
            }
            if (level.filterFunction != null && match) {
                //match filterFunction
                match = level.filterFunction(obj);
            }
            if (!level.nextLevel)
                return match;
            //now we know if I matched.
            if (match) {
                if (hideIfNoChildren) {
                    var children = level.getChildren(obj);
                    for (var i = 0; i < children.length; i++) {
                        const child = children[i];
                        if (UIUtils.filterRecursive(child, filter, level.nextLevel, hideIfNoChildren)) {
                            return true; //I matched and atleast one of my children matched.
                        }
                    }
                    return false;//either I do not have children, or none of them matched.
                } else
                    return true;//hideIfNoChildren=false and I matched
            } else {
                let recurse = false;
                var items = filter.getFilterExpressions();
                for (var i = 0; i < items.length; i++) {
                    const exp = items[i];
                    if (exp.recurse) {
                        recurse = true;
                        break;
                    }
                }
                if (recurse) {
                    //i didnt match, but did any of my children match?
                    var children = level.getChildren(obj);
                    for (var i = 0; i < children.length; i++) {
                        const child1 = children[i];
                        if (UIUtils.filterRecursive(child1, filter, level.nextLevel, hideIfNoChildren)) {
                            return true; //I didnt match but atleast one of my children matched.
                        }
                    }
                    return false;//i did not match, none of my children did either
                } else
                    return false;//i did not match, and there are no recursing expressions.

            }
        } else {
            //nothing to filter
            return true;
        }

    }

    /**
     * 
     * @param {*} items 
     * @param {*} filter 
     * @param {*} grid 
     * @param {*} level 
     * @param {*} recursingExpression 
     */
    static recursiveMatch(items, filter, grid, level, recursingExpression) {
        for (const obj of items) {
            if (recursingExpression.isMatch(obj, grid))
                return true;
            else if (level) {
                if (UIUtils.recursiveMatch(level.getChildren(obj), filter, grid, level.nextLevel, recursingExpression)) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Returns sorted array of pages
     * @param {Array} flat array of total pages
     * @param {*} filter instance of filter
     * @param {*} pages number of pages
     * @return {Array} 
     */
    static filterPageSort(flat, filter, pages) {
        if (typeof pages == "undefined") pages = null;

        if (filter.sorts.length > 0) {
            flat = UIUtils.sortArray(flat, filter.sorts);
        }
        if (filter.getFilterExpressions().length > 0) {
            flat = UIUtils.filterArray(flat, filter, null, null, false);
        }
        filter.recordCount = (flat.length);
        if (filter.pageIndex >= 0) {
            if (pages)
                flat = UIUtils.pageArrayByPageNumbers(flat, pages, filter.pageSize);
            else
                flat = UIUtils.pageArray(flat, filter.pageIndex, filter.pageSize);
        }
        return flat;

    }

    /**
     * Returns Zero if passed in input is NAN else returns input as is.
     * @param {Number} input
     * @return {Number}
     */
    static nanToZero(input) {
        return isNaN(input) ? 0 : input;
    }

    /**
     * check whether browser is internet explorer or not
     */
    static isIE() {
        if (UIUtils.prototype.ieVersion = -1) {
            const myNav = navigator.userAgent.toLowerCase();
            UIUtils.prototype.ieVersion = (myNav.includes('trident')) ? 11 : (myNav.includes('msie')) ? parseInt(myNav.split('msie')[1]) : 0;
            if (UIUtils.prototype.ieVersion == 11) {
                UIUtils.ieVersion = 11;
            }
        }
        return UIUtils.prototype.ieVersion;
    }

    /**
     * check whether browser is Mozila Firefox or not
     */
    static isMoz() {
        return navigator.userAgent.toLowerCase().includes('firefox');
    }

    /**
     * check whether browser is Chrome or Webkit version or not
     */
    static isWebKit() {
        return 'WebkitAppearance' in document.documentElement.style;
    }

    /**
     * Creates a gradient fill
     * @param {*} comp 
     * @param {Array} colors 
     * @param {Number} paddingX 
     * @param {Number} paddingY 
     */
    static gradientFill(comp, colors, paddingX, paddingY) {
        if (typeof paddingX == "undefined") paddingX = 0;
        if (typeof paddingY == "undefined") paddingY = 0;
        if (UIUtils.isIE()) {
            if (!colors || colors.length == 0)
                comp.setStyleAttribute("filter", "");
            else {
                if (UIUtils.ieVersion == 10) {
                    comp.setStyleAttribute("background", `-ms-linear-gradient(top, ${UIUtils.decimalToColor(colors[0])}, ${UIUtils.decimalToColor(colors[1])})`);
                } else if (UIUtils.ieVersion > 10) {
                    comp.setStyleAttribute("background", `linear-gradient(to bottom , ${UIUtils.decimalToColor(colors[0])}, ${UIUtils.decimalToColor(colors[1])})`);
                } else {
                    comp.setStyleAttribute("filter", `progid:DXImageTransform.Microsoft.gradient(startColorstr='${UIUtils.decimalToColor(colors[0])}', endColorstr='${UIUtils.decimalToColor(colors[1])}')`);
                }
            }
        } else if (UIUtils.isMoz()) {
            if (!colors || colors.length == 0)
                comp.setStyleAttribute("background", "");
            else
                comp.setStyleAttribute("background", `-moz-linear-gradient(top, ${UIUtils.decimalToColor(colors[0])}, ${UIUtils.decimalToColor(colors[1])})`);
        } else if (UIUtils.isWebKit()) {
            if (!colors || colors.length == 0)
                comp.setStyleAttribute("background", "");
            else
                comp.setStyleAttribute("background", `-webkit-gradient(linear, left top, left bottom, from(${UIUtils.decimalToColor(colors[0])}), to(${UIUtils.decimalToColor(colors[1])}))`);
        }
    }

    /**
     * Returns true if parent exists somwhere in the parent hierarchy of the child.
     * @param {*} child 
     * @param {*} parent 
     */
    static isInUIHierarchy(child, parent) {

        let childParent = child.parent;
        while (childParent) {
            if (childParent == parent)
                return true;
            childParent = childParent.parent;
        }
        return false;

    }

    /**
     * Returns nearest json object
     * @param {*} target 
     * @return {*}
     */
    static getNearestJSObject(target) {

        if (!target) {
            return null;
        }
        else if (target.hasOwnProperty('component')) {
            return target.component;
        } else if (target.parentNode) {
            return UIUtils.getNearestJSObject(target.parentNode);
        } else
            return null;

    }

    /**
     * Returns true if method exists on the object specified.
     * @param {*} obj
     * @param {String} prop
     * @return {Boolean}
     */
    static hasMethod(obj, prop) {
        return typeof obj[prop] == 'function';
    }

    /**
     * Returns true if method or property with name prop exists on the object specified.
     * @param {*} obj
     * @param {String} prop
     * @return {Boolean}
     */
    static hasMethodOrProperty(obj, prop) {
        return typeof obj[prop] != 'undefined';
    }

    /**
     * convert data object to xml
     * @param {Object} data
     * @return {Object} return xml object
     */
    static parseXML(data) {
        let xml;
        let tmp;
        try {
            if (window.DOMParser) { // Standard
                tmp = new DOMParser();
                xml = tmp.parseFromString(data, "text/xml");
            } else { // IE
                xml = new ActiveXObject("Microsoft.XMLDOM");
                xml.async = "false";
                xml.loadXML(data);
            }
        } catch (e) {
            xml = undefined;
        }
        if (!xml || !xml.documentElement || xml.getElementsByTagName("parsererror").length) {
            alert(`Invalid XML: ${data}`);
        }
        return xml;
    }

    /**
     * Sort array with property and whether there have any numeric data or not
     * @param {Array} arr 
     * @param {String} prop 
     * @param {Boolean} numeric 
     * 
     * @return {Array} returns sorted array
     */
    static sortOn(arr, prop, numeric) {
        arr.sort(
            (a, b, srt) => {
                let aVal;
                let bVal;
                if (numeric) {

                    aVal = (UIUtils.resolveExpression(a, prop));
                    bVal = (UIUtils.resolveExpression(b, prop));
                }
                else {
                    aVal = UIUtils.toString(UIUtils.resolveExpression(a, prop)).toLowerCase();
                    bVal = UIUtils.toString(UIUtils.resolveExpression(b, prop)).toLowerCase();
                }
                if (aVal < bVal)
                    return -1;
                if (aVal > bVal)
                    return 1;
                return 0;
            });
    }

    /**
     * Add static DOM event listener to the UI component having DOM element
     * @param {flexiciousNmsp.UIComponent} uiComponent 
     * @param {*} domElement 
     * @param {*} evt 
     * @param {Function} f 
     */
    static addDomEventListener(uiComponent, domElement, evt, f) {
        const func = f;
        if (domElement.addEventListener) {
            domElement.addEventListener(evt, f);
        } else if (domElement.attachEvent) {
            if (UIUtils.isIE8()) {
                //func=function(e){e.currentTarget=domElement;f(e)};
                //domElement.attachEvent("on" + evt, func);
                $(domElement).on(evt, f);
            }
            else {
                domElement.attachEvent(`on${evt}`, f);
            }
        }
        uiComponent.domListeners.push([domElement, evt, func]);

    }

    /**
     * 
     * Remove static DOM event listener from the UI component having DOM element using same function signature.
     * 
     * @param {*} domElement 
     * @param {*} evt 
     * @param {Function} f 
     */
    static removeDomEventListener(domElement, evt, f) {
        if (domElement.removeEventListener) {
            domElement.removeEventListener(evt, f);
        } else if (domElement.detachEvent) {
            $(domElement).off(evt, f);
            //domElement.detachEvent("on" + evt, f);
        }
    }

    /**
     * Attach style class to DOM elements
     * @param {*} domElement 
     * @param {String} clsName 
     */
    static attachClass(domElement, clsName) {
        if (domElement.className.indexOf(clsName) < 0) {
            domElement.className += ` ${clsName}`;
        }
    }

    /**
     * 
     * Remove style class to DOM elements
     * @param {*} domElement 
     * @param {String} clsName 
     */
    static detachClass(domElement, clsName) {
        if (domElement.className.indexOf(clsName) >= 0) {
            domElement.className = domElement.className.replace(clsName, "");
        }
    }

    /**
     * Compare two numeric values
     * if both of them NaN return 0
     * if a have NaN or b have NaN value then return 1 or -1 for each respective cases.
     * if a greater than b then return -1 otherwise return 1
     * if both are equal then return 0
     * 
     * @param {Number} a 
     * @param {Number} b 
     * 
     * @return {Number}
     */
    static numericCompare(a, b) {
        if (isNaN(a) && isNaN(b))
            return 0;

        if (isNaN(a))
            return 1;

        if (isNaN(b))
            return -1;

        if (a < b)
            return -1;

        if (a > b)
            return 1;

        return 0;
    }

    /**
     * set plain text to DOM element
     * @param {*} elem 
     * @param {String} text 
     */
    static setText(elem, text) {
        UIUtils.adapter.setText(elem, text);
    }

    /**
     * set html text to DOM element
     * @param {*} elem 
     * @param {String} html 
     */
    static setHtml(elem, html) {
        UIUtils.adapter.setHtml(elem, html);
    }

    /**
     *  Compares the two Date objects and returns an integer value
     *  indicating if the first Date object is before, equal to,
     *  or after the second item.
     *
     *  @param a Date object.
     *
     *  @param b Date object.
     *
     *  @return 0 if <code>a</code> and <code>b</code> are equal
     *  (or both are <code>null</code>);
     *  -1 if <code>a</code> is before <code>b</code>
     *  (or <code>b</code> is <code>null</code>);
     *  1 if <code>a</code> is after <code>b</code>
     *  (or <code>a</code> is <code>null</code>).
     */
    static dateCompare(a, b) {

        if (a == null && b == null)
            return 0;

        if (a == null)
            return 1;

        if (b == null)
            return -1;

        const na = a.getTime();
        const nb = b.getTime();

        if (na < nb)
            return -1;

        if (na > nb)
            return 1;

        return 0;

    }

    /**
     *  Compares two numeric values.
     *
     *  @param a First number.
     *
     *  @param b Second number.
     *
     *  @return 0 is both numbers are NaN.
     *  1 if only <code>a</code> is a NaN.
     *  -1 if only <code>b</code> is a NaN.
     *  -1 if <code>a</code> is less than <code>b</code>.
     *  1 if <code>a</code> is greater than <code>b</code>.
     *
     */
    static numericCompare(a, b) {

        if (isNaN(a) && isNaN(b))
            return 0;

        if (isNaN(a))
            return 1;

        if (isNaN(b))
            return -1;

        if (a < b)
            return -1;

        if (a > b)
            return 1;

        return 0;

    }

    /**
     *  Compares two String values.
     *
     *  @param a First String value.
     *
     *  @param b Second String value.
     *
     *  @param caseInsensitive Specifies to perform a case insensitive compare,
     *  <code>true</code>, or not, <code>false</code>.
     *
     *  @return 0 is both Strings are null.
     *  1 if only <code>a</code> is null.
     *  -1 if only <code>b</code> is null.
     *  -1 if <code>a</code> precedes <code>b</code>.
     *  1 if <code>b</code> precedes <code>a</code>.
     *
     */
    static stringCompare(a, b, caseInsensitive) {

        if (a == null && b == null)
            return 0;

        if (a == null)
            return 1;

        if (b == null)
            return -1;

        // Convert to lowercase if we are case insensitive.
        if (caseInsensitive) {
            a = a.toLocaleLowerCase();
            b = b.toLocaleLowerCase();
        }

        let result = a.localeCompare(b);

        if (result < -1)
            result = -1;
        else if (result > 1)
            result = 1;

        return result;

    }

    /**
     * Returns the string value of the passed in object, or if null, returns an empty string.
     */
    static toStringSafe(host) {
        if (host != null) return host.toString();
        return "";
    }

    /**
     * Displays a tooltip for the control in question. The tooltip will disappear if the mouse
     * moves over an area that is not the 'relativeTo' component or the tooltip component..
     * @param relativeTo Which component to position the popup relative to
     * @param tooltip The popup to display
     * @param dataContext If the popup has a data property, it will be set to this value
     * @param point If you specify this, the relativeTo is ignored, and the popup appears at the exact point you specify. Please ensure that the X and Y are relative to the Grid.
     * @param leftOffset Whether to shift the popup left after calculating the positions, for customizing the actual position
     * @param topOffset	Whether to shift the popup top after calculating the positions, for customizing the actual position
     * @param offScreenMath	Whether or not to adjust the popup if it appears off screen
     * @param where		One of three values, left, right or none. If left, positions to bottom left, if right, positions to bottom right, if none, positions right below the relativeTo component.
     * @param container	The holder for the tooltip, defaults to UIUtils.getTopLevelApplication(). You may need to override in multi window Air apps.
     * @param bounds
     * By default, the tooltip will go away once you hover the mouse out of the trigger cell or the tooltip and stayed that way for tooltipWatcherTimeout. You may also
     * manually remove the tooltip by calling the hideToolTip() function.
     */
    static showTooltip(
        relativeTo,
        tooltip,
        dataContext,
        point,
        leftOffset,
        topOffset,
        offScreenMath,
        where,
        container,
        bounds) {

        UIUtils.adapter.showTooltip(relativeTo, tooltip, dataContext, point, leftOffset,
            topOffset, offScreenMath, where, container, bounds);
    }

    static positionComponent(relativeTo, itemToPosition, my, at, leftOffset, topOffset) {

        UIUtils.adapter.positionComponent(relativeTo, itemToPosition, my, at, leftOffset, topOffset);
    }

    static hasLocalStorage() {

        try {
            return window['localStorage'];
        } catch (e) {
            return false;
        }

    }

    /**
     * Check whether attributes is already set and set value to it
     * @param {*} target 
     * @param {String} attrName 
     * @param {*} val 
     */
    static checkSetterAndApply(target, attrName, val) {
        if (typeof target[`set${UIUtils.capitalizeFirstLetterIfPrefix("set", attrName)}`] == 'function') {
            target[`set${UIUtils.capitalizeFirstLetterIfPrefix("set", attrName)}`](val);
        }
        else
            target[attrName] = val;
    }

    /**
     * Check whether attributes is already set and return value of that attribute
     * @param {*} target 
     * @param {String} attrName 
     */
    static checkGetterAndRetrieve(target, attrName) {
        if (!UIUtils.getterCache[attrName])
            UIUtils.getterCache[attrName] = `get${UIUtils.capitalizeFirstLetterIfPrefix("get", attrName)}`;
        if (typeof target[UIUtils.getterCache[attrName]] == 'function') {
            return target[UIUtils.getterCache[attrName]]();
        }
        else
            return target[attrName];
    }

    /**
     * check whether the event is a mouse event or not
     * @param evt 
     * @return {Boolean} returns true if it is mouse event othetwise false
     */
    static isMouseEvent(evt) {
        return (evt.type.toLowerCase().includes('mouse') || evt.type.toLowerCase() == "click") ? evt : null;
    }
    static isRightClickEvent(evt){
        return (evt.type.toLowerCase().indexOf('mouse')>=0 && evt.eventPhase && evt.eventPhase == 3 && evt.button == 2)?evt:null;
    };
    /**
     * Merges the props of two objects.
     * @param to
     * @param from
     */
    static mergeObjects(to, from) {
        for (const attrname in from) { to[attrname] = from[attrname]; }
    }
    /**
     * The MIT License (MIT)

Copyright (c) 2015 Rubén Norte

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
     */
    static objectAssign(target, firstSource) {
        if (target === undefined || target === null) {
            throw new TypeError('Cannot convert first argument to object');
        }

        var to = Object(target);
        for (var i = 1; i < arguments.length; i++) {
            var nextSource = arguments[i];
            if (nextSource === undefined || nextSource === null) {
                continue;
            }

            var keysArray = Object.keys(Object(nextSource));
            for (var nextIndex = 0, len = keysArray.length; nextIndex < len; nextIndex++) {
                var nextKey = keysArray[nextIndex];
                var desc = Object.getOwnPropertyDescriptor(nextSource, nextKey);
                if (desc !== undefined && desc.enumerable) {
                    to[nextKey] = nextSource[nextKey];
                }
            }
        }
        return to;
    }

    /**
     * convert string to date object
     * @param {String} valueString 
     * @param {String} inputFormat 
     * 
     * @return returns date object
     */
    static stringToDate(valueString, inputFormat) {

        let mask;
        let temp;
        let dateString = "";
        let monthString = "";
        let yearString = "";
        let j = 0;

        const n = inputFormat.length;
        for (var i = 0; i < n; i++ , j++) {
            temp = `${valueString.charAt(j)}`;
            mask = `${inputFormat.charAt(i)}`;

            if (mask == "M") {
                if (isNaN((temp)) || temp == " ")
                    j--;
                else
                    monthString += temp;
            }
            else if (mask == "D") {
                if (isNaN((temp)) || temp == " ")
                    j--;
                else
                    dateString += temp;
            }
            else if (mask == "Y") {
                yearString += temp;
            }
            else if (!isNaN((temp)) && temp != " ") {
                return null;
            }
        }

        temp = `${valueString.charAt(inputFormat.length - i + j)}`;
        if (!(temp == "") && (temp != " "))
            return null;

        const monthNum = (monthString);
        const dayNum = (dateString);
        let yearNum = (yearString);

        if (isNaN(yearNum) || isNaN(monthNum) || isNaN(dayNum))
            return null;

        if (yearString.length == 2 && yearNum < 70)
            yearNum += 2000;

        const newDate = new Date(yearNum, monthNum - 1, dayNum);

        if (dayNum != newDate.getDate() || (monthNum - 1) != newDate.getMonth())
            return null;

        return newDate;

    }

    /**
     * get theme by theme ID
     * @param {*} themeID 
     */
    static getThemeById(themeID) {
        for (const o of flexiciousNmsp.themes) {
            if (o.id == themeID) {
                return o;
            }
        }

        return null;
    }

    /**
     * check whether browser is internet explorer 8.0 or not
     */
    static isIE8() {
        if (UIUtils.isIE8Cached == -1) {
            const rv = -1; // Return value assumes failure.
            const myNav = navigator.userAgent.toLowerCase();
            if (navigator.appName == 'Microsoft Internet Explorer') {
                const ua = navigator.userAgent;
                const re = new RegExp("MSIE ([0-9]{1,}[\.0-9]{0,})");
                if (re.exec(ua) != null) {
                    UIUtils.ieVersion = (parseFloat(RegExp.$1));
                    UIUtils.isIE8Cached = (UIUtils.ieVersion < 9) ? 1 : 0;
                }

            }
            else if ((myNav.includes('trident'))) {
                UIUtils.isIE8Cached = 0;
                UIUtils.ieVersion = 11;
            }
            else
                UIUtils.isIE8Cached = 0;

        } else {
            return UIUtils.isIE8Cached;
        }
    }

    /**
     * Disable a DOM element
     * @param {*} elem 
     */
    static disableElement(elem) {
        elem.disabled = true;
        const nodes = elem.getElementsByTagName('*');
        for (let i = 0; i < nodes.length; i++) {
            nodes[i].disabled = true;
        }
        elem.style.opacity = 0.4;
    }

    static showToaster(
        message,
        title,
        type,
        toasterPosition,
        animationDuration,
        visibleDuration,
        moveAnimate,
        fadeAnimate) {
        if (UIUtils.adapter)
            UIUtils.adapter.showToaster(message
                , title
                , type
                , toasterPosition
                , animationDuration
                , visibleDuration
                , moveAnimate
                , fadeAnimate);
    }

    /**
     * convert decimal color code to hexadecimal color code
     * @param {Number} d 
     * @param {*} padding 
     */
    static decimalToColor(d, padding) {
        if (typeof (UIUtils.colorMap[d]) != "undefined") return UIUtils.colorMap[d];
        let hex = Number(d).toString(16);
        hex = "000000".substr(0, 6 - hex.length) + hex;
        return `#${hex}`;
    }

    //    UIUtils.getClassName=function(elem) {
    //        if (elem && elem["getClassNames"]) {
    //            return elem.getClassNames().join();
    //        }
    //        else if (elem && elem["typeName"]) {
    //            return elem.typeName;
    //        }
    //        else if (elem && elem.constructor) {h
    //            return elem.constructor.toString().match(
    //                /function\s*(\w+)/)[1];
    //        }
    //
    //        return "Not a JavaScript object";
    //    }


    /**
     * Escaped the escape character
     * @param {String} str 
     * @return {String}
     */
    static escapeRegExp(str) {
        return str.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&");
    }

    /**
     * Fetch the first DOM element by class name
     * @param {*} parent 
     * @param {String} containerClassName 
     * 
     * @return {Object}
     */
    static findElementWithClassName(parent, containerClassName) {
        if (!parent) return null;
        const result = parent.querySelector(`.${containerClassName}`);
        return result && result.length ? result[0] : result;
    }

    /**
     * Fetch all the DOM elements by class name
     * @param {*} parent 
     * @param {String} containerClassName 
     * 
     * @return {Array}
     */
    static findElementsWithClassName(parent, containerClassName) {
        if (!parent) return [];
        const result = parent.querySelectorAll(`.${containerClassName}`);
        return Array.prototype.slice.call(result)
    }

    /**
     * Fetch the first DOM element by tag name
     * @param {*} parent 
     * @param {String} tagName 
     * 
     * @return {Object}
     */
    static findFirstElementByTagName(parent, tagName) {
        if (!parent) return null;
        const result = parent.getElementsByTagName(tagName.toLowerCase());
        return result && result.length ? result[0] : result;
    }

    /**
     * 
     * @param {*} input 
     * @param {*} options 
     */
    static setupInputMask(input, options) {
        if (UIUtils.adapter)
            UIUtils.adapter.setupInputMask(input, options);
    }

    /**
     * 
     * @param {*} input 
     * @param {*} options 
     */
    static setupWaterMark(input, options) {
        if (options.watermark)
            options.watermark = options.watermark;
        if (options.watermarkStyle)
            options.watermarkStyle = options.watermarkStyle;
        input.placeholder = options.watermark;
    }

    /**
     * get offset values of an element
     * @param {*} el 
     * @return {Object} return offset object contains top-left values
     */
    static offset(el) {
        if (el === document) {
            return {
                top: 0, left: 0
            }
        }
        var rect = el.getBoundingClientRect();
        return {
            top: rect.top + document.body.scrollTop,
            left: rect.left + document.body.scrollLeft
        }
    }
    
    /**
     * find anscestor element of an element by its class name
     * @param {*} elem 
     * @param {*} className 
     * 
     * @return return ancestor element or null
     */
    static findAncestorByClassName(elem, className) {
        do {
            if (el.className.indexOf(className) >= 0) {
                // tag name is found! let's return it. :)
                return el;
            }
        } while (el = el.parentNode);

        // not found :(
        return null;
    }

    /**
     * convert xml to json string or data object
     * @param {*} xml 
     * @param {*} tab add tab character
     * @param {*} returnObject return as string or object 
     * 
     * @return {Object|String} returns json formatted data
     */
    static xml2json(xml, tab, returnObject) {
        if (!tab) tab = "\t";
        const X = {
            toObj(xml) {
                let o = {};
                if (xml.nodeType == 1) {   // element node ..
                    if (xml.attributes.length)   // element with attributes  ..
                        for (let i = 0; i < xml.attributes.length; i++)
                            o[`@${xml.attributes[i].nodeName}`] = (xml.attributes[i].nodeValue || "").toString();
                    if (xml.firstChild) {
                        // element has child nodes ..
                        let textChild = 0;

                        let cdataChild = 0;
                        let hasElementChild = false;
                        for (var n = xml.firstChild; n; n = n.nextSibling) {
                            if (n.nodeType == 1) hasElementChild = true;
                            else if (n.nodeType == 3 && n.nodeValue.match(/[^ \f\n\r\t\v]/)) textChild++; // non-whitespace text
                            else if (n.nodeType == 4) cdataChild++; // cdata section node
                        }
                        if (hasElementChild) {
                            if (textChild < 2 && cdataChild < 2) { // structured element with evtl. a single text or/and cdata node ..
                                X.removeWhite(xml);
                                for (var n = xml.firstChild; n; n = n.nextSibling) {
                                    if (n.nodeType == 3)  // text node
                                        o["#text"] = X.escape(n.nodeValue);
                                    else if (n.nodeType == 4)  // cdata node
                                        o["#cdata"] = X.escape(n.nodeValue);
                                    else if (o[n.nodeName]) {  // multiple occurence of element ..
                                        if (o[n.nodeName] instanceof Array)
                                            o[n.nodeName][o[n.nodeName].length] = X.toObj(n);
                                        else
                                            o[n.nodeName] = [o[n.nodeName], X.toObj(n)];
                                    }
                                    else  // first occurence of element..
                                        o[n.nodeName] = X.toObj(n);
                                }
                            }
                            else { // mixed content
                                if (!xml.attributes.length)
                                    o = X.escape(X.innerXml(xml));
                                else
                                    o["#text"] = X.escape(X.innerXml(xml));
                            }
                        }
                        else if (textChild) { // pure text
                            if (!xml.attributes.length)
                                o = X.escape(X.innerXml(xml));
                            else
                                o["#text"] = X.escape(X.innerXml(xml));
                        }
                        else if (cdataChild) { // cdata
                            if (cdataChild > 1)
                                o = X.escape(X.innerXml(xml));
                            else
                                for (var n = xml.firstChild; n; n = n.nextSibling)
                                    o["#cdata"] = X.escape(n.nodeValue);
                        }
                    }
                    if (!xml.attributes.length && !xml.firstChild) o = null;
                }
                else if (xml.nodeType == 9) { // document.node
                    o = X.toObj(xml.documentElement);
                }
                else
                    alert(`unhandled node type: ${xml.nodeType}`);
                return o;
            },
            toJson(o, name, ind) {
                let json = name ? (`"${name}"`) : "";
                if (o instanceof Array) {
                    for (let i = 0, n = o.length; i < n; i++)
                        o[i] = X.toJson(o[i], "", `${ind}\t`);
                    json += `${(name ? ":[" : "[") + (o.length > 1 ? ("\n" + ind + "\t" + o.join(",\n" + ind + "\t") + "\n" + ind) : o.join(""))}]`;
                }
                else if (o == null)
                    json += `${name && ":"}null`;
                else if (typeof (o) == "object") {
                    const arr = [];
                    for (const m in o)
                        arr[arr.length] = X.toJson(o[m], m, `${ind}\t`);
                    json += `${(name ? ":{" : "{") + (arr.length > 1 ? ("\n" + ind + "\t" + arr.join(",\n" + ind + "\t") + "\n" + ind) : arr.join(""))}}`;
                }
                else if (typeof (o) == "string")
                    json += `${name && ":"}"${o.toString()}"`;
                else
                    json += (name && ":") + o.toString();
                return json;
            },
            innerXml(node) {
                let s = "";
                if ("innerHTML" in node)
                    s = node.innerHTML;
                else {
                    const asXml = n => {
                        let s = "";
                        if (n.nodeType == 1) {
                            s += `<${n.nodeName}`;
                            for (let i = 0; i < n.attributes.length; i++)
                                s += ` ${n.attributes[i].nodeName}="${(n.attributes[i].nodeValue || "").toString()}"`;
                            if (n.firstChild) {
                                s += ">";
                                for (let c = n.firstChild; c; c = c.nextSibling)
                                    s += asXml(c);
                                s += `</${n.nodeName}>`;
                            }
                            else
                                s += "/>";
                        }
                        else if (n.nodeType == 3)
                            s += n.nodeValue;
                        else if (n.nodeType == 4)
                            s += `<![CDATA[${n.nodeValue}]]>`;
                        return s;
                    };
                    for (let c = node.firstChild; c; c = c.nextSibling)
                        s += asXml(c);
                }
                return s;
            },
            escape(txt) {
                return txt.replace(/[\\]/g, "\\\\")
                    .replace(/[\"]/g, '\\"')
                    .replace(/[\n]/g, '\\n')
                    .replace(/[\r]/g, '\\r');
            },
            removeWhite(e) {
                if (e.normalize)
                    e.normalize();
                for (let n = e.firstChild; n;) {
                    if (n.nodeType == 3) {  // text node
                        if (!n.nodeValue.match(/[^ \f\n\r\t\v]/)) { // pure whitespace text node
                            const nxt = n.nextSibling;
                            e.removeChild(n);
                            n = nxt;
                        }
                        else
                            n = n.nextSibling;
                    }
                    else if (n.nodeType == 1) {  // element node
                        X.removeWhite(n);
                        n = n.nextSibling;
                    }
                    else                      // any other node
                        n = n.nextSibling;
                }
                return e;
            }
            , cleanResult(o) {
                //if(!o.children)o.children=[];
                for (const prop in o) {
                    if (o[prop] instanceof Array) {
                        o.children = o[prop];
                        for (let i = 0; i < o[prop].length; i++) {
                            if (prop != "children" && prop != "parent")
                                X.cleanResult(o[prop][i]);
                            // o[prop][i].parent = o;
                        }
                    } else if (!UIUtils.isPrimitive(o[prop])) {
                        o.children = [o[prop]];
                        if (prop != "children" && prop != "parent")
                            X.cleanResult(o[prop]);
                        //o[prop].parent = o;
                    }
                }
            },
            parseXML(xml) {
                {
                    let dom = null;
                    if (window.DOMParser) {
                        try {
                            dom = (new DOMParser()).parseFromString(xml, "text/xml");
                        }
                        catch (e) { dom = null; }
                    }
                    else if (window.ActiveXObject) {
                        try {
                            dom = new ActiveXObject('Microsoft.XMLDOM');
                            dom.async = false;
                            if (!dom.loadXML(xml)) // parse error ..

                                window.alert(dom.parseError.reason + dom.parseError.srcText);
                        }
                        catch (e) { dom = null; }
                    }
                    else
                        alert("cannot parse xml string!");
                    return dom;
                }
            }
        };
        if (xml.nodeType == 9) // document node
            xml = xml.documentElement;
        xml = X.parseXML(xml);
        const obj = X.toObj(X.removeWhite(xml));
        if (returnObject) {
            X.cleanResult(obj);
            return obj;
        }
        const json = X.toJson(obj, xml.nodeName, "\t");
        try {
            let result = JSON.parse(`{\n${tab ? tab : ''}${tab ? json.replace(/\t/g, tab) : json.replace(/\t|\n/g, "")}\n}`);
            result = result["#document"];
            X.cleanResult(result);
            return result;
        } catch (e) {
            alert('Your browser does not seem to support JSON parse = ' + e);
        }
    }
}

flexiciousNmsp.UIUtils = UIUtils;



/**
 * Copy properties from source to target
 * @param {Object} properties 
 * @param {Object} sourceStyle 
 * @param {Object} targetStyle 
 */
function applyStyles(properties, sourceStyle, targetStyle) {
    for (const prop of properties) {
        if (sourceStyle.getStyle(prop))
            targetStyle.setStyle(prop, sourceStyle.getStyle(prop));
    }
}

const _openWindowCount = 0;

/**
 * Returns Date range of a date
 * @param val 
 * @return {Array} returns array contains two element start date and end date
 */
function getDateRange(val) {
    const dr = new flexiciousNmsp.DateRange(val);
    return ([dr.startDate, dr.endDate]);

}

UIUtils.capitalizeFirstLetterIfPrefixCache = {};

UIUtils.wheelDeltaMultiplier = 20;

UIUtils.expressionCache = {};
UIUtils.getterCache = {};
UIUtils.isIE8Cached = -1;
UIUtils.ieVersion = 9;

UIUtils.colorMap = {};



flexiciousNmsp.UIUtils = UIUtils;
